# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Apache::SessionGate;
# File: VMware/Apache/SessionGate.pm
# ---------------------------------------------------------------------------
#  1. Get/Set login and session cookies.
#  2. If the current client has not been authenticated, squirrel away the
#     original URI and request parameters; then redirect the request to the
#     authentication URI. (Redirection is probably -different- than simply
#     handing the request over.)
#  3. Determine which Perl module should be fielding the current request.
#  4. Hand the request over for processing; return results to the client.
#
#  <Location /foo>
#     SetHandler perl-script
#     PerlHandler VMware::Apache::SessionGate
#     PerlSetEnv style_DEFAULT /var/www/vmware/themes/default/foo.xsl
#     PerlSetEnv vmware_HANDLER VMware::Management::Foo
#  </Location>
#
# ---------------------------------------------------------------------------

use strict;
use Apache::Constants qw( OK REDIRECT M_GET );
use VMware::Management::Util qw( trace
                                 hash_cookies
				 );

use VMware::Management::Session qw(validate_session
                                   session_is_valid);

use VMware::Control::Profiler;
use MIME::Base64 ();
use URI::Escape ();

my $r;
my %params;
my %cookies;
my $p = VMware::Control::Profiler::getInstance("mui");

sub handler {
  my $stub = $p->profile();
  %params = ();
  $r = shift( );
  my @args = $r->method( ) eq 'GET' ? $r->args() : $r->content();

  trace( "SessionGate handler( ) called '".$r->uri( )."', method '".
         $r->method( )."', args '@args'", "debug" );

  while(my ($name, $value) = splice(@args, 0, 2)) {
    $params{$name} .= ($params{$name} ? ',' : '') . $value;
  }

  if ($r->method( ) eq 'POST') {
#     my $content = $r->content( );
#     $r->args( $content );
    $r->method( "GET" );
    $r->method_number( M_GET );
    $r->headers_in->unset( "Content-length" );
  }
#   %params = $r->args( );
  %cookies = hash_cookies( );

  direct_request( );
  return OK;
}

sub direct_request {
  my $stub = $p->profile();
  my $s;
  my %h;
  # trace(env_to_string(), "debug");
  $r->header_out( "Last-Modified", VMware::Management::Util::rfc_1123_time( ) );

  # If the client is authenticated and if the current URI is the login URI,
  # forward the client to the default URI; if the current URI is something
  # other than the login URI, direct the client's request appropriately. If
  # the client is not authenticated, ask to see some credentials.
  if ( session_is_valid( $cookies{ "vmware.mui.sid" } ) ) {
    unless ( $params{ "auto-refresh" } ) {
      $r->headers_out->add( "Set-Cookie" => validate_session( $cookies{ "vmware.mui.sid" } ) );
    }
    if ( $ENV{ "SCRIPT_NAME" } eq $ENV{ "vmware_LOGIN_URI" } ) {
      forward_request( $ENV{ "vmware_DEFAULT_URI" } );
    } else {
      ( $s, %h ) = call_handler( );
    }
  } else {
    if ( $ENV{ "SCRIPT_NAME" } eq $ENV{ "vmware_LOGIN_URI" } ) {
      ( $s, %h ) = call_handler( );
    } else {
      force_login( );
    }
  }

  foreach my $k ( keys( %h ) ) {
    my $l = $k;
    $l =~ tr/A-Z/a-z/;
    if ( $l eq "content-type" ) {
      $r->content_type( $h{ $k } );
    } else {
      if ( $l eq "location" ) {
        forward_request( $h{ $k } );
      }
      if ( $l =~ /cookie/ ) {
	 if (defined($h{$k})) {
	    $r->headers_out->add( "Set-Cookie" => $h{ $k } );
	    trace( "$k is ".$h{ $k }, "debug" );
	 }
	 next;
      }
      $l = $h{ $k };
      $l =~ tr/A-Z/a-z/;
      if ( $l eq "no-cache" ) {
        $r->no_cache( 1 );
      }
      $r->header_out( "$k", $h{ $k } );
    }
  }

  if ( $ENV{ "SCRIPT_NAME" } eq $ENV{ "vmware_LOGIN_URI" } ) {
    $r->headers_out->add( "Set-Cookie" => "vmware.mui.test=1;" );
  }
  $r->send_http_header( );

  #trace("\nprint:\n'$s'");

  $r->print( $s ? $s : "\n\n");
}

sub call_handler {
  my $stub = $p->profile();
  no strict 'refs';

  my $handler = $ENV{ "vmware_HANDLER" };
  eval( "require $handler" );
  if( $@ ) {
      trace( "Failed to load handler.  Reason: $@" );
      return("Error: Failed to load handler.\n\nReason: $@\n", undef);
  }
  $handler .= "::handler";
  trace( "$handler( ) called.", "debug" );
  return &$handler( \%params );
}

sub force_login {
  my $stub = $p->profile();
  my $uri = $ENV{ "vmware_LOGIN_URI" }."?n=";
  my $next = $ENV{ "SCRIPT_NAME" }.$ENV{ "PATH_INFO" }."?";

  foreach my $key ( keys( %params ) ) {
    unless ( !$params{ $key } ) {
      $next .= "$key=".$params{ $key }."&";
    }
  }
  chop( $next );

  forward_request( $uri.URI::Escape::uri_escape( MIME::Base64::encode_base64( $next, "" ) ) );
}

sub forward_request {
  my $uri = shift( );
  if ( !$ENV{ "HTTP_HOST" } ) {
    $r->header_out( "Location" => $ENV{"URL_PREFIX"}.$ENV{ "SERVER_NAME" }.($ENV{ "SERVER_PORT" } && $ENV{ "SERVER_PORT" } ne "80" ? ":".$ENV{ "SERVER_PORT" } : "" ).$uri );
  } else {
    $r->header_out( "Location" => $ENV{"URL_PREFIX"}.$ENV{ "HTTP_HOST" }.$uri );
  }
  $r->status( REDIRECT );
}

sub env_to_string {
  my $s = $r->uri . "\n";
  foreach my $key ( keys( %params ) ) {
    $s .= "$key = ".$params{ $key }."\n";
  }
  foreach my $key ( keys( %ENV ) ) {
    $s.= "$key = ".$ENV{ $key }."\n";
  }
  return $s;
}

1;
