#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# InetdControl.pm
# 
# A basic class that allows the stopping & starting of
# inetd services (ie ftp, telnet, etc) on the console os.
#
# Supports both inetd and xinetd (xinetd support is minimal)


package VMware::HConfig::InetdControl;
use strict;
use VMware::ExtHelpers qw(&Warning &shell_string &SystemWithReset &mktmpfile &safe_move);
use VMware::Tokenizer;
use VMware::HConfig::Constant qw(c_getFile c_getDir c_getHelper);
				 


my $port_list  = "/etc/services";
my $inetd_conf  = "/etc/inetd.conf";
my $xinetd_dir  = "/etc/xinetd.d";
my $xinetd_conf = "/etc/xinetd.conf";

my %inetd_managedServices = (
   'telnet' => "telnet  stream  tcp   nowait  root   /usr/sbin/tcpd  in.telnetd\n",
   'ftp' =>    "ftp     stream  tcp   nowait  root   /usr/sbin/tcpd  in.ftpd -l -a\n",
);

#####################################################################
# Public functions                                                  #
#####################################################################

sub enable_service($) {
  my ($service) = @_;

  if ( -e $xinetd_conf || -d $xinetd_dir ) {
     return enable_xinetd($xinetd_conf, $xinetd_dir, $service);
  }

  # check for inetd
  if ( -e $inetd_conf ) {
     return enable_inetd($inetd_conf, $service);
  }
  Warning("Couldn't find $xinetd_conf or $inetd_conf.");
  return 0;
}

sub disable_service($) {
  my ($service) = @_;

  if ( -e $xinetd_conf || -d $xinetd_dir ) {
     return disable_xinetd($xinetd_conf, $xinetd_dir, $service);
  }

  # check for inetd
  if ( -e $inetd_conf ) {
     return disable_inetd($inetd_conf, $service);
  }
  Warning("Couldn't find $xinetd_conf or $inetd_conf.");
  return(0);
}


sub inetdQuery($) {
  my ($service) = @_;

  print("inetdQuery: not implemented\n");
  return(1);
}

#####################################################################
# Private functions                                                 #
#####################################################################




#
# inetd related functions
#

# enable an inetd service by uncommenting the line in the inetd.conf file.
#
# returns 1 on sucess, 0 otherwise.
sub enable_inetd($$) {
   my ($conf,$service) = @_;
   local *CONF;
   my @lines;
   my $found = 0;

   my $port_num = getPort($service);
   if (!open(CONF, "<" . $conf)) {
      Warning("Couldn't open $conf for reading. Reason $!");
      return;
   }

   @lines = <CONF>;
   close(CONF);
   if (grep(/^\s*($port_num|$service)\s+/, @lines)) {
#      Warning("$service is already configured.  Leaving current configuration unchanged");
      return(1);
   }

   my @matches = grep(/^\s*#\s*($port_num|$service)\s+/, @lines);
   foreach (@lines) {
     if (m/^\s*#(\s*($port_num|$service)\s+.*$)/) {
       if (!$found) {
          print("found match\n");
          $found = 1;
          $_ = "$1\n";   #remove '#' from current line w
       }else {
         Warning("Found multiple entries for $service.  Enabling only the first\n");
         last;
       }
     }
   }

   if (!$found) {
      if ( defined($inetd_managedServices{$service}))  {
         # We know how to create an entry of this type, so do it
         push(@lines,$inetd_managedServices{$service});
      } else  {
         Warning("Cannot find an inetd entry for $service.  Is the correct package installed?");
         return(0);
      }
   }

   saveTo($conf, @lines);

   if (SystemWithReset(c_getHelper('restart_inetd'))) {
      Warning("Could not restart inetd.");
      return(0);
   }
   return 1;
}


# disable_inetd
#
# comments out all lines corresponding to $service in inetd.conf
# returns 1 on sucess, 0 otherwise.
sub disable_inetd($$) {
   my ($conf,$service) = @_;
   my $port_num = getPort($service);
   local *FD;
   
   if (!open(FD, "<" . $conf)) {
      Warning("Couldn't open $conf for reading. Reason $!");
      return(0);
   }

   my @lines = <FD>;
   close(FD);
   foreach (@lines) {
      s/(^\s*($service|$port_num)\s+.*$)/#$1/;
   }

   saveTo($conf, @lines);

   if (SystemWithReset(c_getHelper('restart_inetd'))) {
      Warning("Could not restart inetd.");
      return(0) ;
   }
   return(1);
}



# xinetd related functins functions
#
#
# These are really brittle -- will only work on rh7.x or related distros.
# (on rh7.x /usr/sbin/chkconfig actually stopping / starting of
# xinetd services -- that would probably be a better method.)
# In the long run, we will have to implement full parsing of
# the xinetd conf files.
#
#
# These will be rewritten once I port ssl changes back to tot.

sub enable_xinetd($$$) {
  my ($conf,$dir,$service) = @_;

  return set_xinetd($conf, $dir, $service, "no");
}

sub disable_xinetd($$$){
  my ($conf,$dir,$service) = @_;

  return set_xinetd($conf, $dir, $service, "yes");
}


# set_xinetd
#
# helper function that actually reads all of the files /etc/xinet.d/ and
# looks for a given service.  if the file contains the service, it
# toggles the state of the "disabe = xxx" line.
#
sub set_xinetd($$$$){
  my ($conf, $dir, $service, $state) = @_;
  my $port_num = getPort($service);
  my $found = 0;
  my $file;
  my @lines;

  if (!-d $dir) {
     Warning("$dir not a directory");
     return(0);
  }

  # XXX use readdir
  foreach $file (`ls $dir/*`) {
     local *FD;
     if (!open(FD, "<" . $file)) {
        Warning("Couldn't open $file for reading. Reason $!");
        next;
     }
     @lines = <FD>;
     close(FD);

     #Check for "service xxx" or "port =  #"
     #

     if (grep(/(^\s*service\s+$service\s*$)|(^\s*port\s*=\s*$port_num\s*$)/, @lines)) {
        my @enableStatus = grep(s/(^\s*disable\s*=).*$/$1 $state/, @lines);

        #print("Found match in file $file");
        if (@enableStatus > 1) {
           Warning("$file is malformed: Multiple \"disable = xxx\" lines.");
        }
        #XXX Eventually need to make this smarter.  (ie actually
        # read in & understand the format.)
        if (@enableStatus < 1) {
           Warning("$file is malformed: Failed to modify $service\n");
           next;
        }

        #XXX potential race.  File could have
        # been over written & replaced with symlink.
        if (!open(FD, ">" . $file)) {
           Warning("Couldn't open /tmp/$file for writing. Reason $!");
           next;
        }
        print FD @lines;
        close(FD);
        if (SystemWithReset(c_getHelper('restart_xinetd'))) {
           Warning("Could not restart xinetd.");
        }
        return(1);
     }
  }

  Warning("Cannot find an xinetd entry for $service in $dir." .
      "Is the correct package installed?");
  return(0);
}



###########################################################################
# Generic functions                                                       #
###########################################################################

# saveTo
# 
# saves @lines to filename. 
# returns 1 on success, 0 otherwise.
sub saveTo($$) {
   my ($saveFile, @lines) = @_;
   my $tmpfile = mktmpfile();
   local *FD;

   if (!open(FD, ">" . $tmpfile)) {
      Warning("Couldn't open $tmpfile for writing.  Reason $!");
      return(0);
   }
   print FD @lines;
   close(FD);

   if (!safe_move($tmpfile,$saveFile)) {
      Warning("Warning: Cannot move $tmpfile to $saveFile. Reason $!");
      return(0);
   }
}

# look in /etc/services for a port number corresponding to $service
sub getPort($) {
  my ($service) = @_;
  local *FD;

  if (!open(FD, "<" . $port_list)) {
     Warning("Couldn't open $port_list");
     return(0);
  }

  while (<FD>) {
     if (/^\s*$service\s*(\d*)\//) {
       close(FD);
       return($1);
     }
  }
  Warning("Didn't find entry for $service in $port_list");
  return(0);
}




1;
