#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# LILOConfig.pm
# 
# A basic class that contains information representing the lilo.conf
# configuration file.  Information includes a series of options, both
# global and local.

package VMware::HConfig::LILOConfig;
use strict;
use VMware::ExtHelpers qw(&Warning &shell_string);
use VMware::HConfig::LILOBootImage;
use VMware::Tokenizer;

#####################################################################
# Public Static Methods (indep. of each object)                     #
#####################################################################

sub new() {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($self) = create($class);
    return($self);
}

sub newFile($) {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($lilo) = @_;
    my ($self) = create($class);
    $self->parseLILOConf($lilo);
    return($self);
}


#####################################################################
# Public Methods                                                    #
#####################################################################

# Accessor functions
sub getDefaultImage() {
    my ($self) = shift;
    return($self->{DFL_IMAGE});
}

sub getBootImages() {
    my ($self) = shift;
    return($self->{LOCALS});
}

# Locates an image given a label.  Returns the boot image matching
# it.
sub getBootImageByLabel($) {
    my ($self) = shift;
    my ($label) = @_;

    my @images;

    # Finds all the ones that match
    @images = grep { $_->getLabel() eq $label } @{$self->{LOCALS}};

    if( $#images > -1 ) {
	# There should be only one.  Return only one
	return($images[0]);
    }
    else {
	return(undef);
    }
}

# Locates an image given an image name.  Returns the boot images
# that match.  There may be more than one.
sub getBootImageByImage($) {
    my ($self) = shift;
    my ($name) = @_;

    my @images;

    # Finds all the ones that match
    @images = grep { $_->getImage() eq $name } @{$self->{LOCALS}};

    return(\@images);
}

# Returns a list of all the boot image labels
sub getBootImageLabels() {
    my ($self) = shift;
    # Create a list of the labels
    my @labels = map { $_->getLabel() } @{$self->{LOCALS}};
    return(\@labels);
}

# Returns a list of all the boot image file names
sub getBootImageImages() {
    my ($self) = shift;
    # Create a list of the image names
    my @name = map { $_->getImage() } @{$self->{LOCALS}};
    return(\@name);
}

# Check to see if an existing VMnix configuration exists
sub getVMnixConfigs() {
    my ($self) = shift;

    # Get the list of boot images that run the VMnix available
    my @images = @{$self->getBootImages()};
    @images = grep { $_->usesVMnixImage() } @images;

    return(\@images);
}

# Gets only the labels of VMnix boot images
sub getVMBootImageLabels() {
    my ($self) = shift;
    # Create a list of the labels
    my @images = @{$self->getVMnixConfigs()};
    my @labels = map { $_->getLabel() } @images;
    return(\@labels);
}


sub getUniqVMBootName() {
    my ($self) = shift;
    my @labels = @{$self->getBootImageLabels()};

    # Check the default first 
    my @ls = grep { $_ eq "vmnix" } @labels;
    if( $#ls == -1 ) {
	return("vmnix");
    }

    my $i = 2;
    while( 1 ) {
        @ls = grep{ $_ eq "vmnix$i" } @labels;
	if( $#ls == -1 ) {
	    return("vmnix$i");
	}
	$i++;
    }
}


sub toString() {
    my $self = shift;   

    my $s = $self->{GLOBALS} . 
	"default=" . $self->{DFL_IMAGE} . "\n\n";
    my $imgref = $self->{LOCALS};
    my @imgs = @$imgref;

    my $img;

    foreach $img (@imgs) {
	$s .= $img->toString();
    }
	   
    return($s);
}


# Mutator functions
# Succeeds if the label exists.  Returns 1 on success, 0 on failure.  
sub setDefaultImage($) {
    my ($self) = shift;
    my ($def) = @_;

    # Create a list of the labels
    my @labels = map { $_->getLabel() } @{$self->{LOCALS}};

    # See if there is a label that matches exactly
    @labels = grep { /^$def$/ } @labels;

    if( $#labels > -1 ) {
	$self->{DFL_IMAGE} = $def;
	return(1);
    }

    return(0);
}

# Pass in the boot image to be added.
sub addBootImage($) {
    my ($self) = shift;
    my ($img) = @_;
    
    push(@{$self->{LOCALS}}, $img);
}

# Pass in the label of the image that is to be deleted
sub deleteBootImage($) {
    my ($self) = shift;
    my ($label) = @_;
    
    my $img;
    my $i = 0;
    foreach $img (@{$self->{LOCALS}}) {
	if( $label eq $img->getLabel() ) {
	    splice(@{$self->{LOCALS}}, $i, 1);
	    return;
	}
	$i++;
    }
}

# Factory methods
sub createBootImage($$) {
    my($self) = shift;
    my($type, $image) = @_;
    return(VMware::HConfig::LILOBootImage->new($type, $image));
}


#####################################################################
# Methods not intended for use by user                              #
#####################################################################

sub create($) {
    my $class = shift;
    my $self = {};
    $self->{GLOBALS} = "";     # String of config data for global options
    $self->{DFL_IMAGE} = "";   # String containing label of global
    $self->{LOCALS} = [];      # Ref to an array of LILOBootImage objects

    bless($self, $class);        
    return($self);
}

# Describes the mapping for escape sequences
my $EscapeChar = "\\";
my %EscapeSequenceMapping = ( 
			      "\\\\", "\\",
			      '\n', "\n",
			      '\t', "\t",
			      '\"', "\""
			      );
my $Delimiters = "=";

sub parseLILOConf($) {
    my ($self) = shift;
    my ($lilo) = shift;

    my $global = 1;  # State flag indicating that parser is looking for globals
    my $curlocal = {};  # Local object

    local *FD;

    my $numimages = -1;

    # First read the lilo.conf file to get the PCI information
    if( !open(FD, '< ' . $lilo) ) {
	# Error reading file
	Warning("Unable to open $lilo");
	return(0);
    }

    # Set up the tokenizer to understand the lilo.conf syntax
    my $tokenizer = VMware::Tokenizer->new();
    $tokenizer->setEscapeChar($EscapeChar);
    $tokenizer->setEscapeMapping(\%EscapeSequenceMapping);
    $tokenizer->setDelimiters($Delimiters);

    # First process the global options
    # Basic approach:
    #
    # For the global options, we are just going to copy everything
    # verbatim from /etc/lilo.conf with the exception of the default
    # tag.  We are going to keep the default tag in case we need to
    # change it at the request of the user.
    # 
    # For the local options, we are going to save information regarding
    # the type of image, the name of the image, the label, and the append
    # options.  If we are modifying an existing VM configuration, the 
    # append line will have values that we wish to keep.  Also, the
    # user may have some append options there that we should keep.
    while(<FD>) {
	chomp();      # Remove trailing newlines

	# Prepare to tokenize by reseting the count to zero before tokenizing
	$tokenizer->resetCount();    
	my ($rc, $toks) = $tokenizer->tokenize($_);

	my @tokens = @$toks;

	if( !$rc ) {
	    # Error in parsing
	    Warning("parseLILOConf: Error in parsing of string.");
	}

	if( $global ) {       # Are we looking for global config vars?
	    # Read global options until we get to the per-image options.
            if( $#tokens == -1 ) { 
		next;       # Skip if its just white space
	    }
	    elsif( $#tokens == 2 && $tokens[1]->token() eq '=' ) {
		my $key = $tokens[0]->token();
		my $value = $tokens[2]->token();
		
		# Have we reached the start of the per-image options?
		if( VMware::HConfig::LILOBootImage::isValidType($key) ) {
		    # Yes, break out and start handling the per-image options.
		    $global = 0;

		    # Make a new boot image
		    $numimages++;
		    $curlocal = VMware::HConfig::LILOBootImage->new($key, $value);
		    $self->{LOCALS}->[$numimages] = $curlocal;
		    next;
		}
		# No, but does this line set the default boot image?
		elsif( $key eq "default" ) {
		    # We always take the last default that is defined in the file
		    # in the event that there are multiple default tags.
		    $self->{DFL_IMAGE} = $value;
		    next;
		}
	    }
	    
	    # Uninteresting global option, save it and go on to the next line
	    $self->{GLOBALS} .= $_ . "\n";
	} # if($global) 
	else {
	    # Global options are done. Handle per-image options.
	    if( $#tokens == 2 && $tokens[1]->token() eq '=' ) {
		my $key = $tokens[0]->token();
		my $value = $tokens[2]->token();

		# Is this the label?
		if( $key eq "label" ) {
		    $curlocal->setLabel($value);
		    next;
		}
		# Is this the label?
		elsif( $key eq "root" ) {
		    $curlocal->setRootDev($value);
		    next;
		}
		# Is this the initrd?
		elsif( $key eq "initrd" ) {
		    $curlocal->setInitrd($value);
		    next;
		}

		# Is this the next image?
		if( VMware::HConfig::LILOBootImage::isValidType($key) ) {
		    # Make a new boot image
		    $numimages++;
		    $curlocal = VMware::HConfig::LILOBootImage->new($key, $value);
		    $self->{LOCALS}->[$numimages] = $curlocal;		    
		    next;
		}

		# Is this the append line?
		if( $key eq "append" ) {
                    # Create a new tokenizer to parse the append line.
                    my $tokenizer2 = VMware::Tokenizer->new();
                    $tokenizer2->setDelimiters('=');

		    my ($rc2, $toks2) = $tokenizer2->tokenize($value);
		    
		    my @tokens2 = @$toks2;
		    
		    if( !$rc2 ) {
			# Error in parsing
			Warning("parseLILOConf: Error in parsing append line string.");
		    }

		    # Examine the append line
		    while( $#tokens2 != -1 ) {
			my $t = shift(@tokens2);

			if( $t->token() eq '"' ) {   # Skip the quotes
			    next;
			}
			elsif( $t->token() eq "mem" ) {
			    if( $#tokens >= 1 ) {
				my $t1 = shift(@tokens2);
				my $t2 = shift(@tokens2);
				if( $t1->token() eq '=' && $t2->token() =~ /(\d+)M/ ) {
				    $curlocal->setMemSize($1);
				}
			    }
			}
			elsif( $t->token() eq "cpci" ) {
			    if( $#tokens >= 1 ) {
				my $t1 = shift(@tokens2);
				my $t2 = shift(@tokens2);
				if( $t1->token() eq '=' && $t2->token() =~ /[0-9:;,-]*/ ) {
				    $curlocal->setCPCI($t2->token());
				}
			    }
			}
			elsif( $t->token() eq "console" ) {
			    if( $#tokens >= 1 ) {
				my $t1 = shift(@tokens2);
				my $t2 = shift(@tokens2);
			    }
			}
			else {
			    # Not of interest.  Just append whatever this data is.
                            my $otherappend = $curlocal->getOtherAppend();
                            if ($t->token() eq "=") {
                                $otherappend .= $t->token();
                                if( @tokens2 ) {
                                    $t = shift(@tokens2);
                                    $otherappend .= $t->token();
                                }
                            } else {
                                if ($otherappend ne "") {
                                    $otherappend .= " ";
                                }
                                $otherappend .= $t->token();
                            }
			    $curlocal->setOtherAppend( $otherappend );
			}
		    } # while( $#tokens2 != -1 )  Handling of the append line

		    next;
		}
	    } # if( $#tokens == 2 &&

	    # Uninteresting image option, save it and go on to the next line
	    $curlocal->{OPTIONS} .= $_ . "\n";
	} # Per-image handler statement
    } # while(<FD>)

    close(FD);

    return(1);      # Success
}

1;
