#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# MFSFileSystem.pm
#
# A basic class that contains basic information for each file system.
# and provides ways to do things in a file system.


package VMware::HConfig::MFSFileSystem;
use strict;
use VMware::ExtHelpers qw(&Warning &System &shell_string &Debug &round);
use VMware::HConfig::Constant qw(c_getHelper $gVmkDeviceType_hba);
use VMware::HConfig::RawDisk;

#####################################################################
# Public Static Methods (indep. of each object                      #
#####################################################################

# Takes in a RawDisk object and partition which is information about
# where this file system can be found.
sub new($$) {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my ($disk, $part) = @_;
    my $self = {};

    if (VMware::HConfig::RawDisk::isLogVmkDiskName($disk)) {
	# Returns successfully on if the RawDisk's name is of the form "scsi#:# and the
	# partition indicated refers to a valid MFS partition
	my $self = $proto->create($disk, $part);

        # Irrelevant to an MFS partition
        #	$self->{MOUNT_POINT} = "";   # Location where FS is mounted

	bless($self, $class);
	return($self);
    }
    else {
	return(undef);
    }
}


sub create($$) {
    my $proto = shift;
    my $self = {};
    my ($disk, $part) = @_;

    $self->{DISK} = $disk;       # Disk name where fs is located
    $self->{PARTITION} = $part;  # Partition number, according to Linux/VM convention
    resetFSInfo($self);

    return($self);
}

sub resetFSInfo {
    my $self = shift;
    $self->{VERSION} = undef;       # VMFS major version (1 or 2 for now)
    $self->{MINORVERSION} = undef;  # VMFS minor version (between 00 and 99)
    $self->{CAPACITY} = -1;      # Total size of the file system in MBytes
    $self->{FREE_CAPACITY} = -1; # Total size of the file system in MBytes
    $self->{BLOCK_SIZE} = -1;    # Block size of the file system in MBytes
    $self->{LOG_NAME} = undef;   # The logical name of the VMFS
    $self->{ACCESSIBILITY} = undef; # VMFS accessibility (private/shared/public)
    $self->{FILES} = [];
    $self->{UUID} = undef;       # VMFS UUID
    $self->{EXTENTS} = undef;
}

#####################################################################
# Public Methods                                                    #
#####################################################################

sub setMountPoint($) {
    my $self = shift;
    my $mntpt = shift;

    $self->{MOUNT_POINT} = $mntpt;
}

sub getMountPoint() {
    my $self = shift;
    return ($self->{MOUNT_POINT});
}

# Accessor functions
sub getDisk() {
    my $self = shift;
    return($self->{DISK});
}

sub getPartition() {
    my $self = shift;
    return($self->{PARTITION});
}

sub getVmkName() {
    my $self = shift;
    return($self->getDisk() . ":" . $self->getPartition());
}

sub getLogName() {
    my $self = shift;

    if (!defined($self->{LOG_NAME})) {
	$self->loadFSInfo();
    }

    return($self->{LOG_NAME});
}

sub getCapacity() {
    my $self = shift;

    if ($self->{CAPACITY} == -1) {
	$self->loadFSInfo();
    }

    return($self->{CAPACITY});
}

sub getFreeCapacity() {
    my $self = shift;

    if ($self->{FREE_CAPACITY} == -1) {
	$self->loadFSInfo();
    }

    return($self->{FREE_CAPACITY});
}

sub getBlockSize() {
    my $self = shift;

    if ($self->{BLOCK_SIZE} == -1) {
	$self->loadFSInfo();
    }

    return($self->{BLOCK_SIZE});
}

sub getAccessibility() {
    my $self = shift;

    if (!defined($self->{ACCESSIBILITY})) {
	$self->loadFSInfo();
    }

    return($self->{ACCESSIBILITY});
}

sub getVersion() {
   my $self = shift;

   if (!defined($self->{VERSION})) {
      $self->loadFSInfo();
   }

   return($self->{VERSION});
}

sub getMinorVersion() {
   my $self = shift;

   if (!defined($self->{MINORVERSION})) {
      $self->loadFSInfo();
   }

   return($self->{MINORVERSION});
}

sub getUuid() {
   my $self = shift;

   if (!defined($self->{UUID})) {
      $self->loadFSInfo();
   }

   return($self->{UUID});
}

sub getExtents() {
   my $self = shift;

   if (!defined($self->{EXTENTS})) {
      $self->loadFSInfo();
   }

   return($self->{EXTENTS});
}

sub toString() {
    my $self = shift;

    my $s = "";

    # Print out the fs information
    $s .=
       ("File system " . $self->getVmkName() .
	" logical name='" . $self->getLogName() . "'" .
	" capacity=" . $self->getCapacity() . "Mb" .
	" free=" . $self->getFreeCapacity() . "Mb" .
        " version=" . $self->getVersion() .
        " minorversion=" . $self->getMinorVersion() .
        " extents=" . join(',', @{$self->getExtents()}) .
        "\n");

    my $fileEntry;
    foreach $fileEntry (@{$self->listFiles()}) {
	my $name = $fileEntry->[0];
	my $size = $fileEntry->[1];
	my $perm = $fileEntry->[2];
	my $uid  = $fileEntry->[3];
	my $gid  = $fileEntry->[4];
	my $attr = $fileEntry->[5];
	my $mtime = $fileEntry->[6];

	$s .= "  File $name, size = $size MB, owner = $uid, access = $perm, ";
	$s .= "group = $gid, attribute = $attr, modif time = $mtime \n";
    }

    return($s);
}

# Returns information about files in the directory.  Information is
# returned as an array containing fields name, size, permission, uid, gid
# attribute (cow) and last modif time.
sub listFiles() {
    my $self = shift;

    if ($self->{CAPACITY} == -1) {
	$self->loadFSInfo();
    }

    return($self->{FILES});
}

# Returns TRUE if the given file exists in the VMFS.
sub fileExists($) {
    my $self = shift;
    my ($filename) = @_;

    my @files = @{ $self->listFiles() };

    for my $f (@files) {
	my ($fname, $size, $type) = @$f;

	if ($fname eq $filename) {
	    return 1;
	}
    }

    return 0;
}

# Returns information about all files in the VMFS
sub getAllFiles() {
    my $self = shift;

    my @files = @{ $self->listFiles() };

    my $hash = {};
    for my $f (@files) {
	my ($fname, $size, $perm, $uid, $gid, $attr, $mtimeStr) = @$f;

	if ($perm !~ /^(.)(.)(.)(.)(.)(.)(.)(.)(.)/) {
	    Warning("MFSFileSystem::getAllFiles: invalid permissions $perm");
	}

	my $om = ($1 eq "r")*4 + ($2 eq "w")*2 + ($3 eq "x")*1;
	my $gm = ($4 eq "r")*4 + ($5 eq "w")*2 + ($6 eq "x")*1;
	my $wm = ($7 eq "r")*4 + ($8 eq "w")*2 + ($9 eq "x")*1;

	my $modeString = "0" . $om . $gm . $wm;
	my $mode = oct($modeString);

	$hash->{$fname} = [$size, $mode, $uid, $gid, $attr, $mtimeStr];
    }

    return $hash;
}

sub getFile($) {
    my $self = shift;
    my ($filename) = @_;

    my $allFiles = $self->getAllFiles();
    my $arr = $allFiles->{$filename};
    if (!defined($arr)) {
	return undef;
    }

    my ($size, $mode, $uid, $gid, $attr, $mtimeStr) = @{ $arr };

    return ($filename, $size, $mode, $uid, $gid, $attr, $mtimeStr);
}

# Mutator functions

# XXX rbosch: Since we're now caching the data in this module, the
#             mutator functions must call resetFSInfo() on success
#             to invalidate the data and force a reload on the next
#             access.

# Export a file from src to dest
sub exportFile($$) {
    my $self = shift;
    my ($src, $dest) = @_;

    if( !c_getHelper('vmfs_export_file') ) {
	Warning("Cannot execute vmfs_export_file.");
	return(0);
    }

    my $command = c_getHelper('vmfs_export_file') . ' ' . shell_string($dest) . ' ' . shell_string($self->getVmkName() . ":$src");

    # We'll keep stderr messages.  They'll either appear on the console or the
    # web server logs
    if(System($command)) {
	# Error
	Warning("Error in MFSFileSystem::exportFile: $!");
	return(0);
    }

    $self->resetFSInfo();
    return(1);
}

# Import a file from src to dest
sub importFile($$) {
    my $self = shift;
    my ($src, $dest) = @_;

    if( !c_getHelper('vmfs_import_file') ) {
	Warning("Cannot execute vmfs_import_file.");
	return(0);
    }

    my $command = c_getHelper('vmfs_import_file') . ' ' . shell_string($src) . ' ' . shell_string($self->getVmkName() . ":$dest");

    # We'll keep stderr messages.  They'll either appear on the console or the
    # web server logs
    if(System($command)) {
	# Error
	Warning("Error in MFSFileSystem::importFile: $!");
	return(0);
    }

    $self->resetFSInfo();
    return(1);
}

# Remove a file
sub removeFile($) {
    my $self = shift;
    my ($filename) = @_;

    if( !c_getHelper('vmfs_remove_file') ) {
	Warning("Cannot execute vmfs_remove_file.");
	return(0);
    }

    my $command = c_getHelper('vmfs_remove_file') . " " .
        shell_string($self->getPathname($filename));

    # We'll keep stderr messages.  They'll either appear on the console or the
    # web server logs
    if(System($command)) {
	Warning("Error in MFSFileSystem::removeFile: $!");
	return(0);
    }

    $self->resetFSInfo();
    return(1);
}

# Create a file
#
# Size is in megabytes
sub createFile($$) {
    my $self = shift;
    my ($filename, $size) = @_;

    if( !c_getHelper('vmfs_create_file') ) {
	Warning("Cannot execute vmfs_create_file.");
	return(0);
    }

    my $command = c_getHelper('vmfs_create_file') . " " . shell_string($size . "M") . ' ' .
	shell_string($self->getVmkName() . ":$filename");

    # We'll keep stderr messages.  They'll either appear on the console or the
    # web server logs
    if(System($command)) {
	Warning("Error in MFSFileSystem::createFile: $!");
	return(0);
    }

    $self->resetFSInfo();
    return(1);
}

# Changes access permissions on the given file
#
# Input: Name of file, uid and gid of user and mode for file
sub changeAccessPerm($$$$) {
    my $self = shift;
    my ($filename, $uid, $gid, $mode) = @_;
    my $command;

    if (!c_getHelper('vmfs_change_owner') || !c_getHelper('vmfs_change_access')) {
	Warning("Cannot change access permissions on '$filename'.");
	return(0);
    }

    $command = c_getHelper('vmfs_change_owner') . " " .
        shell_string($uid . ":" . $gid) . " " .
        shell_string($self->getPathname($filename));

    if (System($command)) {
	Warning("Error in MFSFileSystem::changeAccessPerm: $!");
	return (0);
    }


    # $mode is an integer. Convert it to Octal string
    my $modeString = sprintf("%o", $mode);
    $command = c_getHelper('vmfs_change_access') . " " .
        shell_string($modeString) . " " .
        shell_string($self->getPathname($filename));

    if (System($command)) {
	Warning("Error in MFSFileSystem::changeAccessPerm: $!");
	return (0);
    }

    $self->resetFSInfo();
    return (1);
}

# Changes ownership of the given file
#
# Input: Name of file and uid for owner
sub changeOwner($$) {
    my $self = shift;
    my ($filename, $uid) = @_;

    if (!(defined($filename) && defined($uid))) {
	Warning("MFSFileSystem::changeOwner: invalid input");
	return (0);
    }

    my ($fname, $size, $mode, $fuid, $fgid) = $self->getFile($filename);

    if (!defined($fname)) {
	Warning("MFSFileSystem::changeOwner: no such file $filename");
	return (0);
    }

    return $self->changeAccessPerm($filename, $uid, $fgid, $mode);
}

# Changes group of the given file
#
# Input: Name of file and gid for group
sub changeGroup($$) {
    my $self = shift;
    my ($filename, $gid) = @_;

    if (!(defined($filename) && defined($gid))) {
	Warning("MFSFileSystem::changeOwner: invalid input");
	return (0);
    }

    my ($fname, $size, $mode, $fuid, $fgid) = $self->getFile($filename);

    if (!defined($fname)) {
	Warning("MFSFileSystem::changeOwner: no such file $filename");
	return (0);
    }

    return $self->changeAccessPerm($filename, $fuid, $gid, $mode);
}


# Changes owner and group of the given file
#
# Input: Name of file, uid for user, gid for group
sub changeOwnerAndGroup($$$) {
    my $self = shift;
    my ($filename, $uid, $gid) = @_;

    if (!(defined($filename) && defined($uid) && defined($gid))) {
	Warning("MFSFileSystem::changeOwner: invalid input");
	return (0);
    }

    my ($fname, $size, $mode, $fuid, $fgid) = $self->getFile($filename);

    if (!defined($fname)) {
	Warning("MFSFileSystem::changeOwner: no such file $filename");
	return (0);
    }

    return $self->changeAccessPerm($filename, $uid, $gid, $mode);
}


# Sets the VMFS logical name
sub setVMFSLogName($) {
    my $self = shift;
    my ($name) = @_;

    Debug(1, "Setting VMFS name to '$name'");

    if( !c_getHelper('vmfs_set_name') ) {
	Warning("Cannot run vmfs_set_file.");
	return(0);
    }

    my $command = c_getHelper('vmfs_set_name') . ' ' . shell_string($name) . ' ' .
	shell_string($self->getVmkName());

    if( System($command) ) {
        $! = $? >> 8;
	Warning("Could not set the name of the partition.  Reason: $!.");
	return(0);
    }

    $self->resetFSInfo();
    return(1);
}

# Sets the VMFS accessibility
sub setAccessibility($) {
    my $self = shift;
    my ($accessibility) = @_;

    Debug(1, "Setting VMFS accessibility to '$accessibility'");

    if ( !c_getHelper('vmfs_set_accessibility') ) {
        Warning("Cannot run vmfs_set_accessibility.");
        return(0);
    }

    my $command = c_getHelper('vmfs_set_accessibility') . ' ' .
        shell_string($accessibility) . ' ' . shell_string($self->getVmkName());

    if( System($command) ) {
        $! = $? >> 8;
	Warning("Could not set the accessibility of the partition.  Reason: $!.");
	return(0);
    }

    $self->resetFSInfo();
    return(1);
}

# Attaches an extent
sub attachExtent($) {
   my $self = shift;
   my ($partname) = @_;

   Debug(1, "Attaching extent $partname to VMFS " . $self->getVmkName());

   if (!c_getHelper('vmfs_attach_extent') ) {
      Warning("Cannot run vmfs_attach_extent.");
      return(0);
   }

   my $command = c_getHelper('vmfs_attach_extent') . ' ' .
      shell_string($partname) . ' ' . shell_string($self->getVmkName());

   if (System($command)) {
      $! = $? >> 8;
      Warning("Could not attach an extent.  Reason: $!.");
      return(0);
   }

   $self->resetFSInfo();

   # XXX Normally we would have to invalidate the VMFS information since
   #     the extent could also be a VMFS.  However, since the caching policy
   #     for VMFS is "never" then we don't have to invalidate. -jhu

   return(1);
}

# Convert VMFS-1 to VMFS-2
sub convertToVmfs2() {
   my $self = shift;

   if ($self->getVersion() ne 1) {
      Warning("VMFS version is not version 1.  Cannot convert to version 2.");
      return 0;
   }

   if (!c_getHelper('vmfs_convert_vmfs2')) {
      Warning("Cannot run vmfs_convert_vmfs2.");
      return 0;
   }

   my $command = c_getHelper('vmfs_convert_vmfs2') . ' ' .
      shell_string($self->getVmkName());

   if (System($command)) {
      $! = $? >> 8;
      Warning("Could not convert to VMFS2.  Reason: $!.");
      return 0;
   }

   $self->resetFSInfo();
   return 1;
}

#####################################################################
# Methods not intended for use by user                              #
#####################################################################

sub getPathname($) {
   my $self = shift;
   my ($filename) = (@_);

   return $self->getMountPoint() . "/" . $filename;
}

sub loadFSInfo() {
    my $self = shift;

    my @files;

    if( !c_getHelper('vmfs_list_files') ) {
	Warning("Cannot execute vmfs_list_files.");
	return(0);
    }

    my $command = c_getHelper('vmfs_list_files') . " " . shell_string($self->getVmkName());

    local *FD;
    if( !open(FD, "$command |") ){
	# Error this should work since we know that the device is legal
	Warning("Did not get the appropriate output from " . c_getHelper('vmfs_list_files') . ".  Reason: $!");
	return(0);
    }

    # Initilize to empty string
    $self->{LOG_NAME} = '';

    # If not specified, file system is private
    $self->{ACCESSIBILITY} = 'private';

  while(<FD>) {
     chomp();
     
     if( /^Capacity (\d+) \(\d+ file blocks \* (\d+)\), (\d+) \(\d+ blocks\) avail/ ) {
         $self->{CAPACITY} = round($1 / 1024 / 1024);
         $self->{FREE_CAPACITY} = round($3 / 1024 / 1024);
         $self->{BLOCK_SIZE} = round($2 / 1024 / 1024);
      } elsif( /^Name: (.*)\s*\(([^()]*)\)$/ ) {
         # VMFS name and optional accessibility info
         my $name = $1;
         my $accessibility = $2;
  
         $name =~ s/^\s*(.*?)\s*$/$1/;
         $self->{LOG_NAME} = $1;
  
         if ($accessibility ne "") {
             $self->{ACCESSIBILITY} = $accessibility;
         }
      } elsif( /^\s*\(([^()]*)\)$/ ) {
         # Accessibility info for unnamed VMFS
         $self->{ACCESSIBILITY} = $1;
      } elsif ( /^([rwx-]+)\s+(\d+)\s+(\d+)\s+(.*?)(\d+).*?\)\s+(\w.{11})\s+(.+)$/) {
         # Note: The construct .*? above is non-greedy form of .*
         # Matches strings of the form:
         #  rw-r--r--   239   201       disk 4294967296 ( 4096)  Jan 23 14:43 blah.vmdk
         #  rw-r--r--     0     0   redo log          0 (    0)  Jan  1  1947 foo.REDO
         #  r--r--r--     0     0                  1024 (    1)  Jan 31  2038 bar
        
        my ($permissions, $uid, $gid, $attr, $size, $mtimeStr, $file) =
         ($1, $2, $3, $4, int(($5 - 1)/1024/1024) + 1, $6, $7);
  
         # Remove leading and trailing spaces in $attr
         $attr =~ s/^\s*(.*?)\s*$/$1/g;
  
         my @f = ($file, $size, $permissions, $uid, $gid, $attr, $mtimeStr);
  
         push( @files, \@f );
     }
  }
    close(FD);
   
   $self->{FILES} = \@files;

    # Get VMFS information about spanned partitions.
    if (!c_getHelper('vmfs_list_partitions') ) {
       Warning("Cannot execute vmfs_list_partitions.");
       return 0;
    }

    $command = c_getHelper('vmfs_list_partitions') . " " . shell_string($self->getVmkName());

    if (!open(FD, "$command |")){
       # Error this should work since we know that the device is legal
       Warning("Did not get the appropriate output from " . c_getHelper('vmfs_list_partitions') .
               ".  Reason: $!");
       return 0;
    }

    $self->{EXTENTS} = [];

    my $parseExtent = 0;
    while(<FD>) {
       chomp();

       my $partname = $self->getVmkName();
       if (/^$partname is a VMFS-(\d+)\.(\d+) volume spanning/) {
          $self->{VERSION} = $1;
          $self->{MINORVERSION} = $2;
       } elsif (/^UUID.*:\s*(\S+)/) {
	  $self->{UUID} = $1;
       } elsif (/^Physical Extents:/) {
          $parseExtent = 1;
       } elsif ($parseExtent && /($gVmkDeviceType_hba\S+)/) {
          push(@{$self->{EXTENTS}}, $1);
       }
    }
    close(FD);
}

1;

