#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################

#####################################################################
# PCIDeviceMap.pm
# 
# A basic class that contains information representing a PCI device.
# Information includes bus #, device/slot #, function #, name, irq,
# owner.
#####################################################################
package VMware::HConfig::PCIDeviceMap;
use strict;
use VMware::ExtHelpers qw(Warning);

# The field definitions for the 2 tables
my %VendorFields = ( "type",       0,
		     "vendor",     1,
		     "short_name", 2,
		     "long_name",  3
		     );

my %DeviceFields = ( "type",   0,
		     "vendor", 1,
		     "device", 2,
		     "name",   3,
		     "desc",   4,
		     "driver", 5,
		     "constraints", 6  # The only optional field
		     );

#####################################################################
# Public Static Methods (indep. of each object)                     #
#####################################################################

sub newEmpty() {
   my $proto = shift;
   my $class = ref($proto) || $proto;

   my $self = create();
   bless($self, $class);
    
   return $self;
}

sub new(@) {
   my $proto = shift;
   my $class = ref($proto) || $proto;

   my $self = create();
   # Needed to bind the data structure $self to this package so
   # Subroutines will then act as the methods for the data structure.
   bless($self, $class);

   foreach my $dev_file (@_) {
      $self->buildDeviceMap($dev_file);
   }

   return $self;
}

sub create() {
   my $self = {};
    
   $self->{CONFIG} = {};        # Table of configuration variables (hash)
   $self->{VENDORS} = [];       # Table of vendors (array)
   $self->{DEVICES} = [];       # Table of devices (array)

   return $self;
}

# Gets the index number of a named field
sub getVendorField($) {
   my $self = shift;
   my ($field_name) = @_;

   if (exists($VendorFields{$field_name})) {
      return $VendorFields{$field_name};
   } else {
      return -1;
   }
}

# Gets the index number of a named field
sub getDeviceField($) {
   my $self = shift;
   my ($field_name) = @_;

   if (exists($DeviceFields{$field_name})) {
      return $DeviceFields{$field_name};
   } else {
      return -1;
   }
}

sub getConfigKeys() {
   my $self = shift;
   return keys(%{ $self->{CONFIG} });
}

sub getConfigValue($) {
   my $self = shift;
   my ($key) = @_;

   return $self->{CONFIG}->{$key};
}

# Returns all the records for which the vendor and device ids match
sub lookupDevice($$) {
   my $self = shift;
   my ($vend, $dev) = @_;
   my @res;

   if ($vend ne "" && $dev ne "") {
      @res = grep( $_->getVendorID() == $vend && $_->getDeviceID() == $dev, 
                   @{ $self->{DEVICES} } );
   }

   return @res;
}

sub lookupVendor($) {
   my $self = shift;
   my ($vend) = @_;
   my @res;

   if ($vend ne "") {
      @res = grep { $_->[$VendorFields{"vendor"}] == $vend } @{ $self->{VENDORS} };
   }

   if ($#res > -1) {
      return $res[0];
   } else {
      return undef;
   }
}

sub getAllDeviceMap() {
   my $self = shift;
   return $self->{DEVICES};
}

sub toString() {
   my $self = shift;
   my $s = "";

   my $rec;

   $s .= "Config Map\n";
   foreach $rec ($self->getConfigKeys() ) {
      $s .= "$rec='" . $self->getConfigValue($rec) . "'\n";
   }

   $s .= "Vendor Map\n";
   foreach $rec (@{ $self->{VENDORS} } ) {
      my $r;
      foreach $r (@$rec) {
         $s .= "'$r' ";
      }
      $s .= "\n";
   }

   $s .= "Device Map\n";
   foreach $rec (@{ $self->{DEVICES} } ) {
      $s .= $rec->toString() . "\n";
   }

   return $s;
}


#####################################################################
# Methods not intended for use by user                              #
#####################################################################

sub buildDeviceMap($) {
   my($self) = shift;
   my ($map) = @_;
   my $lineno = 0;         # Number of lines 
   local *FD;

   my @arr = keys(%VendorFields);
   my $maxvend = $#arr;
   @arr = keys(%DeviceFields);
   my $maxdev = $#arr;

   my $DevMap = {};

   if (! -r $map) {
      # Error file cannot be read
      Warning("Error: Could not read file $map.  Reason: $!");
      return $DevMap;
   }

   if (!open(FD, "< $map")) {
      # Error could not open file
      Warning("Error: Could not open file $map.  Reason: $!");
      return $DevMap;
   }

   while (<FD>) {
      chomp();
      $lineno++;

      # Comments are everything after the occurrence of a # character
      s/\#.*//;

      # Strip off some trailing characters that give us problems
      s/[\s\r]*$//;

      my @fields = split(/,/);

      if ($#fields == -1) {
         next;
      }

      if ($fields[0] eq "config") {
         if ($#fields == 2) {
            my $key = $fields[1];
            my $value = $fields[2];

            # Untaint the values
            if ($key =~ /(.*)/) {
               $key = $1;
            }

            if ($value =~ /(.*)/) {
               $value = $1;
            }
            # Store the configuration variable into the mapping.
            $self->{CONFIG}->{$key} = $value;
         } else {
            # Error
            Warning("Warning: Wrong number of fields for config table on line $lineno.  Skipping.");
         }
      } elsif ($fields[0] eq "vendor") {
         if ($#fields == $maxvend) {
            if ($fields[$VendorFields{"vendor"}] =~ /0x([0-9a-f]+)/i) {
               $fields[$VendorFields{"vendor"}] = hex($1);
            }
		
            # Untaint the values
            my $i;
            for ($i = 0; $i <= $#fields; $i++) {
               if ($fields[$i] =~ /(.*)/) {
                  $fields[$i] = $1;
               }
            }

            push( @{$self->{VENDORS}}, \@fields );
         } else {
            # Error
            Warning("Warning: Wrong number of fields for vendor table on line $lineno.  Skipping.");
         }
      } elsif ($fields[0] eq "device") {
         if ($#fields >= $maxdev - 1) {
            if ($fields[$DeviceFields{"vendor"}] =~ /0x([0-9a-f]+)/i) {
               $fields[$DeviceFields{"vendor"}] = hex($1);
            }
		
            if ($fields[$DeviceFields{"device"}] =~ /0x([0-9a-f]+)/i) {
               $fields[$DeviceFields{"device"}] = hex($1);
            }

            # Untaint the values
            my $i;
            for ($i = 0; $i <= $#fields; $i++) {
               if ($fields[$i] =~ /(.*)/) {
                  $fields[$i] = $1;
               }
            }

            # The device array is an array of strings
            my $entry = VMware::HConfig::PCIDeviceMapEntry->new(@fields);
            push( @{$self->{DEVICES}}, $entry);
         } else {
            # Error
            Warning("Warning: Wrong number of fields for device table on line $lineno.  Skipping.");
         }
      }
      else {
         # Error
         Warning("Warning: Unknown database type encountered on line $lineno.  Skipping.");
      }
   }
   close(FD);
}


#####################################################################
# PCIDeviceMapEntry.pm
# 
# A basic class that contains information representing a PCI device.
# Information includes bus #, device/slot #, function #, name, irq,
# owner.
#####################################################################
package VMware::HConfig::PCIDeviceMapEntry;
use strict;
use VMware::ExtHelpers qw(Warning);

sub new(@) {
   my $proto = shift;
   my $class = ref($proto) || $proto;
   my(@fields) = @_;
   my $self = {};
   $self->{TYPE} = shift;
   $self->{VENDOR} = shift;
   $self->{DEVICE} = shift;
   $self->{NAME} = shift;
   $self->{DESC} = shift;
   $self->{DRIVER} = shift;
   $self->{CONSTRAINTS} = shift;
   bless($self, $class);
   return $self;
}

sub getType() {
   my $self = shift;
   return $self->{TYPE};
}

sub getVendorID() {
   my $self = shift;
   return $self->{VENDOR};
}

sub getDeviceID() {
   my $self = shift;
   return $self->{DEVICE};
}

sub getName() {
   my $self = shift;
   return $self->{NAME};
}

sub getDesc() {
   my $self = shift;
   return $self->{DESC};
}

sub getDriver() {
   my $self = shift;
   return $self->{DRIVER};
}

sub getConstraints() {
   my $self = shift;
   my %constraints;

   if (!$self->{CONSTRAINTS}) {
      return %constraints;
   }

   my @constraintStrings = split(/;/, $self->{CONSTRAINTS});
   my $str;
   foreach $str (@constraintStrings) {
      if ($str =~ /(\S+)=(\S+)/) {
         my $constraint_param = $1;
         my $constraint_value = $2;
         if ($constraint_value =~ /^0x([0-9a-f]+)$/i) {
            $constraint_value = hex($1);
         }
         $constraints{$constraint_param} = $constraint_value;
      } else {
         Warning('Malformed constraint ' . $str);
      }
   }

   return %constraints;
}

sub toString() {
   my $self = shift;
   my $s = "";

   $s .= "'" . $self->{TYPE} . "',";
   $s .= "'" . sprintf("0x%04x", $self->{VENDOR}) . "',";
   $s .= "'" . sprintf("0x%04x", $self->{DEVICE}) . "',";
   $s .= "'" . $self->{NAME} . "',";
   $s .= "'" . $self->{DESC} . "',";
   $s .= "'" . $self->{DRIVER} . "',";

   my $c;
   my %constraints = $self->getConstraints();
   foreach $c (keys(%constraints)) {
      my $v = $constraints{$c};
      $s .= "'$c => " . sprintf("0x%04x", $v) . "' ";
   }

   return $s;
}

1;
