#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# RawDisk.pm
#
# A class that is used to gather and manipulate information about the
# raw disks on the computer.
#
package VMware::HConfig::RawDisk;
use strict;
use VMware::HConfig::Constant qw(&c_getFile &c_getDir &c_getHelper $gVmkDumpPartType $gVMFSPartType
 				 &CACHE_NEVER &CACHE_ONCE &CACHE_CALLBACK $gRawDisk_CacheDefaults
				 $gVmkDeviceType_nic $gVmkDeviceType_hba $gVMFSBlockSizeLimit
                                 $gVMFS2BlockSizeLimit @gMultipathPolicies);
use VMware::ExtHelpers qw(&Warning &System &internal_basename &Debug &Panic $gDEBUG &shell_string);
use VMware::ConfigDB;

my $PartitionIDMap = {
    0x00 => "Empty",
    0x01 => "DOS 12-bit FAT",
    0x02 => "XENIX root",
    0x03 => "XENIX usr",
    0x04 => "DOS 16-bit <32M",
    0x05 => "Extended",
    0x06 => "DOS 16-bit >=32M",
    0x07 => "HPFS/NTFS",
    0x08 => "AIX",
    0x09 => "AIX bootable",
    0x0a => "OS/2 Boot Manag",
    0x0b => "Win95 FAT32",
    0x0c => "Win95 FAT32 LBA",
    0x0e => "Win95 FAT16 LBA",
    0x0f => "Win95 Extended",
    0x10 => "OPUS",
    0x11 => "Hidden FAT12",
    0x12 => "Compaq Diagnostic",
    0x3c => "PartitionMagic",
    0x40 => "Venix 80286",
    0x42 => "SFS",
    0x4d => "QNX4.x",
    0x4e => "QNX4.x 2nd part",
    0x4f => "QNX4.x 3rd part",
    0x51 => "Novell?",
    0x52 => "Microport",
    0x63 => "GNU HURD",
    0x64 => "Novell Netware",
    0x65 => "Novell Netware",
    0x75 => "PC/IX",
    0x80 => "Old MINIX",
    0x81 => "Linux/MINIX",
    0x82 => "Linux swap",
    0x83 => "Linux native",
    0x84 => "OS/2 hidden C:",
    0x85 => "Linux extended",
    0x86 => "NTFS volume set",
    0x87 => "NTFS volume set",
    0x93 => "Amoeba",
    0x94 => "Amoeba BBT",
    0xa0 => "IBM Thinkpad",
    0xa5 => "BSD/386",
    0xa6 => "OpenBSD",
    0xa7 => "NeXTSTEP",
    0xb7 => "BSDI fs",
    0xb8 => "BSDI swap",
    0xc7 => "Syrinx",
    0xdb => "CP/M",
    0xe1 => "DOS access",
    0xe3 => "DOS R/O",
    0xeb => "BeOS fs",
    0xf2 => "DOS secondary",
    0xfd => "Linux raid auto",
    $gVmkDumpPartType => "VMware Core Dump",
    $gVMFSPartType => "VMFS",
    0xfe => "Old VMware Core Dump",
    0xff => "Old VMFS",
		   };

# The min/max partition numbers that are allowed.
my $MIN_PARTITION = 1;
my $MAX_PARTITION = 16;

#####################################################################
# Public Static Methods (indep. of each object)                     #
#####################################################################

sub new($) {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my($vmklogicName) = @_;
    my $self = $proto->create($vmklogicName);

    $self->{CACHE_OPTIONS} = $gRawDisk_CacheDefaults;

    bless($self, $class);

    return($self);
}


sub create($) {
    my $proto = shift;
    my $self = {};
    my($vmkLogicName) = @_;

    $self->{VMK_LOGIC_NAME} = $vmkLogicName;  # Logical name of the disk as
                                              # seen by the vmkernel. ie. scsi0:4
    $self->{VMNIX_LOGIC_NAME} = "";   # The logical name of the disk as seen by Linux. ie. scsi0

    my($host, $target, $lun);
    if( isLogVmkDiskName($vmkLogicName) ) {
	($host, $target, $lun) = unmakeLogVmkDiskName($vmkLogicName);
    } else {
	Warning("Invalid Vmk name '$vmkLogicName' specified in RawDisk constructor.");
    }

    $self->{VMK_LOGIC_NAME_HOST} = $host;     # A copy of the vmkernel host adapter information
    $self->{VMNIX_LOGIC_NAME_HOST} = "";      # A copy of the vmnix host adapter information
    $self->{COMMON_TARGET} = $target; # The target id is common between the vmnix and vmkernel
    $self->{COMMON_LUN} = $lun;       # The lun is common between the vmnix and vmkernel

    $self->{VMNIX_DEV_NAME} = "";     # The VMnix name.  ie. /dev/sda
    $self->{VMNIX_PARTITION_NAME_SEPARTOR} = undef;  # Separator between the disk name and the
                                                     # partition name.  Must be defined at the
                                                     # same time as VMNIX_DEV_NAME
    $self->{GEOMETRY} = undef;        # Geometry is undefined initially
    $self->{PARTITION_TABLE} = undef; # Partition table of the disk

    $self->{MULTIPATH_INFO} = undef;

    return($self);
}

sub lookupPartType($) {
    my($id) = @_;

    my $typeName = $PartitionIDMap->{$id};
    if( defined($typeName) && $typeName ne "" ) {
	return($PartitionIDMap->{$id});
    } else {
	return("Unknown");
    }
}

sub isPartExtended($) {
    my($id) = @_;

    return($id == 0x05 || $id == 0x0f || $id == 0x85 || $id == 0x0f);
}

sub isPartVmware($) {
    my($id) = @_;

    return($id == $gVMFSPartType || $id == 0xff);
}

sub isPartVmkDump($) {
    my($id) = @_;
    return($id == $gVmkDumpPartType || $id == 0xfe);
}

sub printMap() {
    my $d;
    my @keys;
    @keys = sort { $a <=> $b } keys(%$PartitionIDMap);
    foreach $d (@keys) {
	print sprintf("0x%02x", $d) . ": " . $PartitionIDMap->{$d} . "\n";
    }
}

sub isDevName($) {
    my($name) = @_;

    return( $name =~ /^\/dev\/\S+/ );
}

# Determines if the cylinder, head, and sector information is valid
sub isValidGeometry($$$) {
    my($geo_cyl, $geo_head, $geo_sect) = @_;
    return( $geo_cyl > 0 && $geo_head > 0 && $geo_sect > 0 );
}

# Test if partition number corresponds to a physical partition
sub isPhysicalPart($) {
    my($partnum) = @_;
    return( $partnum > 0 && $partnum < 5 );
}

# Test if partition number corresponds to a logical partition
sub isLogicalPart($) {
    my($partnum) = @_;
    return( $partnum > 4 );
}

# Translates a CHS address into a LBA address
# Arguments:
#   cyl: cylinder of CHS address (ranges from 0 -> )
#   head: head of CHS address (ranges from 0 -> 254 (max))
#   sect: sectors of CHS address (ranges from 1 -> 63 (max))
#   geo_cyl: parameter indicating number of cylinders on disk
#   geo_head: parameter indicating number of heads on disk
#   geo_sect: parameter indicating number of sectors on disk
sub fromChsToLba($$$$$$) {
    my($cyl, $head, $sect, $geo_cyl, $geo_head, $geo_sect) = @_;

    if( $geo_cyl < 1 || $geo_head < 1 || $geo_sect < 1 ) {
	Warning("Translating from CHS to LBA with invalid geometry info: ($geo_cyl,$geo_head,$geo_sect)");
	return(0);
    }
    if( $cyl < 0 || $cyl >= $geo_cyl ) {
	Warning("Translating from CHS to LBA with a cylinder $cyl not in range [0," . ($geo_cyl - 1) . "]");
	return(0);
    }
    if( $head < 0 || $head >= $geo_head ) {
	Warning("Translating from CHS to LBA with a head $head not in range [0," . ($geo_head - 1) . "]");
	return(0);
    }
    if( $sect < 1 || $sect > $geo_sect ) {
	Warning("Translating from CHS to LBA with a sector $sect not in range [1," . $geo_sect . "]");
	return(0);
    }

    my $lba = ((($cyl * $geo_head) + $head) * $geo_sect) + $sect - 1;
    return($lba);
}


# Translates a LBA address into a CHS address for the given disk.
# Returns the CHS information as an array given the LBA address.
# Arguments:
#   lba: the logical block address to convert into a CHS address
#   geo_cyl: parameter indicating number of cylinders on disk
#   geo_head: parameter indicating number of heads on disk
#   geo_sect: parameter indicating number of sectors on disk
sub fromLbaToChs($$$$) {
    my($lba, $geo_cyl, $geo_head, $geo_sect) = @_;
    my($cyl, $head, $sect);
    my $mod;

    if( $geo_cyl < 1 || $geo_head < 1 || $geo_sect < 1 ) {
	Warning("Translating from LBA to CHS with invalid geometry info: ($geo_cyl,$geo_head,$geo_sect)");
	return(0);
    }

    # Perform modular operations to make sure that we get integral values
    $mod = $lba % ($geo_head * $geo_sect);
    $cyl = ($lba - $mod) / ($geo_head * $geo_sect);

    $lba = $mod;
    $mod = $lba % $geo_sect;
    $head = ($lba - $mod) / $geo_sect;

    $sect = $mod + 1;

    if( $cyl < 0 || $cyl >= $geo_cyl ) {
	Warning("Translating from LBA to CHS resulted in a cylinder $cyl not in range [0," . ($geo_cyl - 1) . "]");
    }
    if( $head < 0 || $head >= $geo_head ) {
	Warning("Translating from LBA to CHS resulted in a head $head not in range [0," . ($geo_head - 1) . "]");
    }
    if( $sect < 1 || $sect > $geo_sect ) {
	Warning("Translating from LBA to CHS resulted in a sector $sect not in range [1," . $geo_sect . "]");
    }
    return($cyl, $head, $sect);
}

# Returns true if the name is in "logical" vmnix SCSI format
sub isLogVmnixDiskName($) {
    my($name) = @_;
    if( $name =~ /^(\S+[0-9]+):([0-9]+):([0-9]+)$/ ) {
	if( VMware::HConfig::SCSIAdapter::isLogVmnixSCSIName($1) ) {
	    return(1);
	}
    }
    return(0);
}

# Returns true if the name is in "logical" vmkernel SCSI format
sub isLogVmkDiskName($) {
    my($name) = @_;
    if( $name =~ /^($gVmkDeviceType_hba[0-9]+):([0-9]+):([0-9]+)$/ ) {
	return(1);
    }
    return(0);
}

sub makeLogVmkDiskName($$$) {
    my($adapterName, $target, $lun) = @_;
    if( $adapterName =~ /^$gVmkDeviceType_hba[0-9]+$/ && $target =~ /^[0-9]+$/ && $lun =~ /^[0-9]+$/ )  {
	return($adapterName . ':' . $target . ':' . $lun);
    }
    Panic("Could not makeLogVmkDiskName: $adapterName, $target, $lun");
    return(undef);
}

sub unmakeLogVmkDiskName($) {
    my($name) = @_;
    if( $name =~ /^($gVmkDeviceType_hba[0-9]+):([0-9]+):([0-9]+)$/ ) {
	my($adapterName, $target, $lun) = ($1, $2, $3);
	return($adapterName, $target, $lun);
    }
    Panic("Could not unmakeLogVmkDiskName: $name");
    return(undef);
}

sub isLogVmkPartName($) {
    my($name) = @_;
    if( $name =~ /^$gVmkDeviceType_hba[0-9]+:[0-9]+:[0-9]+:[0-9]+$/ ) {
	return(1);
    }
    return(0);
}

sub makeLogVmkPartName($$) {
    my($diskName, $partnum) = @_;
    if( !isLogVmkDiskName($diskName) ) {
	# XXX Should be undef -jhu
	return(0);
    }
    if( $partnum =~ /^[0-9]+$/ )  {
	return($diskName . ':' . $partnum);
    }
    Panic("Could not makeLogVmkPartName: $diskName, $partnum");
    return(undef);
}

sub unmakeLogVmkPartName($) {
    my($name) = @_;
    if( $name =~ /^($gVmkDeviceType_hba[0-9]+:[0-9]+:[0-9]+):([0-9]+)$/ ) {
	my($diskName, $partnum) = ($1, $2);
	return($diskName, $partnum);
    }
    Panic("Could not unmakeLogVmkPartName: $name");
    return(undef);
}

sub makeVmkVmfsFileName($$) {
    my($partName, $fileName) = @_;
    if( !isLogVmkPartName($partName) ) {
	Panic("Could not makeVmkVmfsFileName: $partName, $fileName");
	return(undef);
    }
    if( $fileName !~ /:/ ) {
	return($partName . ':' . $fileName);
    }
    Panic("Could not makeVmkVmfsFileName: $partName, $fileName");
    return(undef);
}

sub unmakeVmkVmfsFileName($) {
    my($name) = @_;
    if( $name =~ /^($gVmkDeviceType_hba[0-9]+:[0-9]+:[0-9]+:[0-9]+):([^:]+)$/ ) {
	my($partName, $fileName) = ($1, $2);
	return($partName, $fileName);
    }
    Panic("Could not unmakeVmkVmfsFileName: $name");
    return(undef);
}

sub isVmkVmfsFileName($) {
    my($name) = @_;
    if( $name =~ /^$gVmkDeviceType_hba[0-9]+:[0-9]+:[0-9]+:[0-9]+:[^:]+$/ ) {
	return(1);
    }
    return(0);
}

#sub makeLogVmfsFileName($$) {
#    my($partName, $fileName) = @_;
#    if( !isLogVmkPartName($partName) ) {
#	Panic("Could not makeVmkVmfsFileName: $partName, $fileName");
#	return(undef);
#    }
#    if( $fileName !~ /:/ ) {
#	return($partName . ':' . $fileName);
#    }
#    Panic("Could not makeVmkVmfsFileName: $partName, $fileName");
#    return(undef);
#}
#
#sub unmakeVmkVmfsFileName($) {
#    my($name) = @_;
#    if( $name =~ /^($gVmkDeviceType_hba[0-9]+:[0-9]+:[0-9]+:[0-9]+):([^:]+)$/ ) {
#	my($partName, $fileName) = ($1, $2);
#	return($partName, $fileName);
#    }
#    Panic("Could not unmakeVmkVmfsFileName: $name");
#    return(undef);
#}
#
#sub isVmkVmfsFileName($) {
#    my($name) = @_;
#    if( $name =~ /^$gVmkDeviceType_hba[0-9]+:[0-9]+:[0-9]+:[0-9]+:[^:]+$/ ) {
#	return(1);
#    }
#    return(0);
#}



#####################################################################
# Public Methods                                                    #
#####################################################################

# Accessor functions
sub getVmkName() {
    my($self) = shift;

    return($self->{VMK_LOGIC_NAME});
}

sub getVmnixDevName() {
    my($self) = shift;
    return($self->{VMNIX_DEV_NAME});
}

sub getVmnixPartSeparator() {
    my($self) = shift;
    return($self->{VMNIX_PARTITION_NAME_SEPARATOR});
}

sub getVmkAdapterName() {
    my ($self) = shift;
    return($self->{VMK_LOGIC_NAME_HOST});
}

sub getVmnixAdapterName() {
    my ($self) = shift;
    return($self->{VMNIX_LOGIC_NAME_HOST});
}

sub getTarget() {
    my ($self) = shift;
    return($self->{COMMON_TARGET});
}

sub getLun() {
    my ($self) = shift;
    return($self->{COMMON_LUN});
}

# Get the geometry of the disk.  Returns an array of 3 elements
# corresponding to the number of cylinders, heads, and sectors.
#
# Note:
# Do not use the geometry functions until all the names for the
# object are set.  Otherwise, may fail to get the geometry info.
sub getGeometry() {
    my($self) = shift;
    my($geo_cyl, $geo_head, $geo_sect);

    if( !defined($self->{GEOMETRY}) || $self->isCachedNever('GEOMETRY') ) {
	($geo_cyl, $geo_head, $geo_sect) = findGeometry($self->getVmnixDevName());

	if( !isValidGeometry($geo_cyl, $geo_head, $geo_sect) ) {
	    Warning("Geometry information for disk " . $self->getVmkName() . " (vmk) " .
		    $self->getVmnixDevName() . " (dev) is not valid: (" . $geo_cyl . "," .
		    $geo_head . "," . $geo_sect . ")");
	    $self->setGeometry(0, 0, 0);
	} else {
	    # Only save the information into the cache if we have valid geometry info
	    $self->setGeometry($geo_cyl, $geo_head, $geo_sect);
	}
    }

    return($self->{GEOMETRY});
}

sub getGeoCylinder() {
    my($self) = shift;
    return($self->getGeometry()->{CYLINDER});
}

sub getGeoHead() {
    my($self) = shift;
    return($self->getGeometry()->{HEAD});
}

sub getGeoSector() {
    my($self) = shift;
    return($self->getGeometry()->{SECTOR});
}

sub getTotalSectors() {
    my($self) = shift;
    my $size = $self->getGeoCylinder() * $self->getGeoHead() * $self->getGeoSector();
    return($size);
}

sub getPartitionTable() {
    my($self) = shift;
    if( !defined($self->{PARTITION_TABLE}) || $self->isCachedNever('PARTITION_TABLE') ) {
	if( !$self->loadPartitionTable() ) {
	    return(undef);
	}
    }
    return($self->{PARTITION_TABLE});
}

# Translates a CHS address into a LBA address for the given disk
# cylinders are from 0 ->
# heads are from 0 -> 254 (max)
# sectors are from 1 -> 63 (max)
sub getChsToLba($$$) {
    my($self) = shift;
    my($cyl, $head, $sect) = @_;
    my($geo_cyl, $geo_head, $geo_sect) = ($self->getGeoCylinder(), $self->getGeoHead(), $self->getGeoSector());

    return( fromChsToLba($cyl, $head, $sect, $geo_cyl, $geo_head, $geo_sect) );
}

# Translates a LBA address into a CHS address for the given disk.
# Returns the CHS information as an array given the LBA address.
sub getLbaToChs($) {
    my($self) = shift;
    my($lba) = @_;
    my($geo_cyl, $geo_head, $geo_sect) = ($self->getGeoCylinder(), $self->getGeoHead(), $self->getGeoSector());

    return( fromLbaToChs($lba, $geo_cyl, $geo_head, $geo_sect) );
}

# Takes a number of kilobytes and converts it to the minimal number of
# cylinders needed to make up that much space.
sub getKBytesToCyl($) {
    my($self) = shift;
    my($kbytes) = @_;
    my $sectPerCyl = $self->getChsToLba(1, 0, 1);
    my $kbytesPerCyl = ($sectPerCyl - ($sectPerCyl % 2)) / 2;
    my $numCyls = ($kbytes - ($kbytes % $kbytesPerCyl)) / $kbytesPerCyl;
    if( $kbytes % $kbytesPerCyl > 0 ) {
	# Round up
	$numCyls++;
    }
    return($numCyls);
}

# Takes a number of cylinders and converts it to the number of KBytes
sub getCylToKBytes($) {
    my($self) = shift;
    my($numCyls) = @_;
    my $sectPerCyl = $self->getChsToLba(1, 0, 1);
    my $kbytesPerCyl = ($sectPerCyl - ($sectPerCyl % 2)) / 2;
    my $kbytes = $numCyls * $kbytesPerCyl;
    return($kbytes);
}

# Determines if the disk has a vmnix name
sub hasVmnixDevName() {
    my($self) = shift;
    my $vmnixDevName = $self->getVmnixDevName();
    my $vmnixPartSep = $self->getVmnixPartSeparator();
    return( defined($vmnixDevName) &&
	    $vmnixDevName =~ /^\/dev/ &&
	    defined($vmnixPartSep) );
}

# Gets the partition numbers
sub getPartNums() {
    my($self) = shift;

    if( !defined($self->{PARTITION_TABLE}) || $self->isCachedNever('PARTITION_TABLE') ) {
	if( !$self->loadPartitionTable() ) {
	    return();
	}
    }
    my @parts = $self->{PARTITION_TABLE}->getPartNums();
    return(@parts);
}

sub getPart($) {
    my($self) = shift;
    my($partnum) = @_;

    if( !defined($self->{PARTITION_TABLE}) || $self->isCachedNever('PARTITION_TABLE') ) {
	if( !$self->loadPartitionTable() ) {
	    return(undef);
	}
    }

    return( $self->{PARTITION_TABLE}->getPart($partnum) );
}

# Gets the device name of the Vmk partition
sub getPartBestDevName($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	if( $self->hasVmnixDevName() ) {
	    my $diskDevName = $self->getVmnixDevName();
	    my $partDevName = $diskDevName . $self->getVmnixPartSeparator() . $partnum;
	    return($partDevName);
	}

	my $name = lookupVsdDevName($self->getVmkName(), $partnum);
	return($name);
    }

    return("");
}

sub getPartStartCyl($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{START_CYL});
    }

    return(-1);
}


sub getPartEndCyl($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{END_CYL});
    }

    return(-1);
}


sub getPartType($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{TYPE});
    }

    return(-1);
}

sub getPartSectors($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{SECTORS});
    }

    return(-1);
}

sub getPartTypeName($) {
    my($self) = shift;
    my($partnum) = @_;

    my $type = $self->getPartType($partnum);
    if( defined($type) && defined($PartitionIDMap->{$type}) ) {
	return($PartitionIDMap->{$type});
    }

    return("Unknown");
}

sub setVmnixDevName($) {
    my($self) = shift;
    $self->{VMNIX_DEV_NAME} = shift;
}

sub setVmnixPartSeparator($) {
    my($self) = shift;
    $self->{VMNIX_PARTITION_NAME_SEPARATOR} = shift;
}

# Creates a partition table object.  If the geometry information for the disk
# is invalid, undef will be returned in place of the partition table.
sub createPartitionTable() {
    my($self) = shift;
    my $geo_cyl = $self->getGeoCylinder();
    my $geo_head = $self->getGeoHead();
    my $geo_sect = $self->getGeoSector();
    my $partTable = undef;
    if( isValidGeometry($geo_cyl, $geo_head, $geo_sect) ) {
	$partTable = VMware::HConfig::PartitionTable->new($self->getVmkName(),
							  $geo_cyl,
							  $geo_head,
							  $geo_sect);
    } else {
	Warning("Cannot create partition table for disk " . $self->getVmkName() .
		" because geometry info is invalid.");
    }
    return($partTable);
}

# Loads the partition table
sub loadPartitionTable() {
    my($self) = shift;
    my $partTable = $self->createPartitionTable();
    if( defined($partTable) ) {
	$self->{PARTITION_TABLE} = $self->findPartitionTable($partTable, $self->getVmkName(),
							     $self->getVmnixDevName());
    } else {
	Warning("Could not load partition table.");
	Warning("Please Rescan.");
	return 0;
    }
    return 1;
}

sub toString() {
    my($self) = shift;

    my $s = "";

    my $vmkName = $self->getVmkName();
    my $devName = $self->getVmnixDevName();
    my $geo_cyl = $self->getGeoCylinder();
    my $geo_head = $self->getGeoHead();
    my $geo_sect = $self->getGeoSector();
    my $total_sec = $self->getTotalSectors();

    $s .=
	"Disk " . $self->getVmkName() . " (vmkernel) " .
	"$devName (dev) " .
	"(cyl,head,sect)=($geo_cyl,$geo_head,$geo_sect) $total_sec total sectors " .
        "has the following partitions:\n";

    my $partTable = $self->getPartitionTable();
    if( defined($partTable) ) {
	# Special case.  In the case the disk is shared, we want to know the partition device names too although it's unnecessary
	if( $self->hasVmnixDevName() ) {
	    my $i;
	    foreach $i ($self->getPartNums()) {
		my $partDevName = $self->getPartBestDevName($i);
		my $type = sprintf("%02x", $self->getPartType($i));
		my $typeName = $self->getPartTypeName($i);
		my $start_cyl = $self->getPartStartCyl($i);
		my $end_cyl = $self->getPartEndCyl($i);
		my $sect = $self->getPartSectors($i);

		$s .= "  $vmkName (vmnix $partDevName):$i type=$typeName ($type) start=$start_cyl end=$end_cyl sect=$sect\n";
	    }
	} else {
	    $s .= $partTable->toString();
	}
    } else {
	$s .= "No partitions found.\n";
    }

    my $policy = $self->getMultipathPolicy();
    my @paths = $self->getMultipathPaths();
    my $active = $self->getMultipathActive();
    my $prefer = $self->getMultipathPrefer();

    $s .= "Multipath info: policy=$policy active=$active prefer=$prefer\n";
    foreach my $p (@paths) {
       $s .= "  " . $p->{PATH} . ": " . $p->{STATUS} . "\n";
    }

    return($s);
}

#####################################################################
# Trigger a LUN IO
#
# This function calls an fdisk on the device specified to cause an IO
#####################################################################
sub triggerLunIO() {
   my $self = shift;
   my $devName = $self->{VMNIX_DEV_NAME};

   if (!c_getHelper('lun_IO')) {
      Warning("Cannot do a lun IO\n");
      return 0;
   }
   my $cmd = c_getHelper('lun_IO') . " -l $devName";
   if (System($cmd)) {
      Warning("Failed to execute $cmd. Reason: $!.");
      return 0;
   }
   return 1;
}


#####################################################################
# Multipathing functions
#####################################################################
sub getMultipathPolicy() {
   my $self = shift;
   return $self->getMultipathInfo()->{POLICY};
}

sub getMultipathPaths() {
   my $self = shift;
   return @{$self->getMultipathInfo()->{PATHS}};
}

sub getMultipathActive() {
   my $self = shift;
   return $self->getMultipathInfo()->{ACTIVE};
}

sub getMultipathPrefer() {
   my $self = shift;
   return $self->getMultipathInfo()->{PREFER};
}

# Loads the multipathing information
sub getMultipathInfo() {
   my($self) = shift;

   if( !defined($self->{MULTIPATH_INFO}) || $self->isCachedNever('MULTIPATH_INFO') ) {
      $self->{MULTIPATH_INFO} = $self->findMultipathInfo();
   }

   return $self->{MULTIPATH_INFO};
}

sub setMultipathPolicy($) {
   my $self = shift;
   my $policy = shift;

   # Check that policy is valid
   if (!grep($_ eq $policy, @gMultipathPolicies)) {
      Warning("Invalid policy '$policy' specified.");
      return 0;
   }

   my $rc = $self->setMultipathInternal(undef, "policy $policy");
   setConfigDB("multipath." . $self->getVmkName() . ".policy", $policy);
   return $rc;
}

sub setMultipathPrefer($) {
   my $self = shift;
   my $path = shift;

   my $rc = $self->setMultipathInternal($path, "preferred $path");
   setConfigDB("multipath." . $self->getVmkName() . ".prefer", $path);
   return $rc;
}

sub setMultipathActive($) {
   my $self = shift;
   my $path = shift;

   my $rc = $self->setMultipathInternal($path, "active $path");
   setConfigDB("multipath." . $self->getVmkName() . ".active", $path);
   return $rc;
}

sub enableMultipathPath($) {
   my $self = shift;
   my $path = shift;

   my $rc = $self->setMultipathInternal($path, "pathon $path");

   my $key = "multipath." . $self->getVmkName() . ".disable";
   my $disable = getConfigDB($key);

   # Take the path to enable out of the disable list.
   my @lst = defined($disable) ? split(/[, ]+/, $disable) : ();
   my @newlst = grep($_ ne $path, @lst);
   $disable = join(', ', @newlst);

   setConfigDB("multipath." . $self->getVmkName() . ".disable", $disable);
   return $rc;
}

sub disableMultipathPath($) {
   my $self = shift;
   my $path = shift;

   my $rc = $self->setMultipathInternal($path, "pathoff $path");

   my $key = "multipath." . $self->getVmkName() . ".disable";
   my $disable = getConfigDB($key);

   # Add the path to disable list only if it isn't already there.
   my @lst = defined($disable) ? split(/[, ]+/, $disable) : ();
   if (!grep($_ eq $path, @lst)) {
      push(@lst, $path);
   }
   $disable = join(', ', @lst);

   setConfigDB("multipath." . $self->getVmkName() . ".disable", $disable);
   return $rc;
}

#####################################################################
# Caching Support Functions                                         #
#####################################################################

# Get all the caching options
sub getCacheOptions() {
    my($self) = shift;
    return($self->{CACHE_OPTIONS});
}

# Get a single caching option by name
sub getCacheOption($) {
    my($self) = shift;
    my($key) = @_;
    if( defined($self->getCacheOptions()->{$key}) ) {
	return($self->getCacheOptions()->{$key});
    }
    Warning("No cache options set for item $key.  Using CACHE_NEVER as default caching policy");
    return(CACHE_NEVER);
}

# Helper function that queries if the option is cached once
sub isCachedNever($) {
    my($self) = shift;
    my($key) = @_;
    return( $self->getCacheOption($key) == CACHE_NEVER );
}

#####################################################################
# End of Caching Support Functions                                  #
#####################################################################

#####################################################################
# Disk Modification Methods.   These methods modify a disk.         #
#####################################################################

# Warning, creates a partition.  Use with caution.
# This function is deprecated.  Use makePartitions instead.
sub createPartition($$$) {
    my $self = shift;
    my ($start_cyl, $num_cyl, $part_id) = @_;

    my ($devName) = $self->getVmnixDevName();

    if( !isDevName($devName) ) {
	# Error
	Warning("Invalid device name '$devName'.  Should be begin with /dev/.");
	return(0);
    }

    if( $part_id < 0 || $part_id > 255 ) {
	# Error
	Warning("Invalid partition type '$part_id'.  Should be between 0 and 255.");
	return(0);
    }

    $part_id = sprintf("%02x", $part_id);

    if( $devName eq "" ) {
	# Error
	Warning("Attempting to delete all partitions with null device name.");
	return(0);
    }

    # Should put some checks in here later

    if( ! c_getHelper('partition_create') ) {
	# Error could not run the partitioning tool
	Warning("Cannot run the partitioning tool. Reason: $!");
	return(0);
    }

    if( s_system(c_getHelper('partition_create') . ' ' . shell_string($devName),
		 "$start_cyl $num_cyl $part_id\n") ) {
	# Error could not run the partitioning tool
	Warning("Could not create partitions.  Reason: $!");
	return(0);
    }

    $self->loadPartitionTable();

    # When partitions change, the vmkernel needs to rescan VMFS.
    rescanVMFS();

    return(1);
}


# Verify the following information about the partitions:
# 1. no more than 4 primary partitions are created
# 2. no more than 1 extended partition is created in one of the primary partitions
# 3. no more than MAX_PARTITIONS partitions are created
# 4. validity of the partition types: must be in range of 0-255
# 5. cylinders are in range of disk geometry
# 6. cylinders do not intersect
# 7. logical partitions must in order that they use cylinders
# .. and many more.
sub checkPartInfo($) {
    my $self = shift;
    my ($partTable) = @_;
    my $geo_cyl = $self->getGeoCylinder();
    my $geo_head = $self->getGeoHead();
    my $geo_sect = $self->getGeoSector();

    Debug(1, "Entered with argument '" . $partTable->toString() . "'");

    # Make sure we have valid geometry information for the disk
    if( !isValidGeometry($geo_cyl, $geo_head, $geo_sect) ) {
	Warning("Disk " . $self->getVmkName() . " (vmk) has invalid geometry information.");
	return(0);
    }

    # Make sure the geometry information is the same as that of the partition table.
    if( $geo_cyl != $partTable->getGeoCylinder() ||
	$geo_head != $partTable->getGeoHead() ||
	$geo_sect != $partTable->getGeoSector() ) {
	Warnign("Disk geometry (" . $geo_cyl . "," , $geo_head . "," . $geo_sect .
		") and partition table geometry (" . $partTable->getGeoCylinder() .
		"," . $partTable->getGeoHead() . "," . $partTable->getSector() .
		") is not the same.");
	return(0);
    }

    return($partTable->checkPartInfo());
}


# Warning, creates a list of partitions.  Use with caution.  Format
# for function, an array of hashes where each array has the keys:
# NUM, START_CYL, END_CYL, and TYPE
#
# Returns 0 on failure, 1 on success.
sub makePartitions($) {
    my $self = shift;
    my($partTable) = @_;
    my $rc = 0;
    my $orgPartTableString = $self->getPartitionTable()->toString();

    Debug(1, "Entered.");

    # Perform some checks to restrict what can be done
    if( !$self->checkPartInfo($partTable) ) {
	Warning("Failed to make partitions because the partition information is invalid.");
	return(0);
    }

    my ($devName) = $self->getVmnixDevName();

    if( !isDevName($devName) ) {
	Warning("Invalid device name '$devName'.  Should be begin with /dev/.");
	return(0);
    }

    if( $devName eq "" ) {
	Warning("Attempting to modify partition of a disk with null device name.");
	return(0);
    }

    my $partString = $partTable->toSfdiskString();

    Debug(3, "Creating partitions on device $devName.  The current configuration is");
    dumpPartitionInfo($devName);

    my $partHelperName = 'partition_create';
    my $dos_compat = isPartDOSCompatible($devName);
    if( $dos_compat == -1 ) {
	Warning("Could not determine if disk $devName contains a partition table this is compatible with DOS.");
	return(0);
    }

    if( !$dos_compat ) {
	$partHelperName = 'partition_create_beta2compat';
	Debug(2, "Detected an old non-DOS compatible partition.  Using beta2 compatible partitioning helper.");
    }

    if( ! c_getHelper($partHelperName) ) {
	Warning("Cannot run the partitioning tool $partHelperName.");
	return(0);
    }

    if( s_system(c_getHelper($partHelperName) . ' ' . shell_string($devName),
		 $partString) ) {
	Warning("Partitioning helper failed.  Make sure that the disk  " .
		"is not in use.  Also make sure that no running virtual " .
		"machines are using the disk.");
    }
    $rc = $?;
    Debug(3, "Partition creator returned error code " . $rc . ".");

    Debug(3, "Done creating partitions on device $devName.  The new configuration is");
    dumpPartitionInfo($devName);

    # Load new partition information now that the tables have changed
    $self->loadPartitionTable();

    my $newPartTableString = $self->getPartitionTable()->toString();
    # verifying that the partition indeed took place. This is to handle the case 
    # when the vmkernel reports an -EBUSY error, which sfdisk ignores. This is also
    # the only place to keep this check as thepartitioning could fail in 
    # vmkernel. The vmkerenel protects the partition table
    # that contain active VMFS and core-dump partitions. If the partitioning is
    # successful then the 2 strings must be different.

    if ($orgPartTableString eq $newPartTableString) {
       Warning("Partition table is not modified as atleast one VMFS volume on " .
               "this disk is in use. Possible causes are :\n" .
               "1. An activated Swap File\n" .
               "2. Running/Migrating VMs \n" .
               "3. Service Console processes are accessing /vmfs");
       return(0);
    }

    if( $rc != 0 ) {
	return(0);
    } else {
        # When partitions change, the vmkernel needs to rescan VMFS.
        rescanVMFS();
	return(1);
    }
}


# Debug function used to print out the current configuration
sub dumpPartitionInfo($) {
    my($devName) = @_;

    if( $gDEBUG != 3 ) {
	return;
    }

    if( !c_getHelper('partition_list') ) {
	Warning("Cannot run the partitioning tool.");
	return(undef);
    }

    if( !open(FD, c_getHelper('partition_list') . ' ' . shell_string($devName) . " |") ) {
	Warning("Could not run the partitioning tool. Reason: $!");
	return(undef);
    }

    while(<FD>) {
	print STDERR "  " . $_;
    }
    close(FD);

    print STDERR "\n";
}


# Warning, deletes all the partition!  Use with caution.
sub deleteAllPartitions() {
    my $self = shift;

    my $partTable = $self->createPartitionTable();

    # Create a partition and make it of type $gVMFSPartType
    if( !$self->makePartitions($partTable) ) {
	Warning("Could not delete partitions.  Reason: $!.\n");
	return(0);
    }

    return(1);
}


# Creates a new MFS partition on a partition.
sub createNewMFS($$$) {
    my $self = shift;
    my ($partition, $blocksize, $version) = @_;

    my ($name) = $self->getVmkName();

    if( $name eq "" ) {
	Warning("Attempting to create an VMFS on a device with null device name.");
	return(0);
    }

    if( !isLogVmkDiskName($name) ) {
	Warning("Must create an VMFS on a device using a logical scsi name. " .
		"Name that was used is '$name'.");
	return(0);
    }

    my $partname = VMware::HConfig::RawDisk::makeLogVmkPartName($name, $partition);
    if( !defined($partname) ) {
	Warning("Invalid partition name $partname.");
	return(0);
    }

    if( ! c_getHelper('vmfs_create_fs') ) {
	Warning("Cannot run the VMFS fstools.");
	return(0);
    }

    # XXX: I'm not sure that we should enforce this limit. First, the limit is
    # arbitrary. Second, it's possible (especially since the block size limit
    # for VMFS-1 file systems has been lowered) that the MUI will have to
    # reformat a file system with an "unsupported" block size. - mikol April 9,
    # 2003 2:05 PM
    if ($blocksize && $blocksize =~ /^(\d+)m$/) {
       my $bsl = $version == 1 ? $gVMFSBlockSizeLimit : $gVMFS2BlockSizeLimit;
       if ($1 > $bsl) {
          Warning("Block size $blocksize exceeds allowed limit.");
          return(0);
       }
    }

    my $cmd = c_getHelper('vmfs_create_fs');
    $cmd .= ' ' . shell_string('vmfs' . $version);
    $cmd .= ' ' . shell_string($partname);
    if ($blocksize) {
       $cmd .= ' ' . shell_string('-b') . ' ' . shell_string($blocksize);
    }

    if(System($cmd)) {
	# Error could not run the partitioning tool
	Warning("Error: Could not create VMFS file system on partition $partname.  " .
                "Reason: $!.  Make sure that the disk is not in use.  " .
                "Also make sure that no running virtual machines are " .
                "using the partition, and that you do not have an active " .
                "swap file on the partition.");
	return(0);
    }

    $self->loadPartitionTable();

    return(1);
}


# Warning, changes the type of a partition.  Use with caution.
# id should be an integer
sub setPartitionID($$) {
    my $self = shift;
    my ($partition, $id_int) = @_;

    my ($devName) = $self->getVmnixDevName();

    if( $devName eq "" ) {
	Warning("Attempting to set id of partition with null device name.");
	return(0);
    }

    if( !isDevName($devName) ) {
	Warning("Invalid device name '$devName'.  Should be begin with /dev/.");
	return(0);
    }

    Debug(1, "Setting partition ID of device " . $self->getVmkName() . " ($devName).\n");

    if( !c_getHelper('partition_settype') ) {
	Warning("Cannot run the partitioning tool.");
	return(0);
    }

    # XXX Need a check here to make sure we aren't changing the partition
    # type to an extended partition from a non-extended partition.
    my $command =
	(c_getHelper('partition_settype') . ' ' .
	 shell_string($devName) . ' ' .
	 shell_string($partition) . ' ' .
	 sprintf("%02x", $id_int)
	 );

    if( System($command) ) {
	Warning("Could not set the partition id of " . $self->getVmkName() .
		":$partition to " . sprintf("%02x", $id_int) . ". Reason: $!.  " .
		"Make sure that the disk is not in use.  Also make sure " .
                "that no running virtual machines are using the partition.");
	return(0);
    }

    $self->loadPartitionTable();

    # When partitions change, the vmkernel needs to rescan VMFS.
    rescanVMFS();

    return(1);
}

#####################################################################
# End of Disk Modification Methods.   These methods modify a disk.  #
#####################################################################

#####################################################################
# Methods not intended for use by user                              #
#####################################################################
# Sets the geometry of the disk
sub setGeometry($$$) {
    my $self = shift;
    my($geo_cyl, $geo_head, $geo_sect) = @_;
    $self->{GEOMETRY} = { CYLINDER => $geo_cyl,
			  HEAD => $geo_head,
			  SECTOR => $geo_sect };
}

sub findGeometry($) {
    # fstools needs the logical name. sfdisk uses the file name.
    my($devName) = @_;
    my($geo_cyl, $geo_head, $geo_sect) = (0, 0, 0);

    local *FD;

    if( !c_getHelper('disk_geometry') ) {
	Warning("Error: Cannot run the partitioning tool.");
	return($geo_cyl, $geo_head, $geo_sect);
    }

    if( !open(FD, c_getHelper('disk_geometry') . ' ' . shell_string($devName) . " |") ) {
	Warning("Could not run the partitioning tool.");
	return($geo_cyl, $geo_head, $geo_sect);
    }

    while(<FD>) {
	chomp();

	if( /^$devName:/ ) {
            if( /^($devName): (\d+) cylinders\, (\d+) heads\, (\d+) sectors\/track/ ) {
		my $devName = $1;
		my $cyl = $2;
		my $head = $3;
		my $sectors = $4;

		$geo_cyl = $cyl;
		$geo_head = $head;
		$geo_sect = $sectors;
	    }
	}
    }
    close(FD);
    return($geo_cyl, $geo_head, $geo_sect);
}


# Find the partition information by using the partitioning tool or
# else the proc file.
#
# On success, return an array containing the partition information.
# If no partitions found, return an empty array.
#
# Definition of the partition fields:
#
# return->[n]->{PARTNUM}: Number of the partition
# return->[n]->{START_CYL}: Logical start cylinder number of the partition
# return->[n]->{END_CYL}: Logical end cylinder number of the partition
# return->[n]->{SECTORS}: Size of partition in sectors.  (not exact)
# return->[n]->{TYPE}: Integer corresponding to the type of the partition.
# return->[n]->{MFS}: Flag indicating if the partition has the MFS file system.
sub findPartitionTable($$$) {
    my $self = shift;
    my ($partTable, $name, $devName) = @_;

    Debug(1, "Entered");

    $partTable = findPartitionsByPartTool($partTable, $name, $devName);
    if( defined($partTable) ) {
	return($partTable);
    }
    return(undef);
}

# Find the partition information by using the partitioning tool
# On success, return an array containing the partition information.
# If an error occurs, return undef.  If no partitions found, return
# an empty array.
#
# Definition of the partition fields: see the comments for findPartition()
sub findPartitionsByPartTool($$$) {
    # fstools needs the logical name. sfdisk uses the file name.
    my ($partTable, $name, $devName) = @_;
    local *FD;

    Debug(1, "Entered");

    if( ! -r $devName ) {
	# Error could not read the device
	Warning("Could not read the device $devName.");
	return(undef);
    }

    if( !c_getHelper('partition_list') ) {
	Warning("Cannot run the partitioning tool.");
	return(undef);
    }

    if (!open(FD, c_getHelper('partition_list') . ' ' . shell_string($devName) .
              " 2>/dev/null |")) {
	Warning("Could not run the partitioning tool. Reason: $!");
	return(undef);
    }

    while(<FD>) {
	chomp();

	# Remove the boot flag if it exists
	s/[\*]//g;

	if( /(^\/dev\/.*)/ ) {
	    # Look for 7 non-white space items
	    if(/(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/) {
		my $partNum = $1;
		my $scyl = $2;
		my $ecyl = $3;
		my $blocks = $5;
		my $type = $6;
		# Don't care about 4,7

		# Remove extra characters that we don't care about.  Have to
		# do this here because some lines may be just the - character
		$partNum =~ s/[+-]//g;
		$scyl =~ s/[+-]//g;
		$ecyl =~ s/[+-]//g;
		$blocks =~ s/[+-]//g;
		$type =~ s/[+-]//g;

		$type = hex($type);

		$partNum = ($partNum eq "" ? 0 : $partNum);
		$scyl = ($scyl eq "" ? 0 : $scyl);
		$ecyl = ($ecyl eq "" ? 0 : $ecyl);
		$blocks = ($blocks eq "" ? 0 : $blocks);
		$type = ($type eq "" ? 0 : $type);

		# Get only the partition number
		$partNum =~ s/$devName[p]?(\d+)/$1/;

		# Skip empty partitions
		if( $type == 0 ) {
		    next;
		}

		$partTable->addPart($partNum,
				    $scyl,
				    $ecyl,
				    $type,
				    $blocks * 2);
	    }
	}
    }

    close(FD);
    return($partTable);
}


# Determines if the partition table leaves the first head on the
# machine free.  If it is, the partition table is DOS compatible.
sub isPartDOSCompatible($) {
    # fstools needs the logical name. sfdisk uses the file name.
    my ($devName) = @_;
    local *FD;

    Debug(1, "Entered");

    if( ! -r $devName ) {
	# Error could not read the device
	Warning("Could not read the device $devName.");
	return(-1);
    }

    if( !c_getHelper('partition_list_sect') ) {
	Warning("Cannot run the partitioning tool.");
	return(-1);
    }

    if( !open(FD, c_getHelper('partition_list_sect') . ' ' . shell_string($devName) . " |") ) {
	Warning("Could not run the partitioning tool. Reason: $!");
	return(-1);
    }

    while(<FD>) {
	chomp();

	if( /^$devName(\S+)\s*: start=\s*([0-9]+), size=\s*([0-9]+),\s*Id=([0-9a-f]+).*/ ) {
	    my $partNum = $1;
	    my $ssect = $2;
	    my $size = $3;
	    my $type = $4;

	    Debug(3, "Got partNum=$partNum ssect=$ssect size=$size type=$type\n");

	    if( $ssect == 1 ) {
		close(FD);
		Debug(1, "Returning 0 for non-DOS compatible.");
		return(0);
	    }
	}
    }

    close(FD);
    Debug(1, "Returning 1 for DOS compatible.");
    return(1);
}


# Uses the fstools program to find the /dev/vsd file the
# corresponds the the vsd device in Linux
sub lookupVsdDevName($$) {
    my ($name, $partnum) = @_;
    local *FD;
    Debug(2, "Lookup on $name, $partnum");
    if( isLogVmkDiskName($name) ) {
	if( !c_getHelper('vmfs_bind_vsd') ) {
	    Warning("Cannot execute vmfs_bind_vsd.");
	    return(0);
	}

	my $command = c_getHelper('vmfs_bind_vsd') . ' ' . shell_string($name . ":$partnum") . " 2>/dev/null";
	Debug(3, "Executing '$command'");

	if( !open(FD, "$command |") ){
	    # Error this should work since we know that the device is legal
	    Warning("Error: did not get the appropriate output from " . c_getHelper('vmfs_bind_vsd') . ".  Reason: $!");
	    return(0);
	}
	my $res = "";
	while(<FD>) {
	    # This check has the added benefit of untainting our data
	    # so that the script using this method can run with setuid.
	    if( /(\/dev\/vsd[0-9]*)/ ) {
		$res = $1;
	    }
	    close(FD);
	    Debug(2, "Returning $res");
	    return($res);
	}
    }

    Warning("Warning: Failed to get the device name of '$name' for some reason.");
    close(FD);
    return("");
}


# Executes the command and prints to stdin for the
# process that is exec'ed.
sub s_system($$) {
    my ($command, $stdin_args) = @_;
    Debug(1, "Executing '$command' with args\n<arg>\n$stdin_args</arg>");
    local *FD;
    my $message = "";

    if(!open(FD, "| $command > /dev/null") ) {
#    if(!open(FD, "| $command > /dev/null 2>&1") ) {
	# Error
	return($!);
    }

    print FD $stdin_args;
    close(FD);

    return($?);
}

# Tell vmkernel to rescan for VMFSs
sub rescanVMFS() {
   my $cmd = c_getHelper('vmfs_rescan');
   if (!$cmd) {
      Warning("Cannot rescan for VMFS.");
      return 0;
   }

   if (System($cmd)) {
      Warning("Failed to rescan for VMFS.  Reason: $!.");
      return 0;
   }

   return 1;
}

# Returns multipath info for a disk.  The example below shows such the
# structure of the info.
#
# { POLICY => 'fixed',
#   PATHS => [ { PATH => 'vmhba0:0:1', STATUS => 'on'} ,
#              { PATH => 'vmhba1:0:1', STATUS => 'on'} ],
#   ACTIVE => 'vmhba0:0:1',
#   PREFER => 'vmhba0;0:1' };
#
sub findMultipathInfo($$$) {
   my $self = shift;

   Debug(1, "Entered");

   my $info = { POLICY => undef,
                PATHS => [],
                ACTIVE => undef,
                PREFER => undef };

   my $vmkName = $self->getVmkName();
   my ($host, $target, $lun) = unmakeLogVmkDiskName($vmkName);
   my $file = c_getDir('proc_vmware_scsi') . "/$host/$target:$lun";

   local *FD;
   if (!open(FD, "< $file")) {
      Warning("Could not open file $file. Reason: $!");
      return undef;
   }

   my $readState = 0;
   while(<FD>) {
      chomp();

      if ($readState == 0 && /^Paths:(\S+)$/) {
         $info->{POLICY} = $1;
         $readState = 1;
      } elsif ($readState == 1 && /\s*(\S+)\s+(\S+?)(\*?)(\#?)$/) {
         my $path = $1;
         my $status = $2;

         if ($3) {
            $info->{ACTIVE} = $path;
         }

         if ($4) {
            $info->{PREFER} = $path;
         }

         push(@{$info->{PATHS}}, { PATH => $path, STATUS => $status });
      } else {
         $readState = 0;
      }
   }
   close(FD);

   return($info);
}

# Sets a new multipath value on a path
sub setMultipathInternal {
   my $self = shift;
   my $path = shift;
   my $cmd = shift;

   Debug(1, "Entered");

   my $vmkName = $self->getVmkName();

   # Check that path is valid if it is defined.
   if ($path && !grep($_->{PATH} eq $path, $self->getMultipathPaths())) {
      Warning("Path $path is not a valid path for $vmkName.");
      return 0;
   }

   # Write the command that was passed in to the proc node
   my ($host, $target, $lun) = unmakeLogVmkDiskName($vmkName);

   my $file = c_getDir('proc_vmware_scsi') . "/$host/$target:$lun";

   local *FD;
   if (!open(FD, "> $file")) {
      Warning("Could not open file $file. Reason: $!");
      return undef;
   }
   print FD $cmd;
   close(FD);

   # triggering a LUN IO
   $self->triggerLunIO();
   # Invalidate multipath info when it gets changed.
   $self->{MULTIPATH_INFO} = undef;

   return 1;
}

# Store one value in the config db
sub setConfigDB($$) {
   my $key = shift;
   my $val = shift;

   my $cfgdb = c_getFile('config_db');
   my $err = ("Could not store settings to $cfgdb.  New settings " .
              "will be lost on the next boot.");

   my $configDB = VMware::ConfigDB->new($cfgdb);
   if (!$configDB->load()) {
      Warning($err);
      return 0;
   }

   $configDB->set($key, $val);

   if ($configDB->hasChanged() && !$configDB->save()) {
      Warning($err);
      return 0;
   }
}


# Get a value from the Config DB.
sub getConfigDB($$) {
   my $key = shift;

   my $cfgdb = c_getFile('config_db');
   my $err = ("Could not store settings to $cfgdb.  New settings " .
              "will be lost on the next boot.");

   my $configDB = VMware::ConfigDB->new($cfgdb);
   if (!$configDB->load()) {
      Warning($err);
      return undef;
   }

   return $configDB->get($key);
}


#####################################################################
# PartitionTable.pm
#
# A basic class that contains information representing a Partition
# Table for a disk.  This class is not to be confused with the
# partition table on a disk which contains only 4 entries for
# physical partitions.
#####################################################################
package VMware::HConfig::PartitionTable;
use strict;
use VMware::ExtHelpers qw(&Warning &Debug);

#####################################################################
# Public Static Methods (indep. of each object)                     #
#####################################################################

sub new($$$$) {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = {};
    my($disk_name, $geo_cyl, $geo_head, $geo_sect) = @_;
    $self->{DISKNAME} = $disk_name;     # Name of VMK disk that owns this partition table
    $self->{PARTITIONS} = {};           # Partitions in the table
    bless($self, $class);

    $self->setGeometry($geo_cyl, $geo_head, $geo_sect);

    return($self);
}

#####################################################################
# Public Methods                                                    #
#####################################################################

# Accessor methods

# Gets the name of the disk that owns this partition table
sub getDiskName() {
    my($self) = shift;
    return($self->{DISKNAME});
}

sub getGeoCylinder() {
    my($self) = shift;
    return($self->{GEOMETRY}->{CYLINDER});
}

sub getGeoHead() {
    my($self) = shift;
    return($self->{GEOMETRY}->{HEAD});
}

sub getGeoSector() {
    my($self) = shift;
    return($self->{GEOMETRY}->{SECTOR});
}

sub getPartNums() {
    my($self) = shift;

    # Must put this after the bless statement since we use a method
    my @partNums = sort { $a <=> $b } keys(%{ $self->{PARTITIONS} });
    return(@partNums);
}

# Gets the partition indicated by the partition number partnum.
# Returns undef if partition does not exist.
sub getPart($) {
    my($self) = shift;
    my($partnum) = @_;

    return($self->{PARTITIONS}->{$partnum});
}

# Gets the starting cylinder of partition indicated by partnum
sub getPartStartCyl($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{START_CYL});
    }

    return(-1);
}

# Gets the ending cylinder of partition indicated by partnum
sub getPartEndCyl($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{END_CYL});
    }

    return(-1);
}

# Gets the ending cylinder of partition indicated by partnum
sub getPartType($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{TYPE});
    }

    return(-1);
}

sub getPartSectors($) {
    my($self) = shift;
    my($partnum) = @_;

    my $part = $self->getPart($partnum);
    if( defined($part) ) {
	return($part->{SECTORS});
    }

    return(-1);
}

sub getPartTypeName($) {
    my($self) = shift;
    my($partnum) = @_;

    my $type = $self->getPartType($partnum);
    if( defined($type) && defined($PartitionIDMap->{$type}) ) {
	return($PartitionIDMap->{$type});
    }

    return("Unknown");
}

# Lookup partitions that match specified fields.  Pass undef to ignore
# a field in the partition table.
sub lookupParts {
   my $self = shift;
   my $start_cyl = shift;     # optional
   my $end_cyl = shift;       # optional
   my $type = shift;          # optional
   my $sectors = shift;       # optional

   my @result;
   foreach my $partnum ($self->getPartNums()) {
      if ((defined($start_cyl) && $start_cyl != $self->getPartStartCyl($partnum)) ||
          (defined($end_cyl) && $end_cyl != $self->getPartEndCyl($partnum)) ||
          (defined($type) && $type != $self->getPartType($partnum)) ||
          (defined($sectors) && $sectors != $self->getPartSectors($partnum))) {
         next;
      }
      push(@result, $partnum);
   }

   return @result;
}

# Sets the geometry of the partition table.  This is an internal
# function which should not be used by the user.
sub setGeometry($$$) {
    my $self = shift;
    my($geo_cyl, $geo_head, $geo_sect) = @_;
    $self->{GEOMETRY} = { CYLINDER => $geo_cyl,
			  HEAD => $geo_head,
			  SECTOR => $geo_sect };
}


# Verify the following information about the partitions:
# 1. no more than 4 primary partitions are created
# 2. no more than 1 extended partition is created in one of the primary partitions
# 3. no more than MAX_PARTITIONS partitions are created
# 4. validity of the partition types: must be in range of 0-255
# 5. cylinders are in range of disk geometry
# 6. cylinders do not intersect
# 7. logical partitions must in order that they use cylinders
# .. and many more.
sub checkPartInfo() {
    my $self = shift;
    my $geo_cyl = $self->getGeoCylinder();
    my $geo_head = $self->getGeoHead();
    my $geo_sect = $self->getGeoSector();
    my $extendedPartNum = 0;
    my $low_cyl;
    my $high_cyl;
    my $part;

    Debug(1, "Entered with argument '" . $self->toString() . "'");

    # Make sure we have valid geometry information for the disk
    if( !VMware::HConfig::RawDisk::isValidGeometry($geo_cyl, $geo_head, $geo_sect) ) {
	Warning("Disk " . $self->getVmkName() . " (vmk) has invalid geometry information.");
	return(0);
    }

    # Create a list of primary partitions and logical partitions
    my %primaryPartInfo;
    my %logicalPartInfo;

    my $partnum;
    foreach $partnum ($self->getPartNums()) {
	my $start_cyl = $self->getPartStartCyl($partnum);
	my $end_cyl = $self->getPartEndCyl($partnum);
	my $type = $self->getPartType($partnum);

	if( $partnum < $MIN_PARTITION ) {
	    Warning("Partition number $partnum too small.  Minimum partition number is " . $MIN_PARTITION . ".");
	    return(0);
	}

	if( $partnum > $MAX_PARTITION ) {
	    Warning("Partition number $partnum too large.  Maximum partition number is " . $MAX_PARTITION. ".");
	    return(0);
	}

	if( $type < 0 || $type > 255 ) {
	    Warning("Invalid partition type " . sprintf("%02x", $type) . " specified for partition $partnum.");
	    return(0);
	}

	if( $start_cyl > $end_cyl ) {
	    Warning("Invalid partition specified for partition $partnum.  Start cylinder cannot be larger than the end cylinder.");
	    return(0);
	}

	if( VMware::HConfig::RawDisk::isPhysicalPart($partnum) ) {
	    if( defined($primaryPartInfo{$partnum}) ) {
		# Should never happen
		Warning("Physical partition $partnum is defined twice.");
		return(0);
	    }
	    $primaryPartInfo{$partnum} = $self->getPart($partnum);
	} elsif( VMware::HConfig::RawDisk::isLogicalPart($partnum) ) {
	    if( defined($logicalPartInfo{$partnum}) ) {
		# Should never happen
		Warning("Logical partition $partnum is defined twice.");
		return(0);
	    }
	    $logicalPartInfo{$partnum} = $self->getPart($partnum);
	} else {
	    Warning("Invalid partition specified.  Partition $partnum corresponds to neither a " .
		    "physical nor a logical partition.");
	    return(0);
	}
    }

    # Check the boundaries of the physical partitions
    $low_cyl = 0;
    $high_cyl = $geo_cyl - 1;

    # Sort each piece of partition information by the start cylinder
    my @sort_phyPartInfo = sort { $a->{START_CYL} <=> $b->{START_CYL} } values(%primaryPartInfo);

    foreach $part (@sort_phyPartInfo) {
	my $partnum = $part->{PARTNUM};
	my $start_cyl = $part->{START_CYL};
	my $end_cyl = $part->{END_CYL};
	my $type = $part->{TYPE};
	if( $start_cyl < $low_cyl ) {
	    if( $low_cyl != 0 ) {
		Warning("Partition $partnum start boundary too low.  It intersects another partition.");
	    } else {
		Warning("Partition $partnum start boundary must be non-negative .");
	    }
	    return(0);
	}

	if( $end_cyl > $high_cyl ) {
	    Warning("Partition $partnum end boundary is too large.  It exceeds the disk geometry.");
	    return(0);
	}

	if( VMware::HConfig::RawDisk::isPartExtended($type) ) {
	    if( $extendedPartNum ) {
		Warning("Partition $extendedPartNum and $partnum are extended partitions.  Only 1 extended " .
			"partition is allowed.");
		return(0);
	    }
	    $extendedPartNum = $partnum;
	}

	# Set the new lower boundary
	$low_cyl = $end_cyl + 1;
    }

    my @logPartNums = keys(%logicalPartInfo);
    if( !$extendedPartNum && $#logPartNums > -1 ) {
	Warning("Cannot have logical partitions unless a physical partition that has an extended type exists.");
	return(0);
    }

    # Done if there are no logical partitions
    if( $#logPartNums == -1 ) {
	return(1);
    }

    # Check the boundaries of the logical partitions
    $low_cyl = $self->getPartStartCyl($extendedPartNum);
    $high_cyl = $self->getPartEndCyl($extendedPartNum);

    my $logPartNum = 5;
    # Sort each piece of partition information by the start cylinder
    my @sort_logPartInfo = sort { $a->{START_CYL} <=> $b->{START_CYL} } values(%logicalPartInfo);
    foreach $part (@sort_logPartInfo) {
	my $partnum = $part->{PARTNUM};
	my $start_cyl = $part->{START_CYL};
	my $end_cyl = $part->{END_CYL};
	my $type = $part->{TYPE};

	if( $partnum != $logPartNum++ ) {
	    Warning("Logical partitions must be numbered sequentially beginning from 5.  " .
		    "Furthermore, those numbers must correspond to the order in which they " .
		    "appear on the disk.  Partition $partnum violates this requirement.");
	    return(0);
	}

	if( $start_cyl < $low_cyl ) {
	    if( $low_cyl != $self->getPartStartCyl($extendedPartNum) ) {
		Warning("Logical partition $partnum start boundary interstects another partition.");
	    } else {
		Warning("Logical partition $partnum start boundary does not begin in an extended partition.");
	    }
	    return(0);
	}

	if( $end_cyl > $high_cyl ) {
	    Warning("Logical partition $partnum end boundary is too large.  It exceeds the end of the extended partition.");
	    return(0);
	}

	if( VMware::HConfig::RawDisk::isPartExtended($type) ) {
	    Warning("Logical partition $partnum cannot be an extended partition.  Only physical partitions " .
		    "may be extended.");
	    return(0);
	}

	# Set the new lower boundary
	$low_cyl = $end_cyl + 1;
    }

    # All checks passed
    return(1);
}

# Queries for the physical partitions that can stil be added.
#
# A physical partition can be added if there are unallocated cylinders on the
# disk and less than 4 physical partitions exist.
sub getCanAddPhyPartNum() {
    my($self) = shift;
    my @result;
    my @physicalPartNums;
    my $extendedPartNum = 0;
    my $partnum;

    # First look for the physical partitions that can be added
    # sort them, look for a gap
    my @partnums = $self->getPartNums();

    for($partnum = $MIN_PARTITION; $partnum <= $MAX_PARTITION;	$partnum++ ) {
	if( VMware::HConfig::RawDisk::isPhysicalPart($partnum) ) {
	    if( !grep($_ eq $partnum, @partnums) ) {
		push(@result, $partnum);
	    } else {
		# Create a list of all the physical partition numbers
		push(@physicalPartNums, $partnum);
	    }
	}
    }

    if( $#result > -1 ) {
	# There is a free partition number.  Look for some unallocated space
	# or else the free partition number won't be usable.
	my @partList = map( $self->getPart($_), @physicalPartNums);

	# Sort each piece of partition information by the start cylinder
	my @sort_phyPartInfo = sort { $a->{START_CYL} <=> $b->{START_CYL} } @partList;

	my $low_cyl = 0;
	my $high_cyl = $self->getGeoCylinder() - 1;
	my $part;
	foreach $part (@sort_phyPartInfo) {
	    my $partnum = $part->{PARTNUM};
	    my $start_cyl = $part->{START_CYL};
	    my $end_cyl = $part->{END_CYL};
	    my $type = $part->{TYPE};

	    if( VMware::HConfig::RawDisk::isPartExtended($type) ) {
		$extendedPartNum = $partnum;
	    }

	    if( $start_cyl > $low_cyl ) {
		# There are unused cylinders
		return(@result);
	    }

	    $low_cyl = $end_cyl + 1;
	}

	if( $low_cyl <= $high_cyl ) {
	    # There are unused cylinders
	    return(@result);
	}
    }

    undef(@result);
    return(@result);
}


# Gets an array of the free ranges in the partition.
#
# Data structure is an array of hashes, each containing
# some elements: START_CYL, END_CYL, and LOGICAL.
#   START_CYL := the low cylinder number for a range
#   END_CYL  := the high cylinder number for a range
#   LOGICAL := 1 if the free range is in an extended partition
#              0 if it can be made into a physical partition.
#
# Partition table must be a well-formed partition table. In
# other words, it must pass the checkPartInfo() test.
#
sub getFreeCylinderRange() {
    my($self) = shift;
    my @result;
    my $extendedPartNum = 0;
    my $low_cyl;
    my $high_cyl;
    my $part;

    # Create a list of primary partitions and logical partitions
    my %primaryPartInfo;
    my %logicalPartInfo;

    my $partnum;
    foreach $partnum ($self->getPartNums()) {
	my $start_cyl = $self->getPartStartCyl($partnum);
	my $end_cyl = $self->getPartEndCyl($partnum);
	my $type = $self->getPartType($partnum);

	if( VMware::HConfig::RawDisk::isPhysicalPart($partnum) ) {
	    $primaryPartInfo{$partnum} = $self->getPart($partnum);
	} elsif( VMware::HConfig::RawDisk::isLogicalPart($partnum) ) {
	    $logicalPartInfo{$partnum} = $self->getPart($partnum);
	}
    }

    # Check the boundaries of the physical partitions
    $low_cyl = 0;
    $high_cyl = $self->getGeoCylinder() - 1;

    # Sort each piece of partition information by the start cylinder
    my @sort_phyPartInfo = sort { $a->{START_CYL} <=> $b->{START_CYL} } values(%primaryPartInfo);

    foreach $part (@sort_phyPartInfo) {
	my $partnum = $part->{PARTNUM};
	my $start_cyl = $part->{START_CYL};
	my $end_cyl = $part->{END_CYL};
	my $type = $part->{TYPE};

	if( VMware::HConfig::RawDisk::isPartExtended($type) ) {
	    $extendedPartNum = $partnum;
	}

	# Check for free ranges between physical partitions
	if( $start_cyl > $low_cyl ) {
	    # There are unused cylinders
	    my %range = ( START_CYL => $low_cyl,
			  END_CYL => $start_cyl - 1,
			  LOGICAL => 0 );
	    push(@result, \%range);
	}

	# Set the new lower boundary
	$low_cyl = $end_cyl + 1;
    }

    # Check the rest of the disk after the last physical partition
    if( $low_cyl <= $high_cyl ) {
	# There are unused cylinders
	my %range = ( START_CYL => $low_cyl,
		      END_CYL => $high_cyl,
		      LOGICAL => 0 );
	push(@result, \%range);
    }

    # Done if there is no extended partition
    if( $extendedPartNum == 0 ) {
	return(@result);
    }

    my @logPartNums = keys(%logicalPartInfo);

    # Check the boundaries of the logical partitions
    $low_cyl = $self->getPartStartCyl($extendedPartNum);
    $high_cyl = $self->getPartEndCyl($extendedPartNum);

    my $logPartNum = 5;
    # Sort each piece of partition information by the start cylinder
    my @sort_logPartInfo = sort { $a->{START_CYL} <=> $b->{START_CYL} } values(%logicalPartInfo);
    foreach $part (@sort_logPartInfo) {
	my $partnum = $part->{PARTNUM};
	my $start_cyl = $part->{START_CYL};
	my $end_cyl = $part->{END_CYL};
	my $type = $part->{TYPE};

	# Check for free ranges between logical partitions
	if( $start_cyl > $low_cyl ) {
	    # There are unused cylinders
	    my %range = ( START_CYL => $low_cyl,
			  END_CYL => $start_cyl - 1,
			  LOGICAL => 1 );
	    push(@result, \%range);
	}

	# Set the new lower boundary
	$low_cyl = $end_cyl + 1;
    }

    # Check the rest of the extended partition after the last logical partition
    if( $low_cyl <= $high_cyl ) {
	# There are unused cylinders
	my %range = ( START_CYL => $low_cyl,
		      END_CYL => $high_cyl,
		      LOGICAL => 1 );
	push(@result, \%range);
    }

    return(@result);
}


# Gets the number of unpartitioned cylinders for both physical and logical space
sub getUnpartSpace() {
    my($self) = shift;
    my @freeRanges = $self->getFreeCylinderRange();
    my $log_cyl = 0;
    my $phy_cyl = 0;

    my @logFreeRange = grep($_->{LOGICAL} == 1, @freeRanges);
    my @phyFreeRange = grep($_->{LOGICAL} == 0, @freeRanges);

    my @logSpace = map( $_->{END_CYL} - $_->{START_CYL} + 1, @logFreeRange );
    my @phySpace = map( $_->{END_CYL} - $_->{START_CYL} + 1, @phyFreeRange );

    my $num;
    foreach $num (@logSpace) {
	$log_cyl += $num;
    }
    foreach $num (@phySpace) {
	$phy_cyl += $num;
    }

    return($phy_cyl, $log_cyl);
}


# Asks if a logical partition can be added
# A logical partition can be added if there are unallocated cylinders in an
# extended partition.  Partition numbers increase monotonically.
sub canAddLogPartNum($) {
    my($self) = shift;
    my @logicalPartNums;
    my @result;
    my $partnum;

    # Look for the logical partitions that can be added
    # look at the extended partition if there was one.  look for gaps

    my $extendedPartNum = $self->getExtendedPartNum();
    if( $extendedPartNum == 0 ) {
	return(0);
    }

    my @partnums = $self->getPartNums();

    for($partnum = $MIN_PARTITION; $partnum <= $MAX_PARTITION;	$partnum++ ) {
	if( VMware::HConfig::RawDisk::isLogicalPart($partnum) ) {
	    if( !grep($_ eq $partnum, @partnums) ) {
		push(@result, $partnum);
	    } else {
		# Create a list of all the logical partition numbers
		push(@logicalPartNums, $partnum);
	    }
	}
    }

    # No free partitions
    if( $#result == -1 ) {
	return(0);
    }

    # There is a free partition number.  Look for some unallocated space
    # or else the free partition number won't be usable.
    my @partList = map( $self->getPart($_), @logicalPartNums);

    # Sort each piece of partition information by the start cylinder
    my @sort_logPartInfo = sort { $a->{START_CYL} <=> $b->{START_CYL} } @partList;

    my $low_cyl = $self->getPartStartCyl($extendedPartNum);
    my $high_cyl = $self->getPartEndCyl($extendedPartNum);
    my $part;
    foreach $part (@sort_logPartInfo) {
	my $partnum = $part->{PARTNUM};
	my $start_cyl = $part->{START_CYL};
	my $end_cyl = $part->{END_CYL};
	my $type = $part->{TYPE};

	if( $start_cyl > $low_cyl ) {
	    # There are unused cylinders
	    return(1);
	}

	$low_cyl = $end_cyl + 1;
    }

    if( $low_cyl <= $high_cyl ) {
	# There are unused cylinders
	return(1);
    }
    return(0);
}

# Gets the number of the extended partition
# Returns 0 if no extended partitions exist.
sub getExtendedPartNum() {
    my($self) = shift;
    my $partnum;

    foreach $partnum ($self->getPartNums()) {
	my $type = $self->getPartType($partnum);
	if( VMware::HConfig::RawDisk::isPartExtended($type) ) {
	    return($partnum);
	}
    }
    return(0);
}

# Creates a string that can be put into sfdisk to make the partitions
# indicated in the partition table
sub toSfdiskString() {
    my($self) = shift;
    my $partString = "";
    my $i;

    my @partNums = $self->getPartNums();
    my $currentPart = 1;
    my $hasExtended = 0;

    while( $#partNums > -1 && $currentPart <= $MAX_PARTITION ) {
	my $partnum = shift(@partNums);
	if( $partnum > $currentPart++ ) {
	    unshift(@partNums, $partnum);
	    $partString .= '0 0 0' . "\n";
	    next;
	}
	my $start_cyl = $self->getPartStartCyl($partnum);
	my $end_cyl = $self->getPartEndCyl($partnum);
	my $part_id = sprintf("%02x", $self->getPartType($partnum));
	my $num_cyl = $end_cyl - $start_cyl + 1;
	$partString .= $start_cyl . ' ' . $num_cyl . ' ' . $part_id . "\n";

        if (VMware::HConfig::RawDisk::isPartExtended($self->getPartType($partnum))) {
           $hasExtended = 1;
        }
    }

    # Make sure that a line exists for at least the first partition table entries.
    for ( ; $currentPart < 5; $currentPart++) {
       $partString .= '0 0 0' . "\n";
    }

    # Fix for bug 23933.  sfdisk only writes what we tell it to write.  If
    # we tell it to only write the extended partition, it will not touch
    # the logical partitions.  In the case that we want to delete the logical
    # partitions, we must explicitly pass in something telling sfdisk to wipe
    # the logical partitions clean.  Passing in '0 0 0' seems to do the trick,
    # but only if it is being passed in for the 5th partition.  If it's passed
    # in for the 6th or later, sfdisk errors out. -jhu

    if ($currentPart == 5 && $hasExtended) {
       $partString .= '0 0 0' . "\n";
    }

    if( $#partNums > -1 ) {
	Warning("ASSERT: Partition number(s) (" . join(',', @partNums) . ") exceed maximum allowed.");
    }

    return($partString);
}


sub toString() {
    my($self) = shift;
    my $s = "";
    my $i;
    my $vmkName = $self->getDiskName();

    foreach $i ($self->getPartNums()) {
	my $type = sprintf("%02x", $self->getPartType($i));
	my $typeName = $self->getPartTypeName($i);
	my $start_cyl = $self->getPartStartCyl($i);
	my $end_cyl = $self->getPartEndCyl($i);
	my $sect = $self->getPartSectors($i);

	$s .= "  $vmkName:$i type=$typeName ($type) start=$start_cyl end=$end_cyl sect=$sect\n";
    }

    return($s);
}

# Mutator methods

# Adds a partition to the partition table.
# partnum: is the partition number
# start_cyl: is the starting cylinder with zero being the first cylinder
# end_cyl: is the ending cylinder of the partition
# type: is the partition type/id
# sector: is the number of sectors in the partition (optional)
sub addPart {
    my($self) = shift;
    my($partnum, $start_cyl, $end_cyl, $type, $sectors) = @_;
    my %part;
    $part{PARTNUM} = $partnum;
    $part{START_CYL} = $start_cyl;
    $part{END_CYL} = $end_cyl;
    $part{TYPE} = $type;
    $part{SECTORS} = defined($sectors) ? $sectors : 0;
    $self->{PARTITIONS}->{$partnum} = \%part;
}

# Remove the partition given the partition number
sub removePart($) {
    my($self) = shift;
    my($partnum) = @_;
    delete($self->{PARTITIONS}->{$partnum});
}

# Reorder the logical partitions in the order of their start cylinders.
# No guarantee that the partitions will be correct.
sub reorderLogicalParts() {
    my($self) = shift;
    my @logParts = grep($_->{PARTNUM} > 4 , values(%{ $self->{PARTITIONS} }));
    my @sortedLogParts = sort { $a->{START_CYL} <=> $b->{START_CYL} } @logParts;
    my $warningFlag = 1;

    my $partnum;

    # Remove all the previous logical partitions
    $self->removeAllLogicalParts();

    # Re-add the logical partitions in order of their start partitions
    $partnum = 5;
    while( $#sortedLogParts > -1 ) {
	my $part = shift(@sortedLogParts);
	if( $part->{PARTNUM} != $partnum ) {
	    if( $warningFlag == 1 ) {
		Warning("Some of your logical partitions were reordered because they were not " .
			"laid out in the order of their cylinders in the extended partition.  " .
			"Saving this partition table may cause corruption to any previously existing " .
			"logical partitions on the disk that you are changing!  Be sure to have " .
			"backed up any important data on this disk before you proceed!");
		$warningFlag = 0;
	    }
	}
	$part->{PARTNUM} = $partnum;
	$self->{PARTITIONS}->{$partnum} = $part;
	$partnum++;
    }
    return($warningFlag);
}


# Remove all logical partitions
sub removeAllLogicalParts() {
    my($self) = shift;
    my @logPartNums = grep($_ > 4 , keys(%{ $self->{PARTITIONS} }));

    my $partnum;

    # Remove all the previous logical partitions
    foreach $partnum (@logPartNums) {
	$self->removePart($partnum);
    }
}


# Indicates if the disk has logical partitions
sub hasLogicalParts() {
    my($self) = shift;
    my @logPartNums = grep($_ > 4 , keys(%{ $self->{PARTITIONS} }));
    return( $#logPartNums > -1 );
}
