#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# SSHost.pm
#
# A class that is used to gather and manipulate information about the
# host being configured to run the Scalable Server.  Information
# includes PCI devices, LILO configuration, and memory size of host.

# Changes
# 1. pci devices should have separate parse and merge methods
# 2. makecpci should take a list of pci devs as args
# 3. add the ability to read the cmdline

package VMware::HConfig::SSHost;
use strict;
use POSIX qw(setsid);                # For making a daemon process
use VMware::HConfig::Constant qw(c_getFile c_getDir c_getHelper
				 &CACHE_NEVER &CACHE_ONCE &CACHE_CALLBACK
				 $gSSHost_CacheDefaults $gSSHost_CacheSource
				 $gVmkDeviceType_nic $gVmkDeviceType_hba
				 $gVmkDeviceType_bond $gVmkDeviceType_tcpip $gVmkDeviceType_migration
                                 %gNonDiskSCSITypes $gVmkDumpPartSize
				 $gUseSIGCHLD $CFG_HOST_SECURITY $CFG_CERTIFICATE
                                 %gCertInfo %gBlockDeviceDir %gBlockDeviceNames);
use VMware::ExtHelpers qw(&Warning &Panic &System &p_number &mktmpfile &Debug &getWarnings &LOG
			  &shell_string &unshell_string &System2 &safe_move &intersection_string
                          &internal_echo);

use VMware::Tokenizer;
use VMware::Config;
use VMware::ConfigDB;
use VMware::HConfig::PCIDevice;
use VMware::HConfig::PCIDeviceMap;
use VMware::HConfig::LILOConfig;
use VMware::HConfig::SCSIAdapter;
use VMware::HConfig::RawDisk;
use VMware::HConfig::MFSFileSystem;
use VMware::HConfig::RunlevelControl;
use VMware::HConfig;

# Error message for missing modules. Variables can be used.
# module is the name of the module.
# type is the type of the module.
my %cMissingModuleMessage = (
			     'linux' => \&MissingModule_linux,
			     $gVmkDeviceType_hba => \&MissingModule_vmhba,
			     $gVmkDeviceType_nic => \&MissingModule_vmnic,
			     'nfshaper' => \&MissingModule_nfshaper,
			     );

#####################################################################
# Public Static Methods (indep. of each object                      #
#####################################################################

sub new() {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self = $proto->create();

    Debug(1, "Entered.");

    # Parameters to control the caching
    $self->{CACHE_OPTIONS} = $gSSHost_CacheDefaults;
    $self->{CACHE_SOURCE} = $gSSHost_CacheSource;

    bless($self, $class);

    return($self);
}

sub create() {
    my $self = {};

    # Initialize all the values to undef signifying that they have not
    # yet been set.  The values are filled in in a lazy manner.
    $self->{KERNEL_INFO} = undef;        # the kernel currently running on the system
    $self->{PCIDEVICE_MAP} = undef;      # Info that maps a device to its driver file
    $self->{PCIDEVICES} = {};            # The PCI device state of the LILO entries
    $self->{DEVICENAMES} = undef;        # Names for all vmk devices
    $self->{NEWNAMES} = undef;           # Device list w/ newly-assigned names
    $self->{LILOCONF} = undef;           # Builds the boot information
    $self->{SNMPCONF} = undef;           # SNMP configuration information
    $self->{BIMAGE_FILES} = undef;       # Finds all vmnix kernel boot images
    $self->{INITRD_FILES} = undef;       # Finds all initrd images for vmnix
    $self->{DIR_MAP} = undef;            # Builds a map from directories to devices
    $self->{VMK_SCSI_ADAPTERS} = undef;  # SCSI disks available
    $self->{VMK_NET_ADAPTERS} = undef;   # Network adapters owned by the VM
    $self->{MFS} = undef;                # MFS file systems found in the SCSI disks
    $self->{NUM_CPUS} = undef;           # Number of CPUs
    $self->{NUM_LCPUS} = undef;          # Number of Logical CPUs
    $self->{TOTAL_MEMSIZE} = undef;      # The total memory on the machine
    $self->{VMNIX_MEMSIZE} = undef;      # The amount of memory allocated to the VMnix
    $self->{VMKDUMP_PARTITION} = undef;  # Partition where VMKDump info will go
    $self->{CURRENT_LABEL} = undef;      # Current label of the lilo entry being used
    $self->{CONFIG_DB} = undef;          # Pointer to config db
    $self->{LOCATIONS_DB} = undef;       # Pointer to locations db
    $self->{RUNNING_WORLDS} = undef;     # List of running worlds

    $self->{ERRORS} = [];                # Error array (not used currently)

    return($self);
}

#####################################################################
# Public Methods                                                    #
#####################################################################

# Accessor functions
sub getConfigDB() {
    my($self) = shift;
    if( $self->needCacheUpdate('CONFIG_DB', 1) ) {
	$self->{CONFIG_DB} = $self->findConfigDB();

	# The configDB file may be created in findConfigDB() so we
	# should update the cache timestamp after the find completes.
	$self->updateCacheSrcTime('CONFIG_DB');

	# XXX Because we update afterwards, we may lose changes if they
        # occur after the find but before the update to the cache. -jhu
    }
    return($self->{CONFIG_DB});
}

sub getLocationsDB() {
    my($self) = shift;
    if( $self->needCacheUpdate('LOCATIONS_DB', 1) ) {
	$self->{LOCATIONS_DB} = $self->findLocationsDB();
    }
    return($self->{LOCATIONS_DB});
}

# XXX For completeness, there should be a get method that gets the current
# state of the system as SSHost thinks it is.  The current getPCIDeviceList
# method gets what's in the config database.  In the case the label is equal
# to the one that is currently running, getPCIDeviceList performs a check
# to make sure that the current system reflects the configuration that
# is returned.

# Gets the PCIDeviceList for a particular configuration by the name of
# label.  If no label is specified, then the actively running one is
# used.  If a label is specified, it must exist or else bad things
# will happen.
sub getPCIDeviceList($) {
    my ($self) = shift;
    my ($label) = @_;

    # VMnix handling from here.
    if( $label eq "" ) {
	$label = $self->getCurrentLabel();
	if( !defined($label) ) {
	    Panic("Could not determine the name of the currently running system");
	}
    }

    # The label may not exist in lilo in the case that it is a new label

    if( $self->needCacheUpdate('PCIDEVICES', 1) || !defined($self->{PCIDEVICES}->{$label}) ) {
	my $PCIDeviceList = $self->findPCIDevices($label);
	$self->{PCIDEVICES}->{$label} = $PCIDeviceList;

	# The configuration database may have changed during the construction of the
	# device list.  Commit the changes.
	$self->commitConfigDB();

	# Perform some sanity checks that will spring some warnings
	if( $label eq $self->getCurrentLabel() ) {
	    $self->checkPCIDevicesWithCurrent($PCIDeviceList, 0);
	}

	# XXX Should do a check and make sure that the cpci settings in lilo.conf
	#     match that of our configuration database, although that would introduce
	#     another dependency on the PCI Device info.. -jhu
    }
    return($self->{PCIDEVICES}->{$label});
}

# Gets an object holding all the device names.  Pass in
# the boot label with desired PCI device configuration.
#
# XXX: Currently, only one set of device names is supported. 
#      If getDeviceNames() is run with different labels, and
#      there is a difference in PCI device configurations,
#      the device names may be overwritten.
sub getDeviceNames($) {
    my($self) = shift;
    my $label = shift;

    if( $self->needCacheUpdate('DEVICENAMES', 1) ) {
        ($self->{DEVICENAMES}, $self->{NEWNAMES}) =
            $self->findDeviceNames($label);
    }

    return($self->{DEVICENAMES});
}

# Gets a list of newly-named devices. Note that this
# must be called after getDeviceNames().
sub getNewlyNamedDevices() {
    my($self) = shift;

    if( !defined($self->{NEWNAMES}) ) {
        return undef;
    } else {
        return($self->{NEWNAMES}->getPCIDevices());
    }
###XXX new code -- oops it doesn't work.
#    if( $self->needCacheUpdate('DEVICENAMES', 1) ) {
#        ($self->{DEVICENAMES}, $self->{NEWNAMES}) =
#            $self->findDeviceNames($label);
#    }

#    return($self->{NEWNAMES}->getPCIDevices());
###XXX end new code
}

# Gets a structure containing LILO information
sub getLILOConfig() {
    my($self) = shift;

    if( $self->needCacheUpdate('LILOCONF', 1) ) {
	$self->{LILOCONF} = $self->findLILOConfig();
    }

    return($self->{LILOCONF});
}

# Gets the contents of the SNMP configuration file as text
sub getSNMPConfig {
    my($self) = shift;

    if( $self->needCacheUpdate('SNMPCONF', 1) ) {
	$self->{SNMPCONF} = $self->findSNMPConfig();
    }

    return($self->{SNMPCONF});
}

# Returns whether SNMP is enabled or not
sub getSNMPEnable {
    my($self) = shift;
    return($self->findSNMPEnable());
}

# Gets the all the vmnix kernel image files that we know about
sub getBootImageFiles() {
    my($self) = shift;

    if( $self->needCacheUpdate('BIMAGE_FILES', 1) ) {
	my @boot_files = $self->findVMBootImageFiles();
	$self->{BIMAGE_FILES} = \@boot_files;
    }

    return($self->{BIMAGE_FILES});
}

# Gets the all the vmnix initrd image files that we know about
sub getInitrdImageFiles() {
    my($self) = shift;

    if( $self->needCacheUpdate('INITRD_FILES', 1) ) {
	my @initrd_files = $self->findVMInitrdImageFiles();
	$self->{INITRD_FILES} = \@initrd_files;
    }

    return($self->{INITRD_FILES});
}

# Gets the file system mapping from directory to disk device
sub getDirMapping() {
    my($self) = shift;

    if( $self->needCacheUpdate('DIR_MAP', 1) ) {
	$self->{DIR_MAP} = $self->findDirMapping();
    }

    return($self->{DIR_MAP});
}

# Gets the build information about the kernel
sub getKernelInfo() {
    my($self) = shift;

    if( $self->needCacheUpdate('KERNEL_INFO', 1) ) {
	$self->{KERNEL_INFO} = $self->findKernelInfo();
    }

    return($self->{KERNEL_INFO});
}

# Gets the type of the kernel
sub getKernelType() {
    my($self) = shift;
    my $kernelType;

    my $info = $self->getKernelInfo();
    my @infoArray = split(/\s+/, $info);

    if( $self->isVMnix() ) {
	$kernelType = "VMnix";
    }
    else {
	if( $#infoArray > -1 ) {
	    $kernelType = $infoArray[0];
	} else {
	    $kernelType = "Unknown";
	}
    }

    return($kernelType);
}

sub getKernelVersion() {
    my($self) = shift;

    my $info = $self->getKernelInfo();
    my @vals = split(/\s+/, $info);

    if( $#vals >= 2 ) {
	return($vals[2]);
    }

    return("Unknown");
}

sub getSystemInfo() {
    my($self) = shift;

    my $info = $self->getKernelInfo();
    my @infoArray = split(/\s+/, $info);

    my $machname = "Unknown";
    if( $#infoArray >= 1 ) {
	$machname = $infoArray[1];
    }

    my $s = $self->getKernelType() . " kernel version " . $self->getKernelVersion() .
	" on machine " . $machname;

    return($s);
}

# Gets the PCIDevice map.
sub getPCIDeviceMap() {
    my ($self) = shift;

    if( $self->needCacheUpdate('PCIDEVICE_MAP', 1) ) {
	$self->{PCIDEVICE_MAP} = $self->findPCIDeviceMap();
    }

    return( $self->{PCIDEVICE_MAP} );
}

# Gets all the SCSI disks available to the VMs
sub getVMSCSIDisks() {
    my($self) = shift;
    my @adapters = @{ $self->getVMSCSIAdapters() };

    my @disks;
    my $a;
    foreach $a (@adapters) {
	push(@disks, @{ $a->getDisks() });
    }

    return(\@disks);
}

# Gets all the SCSI tape drives available to the VMs
sub getVMSCSITapes() {
    my($self) = shift;
    my @adapters = @{ $self->getVMSCSIAdapters() };

    my @tapes;
    my $a;
    foreach $a (@adapters) {
	push(@tapes, $a->getNonDiskDevicesByType('Tape'));
    }

    return(\@tapes);
}

# Gets all the SCSI CDROM drives available to the VMs
sub getVMSCSICdroms() {
    my($self) = shift;
    my @adapters = @{ $self->getVMSCSIAdapters() };

    my @cdroms;
    my $a;
    foreach $a (@adapters) {
	push(@cdroms, $a->getNonDiskDevicesByType('CDROM'));
    }

    return(\@cdroms);
}

# Gets all the SCSI Non Disk items available to the VMs
# Returns a hash of device to type bindings.  For example,
#    $ptdevs = %{ $self->getVMSCSINonDisk() };
sub getVMSCSINonDisks() {
    my($self) = shift;
    my @adapters = @{ $self->getVMSCSIAdapters() };

    my %devList;
    my $a;
    foreach $a (@adapters) {
	my %devs = %{ $a->getNonDiskDevices() };
	my $name;
	foreach $name (keys(%devs)) {
	    my $type = $devs{$name};
	    $devList{$name} = $type;
	}
    }

    return(\%devList);
}

# Gets all the SCSI adapters available to the VMs
sub getVMSCSIAdapters() {
    my($self) = shift;

    if( $self->needCacheUpdate('VMK_SCSI_ADAPTERS', 1) ) {
	if( $self->isVMKernel() ) {
	    $self->{VMK_SCSI_ADAPTERS} = $self->findScsiAdapters($self->getPCIDeviceList(''));
	} else {
	    return([]);
	}
    }

    return($self->{VMK_SCSI_ADAPTERS});
}

# Gets all the network adapters available to the VMs
sub getVMNetAdapters() {
    my($self) = shift;

    if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
	if( $self->isVMKernel() ) {
	    $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
	} else {
	    return([]);
	}
    }

    my @adapters =  map { $_->{"name"} } @{$self->{VMK_NET_ADAPTERS}};
    return(\@adapters);
}

# Takes in a vmkernel NIC name and finds the PCI device info for it
sub getVMNetAdapterPCIInfo($) {
    my($self) = shift;
    my($name) = @_;

    if ($self->isVMKernel()) {
       if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
          $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
       }
       foreach my $adapter ( @{$self->{VMK_NET_ADAPTERS}} ) {
          if ($name eq $adapter->{"name"}) {
             return ($adapter->{"bus"}, $adapter->{"device"}, $adapter->{"function"});
          }
       }
    }

    return (undef, undef, undef);
}

# Takes in a vmkernel NIC name and finds the clustering info for it
sub getVMNetAdapterClustering($) {
    my($self) = shift;
    my($name) = @_;

    if ($self->isVMKernel()) {
       if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
          $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
       }
       foreach my $adapter ( @{$self->{VMK_NET_ADAPTERS}} ) {
          if ($name eq $adapter->{"name"}) {
             return $adapter->{"clustering"};
          }
       }
    }
    return undef;
}

# Takes in a vmkernel NIC name and finds its current speed
sub getVMNetAdapterSpeed($) {
    my($self) = shift;
    my($name) = @_;

    if ($self->isVMKernel()) {
       if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
          $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
       }
       foreach my $adapter ( @{$self->{VMK_NET_ADAPTERS}} ) {
          if ($name eq $adapter->{"name"}) {
             return $adapter->{"speed"};
          }
       }
    }
    return undef;
}

# Takes in a vmkernel NIC name and finds its NIC team
sub getVMNetAdapterNICTeam($) {
    my($self) = shift;
    my($name) = @_;

    if ($self->isVMKernel()) {
       if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
          $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
       }
       foreach my $adapter ( @{$self->{VMK_NET_ADAPTERS}} ) {
          if ($name eq $adapter->{"name"}) {
             return $adapter->{"nicteam"};
          }
       }
    }
    return undef;
}

# Takes in a vmkernel NIC name and finds the list of MAC addresses for it
sub getVMNetAdapterMAC($) {
    my($self) = shift;
    my($name) = @_;

    if ($self->isVMKernel()) {
       if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
          $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
       }
       foreach my $adapter ( @{$self->{VMK_NET_ADAPTERS}} ) {
          if ($name eq $adapter->{"name"}) {
             return $adapter->{"mac"};
          }
       }
    }
    return undef;
}

# Gets all the bond adapters available to the VMs
sub getVMNetBondAdapters() {
   my($self) = shift;

   if ($self->needCacheUpdate('VMK_NET_BOND_ADAPTERS', 1)) {
      if ($self->isVMKernel()) {
         $self->{VMK_NET_BOND_ADAPTERS} = $self->findVmkNetBondAdapters();
      } else {
         return([]);
      }
   }

   my @adapters =  map { $_->{"name"} } @{$self->{VMK_NET_BOND_ADAPTERS}};
   return(\@adapters);
}

# Gets all the names of the slaves for a bond adapter.
sub getVMNetBondAdapterSlaves($) {
   my($self) = shift;
   my($name) = @_;

   if ($self->isVMKernel()) {
      if ($self->needCacheUpdate('VMK_NET_BOND_ADAPTERS', 1)) {
         $self->{VMK_NET_BOND_ADAPTERS} = $self->findVmkNetBondAdapters();
      }
      foreach my $adapter ( @{$self->{VMK_NET_BOND_ADAPTERS}} ) {
         if ($name eq $adapter->{"name"}) {
            return $adapter->{"slaves"};
         }
      }
   }
   return undef;
}

# Gets all the MAC addresses on a bond adapter.
sub getVMNetBondAdapterMAC($) {
   my($self) = shift;
   my($name) = @_;

   if ($self->isVMKernel()) {
      if ($self->needCacheUpdate('VMK_NET_BOND_ADAPTERS', 1)) {
         $self->{VMK_NET_BOND_ADAPTERS} = $self->findVmkNetBondAdapters();
      }
      foreach my $adapter ( @{$self->{VMK_NET_BOND_ADAPTERS}} ) {
         if ($name eq $adapter->{"name"}) {
            return $adapter->{"mac"};
         }
      }
   }
   return undef;
}

# Takes in a vmkernel NIC name and finds its capabilites
sub getVMNetAdapterCapabilities($) {
    my($self) = shift;
    my($name) = @_;

    if ($self->isVMKernel()) {
       if( $self->needCacheUpdate('VMK_NET_ADAPTERS', 1) ) {
          $self->{VMK_NET_ADAPTERS} = $self->findVmkNetAdapters();
       }
       foreach my $adapter ( @{$self->{VMK_NET_ADAPTERS}} ) {
          if ($name eq $adapter->{"name"}) {
             return ($adapter->{"minCapabilities"}, 
                     $adapter->{"hwCapabilities"}, 
                     $adapter->{"maxCapabilities"});
          }
       }
    }
    
    return (undef, undef, undef);
}

# Gets all the network adapters available to be used by a VM.
# The list of available adapters includes all vmnic devices that
# are not part of a NIC team and bond devices that have slaves.
sub getAvailNetAdapters() {
    my($self) = shift;

    my @allVmnics = @{$self->getVMNetAdapters()};
    my @vmnics;
    foreach my $vmnic (@allVmnics) {
       my $nicteam = $self->getVMNetAdapterNICTeam($vmnic);
       if (!$nicteam) {
          push(@vmnics, $vmnic);
       }
    }

    my @allBonds = @{$self->getVMNetBondAdapters()};
    my @bonds;
    foreach my $bond (@allBonds) {
       my @slaves = @{$self->getVMNetBondAdapterSlaves($bond)};
       if (scalar(@slaves) > 0) {
          push(@bonds, $bond);
       }
    }

    my @adapters = (@bonds, @vmnics);
    return(\@adapters);
}

# Gets the TCP/IP information for the vmkernel.
sub getVMNetTcpip() {
   my($self) = shift;
   return $self->findVMNetTcpip();
}

# Gets the configured TCP/IP information for the vmkernel.
sub getVMNetTcpipCfg() {
   my($self) = shift;
   return $self->findVMNetTcpipCfg();
}

# Determines if the configured settings for TCP/IP match the actual settings.
# If so, returns 1.  Otherwise returns 0 indicating a reboot is needed.
sub isVMNetTcpipSync() {
   my($self) = shift;

   my $tcpipCfg = $self->getVMNetTcpipCfg();
   my $tcpipAct = $self->getVMNetTcpip();

   return cmpVMNetTcpipSync($tcpipCfg, $tcpipAct) == 0 ? 1 : 0;
}

# Returns 0 if the settings match.  Non-zero otherwise.
sub cmpVMNetTcpipSync($$) {
   my ($tcpip1, $tcpip2) = @_;
   my %tcpipInfo1 = %{$tcpip1};
   my %tcpipInfo2 = %{$tcpip2};

   my @allkeys;
   push(@allkeys, keys(%tcpipInfo1));
   push(@allkeys, keys(%tcpipInfo2));

   foreach my $name (sort(@allkeys)) {
      if (!defined($tcpipInfo1{$name})) {
         return 1;
      }
      if (!defined($tcpipInfo2{$name})) {
         return -1;
      }

      my $rc =
         ($tcpipInfo1{$name}->{IPADDR} cmp $tcpipInfo2{$name}->{IPADDR} ||
          $tcpipInfo1{$name}->{GATEWAYADDR} cmp $tcpipInfo2{$name}->{GATEWAYADDR} ||
          $tcpipInfo1{$name}->{NICNAME} cmp $tcpipInfo2{$name}->{NICNAME});

      if ($rc) {
         return $rc;
      }
   }

   return 0;
}

# Gets a disk by using the name of the disk "vmhba0:0:5"
sub getVMSCSIDiskByName($) {
    my($self) = shift;
    my($name) = @_;

    my @disks = @{ $self->getVMSCSIDisks() };
    my $disk;
    foreach $disk (@disks) {
	if( $disk->getVmkName() eq $name ) {
	    return($disk);
	}
    }

    return(undef);
}

# Get an MFS by scsi name ie "vmhba0:5:0:1"
# XXX This should be called getMFSByVmkName -jhu
sub getMFS($) {
    my($self) = shift;
    my($name) = @_;

    my @all_mfs = @{$self->getAllMFS()};

    my $mfs;
    foreach $mfs (@all_mfs) {
	if( $mfs->getVmkName() eq $name ) {
	    return($mfs);
	}
    }

    return(undef);
}

# Get an MFS by logical name
# XXX What happens if we have multiple disks with the same logical name? -jhu
sub getMFSByLogName($) {
    my($self) = shift;
    my($name) = @_;
    my $result = undef;

    my @all_mfs = @{$self->getAllMFS()};

    my $mfs;
    foreach $mfs (@all_mfs) {
	if( $mfs->getLogName() eq $name ) {
	    if( !defined( $result ) ) {
		$result = $mfs;
	    } else {
		Warning("Duplicate volume name $name found for VMFS on partition " .
			$mfs->getVmkName());
	    }
	}
    }

    return($result);
}

# Given some name, find the MFS file system
# XXX This should be called getMFS -jhu
sub getMFSLookup($) {
    my($self) = shift;
    my($name) = @_;

    my $vmfs;
    if( VMware::HConfig::RawDisk::isLogVmkPartName($name) ) {
	$vmfs = $self->getMFS($name);
    } else {
	$vmfs = $self->getMFSByLogName($name);
    }
    return($vmfs);
}


# Get modules loaded into the VMnix
# See function findVMnixModules() for a description of the
# return value.
sub getVMnixModules() {
    my($self) = shift;

    # XXX: We don't really care whether or not the running kernel is truely
    # vmnix. We just want a list of the devices the console is using.

    # my $modules = [];

    # if( !$self->isVMnix() ) {
    #   return($modules);
    # }

    my $modules = $self->findVMnixModules();
    return($modules);
}

# Get modules loaded into the VMkernel
# See function findVMKernelModules() for a description of the
# return value.
sub getVMKernelModules() {
    my($self) = shift;

    my $modules = [];

    if( !$self->isVMKernel() ) {
	return($modules);
    }

    $modules = $self->findVMKernelModules();
    return($modules);
}

# Get the current label of the LILO instance that is running
sub getCurrentLabel() {
    my($self) = shift;

    if( $self->needCacheUpdate('CURRENT_LABEL', 1) ) {
	$self->{CURRENT_LABEL} = $self->findCurrentLabel();
    }
    return($self->{CURRENT_LABEL});
}

# Gets the scsi names of all the MFS
sub getAllMFSNames() {
    my($self) = shift;

    my @all_mfs = @{$self->getAllMFS()};
    @all_mfs = map { $_->getVmkName() } @all_mfs;

    return(\@all_mfs);
}

# Gets all the MFS on the system
sub getAllMFS() {
    my($self) = shift;

    if( $self->needCacheUpdate('MFS', 1) ) {
	$self->{MFS} = $self->findMFS();
    }

    return($self->{MFS});
}

# Gets the total amount of memory on the system
sub getTotalMemSize() {
    my($self) = shift;

    if( $self->needCacheUpdate('TOTAL_MEMSIZE', 1) ) {
	$self->{TOTAL_MEMSIZE} = $self->findTotalMemSize($self->isVMnix(), $self->isVMKernel());
    }

    return($self->{TOTAL_MEMSIZE});
}

# Gets the amount of memory used by the Vmnix
sub getVmnixMemSize() {
    my($self) = shift;

    if( $self->needCacheUpdate('VMNIX_MEMSIZE', 1) ) {
	$self->{VMNIX_MEMSIZE} = $self->findVmnixMemSize($self->isVMnix());
    }

    return($self->{VMNIX_MEMSIZE});
}

# Gets the amount of memory available to the vmkernel
sub getVmkMemSize() {
    my($self) = shift;
    return($self->findVmkMemSize($self->isVMnix(), $self->isVMKernel()));
}

# Gets the amount of memory available to the VMs
sub getVMAvailMemSize() {
    my($self) = shift;
    return($self->findVMAvailMemSize($self->isVMnix(), $self->isVMKernel()));
}

# Gets memory sharing statistics
sub getMemSharingInfo() {
    my($self) = shift;
    return($self->findMemSharingInfo($self->isVMnix(), $self->isVMKernel()));
}

# Gets the list of vmkdump partitions.  An optional argument can be supplied
# to filter for the minimum size in Kb for a partitions.
sub getVmkDumpPartitions {
    my($self) = shift;
    my $minSizeKb = shift;

    if (!defined($minSizeKb)) {
       $minSizeKb = 0;
    }

    if( $self->needCacheUpdate('VMKDUMP_PARTITION', 1) ) {
       $self->{VMKDUMP_PARTITION} = findVmkDumpPartition($self->getVMSCSIDisks());
    }

    # Filter the list based on the size of the partitions
    my @partitions;
    foreach my $p (@{$self->{VMKDUMP_PARTITION}}) {
       if ($p->{SIZE} >= $minSizeKb) {
          push(@partitions, $p->{NAME});
       }
    }

    return \@partitions;
}

# Gets the number of CPUs
sub getNumCPUs() {
    my($self) = shift;

    if( $self->needCacheUpdate('NUM_CPUS', 1) ) {
	if( $self->isVMKernel() ) {
	    $self->{NUM_CPUS} = $self->findNumCPUs();
	} else {
	    return(0);
	}
    }
    return($self->{NUM_CPUS});
}


# Gets the number of logical CPUs
sub getNumLogCPUs() {
    my($self) = shift;

    if( $self->needCacheUpdate('NUM_LCPUS', 1) ) {
	if( $self->isVMKernel() ) {
	    $self->{NUM_LCPUS} = $self->findNumLogCPUs();
	} else {
	    return(0);
	}
    }
    return($self->{NUM_LCPUS});
}


# Checks if CPU is hyperthreading enabled.
sub isCPUHyperthreadingEnabled() {
    my $file = c_getFile('proc_cpuinfo');
    local *FD;

    if (! -e $file) {
        Warning("$file does not exist!");
	return 0;
    }
    
    if (!open (FD, "<" . $file)) {
	Warning("Cannot open $file.  Reason $!");
	return 0;
    }
    
    while (<FD>) {
	# Intel Conroe/Woodcrest processors have the ht flag, but aren't really.
	# So here we limit to just Intel processors in the Pentium 4 family.
	if (/^cpu family\s*:/ && !/:\s*15/) {
	    close(FD);
	    return 0;
	}

        if (/flags\s*:.* ht[ \n]/) {
	    close(FD);
	    return 1;
	}
    }

    close(FD);
    return 0;
}

# Accessor functions
sub getRunningVmmWorlds() {
    my($self) = shift;
    if( $self->needCacheUpdate('RUNNING_WORLDS', 1) ) {
	$self->{RUNNING_WORLDS} = $self->findRunningVmmWorlds($self->isVMKernel());
    }
    return($self->{RUNNING_WORLDS});
}

# Rescan luns
sub rescanLuns() {
  my ($self) = shift;
  my $target_name = shift;

  if (!defined($target_name)) {
    # Error : Invalid arguments
    Warning("No target name specified\n");
    return (0);
  }

  if (! c_getHelper('wwpn_list') ) {
    Warning('Cannot execute program "wwpn.pl".');
    return(0);
  }

  # do a special scan on all QLogic vmhbas
  if (System(c_getHelper('wwpn_list') . ' ' . '-s')) {
    # Error
    Warning("Failed to execute qlogic special rescan: Reason: $!");
    return (0);
  }

  if (! c_getHelper('lun_rescan') ) {
    Warning('Cannot execute program "vmkfstools".');
    return(0);
  }

  # Execute vmkfstools rescan
  if (System(c_getHelper('lun_rescan') . ' ' . shell_string($target_name))) {
    # Error
    Warning("Failed to execute rescan luns: Reason: $!");
    return (0);
  }

  # invalidate cached data
  $self->{MFS} = undef;
  $self->{VMK_SCSI_ADAPTERS} = undef;
  return (1);
}


# Persistent bindings support
sub persist_bindings() {
  my ($self) = shift;
  my $adapter = shift;

  if (!defined($adapter)) {
    # Error: Invalid arguments
    Warning("No adapter specified\n");
    return (0);
  }

  if (! c_getHelper('persist_adapter')) {
    Warning('Cannot execute program "pbind".');
    return (0);
  }

  # Execute persist bindings
  if (System(c_getHelper('persist_adapter') . ' ' . shell_string($adapter))) {
    # Error
    Warning("Failed to persist bindings: Reason: $!");
    return (0);
  }

  return (1);
}

# Determines if the VMnix is running.  Returns true if it is.
sub isVMnix() {
    my($self) = shift;
    return(_isVMnixKernel($self->getKernelVersion()));
}

# Determines if the VMnix Module is running.  Required to run the
# vmkernel
sub isVMnixModule() {
    my($self) = shift;
    local *FD;

    if( ! c_getHelper('lsmod') ) {
	Warning('Cannot execute program "lsmod".');
	return(0);
    }

    if( !open(FD, c_getHelper('lsmod') . ' 2>/dev/null |') ) {
	# Error
	Warning('Could not execute program ' . c_getHelper('lsmod'));
	return(0);
    }

    while( <FD> ) {
	if( /^vmnixmod\s+/ || /^vmnix\s+/ ) {
	    close(FD);
	    return(1);
	}
    }

    close(FD);

    return(0);
}

## Determines if the VMKernel is running.
#sub isVMKernel() {
#    my($self) = shift;
#
#    if( !$self->isVMnix() || !$self->isVMnixModule() ) {
#	# Cannot be running if the VMnix nor its module is not running
#	return(0);
#    }
#
#    local *FD;
#
#    if( ! c_getHelper('list_vmkmod') ) {
#	Warning('Cannot execute program "list_vmkmod".');
#	return(0);
#    }
#
#    if( !open(FD, c_getHelper('list_vmkmod') . " 2>/dev/null |") ) {
#	# Error
#	Warning('Could not execute program ' . c_getHelper('list_vmkmod'));
#	return(0);
#    }
#
#    # If the command returns any input on stdout, then the vmkernel is loaded.
#    # Perform a check to see if we see the words "Name" and "Length" which are
#    # a part of the heading for the output of the command.
#    while( <FD> ) {
#	if( /Name\s*Length/ ) {
#	    close(FD);
#	    return(1);
#	}
#    }
#
#    close(FD);
#    return(0);
#}

# Determines if the VMKernel is running.
sub isVMKernel() {
    my($self) = shift;

    if( !$self->isVMnix() || !$self->isVMnixModule() ) {
	# Cannot be running if the VMnix nor its module is not running
	return(0);
    }

    local *FD;

    my $dir1 = c_getDir('proc_vm');
    if( ! $dir1 ) {
	Warning('Cannot open directory "proc_vm".');
	return(0);
    }

    my $dir2 = c_getDir('cpu');
    if( ! $dir2 ) {
	Warning('Cannot open directory "cpu".');
	return(0);
    }

    return( -d $dir1 && -d $dir2 );
}


# Looks for the device on which a directory is mounted.
# Returns "" if none found
sub lookupMountDevice($) {
    my($self) = shift;
    my($dir) = @_;

    my ($dev, $type, $mnt) = $self->lookupMount($dir);

    return defined($dev) ? $dev : "";
}

# Looks for the device on which a directory is mounted.
# Returns (device, type, mnt) if found.
sub lookupMount($) {
    my($self) = shift;
    my($dir) = @_;

    my %map = %{$self->getDirMapping()};

    # Sort the directories by reverse length.  First one that matches completely
    # will be the one
    my @keys = sort { length($b) <=> length($a) } keys(%map);

    while( $#keys > -1 ) {
	my $k = shift(@keys);

	if( $dir =~ /^$k/ ) {
	    return($map{$k}->[0], $map{$k}->[1], $k);
	}
    }

    return undef;
}

# Perform checks to make sure all required modules are loaded
sub checkModules() {
    my($self) = shift;
    $self->checkVmnixModules();
    $self->checkVmkModules();
}

# Checks to make sure that all the required vmnix modules are loaded
sub checkVmnixModules() {
   my($self) = shift;
   my $rc = 1;

   Debug(1, "Entered");
   if (!$self->isVMnix()) {
      # No vmnix modules can be loaded if the vmnix is not running
      Debug(2, "Returning because the VMnix is not running");
      return($rc);
   }

   # Get the modules that should be loaded in the VMnix for shared devices.
   my @vmnixModules = $self->getPCIDeviceList('')->buildVmnixDriverList();

   # Get the modules that are loaded
   my @vmnixModLoaded = grep($_->{LOADED}, @{ $self->getVMnixModules() });
   my @vmnixModulesLoaded = map($_->{MODULE}, @vmnixModLoaded);

   while ($#vmnixModules > -1) {
      my $module = shift(@vmnixModules);
      if (!grep($module eq $_, @vmnixModulesLoaded)) {
         my $modconf = c_getFile('modules_conf');
         my $message =
            ('The ' . $module . ' module is not loaded into the Console OS.  You will not ' .
             'be able to use this device in the Console OS or by the virtual machines unless ' .
             'the Console OS loads this device.  To load this device either reallocate the ' .
             'devices on your machine or manually add the device to the ' . $modconf . ' file.  ' .
             'You may need to remake the initial RAM disk and reboot in order for these manual ' .
             'changes to take effect.');
         Warning($message);
         $rc = 0;
      }
   }

   if ($rc) {
      Debug(2, "All required vmnix modules are loaded.");
   } else {
      Debug(2, "Some required vmnix modules were not loaded.");
   }
   return($rc);
}


# Checks to make sure that all the vmkmodules are loaded
sub checkVmkModules() {
    my($self) = shift;
    my $rc = 1;

    Debug(1, "Entered");
    if( !$self->isVMKernel() ) {
	# No vmk modules can be loaded if the vmkernel is not running
	Debug(2, "Returning because the VMkernel is not running");
	return($rc);
    }

    # Get the modules that should be loaded
    my $driverList = $self->getPCIDeviceList('')->buildDriverList();
    my @vmkModules = $driverList->getDriverModules();

    # Get the modules that are loaded
    my @vmkModulesLoaded = map( $_->{MODULE}, @{ $self->getVMKernelModules() });

    @vmkModules = sort(@vmkModules);
    @vmkModulesLoaded = sort(@vmkModulesLoaded);

    while( $#vmkModules > -1 ) {
	my $module = shift(@vmkModules);

        # Get basename.  Fix for PR#21539.
        my $modulebase = $module;
        $modulebase =~ s/.*\///;
	$modulebase =~ s/\..*//;

	if (!grep($modulebase eq $_, @vmkModulesLoaded)) {
	    my $type = $driverList->getDriverType($module);
	    my $message = &{$cMissingModuleMessage{$type}}($module);
	    Warning($message);
	    $rc = 0;
	}
    }

    if( $rc ) {
	Debug(2, "All required vmk modules are loaded.");
    } else {
	Debug(2, "Some required vmk modules were not loaded.");
    }
    return($rc);
}

# Prints a string version to STDOUT or the file handle passed as a
# parameter to dump
sub dump {
    my($self) = shift;
    local *FD;
    *FD = *STDOUT;
    if( @_ ) {
	(*FD) = @_;
    }

    print FD $self->toString();
}


# Creates a string dump of the data
sub toString {
    my($self) = shift;
    my $s = '';

    my $vmnix_status = "not running";
    if( $self->isVMnix() ) {
	$vmnix_status = "running";
    }

    my $vmnixMod_status = "not loaded";
    if( $self->isVMnixModule() ) {
	$vmnixMod_status = "loaded";
    }

    my $vmkernel_status = "not loaded";
    if( $self->isVMKernel() ) {
	$vmkernel_status = "loaded";
    }

    my $manual_status = "no";
    if ($self->isManualConfig()) {
	$manual_status = "yes";
    }

    $s .= "Kernel version is '" . $self->getKernelVersion() . "'\n";
    $s .= "VMnix Kernel: $vmnix_status\n";
    $s .= "VMnix Module: $vmnixMod_status\n";
    $s .= "VMkernel: $vmkernel_status\n";
    $s .= "Current running configuration: " . $self->getCurrentLabel() . "\n";
    $s .= "Manual config: $manual_status\n";

    if( $self->isVMKernel() ) {
	my $numCPU = $self->getNumCPUs();
	$s .= "Number of CPUs: $numCPU\n";
    }

    my $totalMem = sprintf("%d", $self->getTotalMemSize() / 1024 );
    $s .= "Total Memory: " . p_number($totalMem) . " MB\n";

    if( $self->isVMnix() ) {
	my $vmnixMem = sprintf("%d", $self->getVmnixMemSize() / 1024 );
	$s .= "VMnix Memory: " . p_number($vmnixMem) . " MB\n";

	if( $self->isVMKernel() ) {
	    my $vmkMem = sprintf("%d", $self->getVmkMemSize() / 1024 );
	    $s .= "VMkernel Memory: " . p_number($vmkMem) . " MB\n";

	    my $availMem = sprintf("%d", $self->getVMAvailMemSize() / 1024 );
	    $s .= "VM Avail Memory: " . p_number($availMem) . " MB\n";
	}
    }

    $s .= "=== VMnix Modules Loaded ===\n";
    my ($vmnixMods) = $self->getVMnixModules();
    my $vmMod;
    foreach $vmMod (@$vmnixMods) {
	$s .= "module=" . $vmMod->{MODULE} . " size=" . $vmMod->{SIZE} .
	    " used=" . $vmMod->{USED} . "\n";
    }

    $s .= "=== VMkernel Modules Loaded ===\n";
    my ($vmkernelMods) = $self->getVMKernelModules();
    my $vmkMod;
    foreach $vmkMod (@$vmkernelMods) {
	$s .= "module=" . $vmkMod->{MODULE} .
	    " size=" . $vmkMod->{SIZE} . " id=" . $vmkMod->{ID} . "\n";
    }

    $s .= "=== PCI Devices found ===\n";
    my($devs) = $self->getPCIDeviceList('');
    $s .= $devs->toString() . "\n";

    $s .= "=== VMKernel Device Names ===\n";
    my ($names) = $self->getDeviceNames('vmnix');
    $s .= $names->toString();

    $s .= "=== Newly Named VMKernel Devices ===\n";
    my @namedDevs = $self->getNewlyNamedDevices();

    foreach (@namedDevs) {
        $s .= $_->getHash() . " " . $_->getName() . "\n";
    }

    $s .= "=== LILO Configuration ===\n";

    $s .= $self->getLILOConfig()->toString();

    $s .= "=== Boot Image Files Found ===\n";

    my (@files) = @{ $self->getBootImageFiles() };
    {
	my $f;
	foreach $f (@files) {
	    $s .= "$f\n";
	}
    }

    $s .= "=== Initrd Image Files Found ===\n";
    @files = @{ $self->getInitrdImageFiles() };
    {
	my $f;
	foreach $f (@files) {
	    $s .= "$f\n";
	}
    }

    $s .= "=== Directory Mappings Found ===\n";
    my($dm) = $self->getDirMapping();
    if(defined($dm)) {
	my %dmaps = %$dm;
	my $m;
        my @keys = sort { $a cmp $b } keys(%dmaps);

	foreach $m (@keys) {
	    my $dev = $dmaps{$m}->[0];
	    my $type = $dmaps{$m}->[1];
	    $s .= "Directory $m maps to device " . $dev . " and is a $type file system.\n";
	}
    }
    else {
	$s .= "No Directory Mappings found!\n";
    }

    if( $self->isVMnix() ) {
	$s .= "=== SCSI Adapters found (Owned by VM only) ===\n";

	my $adapters = $self->getVMSCSIAdapters();
	my $a;
	foreach $a (@$adapters) {
	    $s .= $a->toString() . "\n";
	}
    }


    if( $self->isVMnix() ) {
	$s .= "=== Net Adapters found (Owned by VM only) ===\n";

	my $adapters = $self->getVMNetAdapters();
	if(defined($adapters)) {
	    my $a;
	    foreach $a (@$adapters) {
               my ($bus, $dev, $func) = $self->getVMNetAdapterPCIInfo($a);
               my $clustering = $self->getVMNetAdapterClustering($a);
               my $speed = $self->getVMNetAdapterSpeed($a);
               my $nicteam = $self->getVMNetAdapterNICTeam($a);
               my @macs = @{ $self->getVMNetAdapterMAC($a) };
               my ($minCap, $hwCap, $maxCap) = $self->getVMNetAdapterCapabilities($a);
               
               $s .= "$a\n";
               $s .= "  PCI: $bus:$dev:$func\n";
               $s .= "  Clustering: $clustering\n";
               $s .= "  Speed: $speed\n";
               $s .= "  NIC Team: $nicteam\n";
               $s .= "  MAC: " . join(", ", @macs) ."\n";
               $s .= sprintf "  Capabilities: (min:hw:max) 0x%lx:0x%lx:0x%lx\n", $minCap, $hwCap, $maxCap;
	    }
	}
    }

    if( $self->isVMnix() ) {
	$s .= "=== Bond Adapters found ===\n";

	my $adapters = $self->getVMNetBondAdapters();
	if(defined($adapters)) {
	    my $a;
	    foreach $a (@$adapters) {
               my @slaves = @{ $self->getVMNetBondAdapterSlaves($a) };
               my @macs = @{ $self->getVMNetBondAdapterMAC($a) };

               $s .= "$a\n";
               $s .= "  NIC Team: " . join(", ", @slaves) ."\n";
               $s .= "  MAC: " . join(", ", @macs) ."\n";
	    }
	}
    }

    if( $self->isVMnix() ) {
	$s .= "=== TCP/IP Adapters found ===\n";
	my %tcpipInfo;

        $s .= "Configured values\n";
	%tcpipInfo = %{$self->getVMNetTcpipCfg()};
        foreach my $name (sort(keys(%tcpipInfo))) {
           my $ipAddr = $tcpipInfo{$name}->{IPADDR};
           my $gatewayAddr = $tcpipInfo{$name}->{GATEWAYADDR};
           my $nicName = $tcpipInfo{$name}->{NICNAME};

           $s .= "$name\n";
           $s .= "  IP Address: $ipAddr\n";
           $s .= "  Gateway Address: $gatewayAddr\n";
           $s .= "  NIC Name: $nicName\n";
	}
        $s .= "\n";

        $s .= "Actual configuration\n";
	%tcpipInfo = %{$self->getVMNetTcpip()};
        foreach my $name (sort(keys(%tcpipInfo))) {
           my $ipAddr = $tcpipInfo{$name}->{IPADDR};
           my $gatewayAddr = $tcpipInfo{$name}->{GATEWAYADDR};
           my $nicName = $tcpipInfo{$name}->{NICNAME};

           $s .= "$name\n";
           $s .= "  IP Address: $ipAddr\n";
           $s .= "  Gateway Address: $gatewayAddr\n";
           $s .= "  NIC Name: $nicName\n";
	}
        $s .= "\n";

        my $sync = $self->isVMNetTcpipSync();
        $s .= ("TCP/IP actual and configured settings are " .
               ($sync ? "" : "NOT ") . "synchronized.\n");
    }

    if( $self->isVMnix() ) {
	$s .= "=== MFS File System Information ===\n";

	my $all_mfs = $self->getAllMFS();
	if(defined($all_mfs)) {
	    my $fs;
	    foreach $fs (@$all_mfs) {
		$s .= $fs->toString();
	    }
	}
    }

    $s .= "=== VmkDump Partitions ===\n";

    my $vmkdumpPart = $self->getVmkDumpPartitions();
    my $part;
    foreach $part (@$vmkdumpPart) {
	# Print FD out the fs information
	$s .= "Partition " . $part . "\n";
    }

    $s .= "=== Config DB ===\n";

    my $configDB = $self->getConfigDB();
    my $key;
    foreach $key ($configDB->getKeys()) {
	my $value = $configDB->get($key);
	$s .= "$key => '$value'\n";
    }

    $s .= "=== Running VMM Worlds ===\n";
    my $worlds = $self->getRunningVmmWorlds();
    if( defined($worlds) ) {
	$s .= join(", ", @$worlds) . "\n";
    }

    return($s);
}

# Add a vmnic device as a slave for a bond device
sub addBondSlave($$) {
   my($self) = shift;
   my $bond = shift;
   my $vmnic = shift;

   if (!$self->addRemoveBondSlave($bond, $vmnic, 0)) {
      return 0;
   }

   if (!$self->addRemoveBondSlaveConfigDB($bond, $vmnic, 0)) {
      Warning("Failed to persistently bond $vmnic to $bond.  Setting may not " .
              "take effect on next boot.\n");
   }

   return 1;
}

# Remove a vmnic device from a NIC team determined by the bond
sub removeBondSlave($$) {
   my($self) = shift;
   my $bond = shift;
   my $vmnic = shift;

   if (!$self->addRemoveBondSlave($bond, $vmnic, 1)) {
      return 0;
   }

   if (!$self->addRemoveBondSlaveConfigDB($bond, $vmnic, 1)) {
      Warning("Failed to persistently unbond $vmnic from $bond.  Setting may not " .
              "take effect on next boot.\n");
   }

   return 1;
}

# Remove a vmnic device from a NIC team determined by the bond
sub addRemoveBondSlave($$$) {
   my($self) = shift;
   my $bond = shift;
   my $vmnic = shift;
   my $remove = shift;

   # Make sure that the vmnic and bond devices are valid
   my @bonds = @{ $self->getVMNetBondAdapters() };
   my @vmnics = @{ $self->getVMNetAdapters() };

   if (!grep($bond eq $_, @bonds)) {
      Warning("Cannot bond $vmnic to $bond because $bond does not exist.\n");
      return 0;
   }
   if (!grep($vmnic eq $_, @vmnics)) {
      Warning("Cannot bond $vmnic to $bond because $vmnic does not exist.\n");
      return 0;
   }

   my $nicteam = $self->getVMNetAdapterNICTeam($vmnic);
   if ($remove && ($nicteam eq "" || $nicteam ne $bond)) {
      Warning("Cannot unbond $vmnic from $bond because it is not bonded to $bond.");
      return 0;
   } elsif (!$remove && $nicteam ne "") {
      Warning("Cannot bond $vmnic to $bond because it is already bonded to $nicteam.");
      return 0;
   }

   # A vmnic that is in use cannot be bonded although a vmnic can be added
   # to a bond group if the bond is in use.
   #
   # A bond that is in use cannot have a slave taken away.
   my @macs;
   if (!$remove) {
      @macs = @{ $self->getVMNetAdapterMAC($vmnic) };
   } else {
      @macs = @{ $self->getVMNetBondAdapterMAC($bond) };
   }

   if (scalar(@macs)) {
      if (!$remove) {
         Warning("Cannot bond $vmnic to $bond because $vmnic is in use.\n");
      } else {
         Warning("Cannot unbond $vmnic from $bond because $bond is in use.\n");
      }
      return 0;
   }

   my $pvn_dir = c_getDir('proc_vmware_net');
   if (! -d $pvn_dir) {
      # Error /proc/vmware/net is not a directory
      Warning("Could not open $pvn_dir. Reason: $!");
      return 0;
   }

   my $bondfile = "$pvn_dir/$bond/config";

   local *FD;
   if (!open(FD, "> $bondfile")) {
      Warning("Could not open bond file $bondfile.\n");
      return 0;
   }

   print FD "nicteaming " . (!$remove ? "add" : "delete") . " $vmnic\n";
   close(FD);

   # Need to force a flush of network information
   $self->flushCacheItem('VMK_NET_BOND_ADAPTERS');
   $self->flushCacheItem('VMK_NET_ADAPTERS');

   return 1;
}

# Store NIC teaming settings in the hardware config DB to make it persistent.
sub addRemoveBondSlaveConfigDB($$$) {
   my($self) = shift;
   my $bond = shift;
   my $vmnic = shift;
   my $remove = shift;

   my $configDB = $self->getConfigDB();
   my $key = "nicteam.$vmnic.team";

   if (!$remove) {
      $configDB->set($key, $bond);
   } else {
      $configDB->unset($key);
   }

   return $self->commitConfigDB();
}

# Given a LILOConfig object, writes out to lilo.conf
sub writeLILOConfig($) {
    my($self) = shift;
    my $lilo = shift;
    local *LILO;

    if( !UNIVERSAL::isa($lilo,"VMware::HConfig::LILOConfig") ) {
	Warning("Arg is not of LILOConfig type object");
	return(0);
    }

    # Write to temporary lilo.conf
    my $tmpfile = mktmpfile();
    if( !open(LILO, "> $tmpfile") ) {
	# Error opening file
	return(0);
    }

    print LILO $lilo->toString();

    close(LILO);

    # Change lilo.conf
    if( !safe_move($tmpfile, c_getFile('lilo.conf')) ) {
	# Error
	Warning("Warning: Unable to create new lilo.conf file.  Reason $!");
	return(0);
    }

    return(1);
}


# Executes LILO on the currently running system.
sub executeLILO() {
    if( ! c_getHelper('lilo') ) {
	Warning('Unable to execute "lilo".');
	return(0);
    }

    # Execute LILO
    if( System(c_getHelper('lilo')) ) {
	# Error
	Warning("Warning: Unable to execute lilo.  Reason $!");
	return(0);
    }
    return(1);
}


# Writes the input text to the SNMP configuration file
sub writeSNMPConfig($$) {
    my($self) = shift;
    my($snmp, $enable) = @_;

    # Write to temporary snmpd.conf
    my $tmpfile = mktmpfile();
    if( !open(SNMP, "> $tmpfile") ) {
	# Error opening file
	Warning("Warning: Unable to create new snmpd.conf file.");
	return(0);
    }
    print SNMP $snmp;
    close(SNMP);

    # Change snmpd.conf
    if( !safe_move($tmpfile, c_getFile('snmpd.conf')) ) {
	# Error
	Warning("Warning: Unable to create new snmpd.conf file.");
	return(0);
    }

    # Enable/disable SNMP if appropriate
    if (defined($enable) && ($enable ne '')) {
        # Configure SNMP daemon
        if ($enable) {
            VMware::HConfig::RunlevelControl::runLevelStart("vmware-snmpd");
        } else {
            VMware::HConfig::RunlevelControl::runLevelStop("vmware-snmpd");
        }

        # Configure SNMP trap generation
        my $filename = c_getFile('host_config');
        if (!defined($filename)) {
            Warning("Warning: Unable to open host configuration file.");
            return(0);
        }
        my $config = VMware::Config->new();
        if (!$config->readin($filename)) {
            Warning("Warning: Unable to read host configuration file '$filename'.");
            return(0);
        }
        $config->set("snmp.enable", $enable);
        if (!$config->writeout($filename)) {
            Warning("Warning: Unable to write host configuration file '$filename'.");
            return(0);
        }
    }

    return(1);
}


# Can we run pcidivy?
sub checkCanRunPcidivy() {
    my($self) = shift;
    if ($self->isManualConfig()) {
	return(0,
	       "The system is currently configured to load in modules manually from the " .
	       c_getFile("vmkmodule") . " file.  In order to edit this configuration, " .
	       "you should turn off manual configuration by removing the line MANUAL-CONFIG " .
	       "from the file.");
    }
    return (1, undef);
}

# Determines if manual config is set for the system.
sub isManualConfig() {
    my($self) = shift;
    local *F;

    # Look for the words MANUAL-CONFIG in the vmkmodule file.  This is used to
    # toggle manual configuration.
    if (open(F, "< " . c_getFile('vmkmodule'))) {
	while (<F>) {
	    if (/MANUAL-CONFIG/) {
		close(F);
		return(1);
	    }
	}
	close(F);
    }
    return(0);
}

# Write to /etc/vmkmodules.conf.  Takes a label corresponding to the PCI Device
# list to write out.
sub writeVmkModules($) {
    my($self) = shift;
    my($label) = @_;

    # Don't overwrite vmkmodule file if the existing one has a line
    # containing 'MANUAL-CONFIG' in it.
    if ($self->isManualConfig()) {
	Warning("Cannot write because MANUAL-CONFIG is specified in " .
		c_getFile('vmkmodule'));
	return(0);
    }

    if( !$self->getLILOConfig()->getBootImageByLabel($label) ) {
	Warning("No such label '$label' specified.");
	return(0);
    }

    my $PCIDeviceList = $self->getPCIDeviceList($label);

    # Format for driver list:
    # $driverList is a reference to an array of hashes such that the following is true:
    # $driverList->[n]->{DRIVER} is the object file. ie. acenic.o
    # $driverList->[n]->{NAME} is the name of the driver. ie. vmnic
    # $driverList->[n]->{SHARED} is the parameter to add for sharing

    my $driverList = $PCIDeviceList->buildDriverList();

    my $tmpfile = mktmpfile();
    local *WFD;
    if( !defined($tmpfile) || !open(WFD, "> $tmpfile") ) {
	Warning("Could not open file $tmpfile for writing.  Reason: $!");
	return(0);
    }

    print WFD $driverList->toVmkLoadmodString();
    close(WFD);

    # Replace the previous file
    if( !safe_move($tmpfile, c_getFile('vmkmodule'))) {
	Warning("Warning: Unable to replace vmkmodules file.  Reason $!");
	return(0);
    }

    return(1);
}

# Write a line into the modules.conf file.
sub writeModulesConfLine($) {
   my($self) = shift;
   my($line) = @_;

   my $file = c_getFile('modules_conf');
   if (!$file) {
      Warning("Failed to get modules_conf file.");
      return 0;
   }

   local *WFD;
   if (!open(WFD, ">> $file")) {
      Warning("Failed to append to file $file.  Reason $!");
      return 0;
   }

   if ($line !~ /\n$/) {
      $line .= "\n";
   }

   # Print an extra newline before just in case there isn't one at the
   # end of the file.
   print WFD "\n" .  $line;
   close(WFD);

   return 1;
}

# Run the mkinitrd command on a file and a kernel version.
sub makeInitrdFile($$) {
   my($self) = shift;
   my($kernelver, $initrdFile) = @_;

   my $cmd = c_getHelper('mkinitrd');
   if (!$cmd) {
      Warning("Could not get mkinitrd program.");
      return 0;
   }

   # If $initrdFile was not supplied, then generate one based on the
   # kernel version.
   if (!defined($initrdFile)) {
      my $bootDir = c_getDir('boot');
      if( ! -d $bootDir ) {
         # Error /boot is not a directory
         Warning('Could not open directory ' . $bootDir . '.');
         return 0;
      }
      $initrdFile = $bootDir . '/initrd-' . $kernelver . '.img';
   }

   # XXX The makeinitrd program on RedHat 7.2 actually doesn't always
   # report when it fails.  The program is actually a script that doesn't
   # check exit codes very thoroughly. -jhu
   if (System($cmd . ' ' . shell_string($initrdFile) . ' ' . shell_string($kernelver))) {
      Warning("Execution of $cmd failed.  Reason: $!");
      return 0;
   }

   return (1, $initrdFile);
}


# Commit the database.
sub commitConfigDB() {
    my($self) = shift;

    Debug(1, "Entered");

    # Perform a cleanup on the config database
    if( $self->cleanupPCIDevicesConfigDB() > 0 ) {
	LOG(0, "PCIDevice entries in config db were cleaned.");
    }

    # The configuration database has probably changed since we looked for the devices
    if( $self->getConfigDB()->hasChanged() ) {
	return($self->getConfigDB()->save());
    }
    return(1);
}

# Reboot the machine.
sub rebootSystem() {
    my $self = shift;
    my $message = shift;
    return $self->shutdownSystem('reboot', $message);
}

# Halt the machine.
sub haltSystem() {
    my $self = shift;
    my $message = shift;
    return $self->shutdownSystem('halt', $message);
}

#####################################################################
# Methods not intended for use by user                              #
#####################################################################

#####################################################################
# Reboot/Halt the machine: invoked by rebootSystem/haltSystem       #
#####################################################################

sub shutdownSystem() {
    my $self = shift;
    my ($op, $message) = @_;

    my $cmd = c_getHelper($op);
    if( !$cmd ) {
	Warning("Could not $op system: program not found.");
	return(0);
    }

    # Write message to the log file
    if (defined($message) && ($message ne '')) {
       $message = "UserAction $op: $message";
    } else {
       $message = "UserAction $op";
    }

    my $helper = c_getHelper('logger');
    if(!$helper || System($helper . ' "(' . time() . ')" ' . shell_string($message))) {
	Warning("Could not execute 'logger' -- " .
                "reason for $op not written to log.");
    }

    # Set STDOUT to be flushed on write before the fork to
    # avoid duplicating the output to STDOUT.
    $| = 1;

    my $pid = fork();
    if( $pid > 0 ) {
	# Parent process returns HTML to user
	Debug(1, $cmd . ' > /dev/null');
	return(1);
    } elsif( $pid == 0 ) {
	# Child process sleeps and does the reboot
	if( !chdir('/') ) {
	    Warning("Could not chdir to /.  Reason: $!");
	}

	umask 0;
	close(STDIN);
	if( !open(STDIN, '< /dev/null') ) {
	    Warning("Could not read /dev/null.  Reason: $!\n");
	}
	close(STDOUT);
	if( !open(STDOUT, '> /dev/null' ) ) {
	    Warning("Could not write to /dev/null. Reason: $!\n");
	}
	if( !setsid() ) {
	    Warning("Could not call setsid.  Reason: $!\n");
	}
	close(STDERR);
	if( !open(STDERR, '> /dev/null') ) {
	    Warning("Could not write to /dev/null.  Reason: $!\n");
	}

	# Sleep for 3 seconds and then reboot
	sleep(3);

	# If the installer is running (Will have BOOT in its kernel name)
	if ($self->getKernelInfo() =~ m/BOOT/) {
	    system("killall -USR1 loader");
	}

	# Wrap in a block statement to silence warnings from the interpreter
	{
	    exec($cmd);
	}
	Warning("Could not $op system.  Reason: $!");
	return(0);
    } else {
	Warning("Could not $op system.  Reason: $!");
	return(0);
    }
    return(1);
}

#####################################################################
# Caching Support Functions                                         #
#####################################################################

# Get all the caching options
sub getCacheOptions() {
    my($self) = shift;
    return($self->{CACHE_OPTIONS});
}

# Get default caching options
sub getDefaultCacheOptions() {
    my($self) = shift;
    return($self->{CACHE_OPTIONS_DEFAULT});
}

# Helper function that queries if the option is cached once
sub isCachedNever($) {
    my($self) = shift;
    my($key) = @_;
    return( $self->getCacheOption($key) == CACHE_NEVER );
}

# Get a single caching option by name
sub getCacheOption($) {
    my($self) = shift;
    my($key) = @_;
    if( defined($self->getCacheOptions()->{$key}) ) {
	return($self->getCacheOptions()->{$key});
    }
    Warning("No cache options set for item $key.  Using CACHE_NEVER as default caching policy");
    return(CACHE_NEVER);
}

# Sets the new caching options
sub setCacheOptions($) {
    my($self) = shift;
    my($cache_options) = @_;
    $self->{CACHE_OPTIONS} = $cache_options;
}

# Set a single caching option by name
sub setCacheOption($$) {
    my($self) = shift;
    my($key, $value) = @_;
    if( $value == CACHE_NEVER || $value == CACHE_ONCE || $value == CACHE_CALLBACK ) {
	$self->getCacheOptions()->{$key} = $value;
    } else {
	Warning("Invalid cache option '$value' passed in for cache key $key");
    }
}

# Sets the system to use the default cache options
sub useDefaultCacheOptions() {
    my($self) = shift;
    $self->setCacheOptions($self->getDefaultCacheOptions);
}

# Sets the system to never cache
sub useNoCacheOptions() {
    my($self) = shift;
    my @cache_options = keys(%{ $self->getCacheOptions() });
    my $cache_option;
    foreach $cache_option (@cache_options) {
	$self->setCacheOption($cache_option, CACHE_NEVER);
    }
}

# Sets the system to cache and never invalidate
sub useOnceCacheOptions() {
    my($self) = shift;
    my @cache_options = keys(%{ $self->getCacheOptions() });
    my $cache_option;
    foreach $cache_option (@cache_options) {
	$self->setCacheOption($cache_option, CACHE_ONCE);
    }
}

# Flush only the ones as indicated in the callback
# XXX misnomer: Should be invalidate -jhu
sub flushCacheCallback {
    my($self) = shift;
    my $cache_option;
    foreach $cache_option (keys(%{ $self->getCacheOptions() })) {
	if( $self->getCacheOption($cache_option) == CACHE_CALLBACK ) {
	    # Invalidate the cached items
	    $self->{$cache_option} = undef;

	    # If the cached item uses a source, flush the source timestamp as well
	    if( defined($self->{CACHE_SOURCE}->{$cache_option}) ) {
		$self->{CACHE_SOURCE}->{$cache_option}->[1] = 0;
	    }
	}
    }
}

# Flush the cache item
sub flushCacheItem {
    my($self, $cache_option) = @_;
    # Invalidate the cached items
    $self->{$cache_option} = undef;
}

# Flush the entire cache explicitly
sub flushCache {
    my($self) = shift;
    my $cache_option;
    foreach $cache_option (keys(%{ $self->getCacheOptions() })) {
	# Invalidate the cached items
	$self->{$cache_option} = undef;
    }
}

# Check the cache to see if it needs to be updated.  This function
# stats files defined in CACHE_SOURCE to see if a change has occurred.
#
# If the second argument is not equal to 0, it will assume that the
# information is going to be updated and modify the cache metadata
# accordingly.
sub needCacheUpdate($$) {
    my($self) = shift;
    my($cache_option, $update) = @_;

    if( $self->isCachedNever($cache_option) ) {
	Debug(4, "Cache update needed for $cache_option because it is never cached.");
#	Warning("Cache update needed for $cache_option because it is never cached.");
	return(1);
    }

    # If the item is not in the cache, then return true.
    if( !defined($self->{$cache_option}) ) {
	# Set the timestamp for the source if applicable
	if( defined($self->{CACHE_SOURCE}->{$cache_option}) ) {
	    if( $update ) {
		my $newtime = $self->getCacheSrcTime($cache_option);
		$self->{CACHE_SOURCE}->{$cache_option}->[1] = $newtime;
	    }
	}
	Debug(4, "Cache update needed for $cache_option because it is not defined.");
#	Warning("Cache update needed for $cache_option because it is not defined.");
	return(1);
    }

    # Check the stat time
    if( defined($self->{CACHE_SOURCE}->{$cache_option}) ) {
	my $oldtime = $self->{CACHE_SOURCE}->{$cache_option}->[1];
	my $newtime = $self->getCacheSrcTime($cache_option);
	if( $update ) {
	    $self->{CACHE_SOURCE}->{$cache_option}->[1] = $newtime;
	}

	if( !defined($oldtime) ) {
	    Debug(4, "Cache update needed for $cache_option because timestamp is not defined.");
#	    Warning("Cache update needed for $cache_option because timestamp is not defined.");
	    return(1);
	}

	if( $newtime != $oldtime ) {
	    # The modified time has changed.
	    Debug(4, "Cache update needed for $cache_option because time change detected.");
#	    Warning("Cache update needed for $cache_option because time change detected.");
	    return(1);
	}
    }

    return(0);
}

# Updates the time of the cached item
sub updateCacheSrcTime($) {
    my($self) = shift;
    my($cache_option) = @_;
    if( defined($self->{CACHE_SOURCE}->{$cache_option}) ) {
	my $newtime = $self->getCacheSrcTime($cache_option);
	$self->{CACHE_SOURCE}->{$cache_option}->[1] = $newtime;
    }
}

# Returns the new time
sub getCacheSrcTime($) {
    my($self) = shift;
    my($cache_option) = @_;

    if( defined($self->{CACHE_SOURCE}->{$cache_option}) ) {
	my $src = $self->{CACHE_SOURCE}->{$cache_option}->[0];

	my @stats = stat($src);
	if( $#stats == -1 || !defined($stats[9]) ) {
	    return(0);
	}
	my $mtime = $stats[9];
	return($mtime);
    }
    return(0);
}

#####################################################################
# End of Caching Support Functions                                  #
#####################################################################

#####################################################################
# Missing Module Error Message Functions                            #
#####################################################################

sub MissingModule_linux($) {
    my($module) = @_;

    return('Missing ' . $module . ' module.  This driver must be loaded in order ' .
	   'to load other modules.  If you wish to load this module by hand, you ' .
	   'can type "' . c_getHelper('load_vmkmod') . ' -e ' . c_getDir('vmkmodule') .
	   '/' . $module . ' linux" at the shell prompt.');
}

sub MissingModule_vmhba($) {
    my($module) = @_;

    return('Missing ' . $module . ' module.  This driver is needed to load a SCSI or RAID adapter.  '.
	   'You may not be able to use some disks until you load this module.  If you wish to' .
	   ' load this module by hand, you can type "' . c_getHelper('load_vmkmod') . ' ' .
	   c_getDir('vmkmodule') . '/' . $module . ' ' . $gVmkDeviceType_hba .
	   '" at the shell prompt.');
}

sub MissingModule_vmnic($) {
    my($module) = @_;

    return('Missing ' . $module . ' module.  This driver is needed to load an ethernet adapter.  '.
	   'You may not be able to use some ethernet adapters for your virtual machines ' .
	   'until you load this module.  If you wish to load this module by hand, you can type "' .
	   c_getHelper('load_vmkmod') . ' ' . c_getDir('vmkmodule') . '/' . $module . ' ' .
	   $gVmkDeviceType_nic . '" at the '.
	   'shell prompt.');
}

sub MissingModule_nfshaper($) {
    my($module) = @_;

    return('Missing ' . $module . ' module.  This driver is used to control the network usage of '.
	   'your system.  You may not be able to enforce certain resource management policies ' .
	   'without this module.  If you wish to load this module by hand, you can type "' .
	   c_getHelper('load_vmkmod') . ' ' . c_getDir('vmkmodule') . '/' . $module . ' nfshaper" at the '.
	   'shell prompt.');
}


#####################################################################
# Device Naming functions                                           #
#####################################################################

# SSHost::findDeviceNames() --
#  Retrieves the device names associated with each PCI device, by
#  reading from the config DB.  If no device names are found,
#  creates them by calling createCompatibleNames(), which preserves
#  the old device naming.  If any PCI devices are not named, 
#  name them automatically, and create a list of these newly-named
#  devices, so that the names can be reviewed by the user.
#
# Args: ($label)  - boot label / configuration to look up 
#                   PCI devices.
#
# Returns: ($deviceNames, @newlyNamedDevices)
#
sub findDeviceNames($) {
    my($self) = shift;
    my $label = shift;
    my $PCIDeviceList = $self->getPCIDeviceList($label);
    my $config_db = $self->getConfigDB();

    ## Get device names from config DB
    my $devNames = VMware::HConfig::DevNames->new();
    my $foundNames = $devNames->findNamesInConfigDB($config_db, "");

    ## go through PCI device list and find vmkernel-owned or shared
    ## devices that are unnamed.  These are newly-added devices.
    my $matchFunc = sub { return $_[0]->getOwner() != VMware::HConfig::PCIDevice::HOST
                              && defined($_[0]->getDrivers()->[0])
                              && !$devNames->getDeviceName($_[0]->getHash(),
                                                          $_[0]->getEquiv()) };
    my $matchedDevList = $PCIDeviceList->filterPCIDevices($matchFunc);
    my @unnamedDevices = $matchedDevList->getPCIDevices();

    if ($foundNames) {
        # Go through and suggest a name for unnamed devices
        my $dev;
        foreach $dev (@unnamedDevices) {
            my $newName = $devNames->suggestDeviceName($dev);
            Warning("Assigning name $newName to new device ",
                    $dev->getName(), " (", $dev->getHash(), ")");
            $devNames->setDeviceName($dev->getHash(), $dev->getEquiv(),
                                     $newName);
        }
    } else {
        Warning("No device names found in config DB, creating new ones");

        # this will create a name for every PCI device assigned to
        # the vmkernel.  So all these devices are newly named.
        # This is needed to create backwards-compatible device names
        # the first time through.
        $devNames->createCompatibleNames($PCIDeviceList);
    }

    return ($devNames, $matchedDevList);
}


# HConfig::SSHost::writeDevNamesConfig() --
#   Writes out the device names of VMKernel-owned PCI devices into
#   "devnames.conf", for use with vmkchdev in order to set VMK
#   device names.  Given the device names and PCI device list
#   objects.
#
# Args: ($devNames, $pciDevList)
#
# Returns: 0 if error, 1 upon success.
#
sub writeDevNamesConfig($$) {
    my $self = shift;
    my ($devNames, $PCIDeviceList) = @_;

    if( !UNIVERSAL::isa($devNames,"VMware::HConfig::DevNames") ) {
	Warning("Arg is not of DevNames type object");
	return(0);
    }

    my $fname = c_getFile('devnames_conf');
    my $rc = internal_echo($fname, $devNames->deviceNameConfToString($PCIDeviceList) );
    if (!$rc) {
        Warning("Unable to create $fname, reason $!");
    }
    return $rc;
}


#####################################################################
# PCI Device Handling Functions                                     #
#####################################################################
# Top level find PCI device function.
#
# label is the label for which the device list should be found
# devicemap is the PCIDeviceMap used by the devices to locate
#       their drivers.
# isVmnix is a flag indicating if VMnix is the console OS
# config_db is the configuration database
# lilo is the LILOConfig object for the configuration
#
# Guaranteed to return a PCIDeviceList object containing a list of
# devices.
sub findPCIDevices($) {
    my($self) = shift;
    my ($label) = @_;
    Debug(1, "Entered");
    my $isVmnix = $self->isVMnix();
    my $config_db = $self->getConfigDB();
    my $lilo = $self->getLILOConfig();
    my $deviceMap = $self->getPCIDeviceMap();

    my $PCIDeviceList = $self->findPCIDevicesHelper($label, $isVmnix, $config_db, $lilo);

    if( !$PCIDeviceList ) {
	# XXX This code doesn't actually occur because the helper function
	#     already Panics if it can't find any devices.
	Panic("Could not locate any PCI devices on the system");
    }

    # Fill in the driver info for the given PCI device list.
    $self->findPCIDeviceDrivers($PCIDeviceList, $deviceMap);

    Debug(1, "Exited");
    return($PCIDeviceList);
}


# Helper function that detects what PCI devices exist.
# Returns a list of PCI devices on success.
# Label is defined.
sub findPCIDevicesHelper($$) {
    my($self) = shift;
    my ($label, $isVmnix, $config_db, $lilo) = @_;
    my $PCIDeviceList;
    my $FullPCIDeviceList;

    Debug(1, "Entered looking for PCI devices for label '$label'");

    if( !$isVmnix ) {
	$FullPCIDeviceList = $self->findHostPCIDevices();
    } else {
	# Probe for devices so we can see if the hardware settings have changed.
	# XXX Probing introduces the possibility of a race and should only
	#     be done if the user is warned and agrees. -jhu
	$FullPCIDeviceList = $self->findAllPCIDevicesProbe();
    }

    # Find the devices in the ConfigDB and match the owner for the
    # given label.
    $PCIDeviceList = $self->findPCIDeviceListInConfigDB($config_db, 1, $label);

    if( !defined($FullPCIDeviceList) ) {
	if( !defined($PCIDeviceList) ) {
	    # We couldn't probe for the devices nor get the list from the database.
	    # This is bad.  No choice but to panic.
	    Panic("Could not detect PCI devices and could not find any prior " .
		  "configuration.  Cannot proceed because nothing about the PCI " .
		  "devices could be determined.");
	} else {
	    # We couldn't probe but have prior list of PCI devices, use this list
	    # and warn the user that we were unable to check for any new devices.
	    Warning("Could not detect PCI devices, but a prior configuration " .
		    "was found in the database.  Could not determine if any new " .
		    "PCI devices were added to the machine.  However, if nothing " .
		    "has changed on the machine, this does not matter.");
	    return($PCIDeviceList);
	}
    }

    # We were successful in detecting a list of PCI devices at this point

    if( !defined($PCIDeviceList) ) {
	# Failed to get what we needed from the database file.  We're going to
	# use what we found from the probe and use the LILO entry to find
	# defaults that make sense.  If we are booted into a VMnix, we can
	# do even better and actually look at what PCI devices the VMnix
	# knows about to see what it should be.

	Debug(1, "Could not find a prior configuration.  Attempting to generate " .
	      "information about the PCI devices by detecting the current " .
	      "configuration information.");

	if( !$self->regenerateConfigDB($config_db, $FullPCIDeviceList, $lilo) ) {
	    Warning('Unable to regenerate configuration database.');
	    return(undef);
	}

	# Now use what we found in the newly generated configuration database
	$PCIDeviceList = $self->findPCIDeviceListInConfigDB($config_db, 1, $label);

	# No need to save the configuration since we just created a new one
	# during the regeneration process.

	return($PCIDeviceList);
    }

    # We were successful in probing and finding in the database the PCI device
    # information at this point.

    if( !$PCIDeviceList->equals($FullPCIDeviceList)  ) {
	Warning("The PCI device information has changed.  Attempting to reconcile the changes.");

	# A change has occurred and we must perform the necessary update operations
	# on the config database to keep it valid.  See the reconcile function for
	# what these operations are.
	if( !$self->reconcileConfigDB($config_db, $FullPCIDeviceList, $lilo) ) {
	    Warning("Failed to reconcile the PCI device information.");
	    return(0);
	}

	$PCIDeviceList = $self->findPCIDeviceListInConfigDB($config_db, 1, $label);

	# No need to save the configuration since we just updated it
	# during the regeneration process.

	Debug(2, "The config database had to be updated because a change in the " .
	      "hardware configuration was detected.");

	return($PCIDeviceList);
    }

    # We have a PCI list that is good to go.
    return($PCIDeviceList);
}


# Builds a PCIDeviceList contining all the the devices found through
# probing.
sub findAllPCIDevicesProbe() {
    my($self) = shift;
    my $PCIDeviceList;
    Debug(1, "Entered");

    my $PCIDeviceList_Probe1 = parse_probe_lspci();
    my $PCIDeviceList_Probe2 = parse_probe_lspci2();

    if( $PCIDeviceList_Probe1 && $PCIDeviceList_Probe2 ) {
	$PCIDeviceList = ($PCIDeviceList_Probe1->count() > $PCIDeviceList_Probe2->count() ?
			  $PCIDeviceList_Probe1 :
			  $PCIDeviceList_Probe2);
    } elsif( !$PCIDeviceList_Probe1 ) {
	$PCIDeviceList = $PCIDeviceList_Probe2;
    } else {
	$PCIDeviceList = $PCIDeviceList_Probe1;
    }

    return($PCIDeviceList);
}

# Builds a PCIDeviceList consisting of all the devices that the host
# sees.  If the host is Linux, this list is a complete list.  If the
# host is VMnix, we have to use another method.
sub findHostPCIDevices() {
    my($self) = shift;
    my $PCIDeviceList;

    Debug(1, "Entered");

    $PCIDeviceList = parse_safe_lspci();
    if( $PCIDeviceList ) {
	return($PCIDeviceList);
    }

    $PCIDeviceList = parse_proc_bus_pci();
    if( $PCIDeviceList ) {
	return($PCIDeviceList);
    }

    # Ok, we're desperate.  None of the current methods at finding PCI
    # devices work.  We're going to check a deprecated interface that doesn't
    # always give us all the info we need.
    $PCIDeviceList = parse_proc_file();
    if( $PCIDeviceList ) {
	return($PCIDeviceList);
    }

    Panic('Could not find host PCI devices.  Cannot run lspci or read the file ' .
	  c_getFile('proc_bus_pci') . '.');
}

# Parses the /proc/pci file extracting PCI information.
# % parseProcFile(ref to hash)
#
# Errors:
# 1. Cannot open /proc/pci file
# XXX This function is deprecated.
sub parse_proc_file() {
    local *FD;
    my $PCIDeviceList = VMware::HConfig::PCIDeviceList->new();
    my $dev = undef;     # String for parsing device.  undef if looking for new device

    Debug(1, "Entered");

    # First read the /proc/pci file to get the PCI information
    if( !open(FD, '< ' . c_getFile('proc_pci')) ) {
	Warning('Could not read file ' . c_getFile('proc_pci'));
	return(undef);
    }

    # Processing loop for /proc/pci
    while(<FD>) {
	chomp();      # Remove trailing newlines

	if( /Bus[ ]*([0-9]+),[ ]*device[ ]*([0-9]+),[ ]*function[ ]*([0-9]+):/ ) {
	    my $hash = VMware::HConfig::PCIDevice::hash($1, $2, $3);

	    # Check for existence of this device in the list.  If the device does
	    # not exist, add it.
	    if( !$PCIDeviceList->{$hash} ) {
		# New device.  Add it.
		$dev = VMware::HConfig::PCIDevice->new($1, $2, $3);
		$PCIDeviceList->{$hash} = $dev;
	    }
	    else {
		$dev = $PCIDeviceList->{$hash};

		# If it's in the /proc/pci file, the host has it
		$dev->setOwner(VMware::HConfig::PCIDevice::HOST);
	    }

	    # The name of the device always follows the identification of the device's
	    # bus number, device number, and function number
            my $name = <FD>;
            $name =~ s/^\s*(.+)\s*$/$1/;          # Remove trailing and leading spaces
	    $dev->setName($name);

	    # Remaining lines contain other potentially useful information.  That
	    # information will be handled in the next iterations of the loop.
	}
	else {
	    if( defined($dev) ) {
		if (/IRQ[ ]*([0-9]+)/) {
		    # Extract the IRQ information for this device
		    $dev->setIrq($1);
		}
	    }
	}
    } # while
    close(FD);

    if( $PCIDeviceList->count() == 0 ) {
	Debug(1, "Found 0 PCI devices.");
    }

    return($PCIDeviceList);      # Success
}

# Looks for the PCI devices on the host system by examining the
# configuration space through the /proc/bus/pci interface.
sub parse_proc_bus_pci() {
    my $PCIDeviceList;      # Reference to hash
    Debug(1, "Entered");

    # XXX NOT IMPLEMENTED
    return(undef);
}

# Parses the output from /sbin/lspci program extracting PCI information.
# Returns a hash of the pci devices detected on success.  On failure,
# returns undef
sub parse_safe_lspci() {
    my $helperName = 'safe_lspci';
    my $helperName2 = 'safe_lspci_names';

    Debug(1, "Entered.");
    my $PCIDeviceList = parse_lspci($helperName);

    if( $PCIDeviceList ) {
	# Get the names of the devices
	my $PCIDeviceList_Names = parse_lspci_names($helperName2);
	if( !$PCIDeviceList_Names ) {
	    Warning("Could not get the names of the pci devices");
	} else {
	    my $hdev;
	    foreach $hdev ($PCIDeviceList_Names->getPCIDevices()) {
		my $dev = $PCIDeviceList->getPCIDeviceByHash($hdev->getHash());
		if( !$dev ) {
		    Warning('The PCI device ' . $dev->getHash() .
			    ' generated by lsci_names does not exist for the list generated by ' . $helperName);
		    next;
		}
		$dev->setName($hdev->getName());
	    }
	}
	Debug(1, "Leaving with " . $PCIDeviceList->count() . " devices.");
	return($PCIDeviceList);
    }
    Debug(1, "Leaving with 0 devices.");
    return(undef);
}


# Parse with probe method 1
sub parse_probe_lspci() {
    my $helperName = 'probe_lspci';
    my $helperName2 = 'probe_lspci_names';

    Debug(1, "Entered.");
    my $PCIDeviceList = parse_lspci($helperName);

    if( $PCIDeviceList ) {
	# Get the names of the devices
	my $PCIDeviceList_Names = parse_lspci_names($helperName2);
	if( !$PCIDeviceList_Names ) {
	    Warning("Could not get the names of the pci devices");
	} else {
	    my $hdev;
	    foreach $hdev ($PCIDeviceList_Names->getPCIDevices()) {
		my $dev = $PCIDeviceList->getPCIDeviceByHash($hdev->getHash());
		if( !$dev ) {
		    Warning('The PCI device ' . $dev->getHash() .
			    ' generated by lsci_names does not exist for the list generated by ' . $helperName);
		    next;
		}
		$dev->setName($hdev->getName());
	    }
	}
	Debug(1, "Leaving with " . $PCIDeviceList->count() . " devices.");
	return($PCIDeviceList);
    }
    Debug(1, "Leaving with 0 devices.");
    return(undef);
}


# Parse with probe method 2
sub parse_probe_lspci2() {
    my $helperName = 'probe_lspci2';
    my $helperName2 = 'probe_lspci2_names';

    Debug(1, "Entered.");

    # Currently, we know of no type 2 machines so don't actually perform the probe.
    # Doe not actually do anything here.
    return(undef);

    my $PCIDeviceList = parse_lspci($helperName);

    if( $PCIDeviceList ) {
	# Get the names of the devices
	my $PCIDeviceList_Names = parse_lspci_names($helperName2);
	if( !$PCIDeviceList_Names ) {
	    Warning("Could not get the names of the pci devices");
	} else {
	    my $hdev;
	    foreach $hdev ($PCIDeviceList_Names->getPCIDevices()) {
		my $dev = $PCIDeviceList->getPCIDeviceByHash($hdev->getHash());
		if( !$dev ) {
		    Warning('The PCI device ' . $dev->getHash() .
			    ' generated by lsci_names does not exist for the list generated by ' . $helperName);
		    next;
		}
		$dev->setName($hdev->getName());
	    }
	}
	Debug(1, "Leaving with " . $PCIDeviceList->count() . " devices.");
	return($PCIDeviceList);
    }

    Debug(1, "Leaving with 0 devices.");
    return(undef);
}


# Parses the output from /sbin/lspci program extracting PCI information.
# Specify a helper name to lspci to get it going.
sub parse_lspci($) {
    my ($helperName) = @_;
    local *FD;
    my $dev = undef;     # String for parsing device.  undef if looking for new device
    my $PCIDeviceList = VMware::HConfig::PCIDeviceList->new();

    Debug(1, "Entered with helper $helperName.");

    # Make sure that the program is executable
    if( ! c_getHelper($helperName) ) {
	Warning('Cannot execute $helperName');
	return(undef);
    }

    if( !open(FD, c_getHelper($helperName) . ' 2>/dev/null |') ) {
	# Error with command
	Warning('Could not execute ' . c_getHelper($helperName));
	return(undef);
    }

    while(<FD>) {
	chomp();

	if( /^([0-9a-f]+):([0-9a-f]+)\.([0-9a-f]+) Class ([0-9a-f]+):\s*([0-9a-f]+):([0-9a-f]+)/ ) {
	    my $hash = VMware::HConfig::PCIDevice::hash(hex($1), hex($2), hex($3));
	    my ($class, $vendor, $devid) = ($4, $5, $6);

	    # Check for existence of this device in the list.  If the device does
	    # not exist, add it.
	    if( !$PCIDeviceList->getPCIDeviceByHash($hash) ) {
		# New device.  Add it.
		$dev = VMware::HConfig::PCIDevice->new(hex($1), hex($2), hex($3));
		$PCIDeviceList->addPCIDevice($dev);
	    }
	    else {
		$dev = $PCIDeviceList->getPCIDeviceByHash($hash);
	    }

	    # Set the vendor and device id
	    $dev->setVendorID(hex($vendor));
	    $dev->setDeviceID(hex($devid));

	    $dev->setClass(hex($class) << 8);
	}
	else {   # Still working on the device $dev
	    if( defined($dev) ) {
		if( /IRQ\s*([0-9]+)/ ) {
		    # Extract the IRQ information for this device
		    $dev->setIrq($1);
		}
		if( /Subsystem: ([0-9a-f]+):([0-9a-f]+)/ ) {
		    my $subsys_vendor = $1;
		    my $subsys_device = $2;
		    $dev->setSubsysVendorID(hex($subsys_vendor));
		    $dev->setSubsysDeviceID(hex($subsys_device));
		}
	    }
	}

    } # while

    close(FD);

    if( $PCIDeviceList->count() == 0 ) {
	Debug(1 ,"Found 0 PCI devices.");
	return(undef);
    }

    return($PCIDeviceList);
}


# Parses the output from /sbin/lspci program extracting PCI information.
# Returns a the pci devices detected on success.  On failure, returns undef.
# This functions gives us the names of the devices.
# Specify one of the lspci helpers to this function to get it going.
sub parse_lspci_names($) {
    my ($helperName) = @_;
    local *FD;
    my $dev = undef;     # String for parsing device.  undef if looking for new device
    my $PCIDeviceList = VMware::HConfig::PCIDeviceList->new();

    Debug(1, "Entered with helper $helperName.");

    # Make sure that the program is executable
    if( ! c_getHelper($helperName) ) {
	Warning("Cannot execute $helperName");
	return(undef);
    }

    if( !open(FD, c_getHelper($helperName) . ' 2>/dev/null |') ) {
	# Error with command
	Warning('Could not execute ' . c_getHelper($helperName));
	return(undef);
    }

    while(<FD>) {
	chomp();      # Remove trailing newlines

	if( /^([0-9a-f]+):([0-9a-f]+)\.([0-9a-f]+) ([^:]*): (.+)/ ) {
	    my $hash = VMware::HConfig::PCIDevice::hash(hex($1), hex($2), hex($3));
	    my $type = $4;   # Ignored
	    my $name = $5;

	    # Check for existence of this device in the list.  If the device does
	    # not exist, add it.
	    if( !$PCIDeviceList->getPCIDeviceByHash($hash) ) {
		# New device.  Add it.
		$dev = VMware::HConfig::PCIDevice->new(hex($1), hex($2), hex($3));
		$PCIDeviceList->addPCIDevice($dev);
	    }
	    else {
		$dev = $PCIDeviceList->getPCIDeviceByHash($hash);
	    }

	    # Set the vendor and device id
	    $dev->setName($name);
	}
    } # while

    close(FD);

    if( $PCIDeviceList->count() == 0 ) {
	Debug(1, "Found 0 PCI devices.");
	return(undef);
    }

    Debug(1, 'Exited with ' . $PCIDeviceList->count() . ' devices');
    return($PCIDeviceList);
}

# Remove all the keys for a PCI device from the config database
sub removePCIDeviceConfigDB($$) {
    my($self) = shift;
    my($configDB, $hash) = @_;

    Debug(1, "Removing PCI device $hash");

    my($bus, $device, $function) = VMware::HConfig::PCIDevice::hashToBusDevFunc($hash);
    if( !defined($bus) || !defined($device) || !defined($function) ) {
	Warning("Got null hash value.");
	return(0);
    }

    # Get all the keys that exist in the database for this device
    my @keys = grep(/^device\.$bus\.$device\.$function\./, $configDB->getKeys());
    @keys = sort(@keys);

    my @o_keys = grep(/^device\.[^\.]+\.$bus\.$device\.$function.(owner|options|clustering)$/, $configDB->getKeys());
    push(@keys, @o_keys);

    # Remove the keys from the database
    my $key;
    foreach $key (@keys) {
	$configDB->unset($key);
    }

    Debug(1, "Exited.");
}

# Finds a PCI device in the configuration database.  The argument
# specifies paramters governing how the lookup operation works.
#
# if either owner or label are not set, the function finds the list of
# all the PCI devices that are in the system.  No information about
# the owner is extracted.
#
# if owner and label are defined, function gets the owner information
# for a given configuration named label.
sub findPCIDeviceListInConfigDB($$$) {
    my($self) = shift;
    my ($configDB, $owner, $label) = @_;;
    my $key;

    Debug(1, "Entered");

    # Get all the devices that exist
    my @keys = grep(/^device\./, $configDB->getKeys());
    @keys = sort(@keys);

    my $PCIDeviceList = VMware::HConfig::PCIDeviceList->new();

    # Iterate through and get the list of devices
    foreach $key (@keys) {
	if( $key !~ /^device\.(\d+)\.(\d+)\.(\d+)\./ ) {
	    next;
	}
	my ($bus, $device, $function) = ($1, $2, $3);
	my $hash = VMware::HConfig::PCIDevice::hash($bus, $device, $function);

	my $dev = $PCIDeviceList->getPCIDeviceByHash($hash);
	if( !$dev ) {
	    $dev = VMware::HConfig::PCIDevice->new($bus, $device, $function);
	    $PCIDeviceList->addPCIDevice($dev);
	}

	my $val = $configDB->get($key);

	# Figure out what this key is and what to do with the value
	my $prefix = 'device.' . $bus . '.' . $device . '.' . $function . '.';
	if( $key eq ($prefix . 'vendor') ) {
	    $dev->setVendorID(hex($val));
	} elsif( $key eq ($prefix . 'devID') ) {
	    $dev->setDeviceID(hex($val));
	} elsif( $key eq ($prefix . 'name') ) {
	    $dev->setName($val);
	} elsif( $key eq ($prefix . 'class') ) {
	    $dev->setClass(hex($val));
	} elsif( $key eq ($prefix . 'subsys_vendor') ) {
	    $dev->setSubsysVendorID(hex($val));
	} elsif( $key eq ($prefix . 'subsys_devID') ) {
	    $dev->setSubsysDeviceID(hex($val));
	} else {
	    Warning('Unknown config key ' . $key . ' encountered.');
	}
    }

    if( $PCIDeviceList->count() == 0 ) {
	Debug(1, "Found 0 PCI devices.");
	return(undef);
    }

    # Get the owner/options/clustering information if the flag is set
    if( $owner && $label ) {
	foreach $key ($configDB->getKeys()) {
	    if( $key !~ /^device\.$label\.(\d+)\.(\d+)\.(\d+)\.(owner|options|clustering)$/ ) {
		next;
	    }
	    my $hash = VMware::HConfig::PCIDevice::hash($1, $2, $3);
	    my $val = $configDB->get($key);
	    my $dev = $PCIDeviceList->getPCIDeviceByHash($hash);
	    if( !$dev ) {
		Warning('Non device specified for owner key ' . $key . '.');
		next;
	    }
            if ($4 eq "owner") {
               $dev->setOwnerString($val);
            } elsif ($4 eq "options") {
               $dev->setOptions($val);
            }
	}
    }

    Debug(1, "Exited with " . $PCIDeviceList->count() . " devices.");

    return($PCIDeviceList);
}


# Saves the list of pci devices into the database if it
# sees that there is a change in the database.
#
# if either owner or label are not set, the function finds the list of
# all the PCI devices that are in the system.  No information about
# the owner is extracted.
#
# if owner and label are defined, function gets the owner information
# for a given configuration named label.
sub savePCIDeviceListInConfigDB($$$$) {
    my($self) = shift;
    my ($configDB, $PCIDeviceList, $owner, $label) = @_;
    my %keyUsed;
    my $key;
    my $dev;
    my $rc = 1;

    Debug(1, "Entered");

    # Get all the devices that exist in the database
    my @keys = grep(/^device\.(\d+)\.(\d+)\.(\d+)\./, $configDB->getKeys());
    @keys = sort(@keys);

    # Add the owner/options/clustering keys if we need to do this.
    if( $owner && $label ) {
	my @o_keys = grep(/^device\.$label\.(\d+)\.(\d+)\.(\d+).(owner|options|clustering)$/, $configDB->getKeys());
	push(@keys, @o_keys);
    }

    foreach $key (@keys) {
	$keyUsed{$key} = 0;
    }

    # Iterate through the list of devices and add the info into the db.

    foreach $dev ($PCIDeviceList->getPCIDevices()) {
	my $prefix = 'device.' . $dev->getBus() . '.' . $dev->getDevice() . '.' . $dev->getFunction() . '.';
	my $vendor = sprintf("%04x", $dev->getVendorID());
	my $devid = sprintf("%04x", $dev->getDeviceID());
	my $name = $dev->getName();
	my $class = sprintf("%06x", $dev->getClass());
	my $subsys_vendor = sprintf("%04x", $dev->getSubsysVendorID());
	my $subsys_devid = sprintf("%04x", $dev->getSubsysDeviceID());

	$configDB->set($prefix . 'vendor', $vendor);
	$configDB->set($prefix . 'devID', $devid);
	$configDB->set($prefix . 'name', $name);
	$configDB->set($prefix . 'class', $class);
	$configDB->set($prefix . 'subsys_vendor', $subsys_vendor);
	$configDB->set($prefix . 'subsys_devID', $subsys_devid);

	$keyUsed{$prefix . 'vendor'} = 1;
	$keyUsed{$prefix . 'devID'} = 1;
	$keyUsed{$prefix . 'name'} = 1;
	$keyUsed{$prefix . 'class'} = 1;
	$keyUsed{$prefix . 'subsys_vendor'} = 1;
	$keyUsed{$prefix . 'subsys_devID'} = 1;

	if( $owner && $label ) {
	    my $o_key = 'device.' . $label . '.' . $dev->getBus() . '.' .
		$dev->getDevice() . '.' . $dev->getFunction() . '.owner';
	    my $o_val = $dev->getOwnerString();
	    $configDB->set($o_key, $o_val);
	    $keyUsed{$o_key} = 1;
	}
    }

    # Clean up the db and remove entries for db keys that are not
    # there any more.
    foreach $key (keys(%keyUsed)) {
	if( !$keyUsed{$key} ) {
	    $configDB->unset($key);
	}
    }

    Debug(1, "Exited.");
    return($rc);
}

# Regenerates the config database given the list of pci devices and the lilo
# configuration.
sub regenerateConfigDB($$$) {
    my($self) = shift;
    my($configDB, $ProbePCIDeviceList, $lilo) = @_;
    my $rc = 1;

    # Work on a copy of the list so we don't cause any side effects on the
    # list being passed in.
    my $CopyPCIDeviceList = $ProbePCIDeviceList->copy();

    # Must save the device information in case we don't have a vmnix lilo entry.
    # However, we save only the device information and not the ownership info.
    $self->savePCIDeviceListInConfigDB($configDB, $CopyPCIDeviceList, 0, '');

    my(@boot_imgs) = @{ $lilo->getBootImages };
    my $lilo_img;
    foreach $lilo_img (@boot_imgs) {
	if( $lilo_img->usesVMnixImage() ) {
	    # Only do this for vmnix entries
	    $CopyPCIDeviceList = $self->findPCIDevicesOwnershipByLILO($CopyPCIDeviceList, $lilo_img);
	    my $label = $lilo_img->getLabel();
	    if( $label ) {
		$self->savePCIDeviceListInConfigDB($configDB, $CopyPCIDeviceList, 1, $label);
	    } else {
		Warning("Undefined label.");
		$rc = 0;
	    }
	}
    }

    return($rc);
}


# Reconciles the changes between the new list and the old list.
# Implementation-wise, this function invalidates entries in the config database
# by deleting keys that are no longer valid.  Then it updates the database with
# the new list.
sub reconcileConfigDB($$$) {
    my($self) = shift;
    my($configDB, $newPCIDeviceList, $lilo) = @_;
    my $dev;

    Debug(1, "Entered.");

    # Simple algorithm.  All we do is remove information about a device if we notice that
    # 1) it is no longer there.
    # 2) it has changed.

    my $oldPCIDeviceList = $self->findPCIDeviceListInConfigDB($configDB, 0, $lilo->getDefaultImage());

    if( $oldPCIDeviceList->count() == 0 ) {
	Warning("Could not find any devices in an old configuration.  No need to reconcile " .
		"the config database.");
	return(0);
    }

    my %deviceHash;
    foreach $dev ($oldPCIDeviceList->getPCIDevices()) {
	my $hash = $dev->getHash();
	$deviceHash{$hash}->[0] = $dev;
	$deviceHash{$hash}->[1] = 0;   # Flag indicating if device was touched
    }

    # Iterate through the new list of devices and look for changes
    foreach $dev ($newPCIDeviceList->getPCIDevices()) {
	my $hash = $dev->getHash();
	if( $deviceHash{$hash} ) {
	    my $olddev = $deviceHash{$hash}->[0];
	    if( !$olddev->equals($dev) ) {
		Warning("Device $hash is a different device.  It used to be a " .
			$olddev->getClassName() . ' ' . $olddev->getName() .
			" (" . sprintf("%04x", $olddev->getVendorID()) . ',' .
			sprintf("%04x", $olddev->getDeviceID()) .
			") but is now a " . $dev->getClassName() . ' ' . $dev->getName() .
			" (" . sprintf("%04x", $dev->getVendorID()) . ',' .
			sprintf("%04x", $dev->getDeviceID()) . ').');
		$self->removePCIDeviceConfigDB($configDB, $hash);
	    }
	    $deviceHash{$hash}->[1] = 1;   # Flag indicating device was touched
	}
    }

    # Iterate through the list of devices and look for devices that no longer exist.
    foreach $dev ($oldPCIDeviceList->getPCIDevices()) {
	my $hash = $dev->getHash();
	if( $deviceHash{$hash} && $deviceHash{$hash}->[1] == 0 ) {
	    my $olddev = $deviceHash{$hash}->[0];
	    Warning("Device $hash no longer exists.  It used to be a " .
		    $olddev->getClassName() . ' ' . $olddev->getName() .
		    " (" . sprintf("%04x", $olddev->getVendorID()) . ',' .
		    sprintf("%04x", $olddev->getDeviceID()) .
		    ").");
	    $self->removePCIDeviceConfigDB($configDB, $hash);
	}
    }

    $self->savePCIDeviceListInConfigDB($configDB, $newPCIDeviceList, 0, '');

    return(1);
}

# Performs a check of the configuration database and removes any
# keys that shouldn't be there.
sub cleanupPCIDevicesConfigDB() {
    my($self) = shift;
    my %keyUsed;
    my $key;
    my $dev;
    my $keysCleaned = 0;

    my ($configDB) = $self->getConfigDB();
    my ($lilo) = $self->getLILOConfig();

    Debug(1, "Entered");

    # Build a hash of all the device keys that exist in the database
    my @keys = grep(/^device\./, $configDB->getKeys());
    foreach $key (@keys) {
	$keyUsed{$key} = 0;
    }

    # Get all the devices that exist in the database
    @keys = grep(/^device\.(\d+)\.(\d+)\.(\d+)\./, $configDB->getKeys());
    foreach $key (@keys) {
	$keyUsed{$key} = 1;
    }

    # Add the owner/options/clustering keys for the given LILO labels
    my(@labels) = @{ $lilo->getBootImageLabels() };
    my $label;
    foreach $label (@labels) {
	my @o_keys = grep(/^device\.$label\.(\d+)\.(\d+)\.(\d+).(owner|options|clustering)$/, $configDB->getKeys());
	foreach $key (@o_keys) {
	    $keyUsed{$key} = 1;
	}
    }

    # Iterate through the list of keys and remove the ones that should not be there.
    foreach $key (keys(%keyUsed)) {
	if( $keyUsed{$key} == 0 ) {
	    Debug(1, "Found an unused key '$key'.  Removing from the database.");
	    $configDB->unset($key);
	    $keysCleaned++;
	}
    }

    return($keysCleaned);
}


# Uses the list of PCIDevices and finds out who owns the device by consulting
# the LILO entry for the devices.  This is an attempt to find the best defaults
# but LILO is not authoritative about how the user wants the system to be
# configured.
#
# Side effects:
#   Fills in ownership information for the PCIDeviceList
sub findPCIDevicesOwnershipByLILO($$) {
    my($self) = shift;
    my($PCIDeviceList, $lilo_img) = @_;
    my $label = $lilo_img->getLabel();

    # Consult the LILO entry for suitable defaults
    if( !$PCIDeviceList->fromCPCILine($lilo_img->getCPCI()) ) {
	Warning("Error in building PCI ownership information from the LILO information");
    }

    return($PCIDeviceList);
}


# Perform some sanity checks before returning the device list.
# These are not the best checks, but it's an imperfect world.
# Returns the number of warnings found. 0 otherwise.
# Turn on the fix flag to have the function change the ownership
# settings to match that of the current setup.
sub checkPCIDevicesWithCurrent($) {
    my($self) = shift;
    my($PCIDeviceList, $fix) = @_;
    my $warnings = 0;

    Debug(1, "Entered");

    # Verify this result with the current state of the system
    my $dev;
    my $hdev;

    my $HostPCIDeviceList = findHostPCIDevices();

    # Create a hash indicating which devices were touched
    my %deviceTouched;
    my %deviceManageable;
    my %deviceSharable;

    # Iterate over all the devices initializing the hashes
    foreach $dev ($PCIDeviceList->getPCIDevices()) {
	if( $dev->isVMwareManageable() ) {
	    # Ignore the function.  All we want to know is if the PCI card is manageable.
	    my $cardhash = VMware::HConfig::PCIDevice::hash($dev->getBus(), $dev->getDevice(), 0);
	    $deviceManageable{$cardhash} = 1;
	}

	if( $dev->isVMwareSharable() ) {
	    # Ignore the function.  All we want to know is if the PCI card is sharable.
	    my $cardhash = VMware::HConfig::PCIDevice::hash($dev->getBus(), $dev->getDevice(), 0);
	    $deviceSharable{$cardhash} = 1;
	}

	# Initialize each device touched entry
	$deviceTouched{$dev->getHash()} = 0;
    }

    # Check to make sure that devices that are being managed (or
    # shared) by us are really manageable/sharable
    foreach $dev ($PCIDeviceList->getPCIDevices()) {
	my $cardhash = VMware::HConfig::PCIDevice::hash($dev->getBus(), $dev->getDevice(), 0);

	# Only certain devices can be shared
	if( $dev->getOwner() == VMware::HConfig::PCIDevice::SHARED &&
	    !$dev->isVMwareSharable() ) {

	    # Make sure that that there is not another function on the card that
	    # is manageable.  Otherwise, this is a false error message.
	    if( !defined($deviceSharable{$cardhash}) || $deviceSharable{$cardhash} != 1 ) {
		Warning("Device (" . $dev->getHash() . ") " . $dev->getClassName() .
			" is shared but is not a sharable device.");
		$warnings++;

		if( $fix )  {
		    $dev->setOwner(VMware::HConfig::PCIDevice::HOST);
		}
	    }
	}

	# A device can only be owned by VM or shared if we can manage it
	if( $dev->getOwner() != VMware::HConfig::PCIDevice::HOST &&
	    !$dev->isVMwareManageable() ) {

	    # Make sure that that there is not another function on the card that
	    # is manageable.  Otherwise, this is a false error message.
	    if( !defined($deviceManageable{$cardhash}) || $deviceManageable{$cardhash} != 1 ) {
		Warning("Device (" . $dev->getHash() . ") " . $dev->getClassName() . " is owned " .
			$dev->getOwnerString() . " but it should not be " .
			"because it is an unsupported device.");
		$warnings++;

		if( $fix ) {
		    $dev->setOwner(VMware::HConfig::PCIDevice::HOST);
		}
	    }
	}
    }

    # Check to make sure that devices owned by the host are listed as being
    # owned by the host
    foreach $hdev ($HostPCIDeviceList->getPCIDevices()) {
	$dev = $PCIDeviceList->getPCIDeviceByHash($hdev->getHash());

	if( !$dev ) {
	    Warning('Host reports that it has a PCI device that was not found in a prior ' .
		    'configuration nor during the probe of the devices.  Device is ' .
		    $hdev->getHash() . ".  Ignoring this device.");
	    $warnings++;
	    next;
	}

	# Mark each device as being touched
	$deviceTouched{$dev->getHash()} = 1;
    }

    # Each device that was not touched should not be set to HOST or SHARED
    foreach $dev ($PCIDeviceList->getPCIDevices()) {
	if( $deviceTouched{$dev->getHash()} == 0 ) {
	    if( $dev->getOwner() == VMware::HConfig::PCIDevice::SHARED ||
		$dev->getOwner() == VMware::HConfig::PCIDevice::HOST ) {
		Warning("Device (" . $dev->getHash() . ") " . $dev->getClassName() .
			" is owned by " . $dev->getOwnerString() . " but is not ".
			"detected by the hosts.");
		$warnings++;

		if( $fix )  {
		    $dev->setOwner(VMware::HConfig::PCIDevice::VM);
		}
	    }
	}

	# Mark each device as being touched
	$deviceTouched{$dev->getHash()} = 1;
    }

    return($warnings);
}


#####################################################################
# END OF PCI Device Handling Functions                              #
#####################################################################

#####################################################################
# SCSI Adapter Decision Handling Functions                          #
#####################################################################

###############################
# Common SCSI Adapter Functions
###############################

# Returns the list of adapters as SCSIAdapter objects
sub findScsiAdapters($) {
    my($self) = shift;
    my ($PCIDeviceList) = @_;
    my @adapters;

    Debug(1, "Entered");

    # Get the modules that are loaded
    my @vmkModulesLoaded = map( $_->{MODULE}, @{ $self->getVMKernelModules() });

    # Construct the SCSI devices include a mapping between the vmnix and vmkernel SCSI adapters in
    # the case of shared devices.  Also map SCSI adapters to PCI devices.
    @adapters = @{constructSCSIAdapters($PCIDeviceList, $self->getKernelInfo(),
					\@vmkModulesLoaded)};

    # Find the disks for these adapters
    constructSCSIDisks($PCIDeviceList, \@adapters, \@vmkModulesLoaded);

    return(\@adapters);
}


# Takes in a list of vmkernel SCSI adapter names, vmnix SCSI adapter names, and
# PCI devices.  Uses this information and some other information to build the
# SCSI adapter object, the disks that the adapters host, and the links between
# all the different objects.
sub constructSCSIAdapters($) {
    my ($PCIDeviceList, $kernel, $vmkModules) = @_;
    my $name;
    my $dev;

    Debug(1, "Entered");

    # We only care about devices that the vmkernel can see, but we need to know
    # what SCSI and RAID devices are shared with the console OS.

    # Build a set of SCSI adapters that are usable by the vmkernel
    my @vmkAdapters;

    # Get 2 sorted lists of SCSI adapter names
    my @vmkAdapterNames = findVmkScsiAdapterNames();

    Debug(1, "Found " . ($#vmkAdapterNames + 1) . " vmk adapters: @vmkAdapterNames\n");

    if( $#vmkAdapterNames  == -1 ) {
	# No devices found.
	return(\@vmkAdapters);
    }

    # Get the list of world-wide port names to vmhba adapters so that we can assign
    # them to the adapters.
    my %wwpns = findWWPN();

    # Map the Vmk SCSI adapters to the PCI devices
    my %pci2VmkScsiAdapter;
    foreach $name (@vmkAdapterNames) {
	my ($bus, $device, $function) = findVmkScsiAdapterPCIInfo($name);
	if( !defined($bus) || !defined($device) || !defined($function) ) {
	    next;
	}
	my $hash = VMware::HConfig::PCIDevice::hash($bus, $device, $function);
	Debug(1, "Found PCI Info for a vmkAdapter at $hash\n");
	my $pciDevice = $PCIDeviceList->getPCIDeviceByHash($hash);
	if( $pciDevice ) {
	    my $a = VMware::HConfig::SCSIAdapter->new($name);
	    $a->setPCIDevice($pciDevice);
            if (defined($wwpns{$name})) {
               my @sanDevices = findSanDevices($wwpns{$name}->{WWPN});
               $a->setWWPN($wwpns{$name}->{WWPN});
               $a->setSanDevices(@sanDevices);
            }

	    push(@vmkAdapters, $a);
	    $pci2VmkScsiAdapter{$hash} = $a;

	    # Perform a check to see if the adapter is shared
	    my $shared = checkSharedScsi($name);
	    if( $shared && $pciDevice->getOwner() != VMware::HConfig::PCIDevice::SHARED ) {
		Warning('Vmk ' . $gVmkDeviceType_hba . ' adapter ' . $name . ' is shared, but PCI Device ' . $hash .
			' to which the ' . $gVmkDeviceType_hba . ' Adapter is bound is not shared.  ' .
			'The state of the system is inconsistent.  Did you forget to ' .
			'reboot your machine after allocating PCI devices?' . "\n");

		# Mark it shared since this is a more definitive test
		$pciDevice->setOwner(VMware::HConfig::PCIDevice::SHARED);
	    } elsif( !$shared && $pciDevice->getOwner() == VMware::HConfig::PCIDevice::SHARED ) {
		Warning('Vmk ' . $gVmkDeviceType_hba . ' adapter ' . $name .
			' is not shared, but PCI Device ' . $hash .
			' to which the ' . $gVmkDeviceType_hba . ' Adapter is bound is shared.  ' .
			'The state of the system is inconsistent.  Did you forget to ' .
			'reboot your machine after allocating PCI devices?' . "\n");
		# Mark it as owned by the VM since this is a more definitive test
		$pciDevice->setOwner(VMware::HConfig::PCIDevice::VM);
	    }
	} else {
	    Warning('Vmk ' . $gVmkDeviceType_hba . ' adapter ' . $name . ' is supposed to be a VMkernel ' .
		    $gVmkDeviceType_hba . ' host adapter, but information about the device is not' .
		    ' available in the proc filesystem.' . "\n");
	}
    }

    # Perform a check to see if any PCI Device that should have been
    # picked up were not in the previous loop
    foreach $dev ($PCIDeviceList->getPCIDevices()) {
	# Skip device if we're not interested in it
	if( !defined($pci2VmkScsiAdapter{$dev->getHash()}) ) {
	    next;
	}
	if( $dev->getOwner() == VMware::HConfig::PCIDevice::SHARED ||
	    $dev->getOwner() == VMware::HConfig::PCIDevice::VM ) {
	    if( !defined( $pci2VmkScsiAdapter{$dev->getHash()} ) ) {
		Warning('PCI Device ' . $dev->getHash() . ' is ' .
			$dev->getOwnerString() . ', but the ' . $gVmkDeviceType_hba . ' ' .
			'adapter for this device could not be found.  ' .
			'Did the driver for the device fail to load?' . "\n");
	    }
	}
    }

    return(\@vmkAdapters);
}


# Takes in a list of SCSI Adapter objects and creates the disk objects
# refering to the disks that are attached to these adapters.  In the
# case of shared SCSI adapters, also locates the Vmnix device name for
# the disk if possible.
sub constructSCSIDisks($) {
    my ($PCIDeviceList, $vmkadapt, $vmkModules) = @_;
    my @vmkAdapters = @$vmkadapt;
    my $dev;
    my $a;

    # Find out what Vmk Scsi Disks we have
    foreach $a (@vmkAdapters) {
	my @disks = @{ findVmkScsiDisks($a->getVmkName()) };

	Debug(1, "Have " . ($#disks + 1) . " disks for adapter " . $a->getVmkName() . "");

	@disks = map { VMware::HConfig::RawDisk->new($_) } @disks;
	$a->addDisk( @disks );
    }

    # Some of these "disks" are not really disks.  Some of them are things like
    # Processors and CD-ROMs.  We keep CDROM and Tape drives, but we want to discard
    # everything else.
    filterScsiDisks(\@vmkAdapters);

    #
    # Determine the VMnix device names
    #

    # 1. Create a mapping from VMkernel name to VMnix name

    local *FD;
    my %vmnixAdapterName;
    my $pvs_file = c_getFile('proc_vmware_vmkstor');

    if (!(-r $pvs_file) || !open(FD, "< $pvs_file")) {
	Warning("Could not read file $pvs_file.  Reason: $!");
    } else {
        while (<FD>) {
            if (/((scsi|block)\d+)\s+(vmhba\d+)/) {
                $vmnixAdapterName{$3} = $1;
            }
        }
        close(FD);
    }

    # 2. Create a mapping from VMnix name/target/lun to VMnix device file

    local *DIR;
    my %vmnixDevName;
    my %vmnixPartSeparator;
    my $devFile;
    my @devFiles;

    # 2a. Get list of SCSI devices using /proc/partitions to pare the list
    my $dir = c_getDir('dev');
    if (! -d $dir) {
       Warning("Device directory $dir is not a directory. Reason: $!");
    }

    my $partitionsFile = c_getFile('proc_partitions');
    if (!open(PART, "< $partitionsFile")) {
       Warning("Could not read the $partitionsFile file. Reason: $!");
    } else {
       # Get all files matching /dev/sd.
       while (<PART>) {
          if (/\s*\d+\s*\d+\s*\d+\s*(sd[^0-9])\s*/) {
             push @devFiles, $1;
          }
       }

       close(PART);

       @devFiles = map("$dir/$_", @devFiles);
       foreach $devFile (@devFiles) {
          my ($host, undef, $target, $lun) = VMware::HConfig::GetScsiDeviceInfo($devFile);
          if (defined($host) && defined($target) && defined($lun)) {
             $vmnixDevName{"scsi$host:$target:$lun"} = $devFile;
             # No separator character for partitions in SCSI disks
             $vmnixPartSeparator{"scsi$host:$target:$lun"} = "";
          } else {
             Warning("ASSERT: Couldnt get ScsiDeviceInfo for $devFile\n");
          }
       }
    }

    # 2b. Get list of block devices using /proc/partitions to pare the list 
    if (!open(PART, "< $partitionsFile")) {
       Warning("Could not read the $partitionsFile file. Reason: $!");
    } else {
       # Get all files matching block device names
       @devFiles = ();
       while (<PART>) {
          my $key;
          foreach $key (keys %gBlockDeviceNames) {
             if (/\s*\d+\s*\d+\s*\d+\s*($gBlockDeviceNames{$key}\/c\d+d\d+)\s+/) {
               push @devFiles, $1;
             }
          }
       }

       close(PART);
       @devFiles = map("$dir/$_", @devFiles);
       foreach $devFile (@devFiles) {
          $devFile = $1 if ($devFile =~ /(.*)/); # Untaint the data
          if ($devFile =~ /c(\d+)d(\d+)$/) {
             my @stat = stat($devFile);
             if (!defined($stat[6])) {
                Warning("Failed to stat $devFile.\n");
                next;
             }
             my $major = $stat[6] >> 8;
             my $target = $2;
             my $lun = 0; # block devices do not have LUNs
             $vmnixDevName{"block$major:$target:$lun"} = $devFile;
             # 'p' separates disk name from partition number in block devices
             $vmnixPartSeparator{"block$major:$target:$lun"} = 'p';
          }
       }
    }

    # 3. Use mappings to determine VMnix device name for all SCSI disks

    foreach $a (@vmkAdapters) {
        my $vmkAdapter = $a->getVmkName();
        my $vmnixAdapter = $vmnixAdapterName{$vmkAdapter};
        if (defined($vmnixAdapter)) {
            foreach my $disk (@{$a->getDisks()}) {
                my $target = $disk->getTarget();
                my $lun = $disk->getLun();
                my $vmkDisk = $disk->getVmkName();
                my $vmnixDisk = $vmnixDevName{"$vmnixAdapter:$target:$lun"};
                my $separator = $vmnixPartSeparator{"$vmnixAdapter:$target:$lun"};
                if (defined($vmnixDisk)) {
                    Debug(1, "Mapped SCSI disk with VMkernel name '$vmkDisk' " .
                          "to VMnix name '$vmnixDisk' with partition separator " .
                          "'$separator'.\n");
                    $disk->setVmnixDevName($vmnixDisk);
                    $disk->setVmnixPartSeparator($separator);
                } else {
                    Warning("Could not find VMnix device name for disk '$vmkDisk'.");
                }
            }
        } else {
            Warning("Could not find VMnix name for adapter '$vmkAdapter'.");
        }
    }

    Debug(1, "Returning from constructSCSIDisks");
}


# Scans a file for the string indicating the PCI info for SCSI adapter
# that we control.  This function is used for detecting PCI informaton
# for vmk SCSI disks and vmnix SCSI disks that we control.
#
# Returns an array containing the bus, device, and function
sub readScsiPCIInfo($) {
    my ($file) = @_;
    local *FD;

    if( ! -r $file ) {
	Warning("Could not find file '$file'.");
	return(0);
    }

    if( !open(FD, '< ' . $file) ) {
	Warning("Could not open file '$file'.");
	return(0);
    }

    my $s;
    while( <FD> ) {
	if( /^PCI info for (\S+\d+): ([0-9]+):([0-9]+)\.([0-9]+)/ ) {
	    my ($host, $bus, $device, $function) = ($1, int($2), int($3), int($4));
	    close(FD);
	    Debug(2, "Got scsi information for $file: $host:$bus:$device:$function\n");
	    return($bus, $device, $function);
	}
    }
    close(FD);
    return(undef);
}

# Gets a list of the San Devices for a WWPN
#
# Returns a hash where the array contains the target and WWPN to the target
sub findSanDevices() {
   my ($wwpn) = @_;
   my $adapter = "";
   my @sanDevices;

   # get the adapter info from the wwpn.pl script
  
   if (!c_getHelper('get_san_info')) {
      Warning("Cannot get san info.\n");
      return(@sanDevices);
   }
   my $command = c_getHelper('get_san_info');
   local *FD;
   if (!open(FD, "$command |")) {
      Warning("Did not get the required san info for WWPN " . $wwpn . "\n");
      return(@sanDevices);
   }
   while (<FD>) {
      if ($adapter eq "") {
         if ( /\s*(.*):\s+(.*)\s+\(.*/ ) {
            if ($wwpn eq $2) {
               $adapter = $1;
            }
         }
      } else {
         if ( /\s*$adapter:(.*):\s+(.*)/) {
            push(@sanDevices, { 'target' => $1, 'wwpn' => $2 });
         }
      }
   }
   close(FD);
   return(@sanDevices);
}

# Gets a list of the WWPN for SAN hba adapters.
#
# Returns a hash where the array contains the vmkName of the vmhba.
# Fields in the hash include WWPN (world wide port name), DRIVER
# (SAN driver type), PCI (PCI bus:device:function)
sub findWWPN() {
   my %wwpns;
   local *FD;

   if (!open(FD, c_getHelper('wwpn_list') . ' |')) {
      Warning('Could not run program ' . c_getHelper('wwpn_list') . '.  Reason: ' . "$!");
      return %wwpns;
   }

   while (<FD>) {
      # Match line of the form 'vmhba0: 210000e08b09ee8f (Qlogic) 2:8:0'
      if (/^($gVmkDeviceType_hba\d+):\s*(\S+)\s*\((\S+)\)\s*(\S+)\s*$/) {
         my($vmkName, $wwpn, $driver, $pci) = ($1, $2, $3, $4);
         $wwpns{$vmkName}->{WWPN} = $wwpn;
         $wwpns{$vmkName}->{DRIVER} = $driver;
         $wwpns{$vmkName}->{PCI} = $pci;
      }
   }

   close(FD);
   return %wwpns;
}

######################################
# End of Common SCSI Adapter Functions
######################################


#################################
# VMkernel SCSI Adapter Functions
#################################

# Finds a list of SCSI adapters owned by the VMkernel
# Returns a list of devices in the form vmhba0, vmhba1.
sub findVmkScsiAdapterNames() {
    my @adapters;

    my $pvs_dir = c_getDir('proc_vmware_scsi');
    if( ! -d $pvs_dir ) {
	# Error /proc/vmware/scsi is not a directory
	Warning("Could not open $pvs_dir. Reason: $!");
	return(@adapters);
    }

    if( ! (-x $pvs_dir && -r $pvs_dir) ) {
	# Error /proc/vmware/scsi is not accessible and readable
	Warning("$pvs_dir is not accessible. Reason: $!");
	return(@adapters);
    }

    local *DIR;

    if( !opendir(DIR, $pvs_dir) ) {
	# Error can't opendir /proc/vmware/scsi dir
	Warning("Cannot open $pvs_dir. Reason: $!");
	return(@adapters);
    }

    @adapters = grep { $_ =~ /^$gVmkDeviceType_hba(\d+)/ && -d "$pvs_dir/$_" } readdir(DIR);
    closedir DIR;

    my $i;
    for( $i = 0; $i <= $#adapters; $i++ ) {
	# Extract the number
	if( $adapters[$i] =~ /$gVmkDeviceType_hba(\d+)/ ) {
	    $adapters[$i] = $1;
	}
    }

    # Untaint the data
    for( $i = 0; $i <= $#adapters; $i++ ) {
	if( $adapters[$i] =~ /(.*)/ ) {
	    $adapters[$i] = $1;
	}
    }

    # Sort numerically and prepend the 'vmhba' part back on
    @adapters = sort { $a <=> $b } @adapters;
    @adapters = map { "$gVmkDeviceType_hba$_" } @adapters;

    return(@adapters);
}


# Takes in a vmkernel adapter name and finds the PCI device info for it
sub findVmkScsiAdapterPCIInfo($) {
    my ($scsi) = @_;
    local *FD;

    Debug(1, "Entered with arg $scsi");

    if( !VMware::HConfig::SCSIAdapter::isLogVmkSCSIName($scsi) ) {
	Warning('Looking for the PCI information for an invalid ' . $gVmkDeviceType_hba .
		' adapter ' . $scsi. '.');
	return(undef);
    }

    my $pvs_dir = c_getDir('proc_vmware_scsi');
    if( ! -d "$pvs_dir/$scsi"  ) {
	Warning("Could not find directory '$pvs_dir/$scsi'.");
	return(undef);
    }

    if( ! -r "$pvs_dir/$scsi/stats" ) {
	Warning("Could not find file '$pvs_dir/$scsi/stats'.");
	return(undef);
    }

    if( !open(FD, '< ' . "$pvs_dir/$scsi/stats") ) {
	Warning("Could not open file '$pvs_dir/$scsi/stats'.");
	return(undef);
    }

    return( readScsiPCIInfo("$pvs_dir/$scsi/stats") );
}


# Takes in a vmkernel adapter name and checks to see if it is shared.
sub checkSharedScsi($) {
    my ($scsi) = @_;
    local *FD;

    if( !VMware::HConfig::SCSIAdapter::isLogVmkSCSIName($scsi) ) {
	Warning("Checking for sharing with an invalid SCSI adapter name '$scsi'.");
	return(0);
    }

    my $pvs_dir = c_getDir('proc_vmware_scsi');
    if( ! -d "$pvs_dir/$scsi"  ) {
	Warning("Could not find directory '$pvs_dir/$scsi'.");
	return(0);
    }

    if( ! -r "$pvs_dir/$scsi/stats" ) {
	Warning("Could not find file '$pvs_dir/$scsi/stats'.");
	return(0);
    }

    if( !open(FD, '< ' . "$pvs_dir/$scsi/stats") ) {
	Warning("Could not open file '$pvs_dir/$scsi/stats'.");
	return(0);
    }

    my $s;
    while( <FD> ) {
	if( /^$scsi: shared with Service Console/ ) {
	    close(FD);
	    return(1);
	}
	elsif( /^$scsi: not shared/ ) {
	    close(FD);
	    return(0);
	}
    }

    close(FD);
    return(0);
}

########################################
# End of vmkernel SCSI Adapter Functions
########################################

###############################
# SCSI Disk Locating functions
###############################

# Finds SCSI disk devices that are hosted by a particular adapter on
# the vmkernel.  Returns a list of disk name strings in order by
# target.  Format of names is adapter:target.  Note that adapter will
# be the same for all disks.
sub findVmkScsiDisks($) {
    my ($scsi) = @_;
    my @disks;

    my $pvs_dir = c_getDir('proc_vmware_scsi');
    if( ! -d "$pvs_dir/$scsi" ) {
	# Error /proc/vmware/scsi is not a directory
	Warning("Could not open $pvs_dir/$scsi. Reason: $!");
	return(\@disks);
    }

    if( ! (-x "$pvs_dir/$scsi" && -r "$pvs_dir/$scsi") ) {
	# Error /proc/vmware/scsi is not accessible and readable
	Warning("$pvs_dir/$scsi is not accessible. Reason: $!");
	return(\@disks);
    }

    local *DIR;

    if( !opendir(DIR, "$pvs_dir/$scsi") ) {
	# Error can't opendir /proc/vmware/scsi/scsin dir
	Warning("Cannot open $pvs_dir/$scsi. Reason: $!");
	return(\@disks);
    }
    @disks = grep { /^[0-9]+:[0-9]+$/ && ! -d "$pvs_dir/$scsi/$_" } readdir(DIR);
    closedir DIR;

    # Untaint the data
    my $i;
    for( $i = 0; $i <= $#disks; $i++ ) {
	if( $disks[$i] =~ /(.*)/ ) {
	    $disks[$i] = $1;
	}
    }

    @disks = sortTargetLun(@disks);
    @disks = map { "$scsi:$_" } @disks;

    Debug(2, "Found targets for $scsi: " . join(',', @disks) . "\n");
    return(\@disks);
}


# Some of these "disks" are not really disks.  Some of them are things like
# Processors and CD-ROMs.  We keep CDROM and Tape drives, but we want to discard
# everything else.
#
# Takes in a list of SCSI adapter objects and removes any non "Direct-Access",
# non "Sequential-Access", or non CD-ROM type SCSI devices.  In the case of
# Tape drives or CDROMs, we create a different structure, removing them from
# the list of disks for an adapter.
sub filterScsiDisks($) {
    my ($adapterList) = @_;

    my @adapters = @$adapterList;

    foreach $a (@adapters) {
        my @disks = @{ $a->getDisks() };
	my $disk;
        foreach $disk (@disks) {
            my $diskType = findVmkScsiDevTypes($a->getVmkName(), $disk->getTarget(), $disk->getLun());
            if( $diskType ne "Direct-Access" ) {
                my $diskName = $disk->getVmkName() . ':0';

		my $nonDiskType = $gNonDiskSCSITypes{$diskType};
		if( !defined($nonDiskType) ) {
		    # If not defined, just use what it was before.
		    $nonDiskType = $diskType;
		}

		if( $nonDiskType eq 'ignore' ) {
		    Debug(1, "Removing device " . $diskName .
			  " because its disk type is '$diskType'");
		    # Assume that diskType is not 'Direct-Access'
		    $a->removeDisk($disk);
		    next;
		}

		Debug(1, "Moving device " . $diskName . " ($nonDiskType) to non device list.");
		$a->addNonDiskDevice($diskName, $nonDiskType);
		$a->removeDisk($disk);
            }
        }
    }
}


# Finds SCSI disk types given a particular adapter name and target id.
# An adapter name is of the form scsi<n>, where n is a number.
# Returns a string corresponding to the type of the SCSI device.
sub findVmkScsiDevTypes($$$) {
    my ($scsi, $target, $lun) = @_;
    my $type = "";

    my $pvs_dir = c_getDir('proc_vmware_scsi');
    if( ! -r "$pvs_dir/$scsi/$target:$lun" ) {
	# Error /proc/vmware/scsi/scsi<n>/<target> is not a directory
	Warning("$pvs_dir/$scsi/$target is not readable. Reason: $!");
	return($type);
    }

    local *FD;

    if( !open(FD, '< ' . "$pvs_dir/$scsi/$target:$lun") ) {
	# Error can't open /proc/scsi/scsi<n>/<target> file
	Warning("Cannot open $pvs_dir/$scsi/$target. Reason: $!");
	return($type);
    }

    while( <FD> ) {
	if( /^Type:\s*(.*\S)\s+ANSI/ ) {
	    $type = $1;
	    last;
	}
    }
    close(FD);

    return($type);
}

# Sort the Target:Lun combinations
sub sortTargetLun(@) {
    my(@disks) = @_;
    my @newdisks;
    my $disk;
    foreach $disk (@disks) {
	if( $disk =~ /^([0-9]+):([0-9]+)$/ ) {
	    my $disk_struct = [$1, $2];
	    push(@newdisks, $disk_struct);
	}
    }
    @disks =
	sort {
	    if( $a->[0] <=> $b->[0] ) {
		return($a->[0] <=> $b->[0]);
	    } else {
		return($a->[1] <=> $b->[1]);
	    }
	} @newdisks;
    @disks = map { $_->[0] . ':' . $_->[1] } @disks;
    return(@disks);
}

#####################################
# End of SCSI Disk Locating functions
#####################################

#####################################################################
# END of SCSI Adapter Decision Handling Functions                   #
#####################################################################

# Finds Network controllers that are owned by the VM
# Returns a list of devices in the form vmnic0, vmnic1, etc.
sub findVmkNetAdapters() {
    my $self = shift;
    my @adapters;

    my $pvn_dir = c_getDir('proc_vmware_net');
    if( ! -d $pvn_dir ) {
	# Error /proc/vmware/net is not a directory
	Warning("Could not open $pvn_dir. Reason: $!");
	return(\@adapters);
    }

    if( ! (-x $pvn_dir && -r $pvn_dir) ) {
	# Error /proc/vmware/net is not accessible and readable
	Warning("$pvn_dir is not accessible. Reason: $!");
	return(\@adapters);
    }

    local *DIR;

    if( !opendir(DIR, $pvn_dir) ) {
	# Error can't opendir /proc/vmware/net dir
	Warning("Cannot open $pvn_dir. Reason: $!");
	return(\@adapters);
    }

    my @adapterNames = grep { /^$gVmkDeviceType_nic/ && -d "$pvn_dir/$_" } readdir(DIR);
    closedir DIR;

    # Get additional state information about the adapters
    local *FD;
    foreach my $name (sort @adapterNames) {
       my $adapter = {};

       # Untaint the data
       $name =~ /(.*)/;
       $adapter->{"name"} = $1;

       if( ! -r "$pvn_dir/$name/config" ) {
          Warning("Could not find file '$pvn_dir/$name/config'.");
          next;
       }

       if( !open(FD, '< ' . "$pvn_dir/$name/config") ) {
          Warning("Could not open file '$pvn_dir/$name/config'.");
          next;
       }

       # Initialize a few fields that may not be set.
       $adapter->{"speed"} = "";       # Link could be down
       $adapter->{"nicteam"} = "";     # No team defined
       $adapter->{"minCapabilities"} = 0;
       $adapter->{"hwCapabilities"} = 0;
       $adapter->{"maxCapabilities"} = 0;
       
       while ( <FD> ) {
          if ( /^PCI \(bus:slot\.func\):\s*(\d+):(\d+)\.(\d+)/ ) {
             $adapter->{"bus"} = $1;
             $adapter->{"device"} = $2;
             $adapter->{"function"} = $3;
          } elsif ( /^InterruptClustering\s+(Yes|No)/ ) {
             $adapter->{"clustering"} = ($1 eq "Yes") ? 1 : 0;
          } elsif ( /^Speed:\s+(.*)$/ ) {
             $adapter->{"speed"} = $1;
          } elsif ( /^NICTeamingMaster:\s+(\S+)$/ ) {
             $adapter->{"nicteam"} = $1;
          } elsif ( /^Minimum Capabilities\s+0x(.*)$/ ) {
             $adapter->{"minCapabilities"} = hex($1);
          } elsif ( /^Device Capabilities\s+0x(.*)$/ ) {
             $adapter->{"hwCapabilities"} = hex($1);
          } elsif ( /^Maximum Capabilities\s+0x(.*)$/ ) {
             $adapter->{"maxCapabilities"} = hex($1);
          }
       }
       close(FD);

       # Get the list of MAC addresses on this NIC.  This is useful as a usage count.
       local *MAC;
       my @mac;
       if (opendir(MAC, "$pvn_dir/$name")) {
          @mac = grep { /^[0-9a-fA-F]{2}(:[0-9a-fA-F]{2}){5}/ } readdir(MAC);
          closedir(MAC);
       }
       $adapter->{"mac"} = \@mac;

       push @adapters, $adapter;
    }

    return(\@adapters);
}


# Finds bond adapters for NIC teaming.
sub findVmkNetBondAdapters() {
   my $self = shift;
   my @adapters;

   my $pvn_dir = c_getDir('proc_vmware_net');
   if (! -d $pvn_dir) {
      # Error /proc/vmware/net is not a directory
      Warning("Could not open $pvn_dir. Reason: $!");
      return(\@adapters);
   }

   if (!(-x $pvn_dir && -r $pvn_dir)) {
      # Error /proc/vmware/net is not accessible and readable
      Warning("$pvn_dir is not accessible. Reason: $!");
      return(\@adapters);
   }

   local *DIR;

   if (!opendir(DIR, $pvn_dir)) {
      # Error can't opendir /proc/vmware/net dir
      Warning("Cannot open $pvn_dir. Reason: $!");
      return(\@adapters);
   }

   my @adapterNames = grep { /^$gVmkDeviceType_bond/ && -d "$pvn_dir/$_" } readdir(DIR);
   closedir DIR;

   # Get additional state information about the adapters
   local *FD;
   foreach my $name (sort @adapterNames) {
      my $adapter = {};

      # Untaint the data
      $name =~ /(.*)/;
      $adapter->{"name"} = $1;

      if (! -r "$pvn_dir/$name/config") {
         Warning("Could not find file '$pvn_dir/$name/config'.");
         next;
      }

      if (!open(FD, '< ' . "$pvn_dir/$name/config")) {
         Warning("Could not open file '$pvn_dir/$name/config'.");
         next;
      }

      while (<FD>) {
         if (/^NICTeamingSlaves:\s+No slave device/) {
            $adapter->{"slaves"} = [];
         } elsif ( /^NICTeamingSlaves:\s+(.+)\s+\(\d+\)$/ ) {
            my @slaves = split(/\s+/, $1);
            $adapter->{"slaves"} = \@slaves;
         }
      }
      close(FD);

      # Get the list of MAC addresses on this NIC.  This is useful as a usage count.
      local *MAC;
      if (opendir(MAC, "$pvn_dir/$name")) {
         my @mac = grep { /^[0-9a-fA-F]{2}(:[0-9a-fA-F]{2}){5}/ } readdir(MAC);
         closedir(MAC);
         $adapter->{"mac"} = \@mac;
      }

      push @adapters, $adapter;
   }

   return(\@adapters);
}

# Looks in /proc/vmware/net to find all the tcpip instances
sub findVMNetTcpip() {
   my($self) = shift;
   my @tcpip;
   my %tcpipInfo;

   my $dir = c_getDir('proc_vmware_net');
   local *DIR;
   if (!opendir(DIR, $dir) ) {
      Warning('Could not open directory ' . $dir . '.');
      return \%tcpipInfo;
   }

   @tcpip = grep { /^$gVmkDeviceType_tcpip/ && -d "$dir/$_" } readdir(DIR);
   closedir DIR;

   foreach my $name (sort(@tcpip)) {
      my $ipAddr = "";
      my $gatewayAddr = "";
      my $nicName = "";
      my $file = "$dir/$name/config/if0";

      local *FD;
      if (!open(FD, "< $file")) {
         Warning("Could not open file $file.");
         next;
      }

      while (<FD>) {
         if (/^Interface\s*=\s*(\w+)/) {
            $nicName = $1;
         } elsif (/^IPAddress\s*=\s*([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)/) {
            $ipAddr = $1;
         } elsif (/^Gateway\s*=\s*([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)/) {
            $gatewayAddr = $1;
         }
      }
      close(FD);

      $tcpipInfo{$name} = { NAME => $name,
                            NICNAME => $nicName,
                            IPADDR => $ipAddr,
                            GATEWAYADDR => $gatewayAddr };
   }

   return \%tcpipInfo;
}

# Looks for the configured the tcpip information
sub findVMNetTcpipCfg() {
   my($self) = shift;
   my %tcpipInfo;

   my $file = c_getFile('config_db');

   local *FD;
   if (!open(FD, "< $file")) {
      Warning("Could not open file $file.");
      return \%tcpipInfo;
   }

   # XXX Currently, the vmkernel only supports one TCPIP instance although the
   # interface suggests that there can be more than interface.
   my $configDB = $self->getConfigDB();

   my $ipAddr = $configDB->get("tcpip.ipAddr");
   my $gatewayAddr = $configDB->get("tcpip.gatewayAddr");
   my $nicName = $configDB->get("tcpip.nicName");
   my $name;

   if (!$ipAddr || !$gatewayAddr || !$nicName) {
      return \%tcpipInfo;
   }

   # Strip white spaces from strings.  Should we perform other validation checks?
   $ipAddr =~ s/\s//g;
   $gatewayAddr =~ s/\s//g;
   $nicName =~ s/\s//g;

   # We currently only support one instance so always use the existing one.
   # In the unlikely event that we've configured an instance but the vmkernel
   # doesn't have one, just guess at a name.
   $name = $self->getCurrentTcpipName();
   if (!defined($name) && $nicName =~ /(\d+)\s*$/) {
      $name = $gVmkDeviceType_tcpip . "$1";
   }

   $tcpipInfo{$name} = { NAME => $name,
                         NICNAME => $nicName,
                         IPADDR => $ipAddr,
                         GATEWAYADDR => $gatewayAddr };

   return \%tcpipInfo;
}

#####################################################################
# PCI Driver Decision Handling Functions                            #
#####################################################################

# Uses the driver map to get more complete information about our devices
sub findPCIDeviceDrivers($$) {
    my($self) = shift;
    my ($PCIDeviceList, $deviceMap) = @_;
    Debug(1, "Entered.");
    my $dev;

    foreach $dev ($PCIDeviceList->getPCIDevices()) {
	Debug(2, "Trying device " . $dev->getHash());
	my @mapEntries = $deviceMap->lookupDevice($dev->getVendorID(), $dev->getDeviceID());

	Debug(2, "Matched " . ($#mapEntries + 1) . " entries.");

	if( $#mapEntries == -1 ) {
	    next;
	}

	my @savedMapEntries;

	# Check for any additional constraints
	my $entry;
	while( $#mapEntries > -1 ) {
	    Debug(2, "Checking constraints.");
	    my $pass = 1;
	    $entry = shift(@mapEntries);

	    my %constraints = $entry->getConstraints();

	    my $constraint_param;
	    foreach $constraint_param (keys(%constraints)) {
		my $constraint_value = $constraints{$constraint_param};
		Debug(2, "Checking constraint " . $constraint_param . ' = ' . $constraint_value);
		if( $constraint_param eq "subsys_vendorid" ) {
		    if( $dev->getSubsysVendorID() != $constraint_value ) {
			Debug(2, "Did not match.  device vendor id is " . sprintf("%04x", $dev->getVendorID()) . ' constraint is ' . sprintf("%04x", $constraint_value));
			$pass = 0;
			last;
		    }
		} elsif( $constraint_param eq "subsys_deviceid" ) {
		    if( $dev->getSubsysDeviceID() != $constraint_value ) {
			$pass = 0;
			last;
		    }
		}
	    }

	    # Constraints were passed.  Keep this entry.
	    if( $pass ) {
		push(@savedMapEntries, $entry);
	    }
	}

	if( $#savedMapEntries > -1 ) {
	    # Get the set of drivers.  Drivers will be in order of default preference.
	    my @drivers = map( $_->getDriver(), @savedMapEntries);
	    my @driverNames = map ( $_->getName(), @savedMapEntries);

	    $dev->setDrivers(\@drivers);
	    $dev->setDriverNames(\@driverNames);
	}
    }
    Debug(1, "Exited.");
}

#####################################################################
# END OF PCI Driver Decision Handling Functions                     #
#####################################################################

#####################################################################
# Memory Handling Functions                                         #
#####################################################################

# Finds the actual amount of memory on the machine in megabytes
sub findTotalMemSize($) {
    my($self) = shift;
    my ($isVMnix, $isVMKernel) = @_;
    my $total_mem = 0;
    Debug(1, "Entered with isVMnix set to $isVMnix");
    if( $isVMnix ) {
       (undef, $total_mem) = readMemFromProc();
    } else {
       ($total_mem, undef) = readMemFromProc();
    }
    return $total_mem;
}

# Finds the amount of memory that the VMnix sees on the machine in megabytes
sub findVmnixMemSize($) {
    my($self) = shift;
    my ($isVMnix) = @_;

    if( $isVMnix ) {
	my ($vmnix_mem, $total_mem) = readMemFromProc();
	return($vmnix_mem);
    } else {
	return(0);
    }
}

# Find the amount of memory the vmkernel sees
sub findVmkMemSize($$) {
    my($self) = shift;
    my ($isVMnix, $isVMKernel) = @_;

    if( $isVMnix ) {
	if( $isVMKernel ) {
	    my($vm_memAvail, $vm_memTotal, $vm_memNewVMAvail) = readMemFromVmwareProc();
	    return($vm_memTotal);
	} else {
	    # Make this a debug statement since it's redudant with the total memory message
	    Debug(1, "Cannot determine the amount of VMkernel memory until the " .
		  "VMkernel is loaded.");
	    return(0);
	}
    } else {
	return(0);
    }
}

# Find the amount of available memory for the VMs
sub findVMAvailMemSize($$) {
    my($self) = shift;
    my ($isVMnix, $isVMKernel) = @_;

    if( $isVMnix ) {
	if( $isVMKernel ) {
	    my($vm_memAvail, $vm_memTotal, $vm_memNewVMAvail) = readMemFromVmwareProc();
	    return($vm_memNewVMAvail);
	} else {
	    # Make this a debug statement since it's redudant with the total memory message
	    Debug(1, "Cannot determine the amount of available memory until the " .
		  "VMkernel is loaded.");
	    return(0);
	}
    } else {
	return(0);
    }
}

# Find the memory sharing statistics for the VMkernel
sub findMemSharingInfo($$) {
    my($self) = shift;
    my ($isVMnix, $isVMKernel) = @_;
    my ($total, $actual) = (0, 0);

    if( $isVMnix ) {
	if( $isVMKernel ) {
            (undef, undef, undef, $total, $actual) = readMemFromVmwareProc();
	}
    }

    return ($total, $actual);
}

# Reads the amount of memory from the proc filesystem
# In units of kbytes
sub readMemFromProc() {
    my $mem = 0;
    my $total = 0;
    local *FD;

    my $pm_file = c_getFile('proc_meminfo');
    if( ! -r $pm_file ) {
	# Error
	Warning("File $pm_file is not readable.  Reason: $!");
	return(0);
    }

    if( !open(FD, '< ' . $pm_file) ) {
	# Error
	Warning("Could not open file $pm_file.  Reason: $!");
	return(0);
    }

    while( <FD> ) {
	if( /^MemTotal:\s*(\d+) kB/ ) {
	    $mem = $1;
        } elsif( /^MachineMem:\s*(\d+) kB/ ) {
            $total = $1;
        }
    }

    close(FD);

    return($mem, $total);
}

# Reads the amount of memory the vmkernel see from the proc filesystem.
# In units of kbytes
sub readMemFromVmwareProc() {
    my $vm_memAvail = 0;
    my $vm_memTotal = 0;
    my $vm_memNewVMAvail = 0;
    my $vm_memSharedTotal = 0;
    my $vm_memSharedActual = 0;
    local *FD;

    my $pvm_file = c_getFile('proc_vmware_meminfo');
    if( ! -r $pvm_file ) {
	# Error
	Warning("File $pvm_file is not readable.  Reason: $!");
	return(0, 0);
    }

    if( !open(FD, '< ' . $pvm_file) ) {
	# Error
	Warning("Could not open file $pvm_file.  Reason: $!");
	return(0, 0);
    }

    while( <FD> ) {
	if( /^Machine memory free: (\d+) Mbytes\/(\d+) Mbytes/ ) {
	    $vm_memAvail = ($1 * 1024);
	    $vm_memTotal = ($2 * 1024);
	} elsif( /^Maximum new 1-vcpu VM size: (\d+) Mbytes/ ) {
	    $vm_memNewVMAvail = ($1 * 1024);
        } elsif( /^Shared memory \(shared\/actual\): (\d+) Mbytes\/(\d+) Mbytes/ ) {
            $vm_memSharedTotal = ($1 * 1024);
            $vm_memSharedActual = ($2 * 1024);
        }
    }

    close(FD);

    return($vm_memAvail, $vm_memTotal, $vm_memNewVMAvail,
           $vm_memSharedTotal, $vm_memSharedActual);
}

#####################################################################
# END of Memory Handling Functions                                  #
#####################################################################

#####################################################################
# Other Support Functions                                           #
#####################################################################

# Find the lilo.conf information
sub findLILOConfig() {
    my($self) = shift;
    return( VMware::HConfig::LILOConfig->newFile(c_getFile('lilo.conf')) );
}

# Find the snmpd.conf information
sub findSNMPConfig() {
    my($self) = shift;
    my $file = c_getFile('snmpd.conf');
    local *FD;

    if ( ! -r $file ) {
        Debug(1, "Could not read SNMP configuration file '$file'.");
        return(undef);
    }
    if( !open(FD, "< $file") ) {
	Debug(1, "Could not open SNMP configuration file '$file'.");
	return(undef);
    }
    my $config = join("", <FD>);
    close(FD);
    return $config;
}

# Determine whether SNMP is enabled or not
sub findSNMPEnable() {
    my($self) = shift;

    my $filename = c_getFile('host_config');
    if (!defined($filename)) {
        Warning("Warning: Unable to open host configuration file.");
        return(undef);
    }
    my $config = VMware::Config->new();
    if (!$config->readin($filename)) {
        Warning("Warning: Unable to read host configuration file '$filename'.");
        return(undef);
    }
    # If unspecified, default to disabled
    return $config->get("snmp.enable", 0);
}

# Returns the config DB object
sub findConfigDB() {
    my($self) = shift;
    my $configDB = VMware::ConfigDB->new(c_getFile('config_db'));
    if( !$configDB->load() ) {
	Debug(1, 'Could not load database ' . c_getFile('config_db') . '.');
	# Return an empty config DB if there isn't a previous one.
    }
    return($configDB);
}


# Returns a hash containing the locations DB
sub findLocationsDB() {
    my($self) = shift;
    my $file = c_getFile('locations');
    my %db;

    open(INSTALLDB, '<' . $file)
        or error('Unable to read the installer database ' . $file . "\n\n");
    while (<INSTALLDB>) {
        if (/^remove_(\S+) (\S+)/) {
            delete $db{$1}->{$2};
        } elsif (/^(\S+) (\S+) (\S+)/) {
            $db{$1}->{$2} = $3;
        } elsif (/^(\S+) (\S+)/) {
            $db{$1}->{$2} = ($1 eq 'file') ? 0 : '';
        }
    }
    close(INSTALLDB);
    return \%db;
}


# Looks for available boot image files.  Looks only in the
# /boot directory
sub findVMBootImageFiles() {
    my($self) = shift;
    my @files;

    my $bootDir = c_getDir('boot');
    if( ! -d $bootDir ) {
	# Error /boot is not a directory
	Warning('Could not open directory ' . $bootDir . '.');
	return(@files);
    }

    if( ! (-x $bootDir && -r $bootDir ) ) {
	# Error /boot is not accessible and readable
	Warning('Could not open directory ' . $bootDir . '.');
	return(@files);
    }

    local *DIR;

    if( !opendir(DIR, $bootDir) ) {
	# Error can't opendir boot dir
	Warning('Could not open directory ' . $bootDir . '.');
	return(@files);
    }

    @files = grep { /^vmlinuz-[0-9\.-]*vmnix/ && -f "$bootDir/$_" } readdir(DIR);
    closedir DIR;

    # Untaint the data
    my $i;
    for( $i = 0; $i <= $#files; $i++ ) {
	if( $files[$i] =~ /(.*)/ ) {
	    $files[$i] = $1;
	}
    }

    # Prepend the directory information
    @files = map( "$bootDir/$_", @files );

    return(@files);
}


# Looks for available initrd files.  Looks only in the
# /boot directory
sub findVMInitrdImageFiles() {
    my($self) = shift;
    my @files;

    my $bootDir = c_getDir('boot');
    if( ! -d $bootDir ) {
	# Error /boot is not a directory
	Warning('Could not open directory ' . $bootDir . '.');
	return(\@files);
    }

    if( ! (-x $bootDir && -r $bootDir) ) {
	# Error /boot is not accessible and readable
	Warning('Could not open directory ' . $bootDir . '.');
	return(\@files);
    }

    local *DIR;

    if( !opendir(DIR, $bootDir) ) {
	# Error can't opendir boot dir
	Warning('Could not open directory ' . $bootDir . '.');
	return(\@files);
    }

    @files = grep { /^initrd-[0-9\.-]*vmnix\d*\.img$/ && -f "$bootDir/$_" } readdir(DIR);
    closedir DIR;

    # Untaint the data
    my $i;
    for( $i = 0; $i <= $#files; $i++ ) {
	if( $files[$i] =~ /(.*)/ ) {
	    $files[$i] = $1;
	}
    }

    # Prepend the directory information
    @files = map { "$bootDir/$_" } @files;

    return(@files);
}

sub findVmkDumpPartition($) {
    my($vmkScsiDisks) = @_;
    my @vmkDumpPartNames;
    my $a;

    my $disk;
    foreach $disk (@$vmkScsiDisks) {
	my @partNums = $disk->getPartNums();
	my $partnum;

	foreach $partnum (@partNums) {
	    if( VMware::HConfig::RawDisk::isPartVmkDump($disk->getPartType($partnum)) ) {
		my $name = $disk->getVmkName() . ":" . $partnum;
                my $sizeKb = $disk->getPartSectors($partnum) / 2;
                push(@vmkDumpPartNames, { NAME => $name, SIZE => $sizeKb } );
	    }
	}
    }

    return(\@vmkDumpPartNames);
}

# Builds a directory
sub findDirMapping() {
    my $self = shift;
    my %DirMap;
    my $success = 0;
    
    # One way to do it is to look at the currently mounted dirs
    if( !parseMtabFile(\%DirMap) ) {
	# Should propagate some sort of warning here.
	Warning("Could not parse mtab file.");
    } else {
	$success = 1;
    }

    # Insert better method here

    # XXX Hard code /vmfs to be the path containing all VMFS file systems.
    # We really should use a more dynamic way of determining the location
    # of the VMFS.

    $DirMap{'/vmfs'}->[0] = "vmfs";
    $DirMap{'/vmfs'}->[1] = "vmfs";

    return(\%DirMap);
}

# Data structure
# $DirMap->{'mount-point'}->[0] = device name (/dev/sda1)
# $DirMap->{'mount-point'}->[1] = type (ext2)
sub parseMtabFile($) {
    my ($DirMap) = @_;      # Reference to hash
    local *FD;

    # Can we read the file?
    if( !open(FD, '< ' . c_getFile('mtab')) ) {
	# Error reading file
	Warning("Could not open " . c_getFile('mtab'));
	return(0);
    }

    # Processing loop for mtab file
    while(<FD>) {
	chomp();      # Remove trailing newlines

	if( /^(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)/ ) {
	    my ($dev, $mntpt, $type, $opt, $dmp, $fsck) = ($1, $2, $3, $4, $5, $6);

            $mntpt =~ s/\\040/ /g;  # Unescape spaces in the path, if any.
	    $DirMap->{$mntpt}->[0] = $dev;
	    $DirMap->{$mntpt}->[1] = $type;
	}
    } # while
    close(FD);

    return(1);      # Success
}

sub _isVMnixKernel($) {
    my ($img) = @_;
    if( $img =~ /vmnix/ ) {
	return(1);
    } else {
	return(0);
    }
}

# Finds the name of the kernel
sub findKernelInfo() {
    my($self) = shift;

    my $info = "";
    local *FD;

    if( ! c_getHelper('uname') ) {
	Warning('Cannot run program uname.\n');
	return(0);
    }

    if( !open(FD, c_getHelper('uname') . ' |') ) {
	Warning('Could not run program ' . c_getHelper('uname') . '.  Reason: ' . "$!");
	return("");
    }

    # Read the line
    $info = <FD>;
    close(FD);

    # Do this to avoid taintedness issues
    if( $info =~ /(.*)/ ) {
	$info = $1;
    }
    else {
	$info = "";
    }

    return($info);
}

# Reads the file /proc/cmdline and looks for the name of the label
# that was used in booting.
sub findCurrentLabel() {
    my $self = shift;
    my $label = undef;

    local *FD;
    if( !open(FD, '< ' . c_getFile('proc_cmdline')) ) {
	Warning('Could not open file ' . c_getFile('proc_cmdline'));
	return($label);
    }

    while( <FD> ) {
	if( /BOOT_IMAGE=(\S+)/ ) {
	    $label = $1;
	}
    }

    close(FD);
    return($label);
}

# Find all the MFS partitions on the SCSI disks that we have
sub findMFS() {
   my $self = shift;
   my @mfs;

   my $dir = c_getDir('vmfs');
   local(*DIR);
   if (!opendir(DIR, $dir)) {
      Warning("Cannot open $dir. Reason: $!");
      return(0);
   }

   # Read the /vmfs directory to get the vmhba names of all the VMFSes.
   # This is fast, since none of the VMFSes need to be opened.
   # Ignore all entries without a colon, which are the VMFS logical names.
   # The logical names will be found when we run vmkfstools.
   my @vmfsdir = grep { $_ !~ /^\.\.?/ && $_ =~ /:/ } readdir(DIR);
   closedir(DIR);

   foreach my $name (@vmfsdir) {
      my ($disk, $part) = VMware::HConfig::RawDisk::unmakeLogVmkPartName($name);
      my $vmfs = VMware::HConfig::MFSFileSystem->new($disk, $part);

      if (!defined($vmfs)) {
         next;
      }

      # XXX Hard coded mount point for a VMFS to /vmfs/$name.  The
      #     notion of a mount point per VMFS is somewhat wrong, but we
      #     need to keep this to be able to translate from a vmkernel
      #     name for a VMFS file to the Console OS name. -jhu
      $vmfs->setMountPoint("$dir/$name");

      push(@mfs, $vmfs);
   }

   return(\@mfs);
}


# Returns an array of structures containing information about a VMnix module.
# Format of structure:
#
# $modinfo->{MODULE} = name of module. ie vmnixmod
# $modinfo->{SIZE} = size of module
# $modinfo->{USED} = usage count. 0 means module is not in use and can be unloaded.
# $modinfo->{LOADED} = indicates of the module is loaded on the Console OS.
# $modinfo->{ALIASES} = aliases for module, if any, in /etc/modules.conf
sub findVMnixModules() {
    my($self) = shift;
    local *FD;
    my @modules;

    # Load module info using lsmod

    if( ! c_getHelper('lsmod') ) {
	Warning('Cannot not execute "lsmod".');
	return(\@modules);
    }

    if( !open(FD, c_getHelper('lsmod') . ' |') ) {
	Warning('Could not execute ' . c_getHelper('lsmod') . '. Reason: ' . "$!");
	return(\@modules);
    }

    while( <FD> ) {
	chomp();

	if( /^Module\s+Size\s+Used/ ) {
	    next;
	}
	elsif( /^(\S+)\s+(\d+)\s+(\d+)/ ) {
	    my $mod = $1;
	    my $size = $2;
	    my $used = $3;

	    my $modinfo = {};
	    $modinfo->{MODULE} = $mod;
	    $modinfo->{SIZE} = $size;
	    $modinfo->{USED} = $used;
	    $modinfo->{LOADED} = 1;
            $modinfo->{ALIASES} = [];

	    push(@modules, $modinfo);
	}
    }
    close(FD);

    # Load alias info using /etc/modules.conf

    my $modulesFile = c_getFile('modules_conf');
    if (!(-f $modulesFile) || !(-r $modulesFile)) {
	Warning("$modulesFile is not accessible.  Reason: $!");
	return(\@modules);
    }

    if (!open(FD, "< $modulesFile")) {
        Warning("Cannot open $modulesFile.  Reason: $!");
        return(\@modules);
    }

    while (<FD>) {
        if (/^\s*alias\s+(\S+)\s+(\S+)/) {
            my $alias = $1;
            my $module = $2;
            my $modinfo;
            my $found = 0;
            for (my $i = 0; $i < scalar(@modules); $i++) {
                $modinfo = $modules[$i];
                if ($modinfo->{MODULE} eq $module) {
                    $found = 1;
                    last;
                }
            }
            if (!$found) {
               # Couldn't find an existing entry in the list.  Add it, but
               # don't mark it as being loaded
               $modinfo = {};
               $modinfo->{MODULE} = $module;
               $modinfo->{SIZE} = 0;
               $modinfo->{USED} = 0;
               $modinfo->{LOADED} = 0;
               $modinfo->{ALIASES} = [];
               push (@modules, $modinfo);
            }
            # Add alias to the list of aliases for module
            my $aliases = $modinfo->{ALIASES};
            push(@$aliases, $alias);
        }
    }
    close(FD);

    return(\@modules);
}

# Returns an array of structures containing information about a VMnix module.
# Format of structure:
#
# $modinfo->{MODULE} = name of module. ie aic7xxx
# $modinfo->{SIZE} = size of module
# $modinfo->{ID} = ID of module
sub findVMKernelModules() {
    my($self) = shift;
    local *FD;
    my @modules;

    if( ! c_getHelper('list_vmkmod') ) {
	Warning('Cannot not execute "list_vmkmod".');
	return(\@modules);
    }

    if( !open(FD, c_getHelper('list_vmkmod') . ' 2>/dev/null |') ) {
	Warning('Could not execute ' . c_getHelper('list_vmkmod') . '. Reason: ' ."$!");
	return(\@modules);
    }

    while( <FD> ) {
	chomp();

	# Name     R/O Addr  Length   R/W Addr  Length   ID      Loaded
	if( /^(\S+)\s+0x([0-9a-f]+)\s+0x([0-9a-f]*)\s+0x([0-9a-f]+)\s+0x([0-9a-f]*)\s+(\d+)\s+(Yes)/ ) {
	    my $mod = $1;
	    my $rosize = hex($3);
	    my $rwsize = hex($5);
	    my $id = $6;

	    my $modinfo = {};
	    $modinfo->{MODULE} = $mod;
	    $modinfo->{SIZE} = $rosize + $rwsize;
	    $modinfo->{ID} = $id;

	    push(@modules, $modinfo);
	}
    }
    close(FD);

    return(\@modules);
}


# Finds the number of CPUs
sub findNumCPUs() {
    my($self) = shift;
    my $ncpus = -1;
    local(*FD);

    my $dir = c_getDir('cpu');
    if (!$dir) {
       Warning('Cannot open directory "cpu".');
    } elsif (!open(FD, "$dir/ncpus")) {
       Warning("Cannot open $dir/ncpus. Reason: $!");
    } else {
       while (<FD>) {
          if (/(\d+)\s+physical/) {
             $ncpus = $1;
          }
       }
       close(FD);
    }
    
    return $ncpus;
 }


# Finds the number of logical CPUs
sub findNumLogCPUs() {
    my($self) = shift;
    my $nlogcpus = -1;
    local(*FD);

    my $dir = c_getDir('cpu');
    if (!$dir) {
       Warning('Cannot open directory "cpu".');
    } elsif (!open(FD, "$dir/ncpus")) {
       Warning("Cannot open $dir/ncpus. Reason: $!");
    } else {
       while (<FD>) {
          if (/(\d+)\s+logical/) {
             $nlogcpus = $1;
          }
       }
       close(FD);
    }

    return $nlogcpus;
}


# Finds the number of running Vmm worlds
sub findRunningVmmWorlds($) {
    my($self) = shift;
    my($isVMKernel) = @_;
    my @allworlds;
    my @vmm_worlds;

    if( !$isVMKernel ) {
	return(\@vmm_worlds);
    }

    my $dir = c_getDir('proc_vm');
    if( ! -d $dir ) {
	# Error /proc/vmware/vm is not a directory
	Warning("Could not open $dir. Reason: $!");
	return(\@vmm_worlds);
    }

    local(*DIR);
    if( !opendir(DIR, $dir) ) {
	Warning("Cannot open $dir. Reason: $!");
	return(0);
    }

    @allworlds = grep { $_ =~ /^[0-9]+$/ && -d "$dir/$_" } readdir(DIR);
    closedir(DIR);

    my $world;
    foreach $world (@allworlds) {
	my $file = "$dir/$world/cpu/status";
	if( ! -f "$file" ) {
	    next;
	}

	local(*FD);
	if( !open(FD, "< $file") ) {
	    Warning("Could not open file $file.");
	    next;
	}
	my $isVmm = 0;
	while(<FD>) {
	    if( m/$world vmm/ ) {
		$isVmm = 1;
	    }
	}
	close(FD);
	if( $isVmm ) {
	    push(@vmm_worlds, $world);
	}
    }
    @vmm_worlds = sort { $a <=> $b } @vmm_worlds;
    return(\@vmm_worlds);
}

# Finds the PCI device map
sub findPCIDeviceMap() {
    my($self) = shift;
    my @mapFiles = (c_getFile('pcidev_map'));
    if (-r c_getFile('pcidev_map_local')) {
        # Add local file to front of list so it gets higher priority
        unshift @mapFiles, c_getFile('pcidev_map_local');
    }
    return(VMware::HConfig::PCIDeviceMap->new(@mapFiles));
}


# Does a config database file exist
sub canReadConfigDB() {
    my $file = c_getFile('config_db');
    return( -f $file && -r $file );
}

# Does a check to see if any duplicate VMFS names are found
sub checkDuplicateVmfsNames($) {
    my($self) = shift;
    my($all_vmfs) = @_;

    my $vmfs;
    my %vmfsNames;

    # Build a hash of usage counts
    foreach $vmfs (@$all_vmfs) {
	my $name = $vmfs->getLogName();
	if( defined($name) && $name ne "" ) {
	    if( defined($vmfsNames{$name}) ) {
		$vmfsNames{$name}++;
	    } else {
		$vmfsNames{$name} = 1;
	    }
	}
    }

    foreach $vmfs (keys(%vmfsNames)) {
	my $usage = $vmfsNames{$vmfs};
	if( $usage > 1 ) {
	    Warning("Volume name $vmfs is used by more than one VMFS.  Volume names should be unique, " .
		    "or you may experience problems accessing VMFS disks off these volumes.");
	}
    }
}

# Returns 0 if an error occurred.  1 on success.  -1 if no dump partition was
# found.
sub setVmkDumpPartition() {
    my($self) = shift;

    Debug(1, "Entered.");

    if( !c_getHelper('vmkdump_set') ) {
	Warning("Cannot execute vmkdump.");
	return(0);
    }

    # By policy we always use the first vmkdump partition.  The ordering is done
    # by Vmk SCSI Adapter number followed by Disk number
    my @vmkdumpParts = @{ $self->getVmkDumpPartitions() };
    my $partname;
    if( $#vmkdumpParts == -1 ) {
	# No dump partition
	Debug(2, "No dump partition exists.  Skipping set.");
	return(-1);
    }
    $partname = $vmkdumpParts[0];

    if( System(c_getHelper('vmkdump_set') . " " . shell_string($partname) ) ) {
	Warning("Execution of " . c_getHelper('vmkdump_set') . " failed.  Reason: $!");
	return(0);
    }

    return(1);
}

# Unsets the vmkdump partition.  Returns 1 on success.  0 on failure.
sub unsetVmkDumpPartition() {
    my($self) = shift;
    Debug(1, "Entered.");

    if( !c_getHelper('vmkdump_unset') ) {
	Warning("Cannot execute vmkdump.");
	return(0);
    }

    if( System(c_getHelper('vmkdump_unset')) ) {
	Warning("Execution of " . c_getHelper('vmkdump_unset') . " failed.  Reason: $!");
	return(0);
    }

    return(1);
}

sub checkCanRunLicense() {
    if( $gUseSIGCHLD ) {
	return(checkCanRunLicenseUseSIGCHLD());
    } else {
	return(checkCanRunLicenseNoSIGCHLD());
    }
}

# Checks if a license is needed to run VMware.  Returns (1, message) if a license is not
# needed.  Otherwise return (0, message).
sub checkCanRunLicenseUseSIGCHLD() {
    my($self) = shift;
    my $message = '';;

    Debug(1, "Entered.");

    if( !c_getHelper('license_check') ) {
	return(0, "Could not run 'license_check' program because it does not exist.");
    }

    my $command = c_getHelper('license_check') . ' 2>/dev/null';
    $message = `$command`;
    my $rc = $? >> 8;
    if( $rc == -1 ) {
	return(0, "Could not execute " . c_getHelper('license_check') . ".  Reason: $!");
    }

    if( $rc ) {
	Debug(1, "License is needed ($rc).");
	return(0, $message);
    } else {
	Debug(1, "License is not needed ($rc).");
	return(1, $message);
    }
}

# Checks if a license is needed to run VMware.  Returns (1, message) if a license is not
# needed.  Otherwise return (0, message).
#
# This version of the function is needed if SIGCHLD is ignored.
sub checkCanRunLicenseNoSIGCHLD() {
    my($self) = shift;
    my $message = '';;

    Debug(1, "Entered.");

    if( !c_getHelper('license_check') ) {
	return(0, "Could not run 'license_check' program because it does not exist.");
    }

    # We have to dump the output to a file because of the way that serverd integrates
    # other components such as the signal library.  SIGCHLD is ignored in vmserverd and
    # that seems to wreak havoc on the exit codes of child processes.  System appears to
    # get the child process' exit code correctly, but backticks and open do not. -jhu

    my $tmpfile = mktmpfile();
    my $rc = System2(c_getHelper('license_check') . ' > ' . shell_string($tmpfile) . ' 2>/dev/null');
    if( $rc == 127 ) {
	return(0, "Could not execute " . c_getHelper('license_check') . ".  Reason: $!");
    }

    local(*FD);

    if( !open(FD, "< $tmpfile") ) {
	return(0, "Could not read output file " . c_getHelper('license_check') . ".  Reason: $!");
    }

    # Slurp the entire text string in
    while(<FD>) {
	$message .= $_;
    }
    close(FD);
    unlink($tmpfile);

    $rc = $rc >> 8;

    if( $rc ) {
	Debug(1, "License is needed ($rc).");
	return(0, $message);
    } else {
	Debug(1, "License is not needed ($rc).");
	return(1, $message);
    }
}

sub registerNewLicense($) {
    if( $gUseSIGCHLD ) {
	return(registerNewLicenseUseSIGCHLD(@_));
    } else {
	return(registerNewLicenseNoSIGCHLD(@_));
    }
}

# Registers a new serial number for a license.  Returns (1, message) if a license is not
# needed.  Otherwise return (0, message).
sub registerNewLicenseUseSIGCHLD($) {
    my($self) = shift;
    my($serial_num) = @_;
    my $message = '';

    Debug(1, "Entered.");

    if( !c_getHelper('license_register') ) {
	return(0, "Could not run 'license_register' program because it does not exist.");
    }

    if( !c_getHelper('echo') ) {
	return(0, "Could not run 'echo' program because it does not exist.");
    }

    my $command = c_getHelper('echo') . ' ' . shell_string($serial_num) . ' | ' .
	c_getHelper('license_register');
    $message = `$command 2>&1 1>/dev/null`;
    my $rc = $? >> 8;
    if( $rc == -1 ) {
	return(0, "Could not execute " . c_getHelper('license_register') . ".  Reason: $!");
    }

    if( $rc ) {
	Debug(1, "License is needed ($rc).");
	return(0, $message);
    } else {
	Debug(1, "License is not needed ($rc).");
	return(1, $message);
    }
}

# Registers a new serial number for a license.  Returns (1, message) if a license is not
# needed.  Otherwise return (0, message).
#
# This version of the function is needed if SIGCHLD is ignored.
sub registerNewLicenseNoSIGCHLD($) {
    my($self) = shift;
    my($serial_num) = @_;
    my $message = '';

    Debug(1, "Entered.");

    if( !c_getHelper('license_register') ) {
	return(0, "Could not run 'license_register' program because it does not exist.");
    }

    if( !c_getHelper('echo') ) {
	return(0, "Could not run 'echo' program because it does not exist.");
    }

    # We have to dump the output to a file because of the way that serverd integrates
    # other components such as the signal library.  SIGCHLD is ignored in vmserverd and
    # that seems to wreak havoc on the exit codes of child processes.  System appears to
    # get the child process' exit code correctly, but backticks and open do not. -jhu

    local(*FD);

    my $command = c_getHelper('echo') . ' ' . shell_string($serial_num) . ' | ' .
	c_getHelper('license_register');

    my $tmpfile = mktmpfile();
    my $rc = System2($command . ' > ' . shell_string($tmpfile) . ' 2>&1 1>/dev/null');
    if( $rc == 127 ) {
	return(0, "Could not execute " . c_getHelper('license_register') . ".  Reason: $!");
    }

    if( !open(FD, "< $tmpfile") ) {
	return(0, "Could not read output file " . c_getHelper('license_check') . ".  Reason: $!");
    }

    # Slurp the entire text string in
    while(<FD>) {
	$message .= $_;
    }
    close(FD);
    unlink($tmpfile);

    $rc = $rc >> 8;

    if( $rc ) {
	Debug(1, "License is needed ($rc).");
	return(0, $message);
    } else {
	Debug(1, "License is not needed ($rc).");
	return(1, $message);
    }
}

sub getLicenseInfo() {
    if( $gUseSIGCHLD ) {
	return(getLicenseInfoUseSIGCHLD(@_));
    } else {
	return(getLicenseInfoNoSIGCHLD(@_));
    }
}

# Queries the license for each attribute in the argument list,
# and returns the values in a list.
sub getLicenseInfoUseSIGCHLD() {
    my($self) = shift;

    Debug(1, "Entered.");

    if( !c_getHelper('license_info') ) {
	Warning("Could not run 'license_info' program because it does not exist.");
        return;
    }

    my @results = ();
    foreach my $attr (@_) {
        my $command = c_getHelper('license_info') . ' ' . shell_string($attr);
        my $result = `$command 2>/dev/null`;
        my $rc = $? >> 8;
        if( $rc == -1 ) {
            Warning("Could not execute '$command'.  Reason: $!");
            $result = undef;
        }
        push @results, $result;
    }
    return @results;
}

# Queries the license for each attribute in the argument list,
# and returns the values in a list.
#
# This version of the function is needed if SIGCHLD is ignored.
sub getLicenseInfoNoSIGCHLD() {
    my($self) = shift;

    Debug(1, "Entered.");

    if( !c_getHelper('license_info') ) {
	Warning("Could not run 'license_info' program because it does not exist.");
        return;
    }

    # We have to dump the output to a file because of the way that serverd integrates
    # other components such as the signal library.  SIGCHLD is ignored in vmserverd and
    # that seems to wreak havoc on the exit codes of child processes.  System appears to
    # get the child process' exit code correctly, but backticks and open do not. -jhu

    my @results = ();
    foreach my $attr (@_) {
        my $tmpfile = mktmpfile();
        my $command = c_getHelper('license_info') . ' ' . shell_string($attr);
        my $rc = System2($command . ' > ' . shell_string($tmpfile) . ' 2>/dev/null');
        if( $rc == 127 ) {
            Warning("Could not execute '$command'.  Reason: $!");
            push @results, undef;
            next;
        }

        local(*FD);

        if( !open(FD, "< $tmpfile") ) {
            Warning("Could not read output file " . c_getHelper('license_info') . " $attr.  Reason: $!");
            push @results, undef;
            next;
        }

        # Slurp the entire text string in
        my $result = "";
        while(<FD>) {
            $result .= $_;
        }
        close(FD);
        unlink($tmpfile);

        push @results, $result;
    }
    return @results;
}

# Return swap info: Total, free, used kbytes of swap space, and swap volume/filename
# Returns all zeroes if swapper is not enabled
sub getSwapInfo {
    my($self) = shift;
    my ($total, $unused, $used) = (0, 0, 0);
    my $swapvol = "";
    my $swapfile = "";
    my $unrsvd = 0;
    my $file = c_getFile('swap_stats');

    local *FD;
    if (!open(FD, "< $file")) {
        Warning("Could not open file $file");
    } else {
        while (<FD>) {
            # Slots are 4K each; return size in KB
            if (/device\s*=\s*(\S+)/) {
               $swapvol = $1;
               # When logical name is used, a bogus ':0:0:0' is appended
               $swapvol =~ s/:0:0:0$//;
            } elsif (/filename\s*=\s*(\S+)/) {
               $swapfile = $1;
            }
        }
        close(FD);
    }

    $file = c_getFile('mem_sched');
    local *FD;
    if (!open(FD, "< $file")) {
        Warning("Could not open file $file");
    } else {
        while (<FD>) {
           if (/TOTAL\s/) {
              $_ =~ s/\///g; # remove the unwanted forward slashes
              my @memInfo = split ' ', $_;
              # 9th index contains total swap space used
              my $used = $memInfo[9];
              last;
           }
        }
        close(FD);
    }
    
    $file = c_getFile('proc_vmware_meminfo');
    local *FD;
    if (!open(FD, "< $file")) {
        Warning("Could not open file $file");
    } else {
        my $pattern = "Unreserved swap space";
        while (<FD>) {
          if (/$pattern\s*:\s*(\d+)\s*Mbytes\/\s*(\d+)/) {
             # $1 contains unreserved swap space, and $2 contains Total
             $unrsvd = $1 * 1024;
             $total  = $2 * 1024;
             $unused = $total - $used;
             last;
           }
        }
        close(FD);
    }
    
    return ($total, $unused, $used, $unrsvd, $swapvol, $swapfile);
}

# Starts the swapper using the specified VMFS file and size
sub startSwapper {
    my($self) = shift;
    my($volume, $filename, $size) = @_;

    my $command =
        c_getHelper("start_swapper") . ' ' .
        shell_string("${size}M") . ' ' .
        shell_string("${volume}:${filename}");

    if (System($command)) {
        Warning("Could not enable $size MB swap file '$filename' " .
                "on VMFS partition '$volume'.");
        return 0;
    }
    return 1;
}

# Configures clustering support for a VMkernel NIC
sub setVMNetAdapterClustering {
    my($self) = shift;
    my($name, $clustering) = @_;

    my $file = c_getDir('proc_vmware_net') . "/$name/config";
    if( ! -w $file ) {
        Warning("Could not write file '$file'.");
        return 0;
    }
    if( !open(FD, "> $file") ) {
        Warning("Could not open file '$file'.");
        return 0;
    }
    print FD "InterruptClustering " . ($clustering ? "Yes" : "No");
    close(FD);
    return 1;
}

# Check that TCP/IP parameters are valid
# Returns true if they are valid.  false if not valid.
sub checkTcpipParams {
   my($self) = shift;
   my($ipAddr, $gatewayAddr, $nicName) = @_;

   if ($nicName && !grep($_ eq $nicName, @{$self->getAvailNetAdapters()})) {
      Warning("Cannot use interface $nicName.");
      return 0;
   }

   # Check for malformed IP address
   if ($ipAddr && $ipAddr !~ /^\d+\.\d+\.\d+\.\d+$/) {
      Warning("IP address '$ipAddr' is malformed.");
      return 0;
   }

   if ($gatewayAddr && $gatewayAddr !~ /^\d+\.\d+\.\d+\.\d+$/) {
      Warning("Gateway address '$gatewayAddr' is malformed.");
      return 0;
   }

   return 1;
}

# Get the name of current TCP/IP instance.  Currently, we support only one
# TCP/IP instance.  Get the name of that instance.
sub getCurrentTcpipName() {
   my($self) = shift;
   my %tcpipInfo = %{$self->getVMNetTcpip()};
   my @tcpipNames = keys(%tcpipInfo);
   return $tcpipNames[0];
}

# Adds the TCP/IP setting
# Returns the name of the TCP/IP module that was set
sub addVMNetTcpip {
   my($self) = shift;
   my($ipAddr, $gatewayAddr, $nicName) = @_;

   my $name;
   $name = $self->getCurrentTcpipName();
   if (defined($name)) {
      Warning("TCPIP instance already defined.  Cannot define more than one instance.");
      return undef;
   }

   if (!$ipAddr || !$gatewayAddr || !$nicName) {
      Warning("Invalid TCP/IP instance parameters.");
      return undef;
   }
   if (!$self->checkTcpipParams($ipAddr, $gatewayAddr, $nicName)) {
      return undef;
   }

   # Load the driver if it isn't already loaded
   my $vmkModules = $self->getVMKernelModules();
   if (!grep($_->{MODULE} eq $gVmkDeviceType_tcpip, @$vmkModules)) {
      if (System(c_getHelper('load_vmkmod') . ' ' .
                 shell_string(c_getDir("vmkmodule") . "/tcpip") . ' ' .
                 shell_string($gVmkDeviceType_tcpip) . ' ' .
                 shell_string("ipAddrStr=$ipAddr") . ' ' .
                 shell_string("gatewayAddrStr=$gatewayAddr") . ' ' .
                 shell_string("nicName=$nicName"))) {
         Warning("Failed to load TCP/IP module.\n");
         return undef;
      }
   }

   # Load the migration module if it isn't already loaded
   # XXX this doesn't really belong here, but it should suffice for 
   # the time being (since neither of these modules are useful without the other)
   $vmkModules = $self->getVMKernelModules();
   if (!grep($_->{MODULE} eq $gVmkDeviceType_migration, @$vmkModules)) {
      if (System(c_getHelper('load_vmkmod') . ' ' .
                 shell_string(c_getDir("vmkmodule") . "/migration") . ' ' .
                 shell_string($gVmkDeviceType_migration))) {
         Warning("Failed to load migration module.\n");
         return undef;
      }
   }

   # Now that the module is loaded, look for the TCP/IP info.  Currently, we
   # only support one instance.
   $name = $self->getCurrentTcpipName();
   if (!defined($name)) {
      Warning("Failed to find TCP/IP instance after loading module.\n");
      return undef;
   }

   # Save the TCP/IP configuration to a file so that it takes effect across reboots.
   $self->setVMNetTcpipCfg($name, $ipAddr, $gatewayAddr, $nicName);
   return $name;
}

# Removes the TCP/IP setting for the next boot.
sub removeVMNetTcpip {
   my($self) = shift;
   my($name) = @_;      # Currently ignored

   my $configDB = $self->getConfigDB();

   $configDB->unset("tcpip.nicName");
   $configDB->unset("tcpip.ipAddr");
   $configDB->unset("tcpip.gatewayAddr");

   return $self->commitConfigDB();
}

# Sets the TCP/IP settings.
sub setVMNetTcpip {
   my($self) = shift;
   my($name, $ipAddr, $gatewayAddr, $nicName) = @_;

   # Ensure that only an existing TCPIP instance is set.
   my $file = c_getDir('proc_vmware_net') . "/$name/config/if0";
   if (! -f $file) {
      Warning("Invalid TCPIP instance name $name.");
      return 0;
   }

   if (!$self->checkTcpipParams($ipAddr, $gatewayAddr, $nicName)) {
      return undef;
   }

   my $rc = 1;
   if ($nicName) {
      $rc &= internal_echo($file, "Interface $nicName");
   }
   if ($ipAddr) {
      $rc &= internal_echo($file, "IPAddress $ipAddr");
   }
   if ($gatewayAddr) {
      $rc &= internal_echo($file, "Gateway $gatewayAddr");
   }
  
   if (!$rc) {
      Warning("Failed to write to file $file.");
      return 0;
   }

   if (!$self->setVMNetTcpipCfg($name, $ipAddr, $gatewayAddr, $nicName)) {
      Warning("Failed to save TCP/IP settings to configuration database.  " .
              "Settings may not take effect on next boot.");
   }

   return 1;
}

# Sets the TCP/IP settings.  Ignores values that are not defined.
# Returns true on success.
#
# name is ignored
sub setVMNetTcpipCfg {
   my($self) = shift;
   my($name, $ipAddr, $gatewayAddr, $nicName) = @_;

   # Store settings persistently in configuration database and set
   # proc node if possible.
   my $configDB = $self->getConfigDB();
   if ($nicName) {
      $configDB->set("tcpip.nicName", $nicName);
   }
   if ($ipAddr) {
      $configDB->set("tcpip.ipAddr", $ipAddr);
   }
   if ($gatewayAddr) {
      $configDB->set("tcpip.gatewayAddr", $gatewayAddr);
   }

   if (!$self->commitConfigDB()) {
      Warning("Could not persistently save TCPIP instance settings.");
      return 0;
   }

   return 1;
}

1;
