#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# Serialize.pm
# 
# A class that is used to serialize HConfig objects.

package VMware::HConfig::Serialize;

use VMware::ExtHelpers qw(&Warning &Debug);
use VMware::DOMAccess;

#####################################################################
# XML Serialization Support Functions from SSHostRemote.pm          #
#####################################################################

# XXX We should put all the relevant SSHost (de)serialization functions.
#     here.  Currently they're strewn all about the place. -jhu

# Serializes the LILOConfig information.
#
# Input format:
#   out := the DOMAccess node to write the XML information
#   lilo := the LILOConfig object to be serialized
# 
# Output format:
#   .default := the default configuration
#   .configs.config[] := the boot configuration
#   .configs.config[].label := label of the configuration that is currently booted
#   .configs.config[].type := other | image
#   .configs.config[].isVmnix := true | false
#   .configs.config[].kernel_image := kernel image to be booted
#   .configs.config[].initrd := initrd that should be loaded (optional)
#   .configs.config[].cpci := cpci line
#   .configs.config[].memory := amount of memory to allocate
#   .configs.config[].root := root device
#   .configs.config[].boot_vmkmodules := vmkmodules that should be loaded
#
sub fromLILOConfigToXML($$) {
    my($out, $lilo) = @_;

    my $default = $lilo->getDefaultImage();
    $out->setValue(".default", $default);

    my @labels = @{ $lilo->getBootImageLabels() };

    my $label;
    foreach $label (@labels) {
	my $node = VMware::DOMAccess->new('config');

	my $img = $lilo->getBootImageByLabel($label);
	if( !defined($img) ) {
	    Warning("Invalid label $label specified");
	    return(0);
	}

	$node->setValue(".label", $img->getLabel);
	if( $img->usesVMnixImage() ) {
	    $node->setValue(".isVmnix", "true");
	} else {
	    $node->setValue(".isVmnix", "false");
	}
	$node->setValue(".kernel_image", $img->getImage());
	if( $img->getInitrd() ne "" ) {
	    $node->setValue(".initrd", $img->getInitrd);
	}
	if( $img->usesVMnixImage() ) {
	    $node->setValue(".cpci", $img->getCPCI());
	    $node->setValue(".memory", $img->getMemSize());
	    $node->setAttribute(".memory", "units", "MB");
	}
	$node->setValue(".root", $img->getRootDev());
	$node->setValue(".type", $img->getType());
	
	$out->addSubTree('.configs', $node);
	$node->dispose();
    }
    
    return(1);
}

# Deserializes the LILOConfig information.
#
# Input format:
#   in := the DOMAccess node containing the serialized LILOConfig object.
#   lilo := the LILOConfig object to fill the information
#
# Output:
#   the LILOConfig object on success.  undef on failure.
#
# See fromLILOConfigToXML() for the serialization format.
# 
sub toLILOConfigFromXML($$) {
    my($in, $lilo) = @_;

    my %labelTouched;

    # Build each of the individual configurations
    my @configs = $in->listElementNames(".configs");

    my $name;
    foreach $name (@configs) {
	my $new_flag = 0;
	my $key = ".configs.$name";
	my $isVmnix = ($in->getValue($key . ".isVmnix") eq "true");

	my $type = $in->getValue($key . ".type");
	my $label = $in->getValue($key . ".label");
	my $image = $in->getValue($key . ".kernel_image");
	my $root = $in->getValue($key . ".root");

	my $memory = $in->getValue($key . ".memory");
	my $initrd = $in->getValue($key . ".initrd");
	my $cpci = $in->getValue($key . ".cpci");

	# Make sure this is a unique boot image
	my $img = $lilo->getBootImageByLabel($label);
	if( !defined($img) ) {
	    $new_flag = 1;
	    $img = $lilo->createBootImage($type, $image);
	}

	$img->setLabel($label);
	$img->setRootDev($root);
	$img->setImage($image);
	if( $isVmnix ) {
	    $img->setMemSize($memory);
	    $img->setCPCI($cpci);
	}
	if( $initrd ne "" ) {
	    $img->setInitrd($initrd);
	}

	if( $new_flag ) {
	    $lilo->addBootImage($img);
	}

	# Mark each of the LILO entries that were touched
	$labelTouched{$label} = 1;
    }

    # Remove LILO entries that were not touched.  These are considered
    # to be removed.
    foreach $label (@{ $lilo->getBootImageLabels() }) {
	if( !defined($labelTouched{$label}) ) {
	    $lilo->deleteBootImage($label);

	    Debug(2, "Removed a boot image '$label'");
	}
    }

    my $default = $in->getValue(".default");

    # Make sure that the default image exists in LILO
    if( !defined($lilo->getBootImageByLabel($default)) ) {
	Warning("Default image specifies an invalid LILO configuration.");
    }

    $lilo->setDefaultImage($default);

    return( $lilo );
}


# Serializes the hardware configuration database.
#
# Input format:
#   out := the DOMAccess node to write the XML information
#   configDB := the ConfigDB object to be serialized
# 
# Output format:
#   .config[].name := a config key
#   .config[].value := a config value
#
sub fromConfigDBToXML($$) {
    my($out, $configDB) = @_;
    my $key;
    my $value;

    my @keys = sort( $configDB->getKeys() );

    foreach $key (@keys) {
	my $node = VMware::DOMAccess->new('config');

	$value = $configDB->get($key);

	$node->setAttribute('.', 'name', $key);
	$node->setValue('.', $value);

	$out->addSubTree('.', $node);
	$node->dispose();
    }

    return(1);
}


# Deserializes the hardware configuration database..
#
# Input format:
#   in := the DOMAccess node containing the serialized LILOConfig object.
#   configDB := the configDB object to fill the information
#
# Output:
#   the configDB object on success.  undef on failure.
#
# See fromConfigDBToXML() for the serialization format.
# 
sub toConfigDBFromXML($$) {
    my($in, $configDB) = @_;
    # Build each of the individual configurations
    my @keys = $in->listElementNames(".");

    my $key;
    foreach $key (@keys) {
	my $name = $in->getAttribute(".$key", 'name');
	my $value = $in->getValue(".$key");
	$configDB->set($name, $value);
    }
    return($configDB);
}

#####################################################################
# End of XML Serialization Support Functions from SSHostRemote.pm   #
#####################################################################


#####################################################################
# XML Serialization Support Functions from RawDiskRemote.pm         #
#####################################################################

# Serializes a partition table object.
#
# Input format:
#   out := the DOMAccess node to write the XML information
#   partTable := the LILOConfig object to be serialized
# 
# Output format:
#   .partition[].number := the partition number
#   .partition[].type := the hexadecimal type of the partition
#   .partition[].name := the vmk name of the partition
#   .partition[].typeName := the name for the type
#   .partition[].start_cylinder := the start cylinder from 0
#   .partition[].end_cylinder := the end cylinder (total - 1)
#   .partition[].size := the size of the partition in Mbytes
#
sub fromPartitionTableToXML($$) {
    my($out, $partTable) = @_;

    if( !defined($partTable) ) {
	Warning("Undefined partition table.");
	return(0);
    }

    # Partition Handling functions
    my @partnums = $partTable->getPartNums();
    my $partnum;
    foreach $partnum (@partnums) {
	my $part_node = VMware::DOMAccess->new('partition');
	
	my $part_vmkname = VMware::HConfig::RawDisk::makeLogVmkPartName($partTable->getDiskName(),
									$partnum);
	if( !defined($part_vmkname) ) {
	    Warning("Invalid partition name $part_vmkname.");
	    return(0);
	}
	my $part_type = sprintf("%02x", $partTable->getPartType($partnum));
	my $part_type_name = $partTable->getPartTypeName($partnum);
	my $part_scyl = $partTable->getPartStartCyl($partnum);
	my $part_ecyl = $partTable->getPartEndCyl($partnum);
	my $part_sect = $partTable->getPartSectors($partnum);
	my $part_mb_size = sprintf("%d", $part_sect / 2 / 1024);
	
	$part_node->setValue(".number", $partnum);
	$part_node->setValue(".type", $part_type);
	$part_node->setValue(".name", $part_vmkname);
	$part_node->setValue(".typeName", $part_type_name);
	$part_node->setValue(".start_cylinder", $part_scyl);
	$part_node->setValue(".end_cylinder", $part_ecyl);
	$part_node->setValue(".size", $part_mb_size);
	$part_node->setAttribute(".size", "units", "MB");
	
	$out->addSubTree('.', $part_node);
	$part_node->dispose();
    }
	
    return(1);
}


# Deserializes a partition table object.
#
# Input format:
#   in := the DOMAccess node containing the serialized LILOConfig object.
#   partitionTable := the configDB object to fill the information
#
# Output:
#   the partitionTable object on success.  undef on failure.
#
# See fromPartitionTableToXML() for the serialization format.
# 
sub toPartitionTableFromXML($$) {
    my ($in, $partTable) = @_;

    Debug(1, "Entered");

    my @partitionKeys = $in->listElementNames('.');
    my $partitionKey;

    foreach $partitionKey (@partitionKeys) {
	my $part_node = $in->getSubTree('.' . $partitionKey);

	my $partNum = $part_node->getValue('.number');
	my $scyl = $part_node->getValue('.start_cylinder');
	my $ecyl = $part_node->getValue('.end_cylinder');
	my $size = $part_node->getValue('.size');
	my $type = hex($part_node->getValue('.type'));

	$partTable->addPart($partNum, 
			    $scyl, 
			    $ecyl, 
			    $type, 
			    $size * 1024 * 2);

	$part_node->dispose();
    }
    return($partTable);
}

#####################################################################
# End of XML Serialization Support Functions from RawDiskRemote.pm  #
#####################################################################

1;
