#!/usr/bin/perl -w 

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# VmkConfig.pm
# 
# Code to read/write vmkernel config data
#

package VMware::HConfig::VmkConfig;
use strict;
use Exporter   ();
use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
use VMware::ExtHelpers qw(&Warning &Panic &Debug &shell_string &mktmpfile &safe_move);
use VMware::HConfig::Constant qw(&c_getDir &c_getFile);
use VMware::Config;

BEGIN {
    # set the version for version checking
    $VERSION     = 1.00;
    @ISA         = qw(Exporter);
    @EXPORT_OK   = qw();
}

use vars @EXPORT_OK;

# Specify the valid range for each variable.
#
# XXX This range table should be populated to reflect the current
# state of the config variables. -jhu
my %ConfigValidInfo = (
    'HostIRQSharing' => [0, 1]
);

#####################################################################
# Public Static Methods (indep. of each object)                     #
#####################################################################

# Gets a list of all the vmkernel config items in the proc filesystem
sub getAllItems() {
    local(*CONFIGDIR);
    local(*SUBDIR);
    my @configs;

    my $configDir = c_getDir('proc_vmk_config');
    if (!$configDir) {
	return @configs;
    }

    if (!opendir(DIR, $configDir)) {
	Warning("Cannot open $configDir. Reason: $!");
	return @configs;
    }

    my @subdirs = grep {-d "$configDir/$_" && $_ !~ /^\.\.?$/} readdir(DIR);
    foreach my $subdir (@subdirs) {
       my $subdirPath = $configDir . "/" . $subdir;
       if (!opendir(SUBDIR, $subdirPath)) {
          Warning("Cannot open $subdirPath. Reason: $!");
       } else {
          # Read all the files into an array
          my @subConfigs = 
              grep( $_ !~ /^\.\.?$/ && (! -d "$subdirPath/$_"), readdir(SUBDIR));
          # Add the name of the subdirectory to the config option name
          @subConfigs = map {$subdir . "." . $_} @subConfigs;
          push(@configs, @subConfigs);
          closedir(SUBDIR);
       }
    }
    closedir(DIR);

    @configs = sort { $a cmp $b } @configs;
    return @configs;
}

# Gets the current value and the description if successful.  Returns value of
# the form ($value, $description). $value is undef on failure, and $description
# is actually an error string.
sub getItem($) {
    my ($config) = @_;
    my $subdir;
    my $val;
    my $desc;
    my $default;

    my $configFile = c_getDir('proc_vmk_config');
    if (!$configFile) {
	return (undef, "Could not access vmkernel config directory in the proc filesystem.");
    }
    
    # Separate the subdirectory name from the config option.
    if ($config =~ m/^([^.]+)\.(.+)$/) {
       $config = $2;
       $subdir = $1;
    } else {
       Warning("Could not parse the config option $config.\n");
       return (undef, "Could not parse the config option $config.");
    }

    $configFile .= '/' . $subdir . '/' . $config;
    if (! -r $configFile) {
	return (undef, "Could not access vmkernel config file $configFile for reading.");
    }

    local (*FILE);
    if (!open(FILE, '<' . $configFile)) {
	return (undef, "Could not open vmkernel config file $configFile.");
    }

    my $match = 0;
    while (<FILE>) {
        s/</&lt;/g;
        s/>/&gt;/g;
        chomp;
	if (/^$config\s?(\(.*\))?\s*\[default = "?([^"]*)"?\]:\s*"?([^"]*)"?$/) {
	    $desc = $1;
            $default = $2;
	    $val = $3;
            $match = 1;
            last;
	}
    }
    close(FILE);

    if (!$match) {
	return (undef, "Could not get the value of the config node $subdir.$config.");
    }
    if (!defined($desc) || $desc eq "") {
	$desc = "";
    } else {
	$desc =~ s/^\((.*)\)$/$1/;
    }
    return ($val, $desc, $default);
}

# Sets the value of a vmkernel config parameter.  Returns 1 on success,
# 0 on failure.
sub setItem($$) {
    my ($key, $val) = @_;
    my $configName;
    my $subdir;
    my $newval;
    my $desc;
    
    # Write value into proc node

    local *FILE;
    my $configDir = c_getDir('proc_vmk_config');
    if (!$configDir) {
        Warning("Could not access vmkernel config directory in the proc filesystem.");
        return(0);
    }

    # Separate the subdirectory name from the config option.
    if ($key =~ m/^([^.]+)\.(.+)$/) {
       $configName = $2;
       $subdir = $1;
    } else {
       Warning("Could not parse the key $key.\n");
       return(0);
    }

    my $configFile = "$configDir/$subdir/$configName";
    if (!(-w $configFile) || !open(FILE, "> $configFile")) {
	Warning("Could not write value '$val' to configuration entry '$configFile'.");
        return(0);
    }
    print FILE $val . "\n";
    close(FILE);

    # Check whether update was successful

    ($newval, $desc) = getItem($key);
    if (!defined($newval)) {
        Warning($desc);
        return(0);
    }
    if ($newval != $val) {
        Warning("Illegal value '$val' for VMkernel parameter '$key'.");
        return(0);
    }

    # Save new value to config file

    my $failureMsg = 
        "VMkernel parameter '$key' was set to '$val', but the setting " .
        "could not be saved and will be lost on reboot.";

    $configFile = c_getFile('vmkernel_config');
    if ( !defined($configFile) ) {
        Warning($failureMsg);
        return(0);
    }

    my $config = new VMware::Config;
    $config->preserve_case(1);
    if (-r $configFile) {
       if ( !$config->readin($configFile) ) {
          Warning($failureMsg);
          return(0);
       }
    }

    $config->set("$subdir/$configName", $val);

    if ( !$config->writeout($configFile) ) {
        Warning($failureMsg);
        return(0);
    }

    return(1);
}

# Gets the valid range for a config item
sub getItemRange($) {
    my ($config) = @_;

    if (defined($ConfigValidInfo{$config})) {
	return @{$ConfigValidInfo{$config}};
    }
    return undef;
}

# Indicates if the config value is valid for a config value.  Returns
# 1 if the value is in range.  0 otherwise
sub isInRange($$) {
    my ($config, $val) = @_;

    my @range = getItemRange($config);
    if (defined($range[0]) && $val < $range[0]) {
	return 0;
    }
    if (defined($range[1]) && $val > $range[0]) {
	return 0;
    }
    return 1;
}

#####################################################################
# Public Methods                                                    #
#####################################################################

# Constructor Methods

# Accessor functions

sub dump {
    my @configs = getAllItems();
    local *FD = *STDOUT;
    if (@_) {
	(*FD) = @_;
    }

    my $cfg;
    foreach $cfg (@configs) {
	my ($val, $desc) = getItem($cfg);
	if (!defined($val)) {
	    print FD "$cfg error: $desc\n";
	} else {
	    if (defined($desc) && $desc ne "") {
		$desc = "($desc)";
	    }
	    print FD "$cfg = '$val' $desc\n";
	}
    }
}

#####################################################################
# Methods not intended for use by user                              #
#####################################################################

1;
