#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# ESXConfig.pm
# 
# A collection of common routines for generating config pages
package VMware::Management::SS::ESXConfig;

use strict;

use VMware::HConfig::SSHostRemote;
use VMware::HConfig::Constant qw( $URL_LOGOUT $URL_OVERVIEW $URL_INDEX );
use VMware::ExtHelpers qw( &clearWarnings &getWarnings );

use CGI;

use Carp qw(confess);
$SIG{__DIE__} = \&confess;

require Exporter;
use vars qw( @ISA @EXPORT );
@ISA = qw( Exporter );
@EXPORT = qw(
             @htmlPath %MimeHeaders
             toHTMLHeader toHTMLPath toHTMLSystemInfo toHTMLFooter
             toHTMLPanic toHTMLError toHTMLWarnings
             toHTMLHostFailure toHTMLNotRoot toHTMLNotImplemented
             toHTMLTableHeader toHTMLTableRow toHTMLTableFooter
             toHTMLButton toHTMLNavWizard toWizardURL
             );

use vars qw( @htmlPath %MimeHeaders );
@htmlPath = ( [ "VMware ESX Server Configuration" , $URL_INDEX ] );
%MimeHeaders = ( "Content-Type"  => "text/html; charset=iso-8859-1",
                 "Pragma"        => "no-cache",
                 "Cache-Control" => "no-cache",
                 );

my $themedir = "/vmware/themes/default";
my $spacer = "/vmware/imx/spacer.gif";
my $imxdir = "$themedir/ss/imx";

###
### Global page formatting
###

sub toHTMLHeader($;$$) {
   my ($title, $username, $meta) = @_;
   if (!defined($meta)) {
      $meta = "";
   }

   if (defined($username)) {
      $username = "Management Interface for $username";
   } else {
      $username = "Management Interface";
   }

   my $build = 'Build: 44356';

   my $s = <<END_HTML;
<html>

<head>
  <title>$title</title>
  <link rel="stylesheet" 
        href="$themedir/style.css" 
        type="text/css">
  $meta

  <script type="text/javascript">
    <!--
      function aggressiveUpdate() {
        window.setTimeout('window.location.reload( )', 1000 * 60);
      }
      function openHelp(ref,w,h) {
         window.open(ref,
                     "vmware_mui_help",
                     "location,menubar,resizable,scrollbars,status,toolbar," +
                     "width=" + w + ",height=" + h);
         aggressiveUpdate();
      }
    -->
  </script>
</head>

<body bgcolor="#ffffff" topmargin="6" leftmargin="6" marginwidth="6" marginheight="6"
      link="#3366cc" vlink="#663300">
  <table width="100%" border="0" cellpadding="0" cellspacing="0">
    <tr>
      <td width="451" nowrap="true"><img src="$imxdir/header.gif" width="451" height="62"></td>
      <td align="right" valign="top" nowrap="true" background="$imxdir/header-bg.gif">
        <font color="#ffffff"><small><small>
          <br><nobr>$username&nbsp;</nobr>
          <br><nobr>$build&nbsp;</nobr>
        </small></small></font>
      </td>
    </tr>
  </table>
END_HTML

   return $s;
}

sub toHTMLPath(@$) {
   my ($locations, $title) = (@_);

   my $s = <<END_HTML;
  <table width="100%" border="0" cellpadding="1" cellspacing="0">
    <tr>
      <td><img src=$spacer width="1" height="4"></td>
    </tr>
    <tr>
      <td bgcolor="#999980">
        <table width="100%" border="0" cellpadding="3" cellspacing="0">
          <tr bgcolor="#CCCCB3">
            <td>
              <table border="0" cellpadding="3" cellspacing="0">
                <tr>
END_HTML

   foreach my $location (@$locations) {
      my ($title, $url) = @$location;
      my $entry;
      if (defined($url)) {
         $entry = qq(<a href = "$url">$title</a>);
      } else {
         $entry = $title;
      }

      $s .= <<END_HTML;
                  <td><small><small><nobr>$entry</nobr></small></small></td>
                  <td><small><small><nobr>|</nobr></small></small></td>
END_HTML
   }

   $s .= <<END_HTML;
                  <td><nobr><small><small><b>$title</b></small></small></nobr></td>
                </tr>
              </table>
            </td>
END_HTML

   # XXX rbosch: Should only show overview button if VMkernel is loaded
   $s .= qq(<td width="1%" align="right">\n) .
           toHTMLButton("xxs", "VM Overview", $URL_OVERVIEW) .
           qq(</td>\n);

   my $url_help = "http://www.vmware.com/support/esx15/doc/";
   $url_help = qq(javascript:openHelp('$url_help', 700, 500););
   $s .= qq(<td width="1%" align="right">\n) .
           toHTMLButton("xxs", "Help", $url_help) .
           qq(</td>\n);

   $s .= qq(<td width="1%" align="right">\n) .
           toHTMLButton("xxs", "Logout", $URL_LOGOUT) .
           qq(</td>\n);

   $s .= <<END_HTML;
          </tr>
        </table>
      </td>
    </tr>
  </table>
END_HTML

   return $s;
}

sub toHTMLSystemInfo($) {
   my $host = shift;
   my $system_info = $host->getSystemInfo();
   return <<END_HTML;
<p class="txt">
System information: $system_info
</p>
END_HTML
}

sub toHTMLFooter() {
   my $s = <<END_HTML;
  <table width="100%" border="0" cellpadding="0" cellspacing="0">
    <tr>
      <td align="right" nowrap="true" background="$imxdir/footer-bg.gif">&nbsp;</td>
      <td width="451" nowrap="true"><img src="$imxdir/footer.gif" width="451" height="49"></td>
    </tr>
  </table>

</body>

</html>
END_HTML

   return $s;
}


###
### Handling errors/warnings
###

sub toHTMLPanic($$) {
   my ($title, $message) = @_;
   return
       toHTMLHeader($title) .
       toHTMLError("$title Internal Failure", $message) .
       toHTMLFooter();
}

sub toHTMLError($$) {
   my ($eSummary, $eDesc) = @_;

   my $s = toHTMLWarnings();

   $s .= <<END_HTML;
<h3>Error: $eSummary</h3>
<p class="txt">
  $eDesc
</p>
END_HTML

   return $s;
}

sub toHTMLWarnings() {
   my $s = "";
   my @warnings = getWarnings();
   if ( $#warnings > -1 ) {

      $s .= <<END_HTML;
<h3>Reported Warnings:</h3>
<ul>
END_HTML
        
      foreach my $w (@warnings) {
         $s .= <<END_HTML;
  <li class="txt"> $w
END_HTML
      }
      $s .= <<END_HTML;
</ul>
END_HTML
   }

   $s .= <<END_HTML;
<hr>
END_HTML

   return $s;
}

sub toHTMLHostFailure($) {
   my $title = shift;
   return toHTMLError("$title Internal Failure",
                      "Got an undefined host.  " .
                      "Could not create remote host object.");
}

sub toHTMLNotRoot($) {
   my $url_self = shift;
   return toHTMLError("Operation Not Permitted",
                      "You must be root to perform this operation.  " .
                      "You must <a href='$URL_LOGOUT?n=/$url_self'> " .
                      "log out</a> and then log in as root.");
}

sub toHTMLNotImplemented($) {
   my $req = shift;
   return toHTMLError("Operation Not Implemented",
                      "The requested command '$req' has not been implemented.");
}


###
### Table support
###

# We alternate between two row formats within a table;
# use global variable to keep track of formatting state
my $rowClass = "bdy";

sub toHTMLTableHeader {
   my $s = qq(<table bgcolor="#999980" cellpadding=5 cellspacing=1 border=0>\n);
   if (@_) {
      $rowClass = "hdr";
      $s .= toHTMLTableRow(@_);
   }
   $rowClass = "bdy";
   return $s;
}

sub toHTMLTableRow {
   my $s = qq(<tr>\n);

   foreach my $row (@_) {
      if (ref($row) eq "ARRAY") {
         # Text-attribute pair
         my ($text, $attrs) = @$row;
         $s .= qq(<td class="$rowClass" $attrs>$text</td>\n);
      } else {
         $s .= qq(<td class="$rowClass">$row</td>\n);
      }
   }
   $s .= qq(</tr>\n);

   $rowClass = (($rowClass eq "bdy") ? "bdy2" : "bdy");

   return $s;
}

sub toHTMLTableFooter {
   my $s = "";
   if (@_) {
      $rowClass = "ftr";
      $s .= toHTMLTableRow(@_);
   }
   $s .= qq(</table><br>\n);
   return $s;
}


###
### Navigation support
###

sub toWizardURL($$) {
   my ($url, $wizard) = @_;
   if ($wizard) {
      if ($url =~ /\?/) {
         $url .= "&wizard=1";
      } else {
         $url .= "?wizard=1";
      }
   }
   return $url;
}

sub toHTMLNavWizard {
   my ($back, $next) = @_;

   my ($backURL, $backName, $backAction) = @$back;
   my $backButton = toHTMLButton("txt", "&lt;- Back to $backName", 
                                 toWizardURL($backURL, 1), $backAction);

   my ($nextURL, $nextName, $nextAction) = @$next;
   my $nextButton = toHTMLButton("txt", "Next to $nextName -&gt;", 
                                 toWizardURL($nextURL, 1), $nextAction);

   return <<END_HTML;
<table width="100%" border="0" cellpadding="1" cellspacing="0">
  <tr>
    <td><img src=$spacer width="1" height="4"></td>
  </tr>
  <tr>
    <td bgcolor="#999980">
      <table width="100%" border="0" cellpadding="1" cellspacing="0">
        <tr bgcolor="#ccccb3"> <!-- e6e6cc -->
          <td>
            <table border="0" cellpadding="3" cellspacing="0">
              <tr>
                <td align="right">$backButton</td>
                <td width="1%">$nextButton</td>
              </tr>
            </table>
          </td>
        </tr>
      </table>
    </td>
  </tr>
</table>
<br>
END_HTML
}

###
### Buttons
###

sub toHTMLButton {
   my ($class, $text, $action, $onClick) = @_;
   
   if (defined($action) && $action ne "") {
      $action = qq(href="$action");
   } else {
      $action = qq(name="$text");
   }

   if (defined($onClick) && $onClick ne "") {
      $action .= qq( onClick="$onClick");
   }

   my $textColor = "#000000";
   my ($baseColor, $gutterShadow, $gutterHilite, $edgeShadow, $edgeHilite) =
       ("#ccccb3", "#999980", "#ffffe6", "#999980", "#ffffe6");
       #("#e6e6cc", "#ccccb3", "#ffffe6", "#ccccb3", "#ffffe6");

   return <<END_HTML;
<table border="0" cellpadding="0" cellspacing="0">
  <tr><!-- 9 : 7 -->
    <td colspan=7 bgcolor=$gutterShadow><img src="$spacer" width=1 height=1></td>
  </tr>
  <tr><!-- 9 : 6 --><!-- 0 : 5 --><!-- F : 5 -->
    <td rowspan=6 bgcolor=$gutterShadow><img src="$spacer" width=1 height=1></td>
    <td colspan=5 bgcolor="#000000"><img src="$spacer" width=1 height=1></td>
    <td rowspan=5 bgcolor=$gutterHilite><img src="$spacer" width=1 height=1></td>
  </tr>
  <tr><!-- 0 : 4 --><!-- F : 3 --><!-- 0 : 3 -->
    <td rowspan=4 bgcolor="#000000"><img src="$spacer" width=1 height=1></td>
    <td colspan=3 bgcolor=$edgeHilite><img src="$spacer" width=1 height=1></td>
    <td rowspan=3 bgcolor="#000000"><img src="$spacer" width=1 height=1></td>
  </tr>
  <tr><!-- F : 2 --><!-- C : 1 --><!-- 9 : 1 -->
    <td rowspan=2 bgcolor=$edgeHilite><img src=$spacer width=1 height=1></td>
    <td bgcolor=$baseColor>
      <table border=0 cellpadding=1 cellspacing=0>
        <tr>
          <td align="center" nowrap="true" class=$class><nobr>
            <a style="text-decoration:none;" $action>
              <font color=$textColor>&nbsp;&nbsp;$text&nbsp;&nbsp;</font>
            </a>
          </nobr></td>
        </tr>
      </table>
    </td>
    <td bgcolor=$edgeShadow><img src=$spacer width=1 height=1></td>
  </tr>
  <tr><!-- 9 : 2 -->
    <td colspan=2 bgcolor=$edgeShadow><img src=$spacer width=1 height=1></td>
  </tr>
  <tr><!-- 0 : 4 -->
    <td colspan=4 bgcolor="#000000"><img src=$spacer width=1 height=1></td>
  </tr>
  <tr><!-- F : 6 -->
    <td colspan=6 bgcolor=$gutterHilite><img src=$spacer width=1 height=1></td>
  </tr>
</table>
END_HTML
}


1;
