# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SS::Eula;
# File: VMware/Management/SS/Eula.pm

use strict;
use VMware::HConfig::SSHostRemote ();
use VMware::HConfig::Constant qw(c_getFile);
use VMware::VmdbPerl;
use VMware::Management::VmdbUtil;
use VMware::Management::Session;
use VMware::Management::Util qw(trace hash_credentials);

use CGI ();
my $q;

# XXX These shouldn't be hard-coded, but there's no Perl vm_version.h equivalent
my %gProductName = ( "base" => "VMware ESX Server",
                     "smp" => "VMware Virtual SMP for ESX Server",
                     );
my $gProductVersion = "2.0";

my %gErrorMsg = ( "invalid" => "The serial number is invalid.",
                  "expired" => "The serial number has expired.",
                  "noMatch" => "The serial number did not match any licenses.",
                  "noLicenses" => "No licenses were found.",
                  "noCompatLicenses" => "No compatible licenses were found.",
                  "unknown" => "",
                  );

my %gWarningMsg = ( "smpLacking" => 
                       ("The license for " . $gProductName{"base"} .
                        " does not match the license for " . 
                        $gProductName{"smp"} . ". " . $gProductName{"smp"} .
                        " will be disabled and your virtual machines" .
                        " will run with only one virtual processor."),
                    "smpExpired" => 
                       ("The license for " . $gProductName{"smp"} .
                        " has expired. " . $gProductName{"smp"} .
                        " will be disabled and your virtual machines" .
                        " will run with only one virtual processor."),
                    "noUnlockedLicenses" => 
                       ("No unlocked licenses were found. " .
                        "You will not be able to run any virtual ".
                        "machines until a valid serial number for " .
                        $gProductName{"base"} . " is entered."),
                    );

# ---------------------------------------------------------------------------

sub handler {
   $q = new CGI(shift());

   # The default action is to return the end user license agreement.
   if (scalar($q->param()) < 1) {
      return eula();
   }

   my $ctx = &VMware::Management::Session::get_temp_context();

   # Update the license state stored in VMDB
   my $cmdPath = $ctx->NewArrayIndex("/host2/#_mui/util/license/refreshSnJob/#/");
   $ctx->BeginTransaction();
   $ctx->SetBool("${cmdPath}run", 1);
   my ($ok, $msg) = VMware::Management::VmdbUtil::vmdbEndTransaction($ctx, $cmdPath);
   if (!$ok) {
      # This should never happen in practice -- job always returns VMDB_S_OK
      VMware::Management::Util::trace("refreshSnJob failed: $msg", "error");
   }

   $ctx->Unset($cmdPath);

   # Return info about the current license.
   if ($q->param("op") =~ /^info$/i) {
      my $data = info($ctx);
      return html($data);
   }

   # Try to apply the supplied serial number to the installed license.
   if ($q->param("accept") == 1) {
      # Initialize reboot to false (reboot not needed);
      # if a reboot is needed, $ret will set it to true.
      my $data = qq[var reboot=false;$/];

      my ($esxRet, $esxWrn) = license($ctx, "base", "sn1");
      $data .= $esxRet;
      my ($smpRet, $smpWrn) = license($ctx, "smp", "sn2");
      $data .= $smpRet;

      # A license warning reported by the ESX serial number may be resolved
      # once we enter the SMP serial number; eat the warning in that case.
      # Note that this works because the only warnings we currently report
      # are related to incompatible licenses; if the ESX serial number could
      # issue a standalone warning, we'd need to be a bit more clever here.
      $data .= defined($smpWrn) ? $smpWrn : $esxWrn;

      return html($data);
   }

   return html(error("Please read and accept the End User License Agreement " .
                     "before continuing."));
}


# ---------------------------------------------------------------------------
#
# eula --
#
#      Returns the end user license agreement formatted for presentation in a
#      web page.
#
# Prototype:
#
#      string eula()
#
# Side Effects:
#
#      None.
#
# ---------------------------------------------------------------------------

sub eula {
   local *FD;
   my $eulaText;

   if (open(FD, '<' . c_getFile('eula'))) {
      $eulaText = "";
      while (<FD>) {
         $eulaText .= $_;
      }
      close(FD);
   }

   if (! defined($eulaText)) {
      $eulaText = "Could not find end user license agreement.";
   }

   return (qq(<html><head><title>EULA</title></head><body>) .
           qq(<pre style="font-family: Verdana,sans-serif;font-size: 11px">) .
           qq($eulaText</pre></body></html>),
           "Pragma"        => "no-cache",
           "Cache-Control" => "no-cache",
           "Content-Type"  => "text/html; charset=utf-8");
}


# ---------------------------------------------------------------------------
#
# license --
#
#      Try to apply the supplied serial number to the installed license.
#
# Prototype:
#
#      string license($ctx, $type, $param)
#
# Arguments:
#
#      $ctx ... The VMDB context
#      $type ... Type of license; "base" or "smp"
#      $param ... CGI parameter containing the license
#
# Results:
#
#      An HTML/JavaScript string as generated by html().
#
# ---------------------------------------------------------------------------

sub license {
   my ($ctx, $type, $param) = @_;
   my $ret;
   my $wrn;

   my $sn = uc $q->param($param);
   $sn =~ s/,/-/g;

   # Nothing to do if serial number wasn't specified
   if ($sn =~ /^[\s-]*$/) {
      return ($ret, $wrn);
   }

   # Nothing to do if serial number is unchanged
   my $currsn = uc $ctx->GetString("/host2/#_mui/setup/license/serial/$type");
   if ($sn eq $currsn) {
      return ($ret, $wrn);
   }

   # Validate the serial number format
   if (length($sn) < 23) {
      $ret = error("The serial number you entered ($sn) could not be " .
                   "used. The serial number is too short. A valid " .
                   "serial number must be twenty characters long.");
      return ($ret, $wrn);
   }
   
   if (length($sn) > 23) {
      $ret = error("The serial number you entered ($sn) could not be " .
                   "used. The serial number is too long. A valid " .
                   "serial number must be twenty characters long.");
      return ($ret, $wrn);
   }
   
   if ($sn !~ /^[A-za-z0-9-]+$/) {
      $ret = error("The serial number you entered ($sn) could not be " .
                   "used. The serial number format is not valid. " .
                   "A valid serial number must be composed of letters " .
                   "and numbers only.");
      return ($ret, $wrn);
   }

   # Check the current host CPU maximum
   my $pcpus = $ctx->GetInt("/host2/#_mui/setup/license/limits/maxHostCpus");

   # Attempt to add the serial number
   $ctx->BeginTransaction();
   my $cmdPath = $ctx->NewArrayIndex("/host2/#_mui/util/license/cmd/##/");
   $cmdPath = $trnCtx->GetAbsPath($cmdPath);
   my $cmdPathCp = $cmdPath;
   $trnCtx->SetString("${cmdPath}op", "newSn");
   $trnCtx->SetString("${cmdPath}user",
                      &VMware::Management::Session::get_usertoken());
   $cmdPath = "${cmdPath}op/newSn/";

   $ctx->SetString("${cmdPath}in/sn", $sn);
   $ctx->SetString("${cmdPath}in/productName", $gProductName{$type});
   $ctx->SetString("${cmdPath}in/productVersion", $gProductVersion);
   $ctx->SetBool("${cmdPath}in/overwriteExisting", 1);

   my ($ok, $msg) = VMware::Management::VmdbUtil::vmdbEndCmd($ctx, $cmdPath);

   # Check for errors
   if ($ok) {
      $ret = $wrn = "";
      # Check if the host CPU maximum has increased and the VMkernel
      # is loaded; if so, prompt for a reboot
      my $pcpusNew = $ctx->GetInt("/host2/#_mui/setup/license/limits/maxHostCpus");
      if (($pcpus != $pcpusNew) && vmkernel()) {
         $ret .= qq[reboot=true;$/];
      }
   } else {
      my $errorId = $ctx->GetString("${cmdPath}status/error/id");
      VMware::Management::Util::trace("newSn failed, id = $errorId", "debug");
      if ($errorId =~ /^newSn.(.*)$/) {
         if (defined($gErrorMsg{$1})) {
            $ret = error("The serial number you entered ($sn) could not be used. " .
                          $gErrorMsg{$1});
         } elsif (defined($gWarningMsg{$1})) {
            $wrn = warning($gWarningMsg{$1});
         } else {
            $ret = error("The serial number you entered ($sn) could not be used.");
         }
      }
   }

   $ctx->LocalArrayUnset($cmdPathCp, 1);
   return ($ret, $wrn);
}


# ---------------------------------------------------------------------------

sub vmkernel {
   my %credentials = hash_credentials();
   my $user = $credentials{"l"};
   my $pass = $credentials{"m"};
   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $pass);
   return ($host && $host->isVMKernel());
}

# ---------------------------------------------------------------------------

sub info {
   my $ctx = shift();
   my $data = "";

   $ctx->SetCurrentPath("/host2/#_mui/setup/license/");

   # Expiration date
   my $ed = $ctx->GetString("expiration") || "None";
   $data .= qq{var ed="$ed";};

   # Serial numbers
   $ctx->SetCurrentPath("serial");
   my $sn1 = $ctx->GetString("base");
   my $sn2 = $ctx->GetString("smp");
   $data .= qq{var sn=["$sn1","$sn2"];};
   $ctx->SetCurrentPath("..");

   # Resource limits
   $ctx->SetCurrentPath("limits");
   $data .= limit($ctx, "maxRunningVms", "vms");
   $data .= limit($ctx, "maxHostCpus", "pcpus");
   $data .= limit($ctx, "maxVcpusPerVm", "vcpus");

   return $data;
}


# ---------------------------------------------------------------------------

sub limit {
   my ($ctx, $path, $var) = @_;
   my $val = $ctx->GetInt($path);
   if ($val < 0) {
      return qq{var $var="Unknown";};
   } elsif ($val == 0) {
      return qq{var $var="Unlimited";};
   } else {
      return qq{var $var=$val;};
   }
}


# ---------------------------------------------------------------------------

sub error {
   my $msg = shift();
   VMware::Management::Util::trace($msg, "error");
   return "err.push(" . jsStringLiteral($msg) . ");$/";
}


# ---------------------------------------------------------------------------

sub warning {
   my $msg = shift();
   VMware::Management::Util::trace($msg, "warning");
   return "wrn.push(" . jsStringLiteral($msg) . ");$/";
}


# ---------------------------------------------------------------------------
#
# jsStringLiteral --
#
#      Make sure that $msg is safe for consumption in a JavaScript program.
#
# ---------------------------------------------------------------------------

sub jsStringLiteral {
   my $msg = shift();
   $msg =~ s/\\/\\\\/g;
   $msg =~ s/("|')/\\$1/g;
   my $lf = $/;
   $msg =~ s/$lf/\\n/g;

   return qq{"$msg"};
}


# ---------------------------------------------------------------------------

sub html {
   my $data = shift();
   my $cb = shift() || "if (parent.opCb) parent.opCb(self);";

   my $html = qq[<html><head><title></title><script>$/] .
      qq[var err=[];var wrn=[];$/] .
      qq[$data$/</script></head><body onload="$cb"></body></html>];

   return($html,
          "Pragma"        => "no-cache",
          "Cache-Control" => "no-cache",
          "Content-Type"  => "text/html; charset=utf-8");
}


1;
