#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# License.pm
# 
# A script that allows users to view and update the license.
package VMware::Management::SS::License;

use strict;

use VMware::HConfig::SSHostRemote;
use VMware::HConfig::Constant qw(c_getFile
                                 $URL_INDEX $URL_VMCONFIG
                                 $URL_NICCONFIG $URL_REBOOT);

use VMware::ExtHelpers qw( &Warning &clearWarnings );
use VMware::Management::Util qw( &hash_credentials );

use VMware::Management::SS::ESXConfig;
my $htmlTitle = "License Information";

use Carp qw(confess);
$SIG{__DIE__} = \&confess;
use CGI;

my $URL_PREV = $URL_NICCONFIG;
my $URL_NAME_PREV = "Network Configuration";
my $URL_NEXT = $URL_VMCONFIG;
my $URL_NAME_NEXT = "Server Configuration";

my $RCom = "request";
my $RCom_AgreeEULA = "agreeEULA";
my $RCom_DisagreeEULA = "disagreeEULA";
my $RCom_Update = "update";

sub toLicenseHTML($$) {
   my ($cgi, $host) = @_;
   my $url_self = $cgi->url(-relative=>1);

   my ($canRun, $message, $serial, $expires);
   my $maxCPUs = 0;
   my $clustering = 0;

   my $s = toHTMLSystemInfo($host) . toHTMLWarnings();

   if ($cgi->param($RCom) eq $RCom_Update) {
      ($canRun, $message) = saveLicense($cgi, $host);
      if (!$canRun || ($message =~ /[Ee]rror/)) {
         Warning("The serial number you typed could not be applied to the license.\n" .
                 "<pre>$message</pre>\n");
      } else {
         $s .= "<p class='txt'>\n";
         $s .= "<b>Your new serial number has been accepted!</b>\n";
         # Need to reload VMkernel if the number of CPUs has changed
         my ($hmax) = $host->getLicenseInfo("maxCPUs");
         my $cmax = $cgi->param("maxCPUs");
         if ($host->isVMKernel() && ($hmax != $cmax)) {
            $s .= "You must <a href='${URL_REBOOT}?message=License%20changes'> reboot " .
                "your ESX server</a> in order for the new license to take effect.\n";
         }
         $s .= "</p>\n";
      }
   } else {
      ($canRun, $message) = $host->checkCanRunLicense();
   }
   
   my $wizard = ($cgi->param("wizard"));
   my $name = $cgi->param("NAME");
   my $org = $cgi->param("ORGANIZATION");

   if ($canRun) {
      ($serial, $expires, $maxCPUs, $clustering) = 
          $host->getLicenseInfo("Serial", "expiration", "maxCPUs", "clustering");

      $s .= qq(<h3>Current License</h3>);

      $s .= toHTMLTableHeader();

      $s .= toHTMLTableRow("<b>Serial Number</b>", $serial);
      if (defined($name) && ($name ne "")) {
         $s .= toHTMLTableRow("<b>Name</b>", $name);
      }
      if (defined($org) && ($org ne "")) {
         $s .= toHTMLTableRow("<b>Organization</b>", $org);
      }
      if (defined($expires) && ($expires ne "")) {
         $s .= toHTMLTableRow("<b>Expiration</b>", $expires);
      }
      if (defined($maxCPUs) && $maxCPUs > 0) {
         $s .= toHTMLTableRow("<b>Processor Limit</b>", $maxCPUs);
      }
      if (defined($clustering)) {
         $s .= toHTMLTableRow("<b>Clustering Support</b>", $clustering ? "Yes" : "No");
      }

      $s .= toHTMLTableFooter() . "<hr>";
   }

   $s .= <<END_HTML;
<form method="POST" action="$url_self">

<h3>New License</h3>

<p class="txt">
Please enter the 20-character serial number from the 
registration card supplied by the vendor that provided 
this copy of VMware ESX Server, along with your name and 
the name of your organization.
</p>

<table border=0 cellpadding=2>
  <tr>
    <td class="txt"><b>Serial Number:</b></td>
    <td><table border=0>
        <tr>
          <td><input name="SERIAL_NUMBER_0" size="5" maxlength="5" 
                     style="font-family:monospace; font-size:12px;"></td>
          <td>-</td>
          <td><input name="SERIAL_NUMBER_1" size="5" maxlength="5" 
                     style="font-family:monospace; font-size:12px;"></td>
          <td>-</td>
          <td><input name="SERIAL_NUMBER_2" size="5" maxlength="5" 
                     style="font-family:monospace; font-size:12px;"></td>
          <td>-</td>
          <td><input name="SERIAL_NUMBER_3" size="5" maxlength="5" 
                     style="font-family:monospace; font-size:12px;"></td>
        </tr>
      </table>
    </td>
  </tr>

  <tr>
    <td class="txt"><b>Name:</b></td>
    <td>
      <table border=0>
        <tr>
          <td><input name="NAME" value="$name" size="40"></td>
        </tr>
      </table>
    </td>
  </tr>

  <tr>
    <td class="txt"><b>Organization:</b></td>
    <td>
      <table border=0>
        <tr>
          <td><input name="ORGANIZATION" value="$org" size="40"></td>
        </tr>
      </table>
    </td>
  </tr>

</table>

<p>
  <input type="submit" value="Update">
  <input type="hidden" name="$RCom" value="$RCom_Update">
  <input type="hidden" name="maxCPUs" value="$maxCPUs">
  <input type="hidden" name="clustering" value="$clustering">
  <input type="hidden" name="canRun" value="$canRun">
  <input type="hidden" name="wizard" value="$wizard">
</p>

</form>
END_HTML

   if ($canRun && $wizard) {
      my $url_next = $URL_NEXT;
      #$url_next .= "?NAME=$name&ORGANIZATION=$org";
      $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                             [ $url_next, $URL_NAME_NEXT ] );
   }

   return($s);
}

sub toEulaHTML($$) {
   my ($cgi, $host) = @_;
   my $url_self = $cgi->url(-relative=>1);
   my $wizard = $cgi->param("wizard");

   local *FD;
   my $eulaText;
   my $eulaFile = c_getFile('eula');
   if (open(FD, '<' . $eulaFile)) {
      $eulaText = "";
      while (<FD>) {
         $eulaText .= $_;
      }
      close(FD);
   }

   if (!defined($eulaText)) {
      return(toErrorHTML("End User License Agreement",
                         "Could not find End User License Agreement."));
   }
   
   my $title = "End User License Agreement";
   my @path = @htmlPath;
   push @path, [ $htmlTitle, $url_self ];
   
   my $wiz = $wizard ? qq(<input type="hidden" name="wizard" value="$wizard">) : "";

   my $s = 
        toHTMLPath(\@path, $title) .
        toHTMLSystemInfo($host) .
        toHTMLWarnings();

   $s .= <<END_HTML;
<h3>Please Read the End User License Agreement.</h3>
<p class='txt'>If you agree, click <i>Accept</i>.</p>

<table border=0>
  <tr>
    <td>
      <form name="accept" method="POST" action="$url_self">
        <input type="hidden" name="$RCom" value="$RCom_AgreeEULA">
        $wiz
        <input type="submit" value="Accept">
      </form>
    </td>
    <td>
      <form name="accept" method="POST" action="$url_self">
        <input type="hidden" name="$RCom" value="$RCom_DisagreeEULA">
        $wiz
        <input type="submit" value="Decline">
      </form>
    </td>
  </tr>
</table>

<pre>$eulaText</pre>
   
<table border=0>
  <tr>
    <td>
      <form name="accept" method="POST" action="$url_self">
        <input type="hidden" name="$RCom" value="$RCom_AgreeEULA">
        $wiz
        <input type="submit" value="Accept">
      </form>
    </td>
    <td>
      <form name="accept" method="POST" action="$url_self">
        <input type="hidden" name="$RCom" value="$RCom_DisagreeEULA">
        $wiz
        <input type="submit" value="Decline">
      </form>
    </td>
  </tr>
</table>
END_HTML

   return $s;
}

sub saveLicense($$) {
   my ($cgi, $host) = @_;
   my ($canRun, $message);
   
   my $license = 
       $cgi->param('SERIAL_NUMBER_0') . '-' . 
       $cgi->param('SERIAL_NUMBER_1') . '-' . 
       $cgi->param('SERIAL_NUMBER_2') . '-' . 
       $cgi->param('SERIAL_NUMBER_3');
   
   if( length($license) != 23 ) {
      return(0, "Invalid serial number. " .
             "Serial number must contain 20 characters.");
   }
   if( $license !~ /^[A-za-z0-9-]+$/ ) {
      return(0, "Invalid serial number format. " .
             "Serial number may contain only letters and numbers.");
   }
   
   # Untaint the data
   if( $license =~ /^(\S+)$/ ) {
      $license = $1;
   }
   
   ($canRun, $message) = $host->registerNewLicense($license);
   return($canRun, $message);
}

sub handler {
   # Delete the PATH to get by the tainted security requirements of Perl
   delete($ENV{'PATH'}); 
   
   umask 022;   # Create files that are not writable by group
   
   &clearWarnings();
   
   my $cgi = new CGI( shift( ) );
   my %credentials = hash_credentials( );
   
   my $user = $credentials{ "l" };
   my $password = $credentials{ "m" };

   my $s = toHTMLHeader($htmlTitle, $user);
   
   if ($user ne 'root') {
      my $url_self = $cgi->url(-relative=>1);
      $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLNotRoot($url_self);
      goto footer;
   }

   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $password);
   if (!defined($host)) {
      $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLHostFailure($htmlTitle);
      goto footer;
   }

   my $req = $cgi->param($RCom);

   if ($req eq $RCom_DisagreeEULA) {
      my $summary = "End User License Agreement Declined";
      my $msg = "You must read and accept the End User License Agreement to continue.";
      $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLError($summary, $msg);
      goto footer;
   }

   my $db = $host->getLocationsDB();
   my $eulaAgreed = $db->{'answer'}->{'EULA_AGREED'} =~ /yes/i;
   $eulaAgreed ||= $req eq $RCom_AgreeEULA;
   $eulaAgreed ||= $req eq $RCom_Update;

   if ($eulaAgreed) {
      $s .= toLicenseHTML($cgi, $host);
   } else {
      $s .= toEulaHTML($cgi, $host);
   }

footer:
   $s .= toHTMLFooter();
   return ($s, %MimeHeaders);
}

1;
