#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# NICConfig.pm
# 
# A script that allows users to configure NIC settings
package VMware::Management::SS::NICConfig;

use strict;

use VMware::Control '1.00';
use VMware::HConfig::SSHostRemote;
use VMware::HConfig::Constant qw( $URL_PCIDIVY $URL_LICENSE $URL_REBOOT
                                  $gVmkDeviceType_nic @gNICSettings 
                                  $gVmkDeviceType_bond
                                  %gNICConfigurationOptions 
                                  @gNICClusteringDisabled );

use VMware::ExtHelpers qw( &Warning &clearWarnings );
use VMware::Management::Util qw( trace hash_credentials );

use VMware::Management::SS::ESXConfig;
my $htmlTitle = "Network Configuration";

use CGI;
use Carp qw(confess);
$SIG{__DIE__} = \&confess;

my $URL_PREV = $URL_PCIDIVY;
my $URL_NAME_PREV = "Device Allocation";
my $URL_NEXT = $URL_LICENSE;
my $URL_NAME_NEXT = "License Information";

my $RCom = "request";
my $RCom_Update = "update";

sub toNICConfigHTML {
   my ($cgi, $host) = @_;
   my $url_self = $cgi->url(-relative=>1);
   my $wizard = ($cgi->param("wizard"));
   my $configDB = $host->getConfigDB();
   my $s = "";

   # XXX rbosch: Just do current configuration for now
   my $label = $host->getCurrentLabel();

   #
   # Gather information about NICs assigned to VMkernel
   #

   my @vmnics = ();
   my @modules = ();

   foreach my $dev ($host->getPCIDeviceList($label)->getPCIDevices()) {
      next if !$dev->isNetClassDevice();
      next if $dev->getOwnerString eq "host";

      my $id = $dev->getBus() . '.' . $dev->getDevice() . '.' . $dev->getFunction();
      my $module = $dev->getDrivers()->[0];
      my $vmnic = { "id" => "$id", "device" => $dev, "current" => "Unknown", 
                    "name" => undef, "module" => $module, "speed" => "Unknown",
                    "clustering" => "&nbsp;" };
      push @vmnics, $vmnic;
      push @modules, $module if !grep { $module eq $_ } @modules;
   }      

   if ($host->isVMKernel()) {
      # Get current vmnic mappings
      foreach my $adapter (@{$host->getVMNetAdapters()}) {
         my ($bus, $device, $function) = $host->getVMNetAdapterPCIInfo($adapter);
         my $id = "$bus.$device.$function";
         foreach my $vmnic (grep {$_->{"id"} eq $id} @vmnics) {
            $vmnic->{"name"} = $adapter;
            # If speed is undefined, link is down
            $vmnic->{"speed"} = $host->getVMNetAdapterSpeed($adapter) || "Link down";
            $vmnic->{"nicteam"} = $host->getVMNetAdapterNICTeam($adapter) || "none";
         }
      }
   }

   #
   # Apply changes to NIC settings
   #

   my $req = $cgi->param($RCom);
   if (defined($req) && ($req eq $RCom_Update)) {
      foreach my $vmnic (@vmnics) {
         my $id = $vmnic->{"id"};
         my $options = $cgi->param($id);
         my $name = $vmnic->{"name"};
         $configDB->set("device.$label.$id.options", $options);

         my $clustering = $cgi->param("$id-clustering");
         if ($clustering ne "na") {
            $clustering = ($clustering eq "on") ? 1 : 0;
            $configDB->set("device.$label.$id.clustering", $clustering);
            if (defined($name)) {
               $host->setVMNetAdapterClustering($name, $clustering);
            }
         }

         my $bond = $cgi->param("$id-nicteam");
         if (defined($name)) {
            my $success = 1;
            my $nicteam = $host->getVMNetAdapterNICTeam($name);

            if ($nicteam && ($bond eq "none" || $nicteam ne $bond)) {
               $success = $host->removeBondSlave($nicteam, $name);
            }
            if ($success && $bond ne "none") {
               $host->addBondSlave($bond, $name);
            }

            # Refetch the NIC team for this NIC.
            $vmnic->{"nicteam"} = $host->getVMNetAdapterNICTeam($name) || "none";
         }
      }

      # XXX rbosch: need to reload changed drivers here

      if (!$host->commitConfigDB()) {
         Warning("Failed to commit changes to config database.\n");
      }
   }

   #
   # Display system info and any warnings caused by update
   #

   $s .= toHTMLSystemInfo($host) . toHTMLWarnings();

   #
   # Display current settings
   #

   foreach my $vmnic (@vmnics) {
      my $id = $vmnic->{"id"};
      my $module = $vmnic->{"module"};
      next if !defined($gNICConfigurationOptions{$module});

      # Load option list from ConfigDB
      my $name = "device.$label.$id.options";
      my $options = $configDB->get($name);

      # Generate HTML selector for device
      my $select = qq(<select name="$id">\n);
      foreach my $i (0 .. $#gNICSettings) {
         my $myLabel = $gNICSettings[$i];
         my $myOptions = $gNICConfigurationOptions{$module}->[$i];
         if (defined($myOptions)) {
            if ($options eq $myOptions) {
               $vmnic->{"current"} = $myLabel;
            }
            my $selected = ($options eq $myOptions) ? "selected" : "";
            $select .= qq(  <option value="$myOptions" $selected>$myLabel</option>\n);
         }
      }
      $select .= qq(</select>\n);
      $vmnic->{"select"} = $select;

      # Generate clustering checkbox for device
      if (grep { $_ eq $module } @gNICClusteringDisabled) {
         $vmnic->{"clustering"} = 
             qq(<input type="hidden" name="$id-clustering" value="na">N/A);
      } else {
         my $clustering; 
         if (defined($vmnic->{"name"})) {
            $clustering = $host->getVMNetAdapterClustering($vmnic->{"name"});
         } else {
            $clustering = $configDB->get("device.$label.$id.clustering");
         }
         my $chk = $clustering ? "checked" : "";
         $vmnic->{"clustering"} = qq(<input type="checkbox" name="$id-clustering" $chk>);
      }

      # Generate NIC teaming option for device
      my @bond = @{$host->getVMNetBondAdapters()};
      unshift(@bond, "none");
      my $select = qq(<select name="$id-nicteam">\n);
      foreach my $bnd (@bond) {
         my $selected = $vmnic->{"nicteam"} eq $bnd ? "selected" : "";
         $select .= qq(  <option value="$bnd" $selected>$bnd</option>\n);
      }
      $select .= qq(</select>\n);
      $vmnic->{"nicteam"} = $select;
   }

   $s .= <<END_HTML;
<h3>$htmlTitle</h3>
<p class="txt">
This page allows you to configure the network adapters 
used by the virtual machines; see the text below for
detailed descriptions of the configuration options. 
Click "Update Network Configuration" to save the new settings;
changes will only be applied to the current boot configuration.
</p>
END_HTML

   if ($host->isVMKernel()) {
       $s .= <<END_HTML;
<p class="txt">
<b>Note</b>: Interrupt clustering changes will take effect 
immediately, but speed/duplex changes will not take effect 
until you <a href="${URL_REBOOT}?message=Network%20configuration%20changes">
reboot your machine</a>.
</p>
END_HTML
   }

   if (scalar(@vmnics) == 0) {
      $s .= <<END_HTML;
<p class="txt">
There are currently no network adapters assigned to the virtual machines.
You can use the <a href="$URL_PCIDIVY">device allocation page</a>
to switch adapters from the Console OS to the virtual machines.
</p>
END_HTML
   } else {
      $s .= qq(<form method="POST" action="$url_self">\n);
      my @headers = ();
      if ($host->isVMKernel()) {
         push @headers, ["Name", qq(align="right")];
      }
      push @headers, ["PCI", qq(align="right")];
      push @headers, "Device Name";
      push @headers, ["Module", qq(align="right")];
      push @headers, "Configured Speed/Duplex";
      if ($host->isVMKernel()) {
         push @headers, "Actual Speed/Duplex";
      }
      push @headers, ["Interrupt Clustering", qq(align="center")];
      push @headers, ["NIC Team", qq(align="center")];

      $s .= toHTMLTableHeader(@headers);
      
      foreach my $vmnic (@vmnics) {
         my @data = ();
         if ($host->isVMKernel()) {
            my $name = defined($vmnic->{"name"}) ? $vmnic->{"name"} : "&nbsp;";
            push @data, [$name, qq(align="right")];
         }
         push @data, [$vmnic->{"id"}, qq(align="right")];
         push @data, $vmnic->{"device"}->getName();
         push @data, [$vmnic->{"module"}, qq(align="right")];
         push @data, $vmnic->{"select"};
         if ($host->isVMKernel()) {
            push @data, $vmnic->{"speed"};
         }
         push @data, [$vmnic->{"clustering"}, qq(align="center")];
         push @data, [$vmnic->{"nicteam"}, qq(align="center")];

         $s .= toHTMLTableRow(@data);
      }
      $s .= toHTMLTableFooter();
      
      $s .= 
          qq(<input type="hidden" name="$RCom" value="$RCom_Update">\n) .
          ($wizard ? qq(<input type="hidden" name="wizard" value="$wizard">\n) : "") .
          qq(<input type="submit" value="Update Network Configuration">\n) . 
          qq(</form>\n);
   }

   $s .= <<END_HTML;
<h4>Speed/Duplex</h4>
<p class="txt">
This setting allows you to configure the speed and duplex values used by the 
network adapters.  The "Autonegotiate" option enables the adapter to attempt
to determine the optimal settings on its own; if autonegotiation fails, you
can explicitly specify the settings for the adapter.
</p>
<p class="txt">
<b>Suggested settings</b>: For Gigabit Ethernet cards, we recommend that you
use the Autonegotiate setting.  For all other cards, we recommend that you use
100Mbps with full duplex; in some cases autonegotiation can erroneously choose
half duplex mode, leading to a significant degradation in network performance.
</p>

<h4>Interrupt Clustering</h4>
<p class="txt">
Interrupt clustering can improve performance under heavy network
loads.  Loads that are composed of many connections should benefit
most from this feature.  Note that this option is disabled for Gigabit
Ethernet cards, which handle interrupt clustering in hardware.
END_HTML

   if ($host->isVMKernel()) {
      $s .= <<END_HTML;
 Fine-tuning parameters for this feature are available
on the <a href="/vmkernel-config">VMkernel configuration page</a>; the
relevant parameters begin with "NetRXCluster."
</p>
END_HTML
   }

   $s .= <<END_HTML;
<h4>NIC Teaming</h4>
<p class="txt">
This section allows you to setup NIC Teaming which defines a
$gVmkDeviceType_bond device that can have multiple slave
$gVmkDeviceType_nic devices.  You can only add a slave to a team if
the slave device is not in use.  You can only remove a slave from a
team if the team is not in use.
</p>
END_HTML
   
   if ($wizard) {
      $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                             [ $URL_NEXT, $URL_NAME_NEXT ] );
   }
   return $s;
}

sub handler {
   # Delete the PATH to get by the tainted security requirements of Perl
   delete($ENV{'PATH'}); 
   
   &clearWarnings();
   
   my $cgi = new CGI(shift());
   
   my %credentials = hash_credentials( );
   my $user = $credentials{ "l" };
   my $password = $credentials{ "m" };
   
   my @path = ();
   my $s = toHTMLHeader($htmlTitle, $user) . toHTMLPath(\@htmlPath, $htmlTitle);
   
   if($user ne 'root') {
      my $url_self = $cgi->url(-relative=>1);
      $s .= toHTMLNotRoot($url_self);
      goto footer;
   }
   
   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $password);
   
   if(!defined($host)) {
      $s .= toHTMLHostFailure($htmlTitle);
      goto footer;
   }

   $s .= toNICConfigHTML($cgi, $host);
   
 footer:
   $s .= toHTMLFooter();
   return ($s, %MimeHeaders);
}

1;
