#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# Pcidivy.pm
# 
# A script that goes through and configures what devices you have.
package VMware::Management::SS::Pcidivy;

use strict;

my $gBETA1 = 0;
my $gUseDelete = 1;

use VMware::DOMAccess;
use VMware::Management::Util qw( hash_credentials );

use VMware::Control::Profiler;
use VMware::HConfig::PCIDevice;
use VMware::HConfig::SSHostRemote;
use VMware::HConfig::Constant qw($gDefaultMemSize 
                                 $gMaxConsoleMemSize $gVmkDeviceType_hba
				 $URL_INDEX $URL_PCIDIVY $URL_NICCONFIG
				 );
use VMware::ExtHelpers qw(&p_number &Warning $gDEBUG &Debug &getWarnings $gPanicCallback &clearWarnings);

use VMware::Management::SS::ESXConfig;
my $htmlTitle = "Edit Boot Configuration";

my $URL_PREV = $URL_INDEX;
my $URL_NAME_PREV = "Index";
my $URL_NEXT = $URL_NICCONFIG;
my $URL_NAME_NEXT = "Network Configuration";
#my $URL_NEXT = "/license";
#my $URL_NAME_NEXT = "Licensing";

my $RCom = "request";
my $RCom_Create = "create";
my $RCom_Debug = "debug";
my $RCom_Delete = "delete";
my $RCom_Edit = "edit";
my $RCom_Reboot = "reboot";
my $RCom_Save = "save";
my $RCom_Select = "select";

use CGI ();

# XXX Ugh.  A global variable.  It's used here to be convenient, but it's not
#     a good idea for modularity's sake.
my %mimeHeaders;       

my $p = VMware::Control::Profiler::getInstance("mui");

#####################################################################
# Common Functions used by both the web and command line version    #
#####################################################################

# Succeeds if it finds a prior kernel configuration
# If it does not succeed, returns the null string
# Does its best and uses a few rules to find what we believe to be the
# best default boot image to use.
sub findDefaultBootImage($) {
    my $stub = $p->profile();
    my ($host) = @_;
    my $default = "";

    my @images = @{$host->getLILOConfig()->getVMnixConfigs()};

    if( $#images > -1 ) {
	$default = $images[0]->getLabel();   # Since one exists, set it to the default

	# See if there is a better default. ie one that is run by default
	my $def = $host->getLILOConfig()->getDefaultImage();

	# See if it exists and is a VMnix image
	my $img = $host->getLILOConfig()->getBootImageByLabel($def);
	if( $img && $img->usesVMnixImage() ) {
	    $default = $def;
	}
    }

    return($default);
}

sub possibleDev($) {
    my ($dev) = @_;
    return($dev->isVMwareManageable());
}

sub canShareDev($) {
    my ($dev) = @_;
    return($dev->isVMwareSharable());
}

sub isStorageDev($) {
    my ($dev) = @_;
    my $driverNames = $dev->getDriverNames();

    if( defined($driverNames) ) {
	my $name;
	foreach $name (@$driverNames) {
	    if( $name eq $gVmkDeviceType_hba ) {
		return(1);
	    }
	}
    }
    return(0);
}

sub isImageDeletable($$) {
    my( $host, $img ) = @_;
    return( $img->getLabel() ne $host->getCurrentLabel() );
}

sub isInSet($$) {
    my ($input, $set) = @_;

    my @res = grep { $input eq $_ } @$set;
    return( $#res > -1 );
}

sub isNotInSet($$) {
    my ($input, $set) = @_;

    my @res = grep { $input eq $_ } @$set;

    return( $#res == -1 );
}

sub getKernelVersion($) {
    my($kernel) = @_;
    return $kernel =~ /vmlinuz-(\S+)/ ? $1 : undef;
}

sub findMatchInitrd($$) {
    my ($initrdList, $kernel) = @_;
    my $version = getKernelVersion($kernel);
    my $i;

    if (defined($version)) {
	foreach $i (@$initrdList) {
	    if ($i =~ /$version/) {
		return $i;
	    }
	}
    }
    return undef;
}


# ---------------------------------------------------------------------------
#
# getScsiHostAdapterNumber --
#
#      Given a list of vmnix modules, search the list of aliases of the 
#      modules for aliases of the form "scsi_hostadapter#". Return  the
#      largest "#" found + 1. If an alias of exactly 'scsi_hostadapter" is
#      found, it is considered to have a "#" of 0.
#
# Prototype:
#
#      int getScsiHostAdapterNumber(PerlHash \@vmnixMods )
#
# Arguments:
#
#        \@vmnixMods ... A reference to an array of module data hashes
#
# Results:
#
#      the max. number found in any "scsi_hostadapter#" aliases + 1.
#
# ---------------------------------------------------------------------------

sub getScsiHostAdapterNumber($) {
   my ($vmnixMods) = @_;
   my $searchStr = "scsi_hostadapter";
   my $highestScsiHostAdapterNumber = -1;
   my $s;

   #
   # look at all the aliases for all the modules and find
   # the largest value for "#" in the "scsi_hostadapter#" aliases
   foreach $s (@$vmnixMods) {
      if (defined($s->{ALIASES})) {
         my @aliases = @{$s->{ALIASES}};
         foreach $a (@aliases) {
            if ((index($a,$searchStr) >= 0) and
             (length($a) > length($searchStr))) {
               my $thisScsiHostAdapterNumber = 
                 substr($a,length($searchStr),length($a) - length($searchStr));
               if ($thisScsiHostAdapterNumber > $highestScsiHostAdapterNumber) {
                  $highestScsiHostAdapterNumber = $thisScsiHostAdapterNumber;
               }
            } elsif ($a eq $searchStr) {
               my $thisScsiHostAdapterNumber =  0;
               if ($thisScsiHostAdapterNumber > $highestScsiHostAdapterNumber) {
                  $highestScsiHostAdapterNumber = $thisScsiHostAdapterNumber;
               }
            }
         }
      }
   }

   return( $highestScsiHostAdapterNumber + 1);
}

# Takes in the list of VMnix modules, a vmnix module name, and an alias.
# Return 1 if the alias for the vmnix module exists.  0 otherwise.
sub hasVMnixAlias($$$) {
   my ($vmnixMods, $vmnixModName, $alias) = @_;

   # The left parens are a Perl idiom to take only the first array element.
   # If we don't do this and the grep yields more than 1 result, $modinfo
   # will be set to the number of elements.  In any case, there should only
   # be one module that matches.

   my ($modinfo) = grep($_->{MODULE} eq $vmnixModName, @$vmnixMods);
   if (!defined($modinfo)) {
      return 0;
   }

   my @aliases = @{$modinfo->{ALIASES}};
   if (!grep(substr($_,0,length($alias)) eq $alias, @aliases)) {
      return 0;
   }

   return 1;
}

#####################################################################
# HTML producing functions                                          #
#####################################################################

sub makeInputKey($) {
    my ($dev) = @_;
    return("owner" . $dev->getBus() . ":" . $dev->getDevice());
}

sub PanicCallbackHTML(@) {
    my (@error) = @_;
    if ( $error[0] ) {
       print toHTMLPanic($htmlTitle, $error[0]);
    } else {
       print toHTMLPanic($htmlTitle, "");
    }
}

sub toErrorHTML($$) {
    my ($eSummary, $eDesc) = @_;
    return toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLError($eSummary, $eDesc);
}

sub toSelectConfigHTML($$$) {
    my $stub = $p->profile();
    my ($cgi, $host, $default) = @_;

    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");

    # Get the list of boot images that run the VMnix available
    my @images = @{$host->getLILOConfig()->getVMnixConfigs()};

    my $useDelete = 0;        # Flag to indicate if we need to show the delete option
    my @deletableImages = grep( isImageDeletable($host, $_), @images );
    if( $#deletableImages > -1 ) {
	# We have at least one deletable boot configuration
	$useDelete = 1;
    }
    @deletableImages = map { '"boot-' . $_->getLabel() . '"' } @deletableImages;
    my $deleteImages = join(', ', @deletableImages);

    my $title = "Select Confguration";
    my @path = @htmlPath;
    push @path, [ $htmlTitle, $url_self ];

    my $s =
        toHTMLPath(\@path, $title) . 
        toHTMLSystemInfo($host) . 
        toHTMLWarnings();

    $s .= <<END_HTML;
<script type="text/javascript">
  <!--
  function getCheckedValue(config) {
     for (var i = 0; i < document.deleteForm.elements.length; i++) {
        if (document.deleteForm.elements[i].name == config ) {
           return(document.deleteForm.elements[i].checked == "1")
        }
      }
      return(false);
  }
  function doConfirmDelete() {
     var deleteArray = new Array($deleteImages);
     var deleteFlag = 0;
     for (var i = 0; i < deleteArray.length; i++ ) {
        if( getCheckedValue(deleteArray[i]) ) {
           deleteFlag = 1;
        }
     }
     if( deleteFlag == 1 ) {
        return confirm("Are you sure you want to delete\\n" +
                       "the selected boot configurations?");
     } else {
        alert("No boot configurations were selected to be deleted.");
           return(false);
        }
     }
  // -->
</script>
END_HTML
	 
    my $createURL = "$url_self?$RCom=$RCom_Create";
    $createURL = toWizardURL($createURL, $wizard);

    $s .=
        ("<form name=\"deleteForm\" method=\"GET\" action=\"$url_self\" onSubmit=\"return doConfirmDelete()\">" .
	 "\n" .
	 "<h3>Select Boot Configuration</h3>\n" .
	 "\n" .
	 "<p class='txt'>\n" .
	 "The following boot configurations were detected on your system.  Select\n" .
	 "the configuration you would like to edit by clicking on its label.\n" .
	 "You may create a new configuration by clicking on \n" .
         "<a href=\"$createURL\">Create New Configuration</a>.\n" .
	 "</p>\n" .
	 ($gUseDelete && $useDelete ? 
	  "<p class='txt'>To delete a boot configuration,\n" .
          "select the configurations you would like\n" .
	  "to delete by clicking on the box next to the configurations to be deleted.\n" .
	  "Then click the Delete button and those configurations will be removed.\n " .
	  "You cannot delete a configuration if it is currently running.</p>" :
	  "" ) .
	 "\n");

    my @headers = ( "Label", "Root", "Memory", "Console Devices", 
                    "Virtual Machine Devices" );
    # Add a leading column for deletion if needed
    if ($gUseDelete && $useDelete) {
       unshift @headers, "&nbsp;";
    }
    $s .= toHTMLTableHeader(@headers);

    # Print the information for each boot image of interest
    my $img;
    foreach $img (@images) {
	my $PCIDeviceList = $host->getPCIDeviceList($img->getLabel());

	my $possPCIDeviceList = $PCIDeviceList->filterPCIDevices(\&possibleDev);
	my @uniq_possdevs = $possPCIDeviceList->getUniqPCIDevice();
	my @host_dev = grep { $_->getOwner() == VMware::HConfig::PCIDevice::HOST } @uniq_possdevs;
	my @vm_dev = grep { $_->getOwner() == VMware::HConfig::PCIDevice::VM } @uniq_possdevs;
	my @shared_dev = grep { $_->getOwner() == VMware::HConfig::PCIDevice::SHARED } @uniq_possdevs;

	my $label = $img->getLabel();
	my $image = $img->getImage();
	my $memory = $img->getMemSize();
	my $rootdev = $img->getRootDev();
	my $isDefault = "";
	if( $label eq $host->getLILOConfig()->getDefaultImage() ) {
	    $isDefault = "(default)";
	}

        my @fields;

        if ($gUseDelete && $useDelete) {
           if (isImageDeletable($host, $img)) {
              push @fields, qq(<input type="checkbox" name="boot-$label">);
           } else {
              push @fields, "&nbsp;";
           }
        }

        my $editURL = "$url_self?$RCom=$RCom_Edit&boot=$label";
        $editURL = toWizardURL($editURL, $wizard);
        push @fields, 
            qq(<a href="$editURL">$label</a>) .
            qq( [<a href="$editURL">Edit</a>]) .
            qq( $isDefault);
        push @fields, $rootdev;
        push @fields, [ "$memory MB", "align='right'" ];

        my $consoleDevices = qq(<ol>\n);

	# Get the list of shared devices.  This list will be added
	# to both the host and VM columns but will be labeled as shared.
	my @sharedDevNames = map { "    <li>" . $_->getName() . " (shared)\n" } @shared_dev;

	# Print the list of devices owned by the host
	my @devs = map { "    <li>" . $_->getName() . "\n" } @host_dev;
	push(@devs, @sharedDevNames);
	my $li;
	if( $#devs > -1 ) {
	    foreach $li (@devs) {
		$consoleDevices .= $li;
	    }
	}
	else {
	    $consoleDevices .= "None";
	}
	    
        $consoleDevices .= qq(</ol>\n);
        push @fields, $consoleDevices;

        my $vmDevices = qq(<ol\n>);

	# Print the list of devices owned by the vmkernel
	@devs = map { "    <li>" . $_->getName() . "\n" } @vm_dev;
	push(@devs, @sharedDevNames);
	if( $#devs > -1 ) {
	    foreach $li (@devs) {
		$vmDevices .= $li;
	    }
	}
	else {
	    $vmDevices .= "None";
	}
	    
	$vmDevices .= qq(</ol>\n);
        push @fields, $vmDevices;

        $s .= toHTMLTableRow(@fields);
    }

    $s .= toHTMLTableFooter();

    $s .= 
	("<input type=\"hidden\" name=\"$RCom\" value=\"delete\">\n" .
         ($wizard ? 
          "<input type=\"hidden\" name=\"wizard\" value=\"$wizard\">\n" : "") .
	 ($gUseDelete && $useDelete ?
	  "<input type=\"submit\" value=\"Delete\">\n" : "" ) .
         "<p class='txt'>\n" .
	 "<a href=\"$createURL\">Create New Configuration</a>\n" .
         "</p>\n" .
	 "</form>\n" .
	 "\n"
         );

    if ($wizard) {
       $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                              [ $URL_NEXT, $URL_NAME_NEXT ] );
    }

    # Restore the CPCI settings
#    $host->fromCPCILine($saved_cpci);

    return($s);
}


# Requires:  (to be checked by calling functions.
# 1. There must exist at least 1 boot image file in the system.
sub toEditConfigHTML($$$$) {
    my $stub = $p->profile();
    my ($cgi, $host, $img, $mode) = @_;

    my $title = "Create Boot Configuration";
    if( $mode eq "edit" ) {
	$title = "Edit Boot Configuration";
    }

    my @boot_files = @{ $host->getBootImageFiles() };

    my $PCIDeviceList = $host->getPCIDeviceList($img->getLabel());

    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");

    my $defaultLabel = $img->getLabel();
    my $defaultMemSize = $img->getMemSize();
    my $defaultBootFile = $img->getImage();
    my $totalMem = $host->getTotalMemSize() / 1024;
    
    # Flag that indicates if the image we are editing is the default in LILO
    my $lilo = $host->getLILOConfig();
    my $isDefault = $lilo->getDefaultImage() eq $defaultLabel;

    # Flag indicating if the image should be the default image in LILO
    my $defaultImg = $lilo->getBootImageByLabel($lilo->getDefaultImage());
    my $shouldDefault = 0;
    if( !defined($defaultImg) || !$defaultImg->usesVMnixImage() ) {
	# We want to make sure that a vmnix boot image is the default
	$shouldDefault = 1;
    }

    # Locate a proper default in case it doesn't exist
    if( $defaultBootFile eq "" ) {
	$defaultBootFile = $boot_files[0];
    }

    # Build the list of devices and organize them into sets by owner.  List is 
    # already sorted by bus:device:function.
    my $possPCIDeviceList = $PCIDeviceList->filterPCIDevices(\&possibleDev);
    my @uniq_possdevs = $possPCIDeviceList->getUniqPCIDevice();

    my @host_dev = grep { $_->getOwner() == VMware::HConfig::PCIDevice::HOST } @uniq_possdevs;
    my @vm_dev = grep { $_->getOwner() == VMware::HConfig::PCIDevice::VM } @uniq_possdevs;
    my @shared_dev = grep { $_->getOwner() == VMware::HConfig::PCIDevice::SHARED } @uniq_possdevs;

    # The list of CGI input names that correspond to disks
    my $storageKeys = "";

    # We need to check to make sure that the user gives at least 1 storage controller
    # to the Console OS unless the user is booting the Console off an IDE disk.
    my @storage_dev = grep { isStorageDev($_) } @uniq_possdevs;

    # XXX This check is not perfect as we are checking against the currently mounted
    # root device rather than the root device that the user selects. -jhu
    if( !$img->usesIdeRootDev() ) {
	if( $#storage_dev > -1 ) {
	    # Turn the device list into a command delimited list of strings 
	    my @storage_dev_inputs = map { '"' . makeInputKey($_) . '"' } @storage_dev;
	    $storageKeys = join(', ', @storage_dev_inputs);
	}
    }
    # If $storageKeys is non-null, then we will check for at least one vmhba device accessible
    # to the Console.


    # We want to warn the user if he appears to be giving the active NIC
    # to the VMs, because he will lose network connectivity on reboot.
    # Active NIC is the lowest-numbered host NIC using the eth0 driver;
    # uniq_possdevs is sorted, so just grab first one that matches.

    # XXX This check uses the current configuration in /etc/modules.conf,
    # not the actual active network adapter; should be OK, since we're 
    # really worried about the active adapter at next reboot, not necessarily
    # the current active network adapter.  -rbosch

    my $activeNICModule = undef;
    my $activeNICKey = "";
    my $activeNIC = undef;

    # Determine which NIC module is being used for eth0
    my ($vmnixMods) = $host->getVMnixModules();
    foreach my $modinfo (@$vmnixMods) {
        my $aliases = $modinfo->{ALIASES};
        if (grep { $_ eq "eth0" } @$aliases) {
            $activeNICModule = $modinfo->{MODULE};
            # Check the host NICs to see if they're using the active module
            foreach my $dev (@host_dev) {
               if ($dev->getDrivers()->[0] eq "$activeNICModule.o") {
                  $activeNICKey = makeInputKey($dev);
                  $activeNIC = $dev;
                  last;
               }
            }
            last;
        }
    }

    my $title = "Update Configuration";
    my @path = @htmlPath;
    push @path, [ $htmlTitle, $url_self ];

    my $s =
        toHTMLPath(\@path, $title) . 
        toHTMLSystemInfo($host) . 
        toHTMLWarnings();

    $s .= <<END_HTML;
<script type="text/javascript">
  <!--
  function getCheckedValue(device) {
     var checkedButton = ""
     for (var i = 0; i < document.save.elements.length; i++) {
        if (document.save.elements[i].name == device ) {
           if (document.save.elements[i].checked == "1") {
              checkedButton=document.save.elements[i].value;
           }
        }
     }
     return(checkedButton);
  }

  function doConfirmSave() {
     var scsiArray = new Array($storageKeys);
     var consoleFlag = 0;
        if( scsiArray.length == 0 ) {
           consoleFlag = 1;
        }
        for (var device in scsiArray ) {
        if( getCheckedValue(scsiArray[device]) != "vm" ) {
           consoleFlag = 1;
        }
     }
     var nicFlag = (getCheckedValue("$activeNICKey") != "vm");
     if( consoleFlag == 0 ) {
        return confirm("All your storage devices have been allocated " +
                       "to the Virtual Machines.\\nYou must make sure " +
                       "that the disk on which the Console OS is installed\\n" +
                       "is accessible to the Console OS or your system may not " +
                       "start up\\nproperly when you reboot.\\n\\nAre you sure " +
                       "you want to save the changes to the boot configuration?");
     } else if ( nicFlag == 0 ) {
        return confirm("You have assigned the active network adapter " +
                       "to the Virtual Machines.\\nIf you proceed without " +
                       "manually changing the active network adapter\\non the " +
                       "Console OS, you will lose network connectivity " +
                       "when you reboot.\\n\\nAre you sure to want " +
                       "to save the changes to the boot configuration?");
     } else {
        return confirm("Save changes to boot configuration?");
     }
  }
  // -->
</script>
END_HTML
	 
    $s .= 
	("<h3>Boot Configuration</h3>\n" .
	 "\n" .
	 "<p class='txt'>\n" .
	 "In this section, you can set the basic boot configuration information that\n" .
	 "you would like to use for your system.\n" .
	 "</p>\n" .
	 "\n" .
	 "<form name=\"save\" method=\"POST\" action=\"$url_self\" onSubmit=\"return doConfirmSave()\">\n" .
	 "<table border=0>\n" .
	 "\n" .
	 "<tr>\n" .
	 "  <td>\n" .
	 "  <dl>\n" .
	 "    <dt class='txt'>Name of configuration:<dt>\n" .
	 "    <dd class='txt'><i>Name must be unique.</i></dd>\n" .
	 "    <dd class='txt'><i>Existing names: ");  
	 my @images = @{$host->getLILOConfig()->getVMnixConfigs()};
	 my $img;
	 my @labels;
	 foreach $img (@images) {
	    my $label = $img->getLabel();
	    if ($defaultLabel ne $label) {	
	        push(@labels, $label);
	    }
	 }
         @labels = sort @labels;

	$s .= join(", ", @labels); 
	$s .= 
	("</i></dd>\n".
	 "  </dl>\n" .
	 "  </td>\n" .
	 "  <td><input type=\"text\" name=\"label\" maxlength=\"15\" value=\"$defaultLabel\"></td>\n" .
	 "</tr>\n" .

	 "<tr>\n" .
	 "  <td>\n" .
	 "  <dl>\n" .
	 "    <dt class='txt'>Amount of memory to allocate to the Console:</dt>\n" .
	 "    <dd class='txt'><i>Your system is reporting approximately " . p_number($totalMem) ." MB of memory.\n" .
	 "        <p>\n" .
	 "        If you plan to run 4 virtual machines or fewer, 128 MB should suffice.\n" .
	 "        Set this to 192 MB for 8 virtual machines,\n" . 
	 "        272 MB for 16 virtual machines, or\n" .
	 "        384 MB for 32 virtual machines, or\n" .
	 "        512 MB for more than 32 virtual machines.\n" .
         (($totalMem > $gMaxConsoleMemSize) ? "        Note: you cannot allocate " .
         "more than $gMaxConsoleMemSize MB to the Console.\n" : "") .
	 "        </p>\n".
	 "\n" .
         "        </i></dd>\n" .
	 "  </dl>\n" .
	 "  </td>\n" .
	 "  <td class='txt'><input type=\"text\" name=\"memory\" value=\"$defaultMemSize\">&nbsp;MB</td>\n" .
	 "</tr>\n" .
	 
	 "<tr>\n" .
	 "  <td>\n" .
	 "  <dl>\n" .
	 "    <dt class='txt'>Name of kernel file to use:</dt>\n" .
	 "    <dd class='txt'><i>If you're not sure which setting to choose, it's safe to leave this untouched.</i></dd>\n" .
	 "  </dl>\n" .
	 "  </td>\n" .
	 
	 "  <td>\n" .
	 "  <select name=\"imgfile\">\n");

    my @b = map { 
	"  <option" . ($defaultBootFile eq $_ ? " selected" : "") . ">" . 
	    $_ . "</option>\n"
	} @boot_files;
    my $b_opt;
    foreach $b_opt (@b) {
	$s .= $b_opt;
    }
	
    $s .=
	("  </select>\n" .
	 "  </td>\n" .
	 "</tr>\n");

    # If this boot configuration is not the default one, ask the user
    # if he wishes to make it the default. 
    if( !$isDefault ) {
	$s .= 
	    ("<tr>\n" .
	     "  <td>\n" .
	     "  <dl>\n" .
	     "    <dt class='txt'>Use as default boot configuration?</dt>\n" .
	     "    <dd class='txt'><i>If set as the default, this configuration will be automatically loaded\n\n" .
	     "    each time you reboot.  The current default configuration is <b>" . 
	     $host->getLILOConfig()->getDefaultImage() . "</b>\n" .
	     (defined($gBETA1) && $gBETA1 ?
	      "        <p class='txt'>This option is set to \"Yes\" for this Beta release.</p>\n" : "") .
	     "        </i>\n" .
	     "    </dd>\n" .
	     "  </dl>\n" .
	     "  </td>\n" .
	     "  <td>\n" .
	     "  <select name=\"default\">\n" .
	     ($shouldDefault ? 
	      "  <option selected>Yes</option>\n" :
	      "  <option>Yes</option>\n") .

	     # For Beta1, do not allow the user to specify "No"
	     (defined($gBETA1) && $gBETA1 ? "" : 
	      (!$shouldDefault ?
	       "  <option selected>No</option>\n" : 
	       "  <option>No</option>\n")) .

	     "  </select>\n" .
	     "  </td>\n" .
	     "</tr>\n");
    } else {
	$s .= "<input type=\"hidden\" name=\"default\" value=\"Yes\">\n";
    }

    $s .=
	("</table>\n" .
	 "<hr>\n" .

	 "<h3>Device Allocation</h3>\n" .
	 "<p class='txt'> For each device, select whether you would like to\n" .
         "allocate it to the Console OS, to the Virtual Machines, or share\n" .
         "it between the two.\n" .
         "<ul>\n" .
         "<li class='txt'> <b>Console</b>: Device can only be used by the\n" .
         "Console OS and is unavailable to your virtual machines.\n" .
         "<li class='txt'> <b>Virtual Machines</b>: Device can only be used by\n" .
         "your virtual machines and is unavailable to the Console OS.\n" .
         "<li class='txt'> <b>Shared</b>: Some devices such as SCSI and RAID\n" .
         "adapters can be shared, enabling them to be used by both the\n" .
         "Console OS and your virtual machines.\n" .
         "</ul></p>\n" .

	 "<p class='txt'><u><b>Important note</b></u> &#151; when allocating devices:</p>\n\n" .

	 "<ul>\n" .
	 "<li class='txt'>Make sure the Console's active network adapter\n" .
         "(typically the first listed network adapter) does not get reassigned\n" .
	 "to the virtual machines.  Otherwise you will lose network connectivity upon\n" .
	 "rebooting the machine after boot configuration is complete.\n" .

	 "<li class='txt'>If the Console OS and Virtual Machines will be using disks that reside\n" .
	 "on the same SCSI/RAID adapter, that adapter must be configured as \"Shared.\"\n" .
	 "</ul>\n\n");

    $s .= toHTMLTableHeader( "Console", "Shared", "Virtual Machines",
                             "Device Name", "Driver", "Bus", "Dev",
                             # No longer produces any useful info
                             # "IRQ"

                             );

    my $dev;
    foreach $dev (@uniq_possdevs) {
	my $bus = $dev->getBus();
	my $device = $dev->getDevice();
	# No longer produces any useful info
	# my $irq = $dev->getIrq();
	my $canShare = canShareDev($dev);
	my $driver = $dev->getDrivers()->[0];

	# We can have only 1 shared SCSI device and it must be the lowest number
	# Use this to indicate when the low device has been allocated
#	if( $canShare ) {
#	    $haveShared = 1;
#	}

	# No longer produces any useful info
	#if( !defined($irq) ) {
	#    $irq = "-";
	#}
	my $name = $dev->getName();
	my $className = $dev->getClassName();
	my $input_name = makeInputKey($dev);

	my $host_checked = "checked";
	my $vm_checked = "";
	my $sh_checked = "";

	if( $dev->getOwner() == VMware::HConfig::PCIDevice::VM ) {
	    $host_checked = "";
	    $vm_checked = "checked";
	}
	elsif( $dev->getOwner() == VMware::HConfig::PCIDevice::SHARED ) {
	    $host_checked = "";
	    $sh_checked = "checked";
	}

        my $label = "$className: $name";
        if (defined($activeNIC) && ($dev == $activeNIC)) {
            $label .= "&nbsp;&#151;&nbsp;<b>Active Network Adapter</b>";
        }

        my @fields;

        push @fields, [ qq(<input type="radio" name="$input_name" ) .
                        qq(value="host" $host_checked>), "align='center'" ];

        if ($canShare) {
           push @fields, [ qq(<input type="radio" name="$input_name" ) .
                           qq(value="shared" $sh_checked>), "align='center'" ];
        } else {
           push @fields, "&nbsp;";
        }

        push @fields, [ qq(<input type="radio" name="$input_name" ) .
                        qq(value="vm" $vm_checked>), "align='center'" ];

        push @fields, "$label";
        push @fields, "$driver";
        push @fields, "$bus";
        push @fields, "$device";
        # No longer produces any useful info
        #push @fields, "$irq";

        $s .= toHTMLTableRow(@fields);
    }

    $s .= toHTMLTableFooter();
    $s .=
	("<p>\n" .
	 "<input type=\"hidden\" name=\"edit_id\" value=\"$defaultLabel\"> \n" .
	 "<input type=\"hidden\" name=\"$RCom\" value=\"$RCom_Save\"> \n" .
         ($wizard ? "<input type=\"hidden\" name=\"wizard\" value=\"$wizard\">\n" : "") .
	 "<input type=\"submit\" name=\"submit\" value=\"Save Configuration\"> \n" .
	 "<input type=\"reset\" name=\"reset\" value=\"Restore Defaults\"> \n" .
	 "</p>\n" .
	 "\n" .
	 "</form>\n");
	 
    return($s);
}


sub toDoneConfigHTML($$$$) {
    my $stub = $p->profile();
    my ($cgi, $host, $isDefault, $isEdit) = @_;

    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");

    my $title = "Configuration Updated";
    my @path = @htmlPath;
    push @path, [ $htmlTitle, $url_self ];

    my $s =
        toHTMLPath(\@path, $title) . 
        toHTMLSystemInfo($host) . 
        toHTMLWarnings();

    my $url_reboot = toWizardURL("$url_self?$RCom=$RCom_Reboot", $wizard);
    my $js_reboot = "return doConfirmReboot()";

    my $label = $cgi->param('label');

    $s .= <<END_HTML;
<script type="text/javascript">
<!--
 function doConfirmReboot() {
 return confirm("Are you sure you want to reboot the machine?");
}
// -->
</script>

<p class='txt'>
The boot configuration "$label" has been successfully
END_HTML

    if ($isEdit == 0) {
        $s .= "created. ";
    } else {
        $s .= "modified. ";
    }
    $s .= <<END_HTML;
     You must <a href="$url_reboot" onclick="$js_reboot"> reboot your machine </a> for the changes to take effect. 
END_HTML

    if (!defined($isDefault)) {
        $s .= <<END_HTML;
<ul>
    <li class='txt'><i>If this configuration is the default boot configuration</i>, 
    your machine will automatically start up with the new configuration. 
    <li class='txt'><i>If this configuration is not the default boot configuration</i>,      
    after rebooting you will need to select the configuration name
    at the LILO boot menu on the server console
    in order to use the new configuration.  
<ul>
END_HTML
    } elsif ($isDefault) {
        $s .= <<END_HTML;
<p class ='txt'>
This configuration is the default boot configuration; as a result, when you 
reboot the machine, it will automatically be selected and loaded.
</p>
END_HTML
    } else {
        $s .= "To use the ";
        if ($isEdit == 0) {
           $s .= "new";
        } else {
           $s .= "modified";
        }
        $s .= " configuration, select the configuration name in the LILO 
                boot menu on the server console.</p> 
              ";
    }

    if ($wizard) {
        $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                               [ $url_reboot, "Reboot Machine", $js_reboot ] );
    } else {
       $s .=
            qq(<p class='txt'>To continue using the current\n) .
            qq(configuration, click here to \n) .
            qq(<a href="$URL_INDEX">return to the server configuration page</a>.</p>\n);
    }

    return($s);
}


sub toRebootHTML($$) {
    my $stub = $p->profile();
    my ($cgi, $host) = @_;

    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");

    my $title = "Rebooting System";
    my @path = @htmlPath;
    push @path, [ $htmlTitle, $url_self ];

    my $s =
        toHTMLPath(\@path, $title) . 
        toHTMLSystemInfo($host) . 
        toHTMLWarnings();

    my $continueMsg;
    if ($wizard) {
        $continueMsg = 
            "you may continue the setup process by clicking the " .
            "'Next to $URL_NAME_NEXT' button.\n";
    } else {
        $continueMsg = 
            "you may <a href=\"$URL_INDEX\">return to the server configuration " .
            "page</a>.\n";
    }

    $s .= 
        ("<p class='txt'>\n" . 
         "Your system is being rebooted.  After the reboot process is complete\n" .
         "(about 5-10 minutes, depending upon your system configuration)\n" .
         $continueMsg .
         "If you attempt to proceed before the system has finished booting,\n" .
         "you may receive an error; if so, you should wait for a few minutes\n" .
         "and then reload the page.\n" .
         "</p>\n"
         );

    if ($wizard) {
       $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                              [ $URL_NEXT, $URL_NAME_NEXT ] );
    }

    return($s);
}



#####################################################################
# Function unique to this version                                   #
#####################################################################

sub createNewBootImage($) {
    my $stub = $p->profile();
    my ($host) = @_;

    my $img;     # Image to return

    $img = $host->getLILOConfig()->createBootImage("image", "");
    $img->setLabel($host->getLILOConfig()->getUniqVMBootName());

    # One option that may be useful for this new image
    $img->setOptions( $img->getOptions() . "\tread-only\n" );

    # Set a default memory.  We must have one.
    $img->setMemSize($VMware::HConfig::Constant::gDefaultMemSize);

    return($img);
}

sub selectVMnixBootImage($$) {
    my ($cgi, $host) = @_;

    # Check for a default boot image
    my $default = findDefaultBootImage($host);

    if($default ne "") {
	# There is a prior default image.  Use this one as the default, but
	# prompt the user just the same.
	my @configs = $host->getLILOConfig()->getVMBootImageLabels();

	return(toSelectConfigHTML($cgi, $host, $default));
    }

    return(undef);
}

sub editConfig($$$$) {
    my $stub = $p->profile();
    my($cgi, $host, $img, $mode) = @_;

    my ($canRun, $message) = $host->checkCanRunPcidivy();
    if (!$canRun) {
	Warning($message);
    }

    # Perform checks here.
    if( $img->getImage() eq "" ) {
	my @boot_files = @{ $host->getBootImageFiles() };
	if( $#boot_files == -1 ) {
	    return(toErrorHTML("Missing Boot Image Files", 
			       "Could not find any VM boot image files on the system.  Are you sure you installed the correct packages?"));
	}
    }
    if( $img->getRootDev() eq "" ) {
	$img->setRootDev($host->lookupMountDevice("/"));
    }

    return(toEditConfigHTML($cgi, $host, $img, $mode));
}


sub deleteConfig($$) {
    my $stub = $p->profile();
    my($cgi, $host) = @_;
    my $rc = 1;
    my $img;

    # Delete the boot image from LILO
    my @deleteConfigs = grep( /^boot-\S+$/, $cgi->param());
    my $label;
    foreach $label (@deleteConfigs) {
	$label =~ s/^boot-(\S+)$/$1/;
    }

    if( $#deleteConfigs == -1 ) {
	Warning("No boot configurations were selected to be deleted.");
	return(1);
    }

    my $current_label = $host->getCurrentLabel();

    my $lilo = $host->getLILOConfig();
    foreach $label (@deleteConfigs) {
	# Do no allow the deletion of the label that is currently running
	if( $label eq $current_label ) {
	    Warning("The boot configuration $label is currently running and cannot be deleted.");
	    next;
	}
	$lilo->deleteBootImage($label);

	# A configuration was deleted.  We should update the config DB as well.
	$host->getConfigDB()->markChanged();
    }

    # Set a new default image for LILO if this was the default
    my $default = $lilo->getDefaultImage();
    if( !$lilo->getBootImageByLabel($default) ) {
	$default = findDefaultBootImage($host);

	if( !defined($default) || $default eq "" ) {
	    my @labels = @{ $lilo->getBootImageLabels() };
	    if( $#labels == -1 ) {
		Warning("No LILO boot image labels remaining.  You must have at least one valid LILO entry.");
		return(0);
	    }
	    $default = $labels[0];
	}
	Debug(2, "Setting default boot image to $default\n");
	$lilo->setDefaultImage($default);
    }

    # Save LILO
    if( !$host->writeLILOConfig($lilo) ) {
	Warning("Failed to writeLILOConfig.");
	$rc = 0;
    }
    if( !$host->executeLILO() ) {
	Warning("Failed to execute LILO.");
	$rc = 0;
    }

    # Commit the hardware config database
    if( !$host->commitConfigDB() ) {
	Warning("Failed to commit hardware configuration database.");
	$rc = 0;
    }

    return($rc);
}


sub saveConfig($$) {
    my $stub = $p->profile();
    my($cgi, $host) = @_;

    my ($canRun, $message) = $host->checkCanRunPcidivy();
    if (!$canRun) {
	return(toErrorHTML("Configuration Error", $message));
    }

    # Get all the information about the save from the form and into Perl data
    # structures that we can manipulate.
    my $label = $cgi->param('label');
    my $memSize = $cgi->param('memory');
    my $bootFile = $cgi->param('imgfile');
    my $edit_id = $cgi->param('edit_id');

    # Get the PCI device corresponding to what we are configuring.
    my $PCIDeviceList = $host->getPCIDeviceList($edit_id);

    # Flag that indicates if we need to set the image to be the default
    my $isDefault = ( defined($cgi->param('default')) && $cgi->param('default') eq "Yes" );

    # Mark the devices that are to be owned by the host vs. the VMs
    my @params = $cgi->param;
    @params = grep { /^owner/ } @params;
    my @host_params = grep { $cgi->param($_) eq "host" } @params;
    my @vm_params = grep { $cgi->param($_) eq "vm" } @params;
    my @sh_params = grep { $cgi->param($_) eq "shared" } @params;

    # Now we reduce the lists so that each element is a device hash
    @host_params = map { /owner(.*)/; $1 } @host_params;
    @vm_params = map { /owner(.*)/; $1 } @vm_params;
    @sh_params = map { /owner(.*)/; $1 } @sh_params;

    # Perform error checks
    # 1. Make sure the boot file exists
    # 2. Make sure the label has only alphanumeric characters and dashes
    # 3. Make sure the label is unique, but it can be the one that was allocated to it previously.
    # 4. Make sure the memory size is a number
    my @boot_files = @{ $host->getBootImageFiles() };
    if( !isInSet($bootFile, \@boot_files) ) {
	return(toErrorHTML("Configuration Error", 
			   "Could not find the boot image file '$bootFile'.  Please specify another.\n"));
    }

    if( $label !~ /^[A-Za-z0-9\-_]+$/ ) {
	return(toErrorHTML("Configuration Error", 
			   "Label '$label' is not a valid boot configuration label because it contains\n" .
			   "invalid characters.  Please specify another label using only letters, numbers,\n" .
			   "dashes, or underscores.\n"));
    }

    my @labels = @{$host->getLILOConfig()->getVMBootImageLabels()};
    my $isEdit = 0;
    if (isInSet($edit_id,\@labels)) {
       $isEdit = 1;
    }

    @labels = grep { $_ ne $edit_id } @labels;
    if( isInSet($label, \@labels) ) {
	return(toErrorHTML("Configuration Error", 
			   "Label '$label' is not unique.  Please specify another.\n"));
    }

    # Error checks for memory.  If the vmnix is running but the vmkernel is not
    # loaded, don't perform the check against the total memory size because we
    # cannot know for sure how much memory we really have.
    my $totalMem = $host->getTotalMemSize() / 1024;
    if( $memSize !~ /^[0-9]+$/ || $memSize <= 0 ) {
	return(toErrorHTML("Configuration Error", 
			   "Memory size '$memSize' is not a positive number.  Please specify the amount of\n" .
			   "memory to allocate to the Console OS as a positive number."));
    } elsif ( $memSize > $totalMem && !($host->isVMnix() && !$host->isVMKernel()) ) {
	return(toErrorHTML("Configuration Error", 
			   "Memory size '$memSize' is larger than the\n" .
                           "reported physical memory size of " . p_number($totalMem) .
                           " MB.\n"));
    } elsif ( $memSize > $gMaxConsoleMemSize ) {
	return(toErrorHTML("Configuration Error", 
			   "Memory size '$memSize' is larger than the\n" .
                           "maximum memory size of $gMaxConsoleMemSize MB.\n"));
    }

    # Perform the save by making changes to the SShost structure and then 
    # saving the SShost structure

    # Get a ref to the boot image we wish to save.  If it's not there, it was 
    # just created.  We have to create it again because HTTP is stateless.
    my $img = $host->getLILOConfig()->getBootImageByLabel($edit_id);
    if( !defined($img) ) { 
	$img = createNewBootImage($host);
	$host->getLILOConfig()->addBootImage($img);
    }
    
    $img->setImage($bootFile);
    $img->setMemSize($memSize);
    $img->setLabel($label);

    # Set root device if it is not set
    # Default to the current root device.
    if( $img->getRootDev() eq "" ) {
	$img->setRootDev($host->lookupMountDevice("/"));
    }

    # Select initrd image file.  Try to select the best initrd image that matches
    # the version of the kernel.  Otherwise, just select the first initrd image.
    my $initrdList = $host->getInitrdImageFiles();
    my $initrd = findMatchInitrd($initrdList, $bootFile);
    my $kernelver = getKernelVersion($bootFile);
    if (!$kernelver) {
        # XXX Not sure what to do.  This has to be valid for later on. -jhu
        $kernelver = "unknown";
    }
    
    if (defined($initrd)) {
	$img->setInitrd($initrd);
    } else {
        my $rc;
        ($rc, $initrd) = $host->makeInitrdFile($kernelver, undef);
        if (!$rc || !defined($initrd)) {
            Warning("Could not make $initrd. System may not be able to boot.\n");
            # XXX Not sure what to do.  This has to be valid for later on. -jhu
            $initrd = "initrd-" . $kernelver . ".img";
        } else {
            $img->setInitrd($initrd);
        }
    }

    # Now to divy up the devices
    my $h;
    foreach $h (@host_params) {
	if( $h =~ /(\d+):(\d+)/ ) {
	    my $bus = $1;
	    my $device = $2;
	
	    my @devs = $PCIDeviceList->getPCIDeviceByBusDev($bus, $device);

	    my $dev;
	    foreach $dev (@devs) {
		$dev->setOwner(VMware::HConfig::PCIDevice::HOST);
	    }
	}
    }

    foreach $h (@vm_params) {
	if( $h =~ /(\d+):(\d+)/ ) {
	    my $bus = $1;
	    my $device = $2;
	
	    my @devs = $PCIDeviceList->getPCIDeviceByBusDev($bus, $device);

	    my $dev;
	    foreach $dev (@devs) {
		$dev->setOwner(VMware::HConfig::PCIDevice::VM);
	    }
	}
    }

    foreach $h (@sh_params) {
	if( $h =~ /(\d+):(\d+)/ ) {
	    my $bus = $1;
	    my $device = $2;
	
	    my @devs = $PCIDeviceList->getPCIDeviceByBusDev($bus, $device);

	    my $dev;
	    foreach $dev (@devs) {
		$dev->setOwner(VMware::HConfig::PCIDevice::SHARED);
	    }
	}
    }

    # Now generate and save the CPCI line
    my $res;
    $res = $PCIDeviceList->makeCPCILine();
    $img->setCPCI($res);

    if( $isDefault ) {
	$host->getLILOConfig()->setDefaultImage($img->getLabel());
    }

    # Check to see if we need to add a vmhba driver to the Console OS's
    # modules.conf file.  We need to make sure of this for devices that
    # are shared between the Console OS and the vmkernel.  If we do need to
    # add a line to modules.conf, add it and rebuild the initrd file for
    # the specific kernel version.

    my @vmnixDriverList = $PCIDeviceList->buildVmnixDriverList();
    my ($vmnixMods) = $host->getVMnixModules(); 
    my $scsiHostAdapterNumber = getScsiHostAdapterNumber($vmnixMods);
    if ($scsiHostAdapterNumber < 0 ) {
       Warning("Could not determine maximum SCSI host adapter number in modules.conf file. " .
                "You may have trouble using SCSI devices.");
    }

    foreach my $vmnixDriver (@vmnixDriverList) {
       # $initrd, and $kernelver are defined above.
       my ($vmnixMods) = $host->getVMnixModules(); 

       if (! hasVmnixAlias($vmnixMods, $vmnixDriver, 'scsi_hostadapter')) {
          my $scsiHostAdapterName = "scsi_hostadapter";
          if ($scsiHostAdapterNumber > 0) {
             $scsiHostAdapterName = $scsiHostAdapterName . $scsiHostAdapterNumber;
          }
   
          my $line = qq(alias $scsiHostAdapterName $vmnixDriver); 

          if (! $host->writeModulesConfLine($line)) {
             Warning("Failed to add line '" . $line . "' to the modules.conf file. You may " .
                     "have trouble using ". $vmnixDriver . " devices.") ; 
          } elsif (! $host->makeInitrdFile($kernelver, $initrd)) {
             Warning("Could not make " .  $initrd . ".  System may not be able to boot.\n");
          }
          $scsiHostAdapterNumber = $scsiHostAdapterNumber + 1; 
       }
    }

    if( !$host->writeLILOConfig($host->getLILOConfig()) ) {
	Warning("Failed to writeLILOConfig.");
	$res = 0;
    }
    if( !$host->executeLILO() ) {
	Warning("Failed to execute LILO.");
	$res = 0;
    }
    if( !$host->savePCIDeviceListInConfigDB($host->getConfigDB(), $PCIDeviceList, 1, $img->getLabel()) ) {
	Warning("Failed to save PCIDevices in Config DB.");
	$res = 0;
    }
    if( !$host->commitConfigDB() ) {
	Warning("Failed to commit hardware configuration database.");
	$res = 0;
    }
    if( !$host->writeVmkModules($img->getLabel()) ) {
	Warning("Failed to save vmkmodules file.");
	$res = 0;
    }
    
    if( $res ) {
	return(toDoneConfigHTML($cgi, $host, $isDefault,$isEdit));
    }
    else {
	return(toErrorHTML("Error Saving Configuration",
			   "There was an error in saving your configuration.  Reason: $!\n"));

    }
}


sub handleRequest($$) {
    my $stub = $p->profile();
    my ($cgi, $host) = @_;
    
    my $req = $cgi->param($RCom);
    my $wizard = $cgi->param("wizard");
    
    my @worlds = @{ $host->getRunningVmmWorlds() };
    if( $#worlds > -1 ) {
	Warning("There are currently " . ($#worlds + 1 ) . " virtual machines running on this machine. " .
		"You should stop those virtual machines before performing this operation.\n");
    }

    # XXX Hack
    $gPanicCallback = \&PanicCallbackHTML;

    if( !defined($req) || $req eq "" || $req eq $RCom_Select) {
	my $result = selectVMnixBootImage($cgi, $host);
	if( defined($result) ) {
	    return($result);
	}
	 
	# If there is no image to select, create a new one.
	my $img = createNewBootImage($host);
	$host->getLILOConfig()->addBootImage($img);
	return(editConfig($cgi, $host, $img, "edit"));
    } elsif($req eq $RCom_Create) {
	# For Beta1, put a restriction that only allows 1 vmnix configuration
	if( defined($gBETA1) && $gBETA1 ) {
	    my @images = $host->getLILOConfig()->getBootImages();

	    if( $#images > -1 ) {
		my $message = 
		    "For this Beta version, you are restricted to just one boot configuration with the web based\n" .
                    "setup applications.  However, you may create additional configurations manually.  Read\n" .
		    "the file /usr/doc/README for more information.\n";
		return(toErrorHTML("Beta Configuration Restriction", $message));
	    }
	}

	my $img = createNewBootImage($host);
	$host->getLILOConfig()->addBootImage($img);
	return(editConfig($cgi, $host, $img, "create"));
    } elsif($req eq $RCom_Delete) {
	if( !deleteConfig($cgi, $host) ) {
	    return(toErrorHTML("Deleting Boot Configuration", 
			       "Error deleting boot configurations.") );
	}
        my $url_pcidivy = toWizardURL($URL_PCIDIVY, $wizard);
	force_redirect(\%mimeHeaders, $url_pcidivy);
	return(toErrorHTML("Deleting Boot Configuration",
			   "Delete succeeded.  You should be redirected to " . 
			   "<a href=\"$url_pcidivy\">this page</a>."));
    } elsif($req eq $RCom_Reboot) {
	$host->rebootSystem("Updating boot configuration");
	return(toRebootHTML($cgi, $host));
    } elsif($req eq $RCom_Edit) {
	my $img_label = $cgi->param('boot');
	my $img = $host->getLILOConfig()->getBootImageByLabel($img_label);
	if( defined($img) ) {
	    return(editConfig($cgi, $host, $img, "edit"));
	}
	else {
	    return(toErrorHTML("Edit failure", 
			       "Invalid boot configuration '$img_label' specified.  Configuration does not exist."));
	}
    } elsif($req eq $RCom_Save) {
	return(saveConfig($cgi, $host));
    } elsif($req eq $RCom_Debug) {
	return("<pre>\n" .
	       $host->toString() .
	       "</pre>\n");
    }
}

sub force_redirect($$) {
    my($mimeHeaders, $url) = @_;
    Debug(1, "Redirecting to '$url'");

    if( !defined($mimeHeaders->{'Location'}) ) {
	$mimeHeaders->{'Location'} = $url;
    }
}

sub handler {
    my $stub = $p->profile();
    # Delete the PATH to get by the tainted security requirements of Perl
    delete($ENV{'PATH'}); 

    umask 022;   # Create files that are not writable by group

    &clearWarnings();
    
    my $cgi = new CGI( shift( ) );
    my %credentials = hash_credentials( );
    
    %mimeHeaders = %MimeHeaders;

    my $user = $credentials{ "l" };
    my $password = $credentials{ "m" };
    
    my $s = toHTMLHeader($htmlTitle, $user);

    if ($user ne 'root') {
	my $url_self = $cgi->url(-relative=>1);
        $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLNotRoot($url_self);
        goto footer;
    }
    
    my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $password);

    if(!defined($host)) {
        $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLHostFailure($htmlTitle);
        goto footer;
    }
    
    $s .= handleRequest($cgi, $host);

footer:
    $s .= toHTMLFooter();
    return ($s, %mimeHeaders);
}

1;
