#!/usr/bin/perl

#############################################################################
# scriptedinstall-setup.pl 
# by J. Paul Reed
# for VMware, Inc.
# 01 Dec 2003


use strict;

use lib qw( /usr/lib/vmware-mui/lib/perl5/site_perl/5.005
            /usr/lib/vmware-mui/lib/perl5/site_perl/5.005/VMware/Management/SS/ScriptedInstall );

use Getopt::Long;

use RedHat::BootImage;
use VMware::HConfig::Constant qw( %EsxScriptedInstall );

sub freeSpaceAvailable {
   my $location = shift;

   ## Location must be an existing directory
   if (not -d $location) {
      return -1;
   }

   my ($installDevNum) = (stat($location))[0];

   open(MTAB, "/etc/mtab") or ( warn "Couldn't open mtab: $!" && return -1 );
   my @mounts = <MTAB>;
   close(MTAB);

   my $wantedMount = '';

   foreach my $mountLine (@mounts) {
      chomp($mountLine); 
      ## The mounted directory is in the second column of mtab lines
      my $mount = (split(/\s+/, $mountLine))[1];
      my @mntInfo = stat($mount);
      if ($mntInfo[0] == $installDevNum) {
         $wantedMount = $mount;
         last;
      }
   }

   if ($wantedMount eq '') {
      ## "Couldn't find mount";
      return -1;
   }

   ## -m gives us megabytes
   my $dfout = `/bin/df -m $wantedMount` or 
    ( warn "Couldn't run df: $!" && return -1 );

   ## First line is just a bunch of header info
   my $infoline = (split(/\n/, $dfout))[1];
   ## Free space is in the fourth column
   my $freespace = (split(/\s+/, $infoline))[3]; 
  
   return $freespace; ## in megabytes
} 

sub checkFreeSpace {
   ## Couple of notes: we want to make sure that copying the image won't
   ## totally fill the partition, since that's bad; also, we check the space
   ## on /var because we're going to put this stuff in  
   my $checkPartition = $EsxScriptedInstall{'BOOTCD_LOCATION'};
   $checkPartition =~ s/(.+)\/.+$/$1/;

   return if (freeSpaceAvailable($checkPartition) <
    $EsxScriptedInstall{'RESERVE_SPACE'} +
    $EsxScriptedInstall{'CDROM_ISO_SIZE'});
   return 1;
}

my $checkInstall = 0;
my $verbose = 0;
my $usage = 0;

sub main() {
   GetOptions('h|help' => \$usage,
    'c|check-install' => \$checkInstall,
    'v|verbose' => \$verbose);

   if ($usage != 0) {
      print "Usage info: \n";
      exit 0;
   }
   elsif ($checkInstall != 0) {
      my $rv = checkInstallation();
      if ($verbose) {
         print "This VMware ESX server " . 
          (($rv <= 0) ? 'IS NOT ' : 'is') . 
          " setup for scripted installations.\n";
      }

      ## We mung the return value because of shell return value semantics
      return ($rv <= 0) ? $EsxScriptedInstall{'ERRORS'}{'ENOTSETUP'} : 0;
   } 
   else {
      my $rv = scriptedInstallerSetup();

      if ($EsxScriptedInstall{'ERRORS'}{'EALREADYSETUP'} == $rv) {
         print STDERR "Scripted installation server already setup!\n";
         exit $rv;
      }
      elsif ($EsxScriptedInstall{'ERRORS'}{'EFREESPACE'} == $rv) {
         print STDERR "Not enough space to install scripted installation server\n";
         exit $rv;
      }
   }
} 

sub checkInstallation {
   my $ksFileLocation =
    "$EsxScriptedInstall{'WEBCONFIG_ROOT'}/$EsxScriptedInstall{'KS_FILE'}";

   return if (not -e $ksFileLocation);
   return 1;
}

sub scriptedInstallerSetup {
   return $EsxScriptedInstall{'ERRORS'}{'EALREADYSETUP'} 
    if (checkInstallation());
   return $EsxScriptedInstall{'ERRORS'}{'EFREESPACE'} 
    if (not checkFreeSpace()); 
   return $EsxScriptedInstall{'ERRORS'}{'ENOSUDO'}
    if (not -e $EsxScriptedInstall{'SUDOERS'});

   ## Check for mounted CDROM?
   ## Try mounting the CD?
   if (not -d $EsxScriptedInstall{'ESXCD_LOCATION'}) {
      print STDERR "Must mount the VMware CD in your system's CD-ROM drive;\n";
      print STDERR "Expected location: /mnt/cdrom\n";
      return -1;
   }

   if( -d $EsxScriptedInstall{'BOOTCD_LOCATION'} ) {
      print "Updating your previous VMware ESX CD-ROM image\n";
      `/bin/rm -rf $EsxScriptedInstall{'BOOTCD_LOCATION'}`;
   }

   print "Copying your VMware ESX CD-ROM to your ESX System\n";
   `/bin/cp -a /mnt/cdrom/ $EsxScriptedInstall{'BOOTCD_LOCATION'}`;

   `/bin/ln -s $EsxScriptedInstall{'BOOTCD_LOCATION'} $EsxScriptedInstall{'HTDOC_ROOT'}/install`;

   #
   # get the kickstart file from the boot floppy
   #
   my $tmpFile = `/bin/mktemp boot-img.XXXXXX` or
    die "Temp file creation error: $!\n";
  
   chomp($tmpFile);

   `/bin/cp -f $EsxScriptedInstall{'BOOTIMG_LOCATION'} $tmpFile`;

   my $img = new RedHat::BootImage($tmpFile);

   $img->mount_image();
   $img->retrieve_file( '/ks.cfg', "$EsxScriptedInstall{'WEBCONFIG_ROOT'}/ks.cfg" );
   $img->umount_image();

   unlink($tmpFile);

   #
   # Add the pertinent lines to the sudo file if necessary
   #

   open(SUDO, "<$EsxScriptedInstall{'SUDOERS'}") or die "Failed to open() sudoers file: $!\n";

   my $modBootFound = 0;
   while (<SUDO>) {
      if ($_ =~ /modifyboot/) {
         $modBootFound = 1;
         last;
      }
   }
   close(SUDO);

   if (0 == $modBootFound) {
      open(SUDO, ">>$EsxScriptedInstall{'SUDOERS'}") or die "Failed to open() sudoers file for write: $!\n";
      print SUDO $EsxScriptedInstall{'SUDO_ENTRY'};
      close(SUDO);
   }
}

main();
