# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SS::ServerConfig;
# File: VMware/Management/SS/ServerConfig.pm

use strict;

use VMware::Control '1.00';
use VMware::Control::Server ();

use VMware::ExtHelpers qw( &Warning &clearWarnings );
use VMware::Management::Util qw( &hash_credentials );
use VMware::HConfig::Constant qw( $URL_PCIDIVY $URL_LICENSE $URL_SECURITY
                                  $URL_VIEWPART $URL_NICCONFIG $URL_SWAPFILE 
                                  $URL_WIZ $URL_OVERVIEW $URL_HSTATUS 
                                  $URL_REBOOT );

use VMware::HConfig::SSHostRemote;

use VMware::Management::SS::ESXConfig;

my $htmlTitle = "VMware ESX Server Configuration";

sub toHeaderHTML {
   return qq(<div align="center"><table width="600"><tr><td align="center">\n) .
       toHTMLTableHeader(@_);

}

sub toEntryHTML {
    my ($url, $title, $description) = @_;
    return toHTMLTableRow([qq(<a href="$url"><nobr><b>$title</b></nobr></a>),
                           qq(valign="top")], $description);
}

sub toFooterHTML {
   return toHTMLTableFooter(@_) . qq(</td></tr></table></div>\n);

}

sub toIndexHTML($$) {
   my ($cgi, $host) = @_;
   my $url_self = $cgi->url(-relative=>1);
   my $s;

   $s = <<END_HTML;
<h3>$htmlTitle</h3>
<p class="txt">
If this is a new installation of ESX Server, you should proceed to the
<a href="${URL_PCIDIVY}?wizard=1">VMware ESX Server Setup Wizard</a>.  
The Setup Wizard will guide you through the stages of the server
configuration process.
</p>
<p class="txt">
If you have already configured ESX Server, you may view or update
the configuration by selecting one of the following server
configuration pages.
</p>
END_HTML


   $s .= toHeaderHTML(["Server Configuration", "colspan=2"]);

   $s .= toEntryHTML($URL_PCIDIVY, 
                     "Update Boot Configuration<br>Allocate Devices",
                     "With this option, you can create and modify " .
                     "ESX Server boot configurations.  For each " .
                     "configuration, you can specify how you wish " .
                     "to allocate your devices: to the virtual machines, " .
                     "to the console OS, or shared between them.");
   
   $s .= toEntryHTML($URL_NICCONFIG,
                     "Network Configuration",
                     "Configure the network adapters " .
                     "assigned to the virtual machines.  This " .
                     "option allows you to change the speed and duplex " .
                     "settings of the adapters, and to enable interrupt " .
                     "clustering for improved performance under heavy loads.");
   
   $s .= toEntryHTML($URL_LICENSE,
                     "License Information" ,
                     "View the current license information for this product. " .
                     "If you have a new serial number, you may enter it here.");
   
   $s .= toEntryHTML($URL_SECURITY,
                     "Security Settings",
                     "Configure ESX Server security properties. " .
                     "In this step you can set up SSL-encrypted " .
                     "web access, ssh, telnet and ftp access to the server.");
   
   $s .= toEntryHTML($URL_VIEWPART,
                     "Edit Disk Partitions<br>Create File Systems",
                     "View and modify the partitions and file systems " .
                     "on your disks.  This page allows the creation of disk " .
                     "partitions that use the VMFS file system, suitable " .
                     "for storing disks for virtual machines.");
   
   $s .= toEntryHTML($URL_SWAPFILE,
                     "Swap File Configuration",
                     "Use this option to create and configure a swap file, " .
                     "which enables your virtual machines to use more memory " .
                     "than is physically available on the server.");

   $s .= toEntryHTML("/snmp-config",
                     "SNMP Configuration",
                     "Configure the ESX Server SNMP agent, allowing you to monitor " .
                     "the health of the host machine and of virtual machines running " .
                     "on the host.");

   $s .= toEntryHTML("/vmkernel-config",
                     "VMkernel Configuration",
                     "View and modify the configuration parameters of the VMkernel.");

   $s .= toFooterHTML();

#   $s .= <<END_HTML;
#<p class="txt">
#The following links allow you to manage ESX Server and your virtual machines.
#</p>
#END_HTML

   $s .= toHeaderHTML(["Server Management", "colspan=2"]);

   $s .= toEntryHTML($URL_WIZ,
                     "Virtual Machine Wizard",
                     "With this wizard, you can get a quick start " .
                     "in creating new virtual machines once the machine " .
                     "is set up.");
   
   $s .= toEntryHTML($URL_OVERVIEW,
                     "Virtual Machine Overview",
                     "With this virtual machine manager, you can start, " .
                     "stop, and open a remote console to any virtual machine " .
                     "that you have created through the wizard.");
   
   $s .= toEntryHTML($URL_HSTATUS,
                     "Machine Status",
                     "View a summary of the current status of the machine. " .
                     "The summary includes information about modules, memory, " .
                     "PCI Devices, SCSI adapters, SCSI disks, and " .
                     "VMFS file systems.");
   
   $s .= toEntryHTML("/viewlog",
                     "Log File Viewer",
                     "View the contents of the system log files.");

   $s .= toEntryHTML("/stats/memory",
                     "Memory Utilization",
                     "This page includes a set of charts that display the " .
                     "overall memory utilization on the server, along with " .
                     "detailed memory allocation statistics for the running " .
                     "virtual machines.");

   $s .= toEntryHTML("/uptime",
                     "Availability Report",
                     "View a report showing the uptime and downtime percentages " .
                     "for this server.");

   $s .= toEntryHTML($URL_REBOOT,
                     "Reboot/Halt System",
                     "Reboot or halt the server.");
   
   $s .= toFooterHTML();

   return $s;
}


sub handler {
   # Delete the PATH to get by the tainted security requirements of Perl
   delete($ENV{'PATH'}); 
   
   &clearWarnings();
   
   my $cgi = new CGI( shift( ) );
   
   my %credentials = hash_credentials( );
   my $user = $credentials{ "l" };
   my $password = $credentials{ "m" };
   
   my @path = ();
   my $s = toHTMLHeader($htmlTitle, $user) . toHTMLPath(\@path, $htmlTitle);
   
   if($user ne 'root') {
      my $url_self = $cgi->url(-relative=>1);
      $s .= toHTMLNotRoot($url_self);
      goto footer;
   }
   
   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $password);
   
   if(!defined($host)) {
      $s .= toHTMLHostFailure($htmlTitle);
      goto footer;
   }

   $s .= toHTMLSystemInfo($host) . toHTMLWarnings();
   $s .= toIndexHTML($cgi, $host);
   
 footer:
   $s .= toHTMLFooter();
   return ($s, %MimeHeaders);
}

1;
