# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SS::StartupProfile;
# File: VMware/Management/SS/StartupProfile.pm

use strict;

use VMware::HConfig::SSHostRemote ();
use VMware::HConfig::Constant qw($gDefaultMemSize
                                 $gMaxConsoleMemSize);
use VMware::HConfig::PCIDevice ();
use VMware::HConfig::DevNames ();
use VMware::Management::Util ();

use CGI ();
my $q;

# Error and Warning Messages ------------------------------------------------

sub HOST_FAILURE () {
   return (1, "Could not create remote host object.");
}

sub IMG_DOES_NOT_EXIST ($) {
   my $label = shift();
   return (2, qq{The specified startup profile "$label" does not exist});
}

sub MISSING_STARTUP_FILES {
   # XXX: This message used to include the question, "Are you sure you
   # installed the correct packages?" That doesn't seem to provide information
   # that would lead to a solution so I've omitted it here. If we want to
   # direct the user to a solution, we should try harder. -mikol February 24,
   # 2003 11:40 AM
   return (3, "Could not find any startup files on the system.");
}

sub KERNEL_DOES_NOT_EXIST ($) {
   my $kernel = shift();
   return (4, qq{The specified kernel "$kernel" does not exist});
}

sub LABEL_IS_INVALID ($) {
   my $label = shift();
   return (5, qq{The specified profile name "$label" cannot be used. } .
      qq{Profile names may only contain letters (a-z and A-Z), } .
      qq{digits (0-9), dashes (-) and underscores (_).});
}

sub PROFILE_EXISTS ($) {
   my $label = shift();
   return (6, qq{A profile named "$label" already exists. } .
   qq{Please specify a different profile name.});
}

sub INVALID_MEMORY ($$) {
   my $memory = shift();
   my $max = shift();
   $max = $max > $gMaxConsoleMemSize ? $gMaxConsoleMemSize : $max;
   return (7, qq{The specified reserved memory value "$memory" } .
      qq{cannot be used. Reserved memory must be between } .
      qq{1 and $max megabytes.});
}

# ---------------------------------------------------------------------------

sub handler {
   $q = new CGI(shift());

   my %credentials = VMware::Management::Util::hash_credentials();
   my $user = $credentials{"l"};
   my $pass = $credentials{"m"};
   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $pass);

   if (! defined($host)) {
      my $data = error(HOST_FAILURE());
      return html($data, "if (parent.opCb) parent.opCb(self);");
   }

   # Return data for the default startup profile.
   if ($q->param("ctx") =~ /^wizard$/i) {
      my @images = @{$host->getLILOConfig()->getVMnixConfigs()};

      if (scalar(@images) > 0) {
         # Grab the default profile name. getDefaultImage resturns a string.
         my $default = $host->getLILOConfig()->getDefaultImage();

         # Make sure the default profile is legit.
         my $ok = 0;
         if ($default ne "") {
            for (@images) {
               if ($_->getLabel() eq $default) {
                  $ok = 1;
                  last;
               }
            }
         }

         # Use a substitute if necessary.
         if (! $ok) {
            # Assume that existing images are always labeled.
            $default = $images[0]->getLabel();
         }

         # Generate the HTML data page.
         $q->param("profile", $default);
      } else {
         # There are no profiles. Fake one before generating the HTML data page.
         $q->param("profile", fakeProfile($host)->getLabel());
      }
   }

   # Honor a save request.
   if ($q->param("op") =~ /^save$/i) {
      return save($host);
   }

   # The default action is to return all startup profiles. If param("profile")
   # is defined, data will be returned for that profile only.
   return data($host);
}


# ---------------------------------------------------------------------------
#
# data --
#
#      Given a serverd host, return the data for the requested context and/or
#      object as an HTML document.
#
# Prototype:
#
#      string data(VMware::HConfig::SSHostRemote $host[, string $data])
#
# Arguments:
#
#      $host ... A remote VMware ESX Server object
#      $data ... Optional extra data as JavaScript code
#
# Results:
#
#      An HTML document containing JavaScript code representing $host and one
#      or more startup profiles. If the optional $data argument is passed in,
#      it will be included in the HTML document.
#
# ---------------------------------------------------------------------------

sub data {
   my $host = shift();
   my $data = shift() || "";

   # ------------------------------------------------------------------------
   # Collate the host data in a string of JavaScript code.
   # ------------------------------------------------------------------------

   # Grab the profile for the running kernel.
   my $img = $host->getLILOConfig()->getBootImageByLabel($host->findCurrentLabel());

   if (! defined($img) && $q->param("ctx") =~ /^wizard$/i) {
      $img = $host->getLILOConfig()->getBootImageByLabel($q->param("profile"));
   }

   if (! defined($img)) {
      $data .= error("Could not find any VMware ESX Server startup profiles.");
      return html($data);
   }

   # Grab the managed PCI devices for the running kernel.
   my @imgPciDevices;
   my $pciDeviceList = $host->getPCIDeviceList($img->getLabel());
   $pciDeviceList = $pciDeviceList->filterPCIDevices(\&isManagedDevice);
   @imgPciDevices = $pciDeviceList->getUniqPCIDevice();

   # All bootable kernels
   my @kernels = @{ $host->getBootImageFiles() };
   if ($img->getImage() eq "") {
      if ($#kernels == -1) {
         return(error(MISSING_STARTUP_FILES()));
      }
   }
   $data .= qq{var krnls=["} . join('","', @kernels) . qq{"];$/};

   # HyperThreading enabled/disabled
   my $cfgDb = $host->getConfigDB();
   my $cpuHt = $cfgDb->get("hyperthreading");
   if ( $cpuHt eq "" ) { # case hyperthreading key non existent in hwconfig file
     $data .= "var cpuHt=false;$/";
   } else { 
     $data .= "var cpuHt=$cpuHt;$/";
   }

   # Total physcial RAM
   my $sysMemSize;
   $sysMemSize = $host->getTotalMemSize() / 1024;
   $data .= "var ram=$sysMemSize;$/";
   $data .= "var maxCosRam=$gMaxConsoleMemSize;$/";

   # Reflect the allocation constants into the client code.
   $data .= "var devAllocTypes={".
      "host:" . VMware::HConfig::PCIDevice::HOST . "," .
      "vm:" . VMware::HConfig::PCIDevice::VM . "," .
      "shared:". VMware::HConfig::PCIDevice::SHARED . "};$/";

   # The device configured to be eth0 at boot
   my %cosNic = consoleNic($host,
                           grep { $_->getOwner() == VMware::HConfig::PCIDevice::HOST } @imgPciDevices);
   if (defined(%cosNic)) {
      $data .= "var cosNic={bus:$cosNic{bus},dev:$cosNic{dev},fun:$cosNic{fun}};$/";
   } else {
      $data .= "var cosNic=null;$/";
   }

   # Make sure we have a root file system.
   if ($img->getRootDev() eq "") {
      $img->setRootDev($host->lookupMountDevice("/"));
   }

   # Profile prototype
   $data .= "function Prfl(name,ram,krnl,devs,isDflt){" .
      "this.name=name;" .         # Profile name
      "this.ram=ram;" .           # RAM reserved for the console OS
      "this.krnl=krnl;" .         # Kernel URL
      "this.devs=devs;" .         # Managed devices
      "this.isDflt=isDflt;}$/";   # If profile will be selected at boot

   # Pci Device prototype
   $data .= "function Dev(cls,dnm,mode,alloc,drvrs,slctDrvr,bus,dev,fun,mf){" .
      "this.cls=cls;" .           # Device class name
      "this.dnm=dnm;" .           # Device name (manufacturer, model, etc.)
      "this.mode=mode;" .         # Device can be allocated to...
      "this.alloc=alloc;" .       # Device is allocated to...
      "this.drvrs=drvrs;" .       # List of compatible drivers...
      "this.slctDrvr=slctDrvr;" . # Index of selected driver
      "this.mf=mf;" .             # Is this a multifunction device in which
                                  # each function is allocated separately?
      "this.bus=bus;" .           # PCI bus number
      "this.dev=dev;" .           # PCI device number
      "this.fun=fun;}$/";         # PCI function number

   # ------------------------------------------------------------------------
   # Collate the profile data in a string of JavaScript code.
   # ------------------------------------------------------------------------

   $data .= "var prfls=new Array();";

   if (! defined($q->param("profile"))) {
      for my $img (@{$host->getLILOConfig()->getVMnixConfigs()}) {
         if (defined($img)) {
            $data .= $/ . profileData($host, $img);
         }
      }
   } else {
      my $img = $host->getLILOConfig()->getBootImageByLabel($q->param("profile"));
      $data .= $/ . profileData($host, $img);
   }

   return html($data, "if (parent.opCb) parent.opCb(self);");
}


# ---------------------------------------------------------------------------
#
# profileData --
#
#      Given a serverd host and profile, return the data for the requested
#      object as a string of JavaScript code.
#
# Prototype:
#
#      string profileData(VMware::HConfig::SSHostRemote $host,
#                         VMware::HConfig::LILOBootImage $img)
#
# Arguments:
#
#      $host ... A remote VMware ESX Server object
#       $img ... The startup profile object
#
# Results:
#
#      An HTML document containing JavaScript code representing $host and one
#      or more startup profiles.
#
# ---------------------------------------------------------------------------

sub profileData {
   my $host = shift();
   my $img = shift();
   my $data = "";

   if (defined($img)) {
      my ($ok, $msg) = $host->checkCanRunPcidivy();
      if (! $ok) {
         # XXX: Should this be an error?
         $data .= warning($msg);
      }

      # Managed devices in the selected hardware profile
      my @imgPciDevices;
      my $pciDeviceList = $host->getPCIDeviceList($img->getLabel());
      $pciDeviceList = $pciDeviceList->filterPCIDevices(\&isManagedDevice);

      # Selected profile name
      my $imgLabel;
      $imgLabel = $img->getLabel();

      # Selected RAM reserved for the console
      my $imgMemSize;
      $imgMemSize = $img->getMemSize();

      # Selected kernel URL
      my $imgKernel;
      $imgKernel = $img->getImage();

      # Selected profile booted by default
      my $isDefault;
      $isDefault = $host->getLILOConfig()->getDefaultImage() eq $imgLabel;

      # Generate the selected profile object string.
      # new Prfl(name,ram,krnl,devs,isDflt)
      $data .= "prfls.push(new Prfl(" .
         qq{"$imgLabel",} .
         qq{$imgMemSize,} .
         qq{"$imgKernel",} .
         deviceData($pciDeviceList) . "," .
         ($isDefault ? "true" : "false") . "));";
   } else {
      $data = error(IMG_DOES_NOT_EXIST($q->param("profile")));
   }
}


# ---------------------------------------------------------------------------
#
# consoleNic --
#
#      Given a serverd host and a list of PCI devices currently assigned to
#      the console operating system, return the PCI bus/device/function numbers
#      of the network adapter that will be eth0 the next time the console
#      operating system is started.
#
# Prototype:
#
#      PerlHash consoleNic(VMware::HConfig::SSHostRemote $host,
#                          VMware::HConfig::PCIDevice @hostDevices)
#
# Arguments:
#
#             $host ... The VMware server to probe
#      @hostDevices ... A list of PCI devices assign to the $host console
#
# Results:
#
#      A Perl hash containing the bus/device/function numbers of the
#      console network adapater if it exists, otherwise -1/-1/-1.
#
# ---------------------------------------------------------------------------

sub consoleNic {
   my $host = shift();
   my @hostDevices = @_;

   if (scalar(@hostDevices < 1)) {
      return ("bus" => -1, "dev" => -1, "fun" => -1);
   }

   my $ethZeroModule = undef;
   my $ethZero = undef;

   # XXX This check uses the current configuration in /etc/modules.conf,
   # not the actual active network adapter; should be OK, since we're
   # really worried about the active adapter at next reboot, not necessarily
   # the current active network adapter.  -rbosch

   # Determine which module is being used for eth0.
   my ($vmnixMods) = $host->getVMnixModules();

   for my $modinfo (@$vmnixMods) {
      my $aliases = $modinfo->{ALIASES};

      if (grep { $_ eq "eth0" } @$aliases) {
         $ethZeroModule = $modinfo->{MODULE};

         # Check the host NICs to see if they're using the eth0 module.
         for my $device (@hostDevices) {
            if ($device->getDrivers()->[0] eq "$ethZeroModule.o") {
               $ethZero = $device;
               last;
            }
         }
         last;
      }
   }

   if (defined($ethZero)) {
      return ("bus" => $ethZero->getBus(),
              "dev" => $ethZero->getDevice(),
              "fun" => $ethZero->getFunction());
   }

   return ("bus" => -1, "dev" => -1, "fun" => -1);
}


# ---------------------------------------------------------------------------
#
# deviceData --
#
#      Given a list of PCI devices, return JavaScript code for an anonymous
#      array of those devices.
#
# Prototype
#
#      string deviceData(VMware::HConfig::PCIDevice @devices)
#
# Arguments:
#
#      @devices ... A list of VMware::HConfig::PCIDevice objects
#
# Results:
#
#      An anonymous JavaScript array of Dev objects.
#
# ---------------------------------------------------------------------------

sub deviceData {
   my $pciDeviceList = shift;
   my @devices = $pciDeviceList->getUniqPCIDevice();

   my $jsDevStr = "";

   for my $device (@devices) {
      my $cls = $device->getClassName();
      my $dnm = $device->getName();

      # $mode can be any combination of 4 VMs, 2 COS, or 1 Shared.
      my $mode = $device->isVMwareSharable() ? 7 : 6;
      my $alloc = $device->getOwner();
      my $drvrs = join('","', @{ $device->getDrivers() });
      # XXX: $slctDrvr is fixed for the moment. But we need to permit selection
      # for certain SAN devices.
      my $slctDrvr = 0;
      my $bus = $device->getBus();
      my $dev = $device->getDevice();
      my $fun = $device->getFunction();
      my $mf = $pciDeviceList->getMatchingPCIDevice($device) > 1 ? "true" : "false";

      # new Dev(cls,dnm,mode,alloc,drvrs,slctDrvr,bus,dev,fun,mf);
      $jsDevStr .= ($jsDevStr ne "" ? "," : "") .
         qq{new Dev("$cls","$dnm",$mode,$alloc,} .
         qq{["$drvrs"],$slctDrvr,$bus,$dev,$fun,$mf)};
   }

   return "[$jsDevStr]";
}

# ---------------------------------------------------------------------------
#
# save --
#
#      Given a serverd host, sanity check and save the startup profile.
#
# Prototype:
#
#      string save(VMware::HConfig::SSHostRemote $host)
#
# Arguments:
#
#      $host ... A remote VMware ESX Server object
#
# Results:
#
#      An HTML document containing JavaScript code for any errors or warnings
#      that may have occurred along the way in addition to the information
#      collated by data() (unless the save is aborted).
#
# ---------------------------------------------------------------------------

sub save {
   my $host = shift();
   my $data = "";

   my ($ok, $msg) = $host->checkCanRunPcidivy();
   if (! $ok) {
      return html(error($msg), "if (parent.opCb) parent.opCb(self);");
   }

   # Console operating system configuration.
   my $oldLabel = $q->param("oldLabel");
   my $newLabel = $q->param("newLabel") || $oldLabel;
   my $ram = $q->param("ram");
   my $kernel = $q->param("krnl");
   my $isDefault = 1; # XXX: Hardcoded!
   my $ht = $q->param("hprThrdTgl");


   # Data Validation --------------------------------------------------------

   # Make sure that the selected kernel exists.
   my @kernels = @{ $host->getBootImageFiles() };
   if (! scalar(grep { $kernel eq $_ } @kernels)) {
      return html(error(KERNEL_DOES_NOT_EXIST($kernel)),
                  "if (parent.opCb) parent.opCb(self);");
   }

   # Make sure the label contains only valid characters.
   if ($newLabel !~ /^[A-Za-z0-9\-_]+$/) {
      return html(LABEL_IS_INVALID($newLabel),
                  "if (parent.opCb) parent.opCb(self);");
   }

   # Make sure the profile name is unique.
   my @labels = @{$host->getLILOConfig()->getVMBootImageLabels()};
   @labels = grep { $_ ne $oldLabel } @labels;
   if (scalar(grep { $newLabel eq $_ } @labels)) {
      return html(error(PROFILE_EXISTS($newLabel)),
                  "if (parent.opCb) parent.opCb(self);");
   }

   # Scrutinize the reserved memory value. If the vmnix is running but the
   # vmkernel is not loaded, don't perform the check against the total memory
   # size because we cannot know for sure how much memory we really have.
   my $totalMem = $host->getTotalMemSize() / 1024;
   if (($ram !~ /^[0-9]+$/ || $ram <= 0) ||
       ($ram > $totalMem && !($host->isVMnix() && !$host->isVMKernel())) ||
       ($ram > $gMaxConsoleMemSize)) {
      return html(error(INVALID_MEMORY($ram, $totalMem)),
                  "if (parent.opCb) parent.opCb(self);");
   }

   # Startup Configuration --------------------------------------------------

   # hyperthreading support - changing hpwconfig file through configDB

   my $cpuHtSprt = $host->isCPUHyperthreadingEnabled();

   if ($cpuHtSprt == 1) {
     # system CPU supports hyperthreading
     my $cfgDb = $host->getConfigDB();
    
     if ($ht == 1) {
       # hprThrd enabled by user
       $cfgDb->set("hyperthreading", "true");
     } else {
       # hprThrd disabled by user
       $cfgDb->set("hyperthreading", "false");
     }
   }

   # Get a reference to the specified startup profile. If it's not there, it
   # was just faked. We have to fake it again.
   my $img = $host->getLILOConfig()->getBootImageByLabel($oldLabel);
   if (!defined($img)) {
      $img = fakeProfile($host);
   }

   $img->setImage($kernel);
   $img->setMemSize($ram);
   $img->setLabel($newLabel);

   # Set root device if it is not set, default to the current root device.
   if ( $img->getRootDev() eq "" ) {
      $img->setRootDev($host->lookupMountDevice("/"));
   }

   # Select initrd image file. Try to select the initrd image that matches the
   # kernel. Otherwise, make an initrd image.
   my $version = ($kernel =~ /vmlinuz-(\S+)/ ? $1 : undef);
   my $initrd;
   if ( defined($version) ) {
      for (@{$host->getInitrdImageFiles()}) {
         if ($_ =~ /$version/) {
            $initrd = $_;
            last;
         }
      }
   } else {
      # XXX Not sure what to do. This has to be valid for later on. -jhu
      $version = "unknown";
   }

   if (defined($initrd)) {
      $img->setInitrd($initrd);
   } else {
      my $rc;
      ($rc, $initrd) = $host->makeInitrdFile($version, undef);
      if (! $rc || ! defined($initrd)) {
         # XXX Not sure what to do. This has to be valid for later on. -jhu
         if (! defined($initrd)) {
            $initrd = "initrd-" . $version . ".img";
         }
         $data .= warning(qq{Could not make initrd image "$initrd." } .
            qq{System may not be able to boot.});
      } else {
         $img->setInitrd($initrd);
      }
   }

   # Hardware Profile -------------------------------------------------------

   # Assign devices.
   my $PCIDeviceList = $host->getPCIDeviceList($oldLabel);
   for my $owner ($q->param()) {
      if ($owner =~ /^(\d+):(\d+).(\d+):owner/) {
         # PCI bus:dev:fun
         my $hash = VMware::HConfig::PCIDevice::hash($1, $2, $3);
         my $dev = $PCIDeviceList->getPCIDeviceByHash($hash);
         if (defined($dev)) {
            foreach my $mdev ($PCIDeviceList->getMatchingPCIDevice($dev)) {
               $mdev->setOwner($q->param($owner));
            }
         }
      }
   }

   # Generate and save the CPCI line.
   my $res;
   $res = $PCIDeviceList->makeCPCILine();
   $img->setCPCI($res);

   if ($isDefault) {
      $host->getLILOConfig()->setDefaultImage($img->getLabel());
   }

   # Get a list of device names
   my $devNames = $host->getDeviceNames($oldLabel);

   # Check to see if we need to add a vmhba driver to the Console OS's
   # modules.conf file. This is required for devices that are shared between
   # the Console OS and the vmkernel. If we do need to add a line to
   # modules.conf, add it and rebuild the initrd file for the specific kernel
   # version.  Be sure to add it as the last SCSI driver loaded so as not to
   # to change the /dev/sd#s of existing devices
   #
   # XXX: buildVmnixDriverList returns a list of devices shared between the
   # console and the virtual machines. At the moment, this list will only
   # include "scsi_hostadapter"s. But that could change.
   my @vmnixDriverList = $PCIDeviceList->buildVmnixDriverList();
   my ($vmnixMods) = $host->getVMnixModules();
   my $scsiHostAdapterNumber = getScsiHostAdapterNumber($vmnixMods);
   if ($scsiHostAdapterNumber < 0 ) {
      $data .= warning("Could not determine maximum SCSI host adapter number in modules.conf file. " .
              "You may have trouble using SCSI devices.");
   }

   for my $vmnixDriver (@vmnixDriverList) {
      my ($vmnixMods) = $host->getVMnixModules();

      if (! hasVmnixAlias($vmnixMods, $vmnixDriver, 'scsi_hostadapter')) {
         my $scsiHostAdapterName = "scsi_hostadapter";
         if ($scsiHostAdapterNumber > 0) {
           $scsiHostAdapterName = $scsiHostAdapterName . $scsiHostAdapterNumber;
         }
   
         my $line = qq(alias $scsiHostAdapterName $vmnixDriver); 

         if (! $host->writeModulesConfLine($line)) {
            $data .= warning(qq{Failed to add line "$line" to the } .
               qq{modules.conf file.  You may have trouble using } .
               qq{devices controlled by the "$vmnixDriver" driver module.});
         } elsif (! $host->makeInitrdFile($version, $initrd)) {
            # $initrd, and $version are defined above.
            $data .= warning(qq{Could not make initrd image "$initrd." } .
               qq{System may not be able to boot.});
         }
         $scsiHostAdapterNumber = $scsiHostAdapterNumber + 1;
      }
   }

   # Commit Profile ---------------------------------------------------------

   if (! $host->writeDevNamesConfig($devNames, $PCIDeviceList) ) {
      $data .= warning("Failed to generate devnames.conf file.");
   }
   if (! $host->writeLILOConfig($host->getLILOConfig())) {
      $data .= warning("Failed to writeLILOConfig.");
   }
   if (! $host->executeLILO()) {
      $data .= warning("Failed to execute LILO.");
   }
   if (! $host->savePCIDeviceListInConfigDB($host->getConfigDB(),
                                            $PCIDeviceList,
                                            1,
                                            $img->getLabel())) {
      $data .= warning("Failed to save PCIDevices in Config DB.");
   }

   $devNames->writeToConfigDB($host->getConfigDB(), "");

   if (! $host->commitConfigDB()) {
      $data .= warning("Failed to commit hardware configuration database.");
   }
   if (! $host->writeVmkModules($img->getLabel())) {
      $data .= warning("Failed to save vmkmodules file.");
   }

   return data($host, $data);
}


# ---------------------------------------------------------------------------
#
# remove --
#
#      Given a serverd host, remove the requested profiles.
#
# Prototype:
#
#      string remove(VMware::HConfig::SSHostRemote $host)
#
# Arguments:
#
#      $host ... A remote VMware ESX Server object
#
# Results:
#
#      An HTML document containing JavaScript code for any errors or warnings
#      that may have occurred along the way in addition to the information
#      collated by data() (unless the removal is aborted).
#
# ---------------------------------------------------------------------------

sub remove {
   my $host = shift;
   my $data = "";

   # Make sure that we've been asked to remove at least one profile.
   my @profiles = split(",", $q->param("profile"));

   if (scalar(@profiles) < 1) {
      return html(error(-1, "Profile removal aborted. " .
                        "No startup profiles were selected."),
                  "if (parent.opCb) parent.opCb(self);");
   }

   my $currentLabel = $host->getCurrentLabel();
   my $lilo = $host->getLILOConfig();
   my $defaultLabel = $lilo->getDefaultImage();

   for (@profiles) {
      # Do not remove the running profile.
      if ($_ eq $currentLabel) {
         $data .= warning(qq{The startup profile "$_" is running. } .
                          qq{It was not removed.});
         next;
      }

      # Remove profile.
      $lilo->deleteBootImage($_);
      # Update the config DB.
      $host->getConfigDB()->markChanged();
   }

   # Ensure that a default profile exists.
   my $default = $lilo->getDefaultImage();
   if (! $lilo->getBootImageByLabel($default)) {
      # Make the next available vmnix profile the default.
      my ($default) = @{$lilo->getVMnixConfigs()};

      if (! defined($default) || $default eq "") {
         my @labels = @{$lilo->getBootImageLabels()};

         # Be sure not to remove the last profile.
         if (scalar(@labels) < 1) {
            $data .= error(-1, qq{Profile removal aborted. If the reuested } .
                           qq{startup profiles had been removed, no } .
                           qq{profiles would remain and your system would } .
                           qq{not boot.});
            return html($data, "if (parent.opCb) parent.opCb(self);");
         }

         $default = $labels[0];
      }

      $lilo->setDefaultImage($default);
   }

   if (! $host->writeLILOConfig($lilo)) {
      $data .= warning("Failed to writeLILOConfig.");
   }
   if (! $host->executeLILO()) {
      $data .= warning("Failed to execute LILO.");
   }
   if (! $host->commitConfigDB()) {
      $data .= warning("Failed to commit hardware configuration database.");
   }

   # Return data for the unremoved profiles.
   $q->param("profile", "");
   return data($host, $data);
}


# ---------------------------------------------------------------------------
#
# error --
#
#      Given an error number and message, return a string of JavaScript code
#      that will add the error to the appropriate arrays for use on the client.
#
# Prototype:
#
#      string error(int $err, string $msg)
#
# Arguments:
#
#      $err ... An "error code"
#      $msg ... The error message
#
# Results:
#
#      A string of JavaScript code that will push the error number and message
#      onto the appropriate client-side arrays.
#
# ---------------------------------------------------------------------------

sub error {
   my $err = shift();
   my $msg = shift();

   # Make sure that $msg is safe for cunsmption in a JavaScript program.
   $msg =~ s/\\/\\\\/g;
   $msg =~ s/("|')/\\$1/g;
   my $lf = $/;
   $msg =~ s/$lf/\\n/g;

   return qq[err.push($err);msg.push("$msg");];
}


# ---------------------------------------------------------------------------
#
# warning --
#
#      Given a warning message, return a string of JavaScript code that will
#      add the warning to the appropriate array for use on the client.
#
# Prototype:
#
#      string warning(string $msg)
#
# Arguments:
#
#      $msg ... The warning message
#
# Results:
#
#      A string of JavaScript code that will push the warning message onto the
#      appropriate client-side array.
#
# ---------------------------------------------------------------------------

sub warning {
   my $msg = shift();

   # Make sure that $msg is safe for cunsmption in a JavaScript program.
   $msg =~ s/\\/\\\\/g;
   $msg =~ s/("|')/\\$1/g;
   my $lf = $/;
   $msg =~ s/$lf/\\n/g;

   return qq[wrn.push("$msg");];
}


# ---------------------------------------------------------------------------
#
# html --
#
#      Given data and an optional JavaScript callback, return an HTML document
#      and appropriate headers.
#
# Prototype:
#
#      PerlList html(string $data[, string $cb])
#
# Arguments:
#
#      $data ... Data as a string of JavaScript code
#        $cb ... JavaScript code for a callback function
#
# Results:
#
#      A perl list including the contents of an HTML document and a hash of
#      HTTP headers.
#
# ---------------------------------------------------------------------------

sub html {
   my $data = shift();
   my $cb = shift();
   $cb ||= ";";

   my $html = qq[<html><head><title></title><script>$/] .
      qq[var err=new Array();var msg=new Array();var wrn=new Array();$/] .
      qq[$data$/</script></head><body onload="$cb"></body></html>];

   return($html,
          "Pragma"        => "no-cache",
          "Cache-Control" => "no-cache",
          "Content-Type"  => "text/html; charset=utf-8");
}


# ---------------------------------------------------------------------------
#
# isManagedDevice --
#
#      Decide whether a given device is/can be managed. This routine wraps
#      $device->isVMwareManageable() so that it can be used as a filter:
#
#           $pciDeviceList->filterPCIDevices(\&isManagedDevice)
#
# Prototype:
#
#      bool isManagedDevice(VMware::HConfig::PCIDevice $device)
#
# Arguments:
#
#      $device ... The PCI device in question
#
# Results:
#
#      1 if the device can be managed. 0 otherwise.
#
# ---------------------------------------------------------------------------

sub isManagedDevice {
   my $device = shift();
   return $device->isVMwareManageable();
}


# ---------------------------------------------------------------------------
#
# getScsiHostAdapterNumber --
#
#      Given a list of vmnix modules, search the list of aliases of the 
#      modules for aliases of the form "scsi_hostadapter#". Return  the
#      largest "#" found + 1. If an alias of exactly 'scsi_hostadapter" is
#      found, it is considered to have a "#" of 0.
#
# Prototype:
#
#      int getScsiHostAdapterNumber(PerlHash \@vmnixMods )
#
# Arguments:
#
#        \@vmnixMods ... A reference to an array of module data hashes
#
# Results:
#
#      the max. number found in any "scsi_hostadapter#" aliases + 1.
#
# ---------------------------------------------------------------------------

sub getScsiHostAdapterNumber($) {
   my ($vmnixMods) = @_;
   my $searchStr = "scsi_hostadapter";
   my $highestScsiHostAdapterNumber = -1;
   my $s;

   #
   # look at all the aliases for all the modules and find
   # the largest value for "#" in the "scsi_hostadapter#" aliases
   foreach $s (@$vmnixMods) {
      if (defined($s->{ALIASES})) {
         my @aliases = @{$s->{ALIASES}};
         foreach $a (@aliases) {
            if ((index($a,$searchStr) >= 0) and
             (length($a) > length($searchStr))) {
               my $thisScsiHostAdapterNumber = 
                 substr($a,length($searchStr),length($a) - length($searchStr));
               if ($thisScsiHostAdapterNumber > $highestScsiHostAdapterNumber) {
                  $highestScsiHostAdapterNumber = $thisScsiHostAdapterNumber;
               }
            } elsif ($a eq $searchStr) {
               my $thisScsiHostAdapterNumber = 0;
               if ($thisScsiHostAdapterNumber > $highestScsiHostAdapterNumber) {
                  $highestScsiHostAdapterNumber = $thisScsiHostAdapterNumber;
               }
            }
         }
      }
   }
   return( $highestScsiHostAdapterNumber + 1);
}


# ---------------------------------------------------------------------------
#
# hasVmnixAlias --
#
#      Given a list of vmnix modules, a module name and an alias, return 1 if
#      the alias for the module exists. 0 otherwise.
#
# Prototype:
#
#      bool hasVmnixAlias(PerlHash \@vmnixMods, string $vmnixModName,
#                         string $alias)
#
# Arguments:
#
#        \@vmnixMods ... A reference to an array of module data hashes
#      $vmnixModName ... The vmnix module name
#             $alias ... The alias for the module named by $vmnixModName
#
# Results:
#
#      1 if $alias exists for $vmnixModName. 0 otherwise.
#
# ---------------------------------------------------------------------------

sub hasVmnixAlias($$$) {
   my ($vmnixMods, $vmnixModName, $alias) = @_;

   # grep returns a PerlList. We only care about the first entry.
   my ($modinfo) = grep($_->{MODULE} eq $vmnixModName, @$vmnixMods);
   if (! defined($modinfo)) {
      return 0;
   }

   my @aliases = @{$modinfo->{ALIASES}};
   if (! grep(substr($_,0,length($alias)) eq $alias, @aliases)) {
      return 0;
   }

   return 1;
}


# ---------------------------------------------------------------------------
#
# fakeProfile --
#
#      Creates a temporary profile that can be used when there is no other
#      profile available. This profile may or may not be saved at some later
#      point.
#
# Prototype:
#
#      VMware::HConfig::LILOBootImage
#        fakeProfile(VMware::HConfig::SSHostRemote $host)
#
# Arguments:
#
#      $host ... The host where the profile in question should be generated.
#
# Results:
#
#      A VMware::HConfig::LILOBootImage object representing the new profile.
#
# ---------------------------------------------------------------------------

sub fakeProfile() {
   my $host = shift();

   my $img = $host->getLILOConfig()->createBootImage("image", "");
   $img->setLabel($host->getLILOConfig()->getUniqVMBootName());

   # Set useful options.
   $img->setOptions($img->getOptions() . "\tread-only\n");

   # Set the required default memory.
   $img->setMemSize($VMware::HConfig::Constant::gDefaultMemSize);

   # Add the profile.
   $host->getLILOConfig()->addBootImage($img);

   return $img;
}

1;
