# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SS::StorageConfiguration;
# File: VMware/Management/SS/StorageConfiguration.pm

use strict;

use VMware::HConfig::SSHostRemote ();
use VMware::HConfig::MFSFileSystemRemote;
use VMware::HConfig::RawDisk;
use VMware::HConfig::Constant qw($gVmkDumpPartSize $gVmkDumpPartType
                                 @gPartitionCreateTypes $gVMFSPartType
                                 $gVMFSMaxKBlocks $gVMFS2MaxKBlocks
                                 $gVMFSBlockSizeLimit $gVMFS2BlockSizeLimit
                                 $gVMFSAbsMaxFileSize $gVMFS2AbsMaxFileSize);
use VMware::Management::SS::ESXConfig;
use VMware::Management::Util ();
use VMware::ExtHelpers ();

use CGI ();
my $q;

my $DISK_SET_VMKDUMP = (1 << 0);

# ---------------------------------------------------------------------------

sub handler {
   $q = new CGI(shift());

   my %credentials = VMware::Management::Util::hash_credentials();
   my $user = $credentials{"l"};
   my $pass = $credentials{"m"};
   my $ctx = &VMware::Management::Session::get_temp_context();
   if (!defined($ctx)) {
      return (&VMware::Management::Util::error_page("Could not get temporary context.",
                                                    __PACKAGE__, __LINE__));
   }
   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $pass, $ctx);

   if (! defined($host)) {
      my $data = error("Could not create remote host object.");
      return html($data);
   }

   if (! $host->isVMKernel()) {
      my $data = error("Could not determine storage configuration: " .
                       "The VMkernel is not loaded.");
      return html($data);
   }

   VMware::Management::Util::trace("OP: " . $q->param("op"), "debug");

   if ($q->param("op") =~ /^(add|rem|typical|alien|save|convertToVmfs2)$/i) {
      my $name = $q->param("disk");
      my $disk = $host->getVMSCSIDiskByName($name);

      if ($q->param("op") =~ /^add$/i) {
         return addPartition($host, $disk);
      } elsif ($q->param("op") =~ /^rem$/i) {
         return remPartition($host, $disk);
      } elsif ($q->param("op") =~ /^typical$/i) {
         return createTypicalPartitionTable($host, $disk);
      } elsif ($q->param("op") =~ /^alien$/i) {
         my $part = $q->param("part");
         remPartition($host, $disk, $part);
         return addPartition($host, $disk, $part);
      } elsif ($q->param("op") =~ /^save$/i) {
         return savePartition($host, $disk);
      } elsif ($q->param("op") =~ /^convertToVmfs2$/i) {
         return convertToVmfs2($host, $disk);
      }
   }

   return data($host);
}


# ---------------------------------------------------------------------------

sub data {
   my $host = shift();
   my $data = shift() || "";

   # Access modes
   $data .= qq{var ca={"public":"Public","shared":"Shared"};$/};

   my $hasCoreDumpPartition = scalar(@{$host->getVmkDumpPartitions($gVmkDumpPartSize)});
   $hasCoreDumpPartition = $hasCoreDumpPartition > 0 ? "true" : "false";
   $data .= "var cdp=$hasCoreDumpPartition;$/";
   $data .= "var defCdpMb=" . ($gVmkDumpPartSize / 1024) . ";$/";

   # Multiply bs by bsm to get max file size in GB.
   $data .= "var bsm={vmfs1:$gVMFSMaxKBlocks,vmfs2:$gVMFS2MaxKBlocks};$/";
   $data .= "var bsl={vmfs1:$gVMFSBlockSizeLimit,vmfs2:$gVMFS2BlockSizeLimit};$/";
   $data .= "var ams={vmfs1:$gVMFSAbsMaxFileSize,vmfs2:$gVMFS2AbsMaxFileSize};$/";

   # Hex codes for VMkernel dump and VMFS partitions
   my $vmfstc = sprintf("0x%02x", $gVMFSPartType);
   my $vmkdtc = sprintf("0x%02x", $gVmkDumpPartType);
   $data .= "var tc={vmfs:$vmfstc,vmkd:$vmkdtc,extended:0x05};$/";
   $data .= "function isExtd(id){return (id==0x05||id==0x0f||id==0x85||id==0x0f);}$/";


   $data .= "function Dsk(id,c,h,s,mb,fs,pt){" .
      "this.id=id;" .         # VMkernel logical name (e.g., vmhba0:0:0)
      "this.c=c;" .           # Cylinders
      "this.h=h;" .           # Heads
      "this.s=s;" .           # Sectors
      "this.mb=mb;" .         # Capacity in megabytes
      "this.fs=fs;" .         # Unpartitioned space
      "this.pt=pt;}$/";       # Partition table as an array of partitions

   $data .= "function Prt(id,sc,ec,mb,tc,tn,fs,bs,vl,ca,sp){" .
      "this.id=id;" .         # Partition number
      "this.sc=sc;" .         # Logical start cylinder
      "this.ec=ec;" .         # Logical end cylinder
      "this.mb=mb;" .         # Capacity in megabytes
      "this.tc=tc;" .         # Partition type code (extended, Dump, VMFS)
      "this.tn=tn;" .         # Partition type name (extended, Dump, VMFS)
      "this.fs=fs;" .         # If formatted
      "this.bs=bs;" .         # Block size (max file size)
      "this.vl=vl;" .         # Volume label
      "this.ca=ca;" .         # Cluster access
      "this.sp=sp;}$/";       # Spanned partitions as an array of Prt objects

   $data .= "var dsks=[];$/";

   if (! defined($q->param("disk"))) {
      for my $disk (@{$host->getVMSCSIDisks()}) {
         if (defined($disk)) {
            $data .= diskData($host, $disk);
         }
      }
   } else {
      my $name = $q->param("disk");
      my $disk = $host->getVMSCSIDiskByName($name);

      if (defined($disk)) {
         $data .= diskData($host, $disk);
      } else {
         $data .= error(qq{The requested disk "$name" does not exist.});
      }
   }

   return html($data);
}


# ---------------------------------------------------------------------------

sub diskData {
   my $host = shift();
   my $disk = shift();
   my $data = "";

   my $name = $disk->getVmkName();
   if (! VMware::HConfig::RawDisk::isLogVmkDiskName($name)) {
      $data .= error(qq{"$name" is not a valid disk identifier.});
      return $data;
   }

   my $cylinders = $disk->getGeoCylinder();
   my $heads = $disk->getGeoHead();
   my $sectors = $disk->getGeoSector();
   my $capacity = $disk->getTotalSectors() / 2 / 1024;
   my $freeSpace = 0;

   my $partitionTable = $disk->getPartitionTable();

   if (! defined($partitionTable)) {
      return $data;
   }

   my @fcr = $disk->getPartitionTable()->getFreeCylinderRange();
   for (@fcr) {
      my $sc = $_->{START_CYL};
      my $ec = $_->{END_CYL};
      my $mb = $disk->getCylToKBytes($ec - $sc + 1) / 1024;

      $freeSpace += $mb;
   }

   # new Dsk(id,c,h,s,mb,fs,pt)
   $data .= qq{"$name",$cylinders,$heads,$sectors,$capacity,$freeSpace,};
   $data .= partitionTableData($host, $disk, @fcr);

   $data = qq{dsks.push(new Dsk($data));};

   return $data;
}


# ---------------------------------------------------------------------------

sub sortByStartCylinder {
   if ($a->{START_CYL} == $b->{START_CYL}) {
      if (defined($a->{PARTNUM}) && defined($b->{PARTNUM})) {
         return $a->{PARTNUM} <=> $b->{PARTNUM};
      }

      if (defined($b->{PARTNUM})) {
         return 1;
      }

      if (defined($a->{PARTNUM})) {
         return -1;
      }
   }
   return $a->{START_CYL} <=> $b->{START_CYL};
}

sub partitionTableData {
   my $host = shift();
   my $disk = shift();
   my @partitionTable = @_; # Remaining args are unpartitioned cylinder ranges.
   my $data = "";

   # Unify unpartitioned cylinders and existing partitions.
   for ($disk->getPartNums()) {
      my %range = (START_CYL => $disk->getPartStartCyl($_),
                   END_CYL   => $disk->getPartEndCyl($_),
                   LOGICAL   => ($_ > 4 ? 1 : 0),
                   PARTNUM   => $_);
      push(@partitionTable, \%range);
   }

   # Sort unpartitioned space and existing partitions into their order on disk.
   @partitionTable = sort sortByStartCylinder @partitionTable;

   for (@partitionTable) {
      my $pd = "";

      if (defined($_->{PARTNUM})) {
         $pd = partitionData($host, $disk, $_->{PARTNUM});
      } else {
         # Generate a fake partition for any free space.
         my $sc = $_->{START_CYL};
         my $ec = $_->{END_CYL};
         my $mb = $disk->getCylToKBytes($ec - $sc + 1) / 1024;
         my $tn = $_->{LOGICAL} ? "logical" : "physical";

         # new Prt(id,sc,ec,mb,tc,tn,fs,bs,vl,ca,sp)
         $pd = qq/new Prt("",$sc,$ec,$mb,null,"$tn",false,0,"","",{id:"",ex:[]})/;
      }

      $data .= ($data ? "," : "") . $pd;
   }

   return "[$data]";
}


# ---------------------------------------------------------------------------

sub partitionData {
   my $host = shift();
   my $disk = shift();
   my $number = shift();
   my $data = "";

   my $isVMwarePartType =
      VMware::HConfig::RawDisk::isPartVmware($disk->getPartType($number));

   # new Prt(id,sc,ec,mb,tc,tn,fs,bs,vl,ca,sp)
   my $id = VMware::HConfig::RawDisk::makeLogVmkPartName($disk->getVmkName(),
                                                         $number);
   my $startCylinder = $disk->getPartStartCyl($number);
   my $endCylinder = $disk->getPartEndCyl($number);
   my $capacity = sprintf("%d", $disk->getPartSectors($number) / 2 / 1024);
   my $typeCode = sprintf("0x%02x", $disk->getPartType($number));
   my $typeName = $disk->getPartTypeName($number);
   my $blockSize = 0;
   my $volumeLabel = "";
   my $clusterAccess = "null";
   my $spanInfo = qq/{id:"",ex:[]}/;

   my ($isVmfs, $isHead, $fs) = (0, 0, undef);

   # Only singleton disks or disk chain heads are reported in getAllMFS. Disks
   # in a broken disk chain are not reported at all.
   for my $mfs (@{$host->getAllMFS()}) {
      if ($id eq $mfs->getVmkName()) {
         ($isVmfs, $isHead, $fs) = (1, 1, $mfs);
      } elsif (grep($_ eq $id, @{$mfs->getExtents()})) {
         ($isVmfs, $isHead, $fs) = (1, 0, $mfs);
      }
   }

   if (defined($fs)) {
      $typeName .= "-" . $fs->getVersion() . "." . $fs->getMinorVersion();
      $volumeLabel = $fs->getLogName();
      $blockSize = $fs->getBlockSize();
      $clusterAccess = $fs->getAccessibility();

      if ($isHead) {
         my $str = "";
         for (@{$fs->getExtents()}) {
            $str .= ($str ? "," : "") . qq/"$_"/;
         }
         $spanInfo = qq/{id:"$id",ex:[$str]}/;
      } else {
         # In this case getVmkName refers to the head of the disk chain. In
         # other words, prt.id will not equal prt.sp.id if the partition is an
         # extent. ex.length will also be 0.
         $spanInfo = qq/{id:"/ . $fs->getVmkName() . qq/",ex:[]}/;
      }

   } elsif ($isVmfs) {
      $data .= warning(qq{Partition "$id" is a VMFS partition, but a } .
         qq{corresponding VMFS object could not be found.});
   }

   $isVmfs = $isVmfs ? "true" : "false";

   $data .= qq{new Prt("$id",$startCylinder,$endCylinder,$capacity,$typeCode,} .
            qq{"$typeName",$isVmfs,$blockSize,"$volumeLabel",} .
            qq{"$clusterAccess",$spanInfo)};
}


# ---------------------------------------------------------------------------
# https://mikol-esx.vmware.com/storage?op=add&disk=vmhba0:1:0&sc=251&ec=400&tn=primary&tc=0xfb

sub addPartition {
   my $host = shift();
   my $disk = shift();
   my $part = shift();
   my $data = "";
   my $number;

   my $startCylinder = $q->param("sc");
   my $endCylinder = $q->param("ec");
   my $typeName = $q->param("tn"); # One of primary, extended or logical
   my $typeCode = $q->param("tc");
   my $volumeLabel = $q->param("vl");
   my $clusterAccess = $q->param("ca");
   my $mbCapacity = $q->param("mb");
   my $blockSize = $q->param("bs");

   if (! isVolumeLabelUnique($host, $volumeLabel)) {
      $data .= error(qq{Create Volume: Label is not unique: $volumeLabel. } .
         qq{The volume was not created.});
      return data($host, $data);
   }


   if (defined($mbCapacity) && $mbCapacity > 0 &&
       $gVmkDumpPartType != $typeCode) {
      my $spc = $disk->getChsToLba(1, 0, 1) + 1; # Sectors per cylinder
      my $kpc = $spc / 2;                        # Kilobytes per cylinder
      my $cylinderLength = int($mbCapacity * 1024 / $kpc);

      # Round cylinder length up.
      if ($mbCapacity % $kpc > 0) {
         $cylinderLength++;
      }

      $endCylinder = $startCylinder + $cylinderLength - 1;
   }

   my $partitionTable = $disk->getPartitionTable();

   if ($typeCode == 0x05) {
      $typeName = "extended";
   }

   # In the absence of a type name, assume that the partition will be primary -
   # unless the cylinder range falls within an extended partition.
   if (! defined($typeName) || $typeName eq "") {
      $typeName = "primary";

      if (isLogicalFreeSpace($partitionTable, $startCylinder, $endCylinder)) {
         $typeName = "logical";
      }
   }

   my $cylinderLength = $endCylinder - $startCylinder + 1;
   my $cylinders = $disk->getGeoCylinder() + 1;
   my $heads = $disk->getGeoHead();
   my $sectors = $disk->getGeoSector();

   # Use default size for VMkernel core dump partitions.
   if (defined($part)) {
      my ($d, $n) = VMware::HConfig::RawDisk::unmakeLogVmkPartName($part);
      $number = $n;
   } elsif ($gVmkDumpPartType == $typeCode) {
      my @range = placeDumpPartition($disk, $startCylinder, $endCylinder);

      if (! defined($range[0])) {
         return html($range[1]);
      } else {
         VMware::Management::Util::trace("$startCylinder - $endCylinder", "debug");
         $startCylinder = $range[0];
         $endCylinder = $range[1];
         $cylinderLength = $range[1] - $range[0];
         VMware::Management::Util::trace("$startCylinder - $endCylinder", "debug");

         # Give core dump partitions a number that is more likely to reflect
         # their position on disk. XXX: Although the required behavior is the
         # same for extended partitions, the following is clearly a kludge.
         ($number) = generatePartitionNumberAndTypeCode($disk,
                                                        $typeName eq "logical" ? $typeName : "extended");
      }
   } else {
      ($number, $typeCode) = generatePartitionNumberAndTypeCode($disk, $typeName);
   }

   my $sectorLength = VMware::HConfig::RawDisk::fromChsToLba($cylinderLength,
                                                             0,
                                                             1,
                                                             $cylinders,
                                                             $heads,
                                                             $sectors);

   if ($sectorLength > 0) {
     $partitionTable->addPart($number, $startCylinder, $endCylinder,
                              hex($typeCode), $sectorLength, 0);

     if (VMware::HConfig::RawDisk::isLogicalPart($number)) {
        if ($partitionTable->reorderLogicalParts() == 0) {
           # If partitions are reordered, we need to get the partition's new index
           my @parts = $partitionTable->lookupParts($startCylinder, $endCylinder, undef, undef);
           $number = $parts[0];
        }
     }

     $data .= preDiskOperation($host, $DISK_SET_VMKDUMP);
     $data .= writePartitionTable($host, $disk, $partitionTable);
     $data .= postDiskOperation($host, $DISK_SET_VMKDUMP);

     # XXX: Need to invalidate cache so FS change is picked up.
     $host->{MFS} = undef;

     # Initialize VMFS partitions.
     if ($typeCode == $gVMFSPartType) {
        $part = VMware::HConfig::RawDisk::makeLogVmkPartName($disk->getVmkName(),
                                                             $number);

        # Pick a block size that allows a single file to occupy the entire volume.
        my $fsSizeMb = $disk->getCylToKBytes($cylinderLength) / 1024 / 1024;

        makeFileSystem($host, $disk, $part,
                       $blockSize || getVmfsBlockSize($fsSizeMb));

        # XXX: Need to invalidate cache so FS change is picked up.
        $host->{MFS} = undef;

        if (defined($volumeLabel) && $volumeLabel ne "") {
           $data .= setVolumeLabel($host, $disk, $part, $volumeLabel);
        }

        if (defined($clusterAccess) &&
            $clusterAccess =~ /^(private|public|shared)/i) {
           setClusterAccess($host, $disk, $part, $clusterAccess);
        }
     }
   } else {
     # Translating from CHS to LBA failed.
     $data .= error("Could not write partition information to disk.");
   }


   $q->delete("op");
   $q->delete("sc");
   $q->delete("ec");
   $q->delete("tn");
   $q->delete("tc");

   return data($host, $data);
}


# ---------------------------------------------------------------------------

sub savePartition {
   my $host = shift();
   my $disk = shift();
   my $data = "";

   my $part = $q->param("part");
   my $blockSize = $q->param("bs");
   my $typeName = $q->param("tn");
   my $typeCode = $q->param("tc");
   my $volumeLabel = $q->param("vl");
   my $clusterAccess = $q->param("ca");

   if (! isVolumeLabelUnique($host, $volumeLabel)) {
      $data .= error(qq{Edit Volume: Label is not unique: $volumeLabel. } .
         qq{The volume has not been changed.});
      return data($host, $data);
   }

   if ($typeCode == $gVMFSPartType) {
      my $fileSystem = $host->getMFS($part) ? 1 : 0;
      my $curBlockSize = 1;

      if ($fileSystem) {
         my $fs = $host->getMFS($part);

         if (defined($fs)) {
            $curBlockSize = $fs->getBlockSize();
         }
      }

      if (! $fileSystem || $blockSize != $curBlockSize) {
         $data .= makeFileSystem($host, $disk, $part, $blockSize);
      }

      # XXX: Need to invalidate cache so FS change is picked up.
      $host->{MFS} = undef;

      if (defined($volumeLabel)) {
         $data .= setVolumeLabel($host, $disk, $part, $volumeLabel);
      }

      if (defined($clusterAccess) &&
          $clusterAccess =~ /^(private|public|shared)/i) {
         $data .= setClusterAccess($host, $disk, $part, $clusterAccess);
      }

      $data .= addExtents($host);
   }

   return data($host, $data);
}


# ---------------------------------------------------------------------------

sub convertToVmfs2 {
   my $host = shift();
   my $disk = shift();
   my $data = "";

   my $part = $q->param("part");
   my ($diskName, $number) = VMware::HConfig::RawDisk::unmakeLogVmkPartName($part);
   my $typeName = $q->param("tn");
   my $typeCode = $disk->getPartType($number);

   if ($typeCode == sprintf("0x%02x", $gVMFSPartType)) {
      my $fileSystem = $host->getMFS($part) ? 1 : 0;
      my $curVersion = 1;

      if ($fileSystem) {
         my $fs = $host->getMFS($part);

         if (defined($fs)) {
            $curVersion = $fs->getVersion();

            if ($curVersion != 2) {
               if (! $fs->convertToVmfs2()) {
                  $data .= error("VMFS Conversion: Failed.");
               }

               # XXX: Need to invalidate cache so FS change is picked up.
               $host->{MFS} = undef;
            }
         } else {
            $data .= error("VMFS Conversion: $part: No file system.");
         }
      } else {
         $data .= error("VMFS Conversion: $part: No file system.");
      }
   } else {
      $data .= error("VMFS Conversion: $part: Not VMFS.");
   }

   return data($host, $data);
}


# ---------------------------------------------------------------------------

sub generatePartitionNumberAndTypeCode {
   my $disk = shift();
   my $typeName = shift();

   my ($number, $typeCode);

   if ($typeName =~ /^extended$/i) {
      $number = 4;

      # Loop until we find an unused number.
      while (isExistingPartitionNumber($disk, $number) && $number > 0) {
         $number--;
      }

      if ($number < 1) {
         # XXX: Must surface this error.
         return (0,
                 qq{No more than four primary partitions, or three primary } .
                 qq{and one extended partition, can be created on a single } .
                 qq{disk.});
      }

      $typeCode = 0x05;
   } elsif ($typeName =~ /^logical$/i) {
      $number = 5;

      # Loop until we find an unused number.
      while (isExistingPartitionNumber($disk, $number)) {
         $number++;
      }

      $typeCode = sprintf("0x%02x", $gVMFSPartType);
   } else {
      # Create a primary partition, even if a type wasn't specified.
      $number = 1;

      # Loop until we find an unused number.
      while (isExistingPartitionNumber($disk, $number) && $number < 5) {
         $number++;
      }

      if ($number > 4) {
         # XXX: Must surface this error.
         return (0,
                 qq{No more than four primary partitions, or three primary } .
                 qq{and one extended partition, can be created on a single } .
                 qq{disk.});
      }

      $typeCode = sprintf("0x%02x", $gVMFSPartType);
   }

   return ($number, $typeCode);
}


# ---------------------------------------------------------------------------
# https://mikol-esx.vmware.com/storage?op=rem&disk=vmhba0:1:0&part=vmhba0:1:0:2

sub remPartition {
   my $host = shift();
   my $disk = shift();
   my $part = shift();
   my $data = "";

   $part ||= $q->param("part");

   my ($ok, $name, $number) = isExistingPartitionName($disk, $part);

   if (! $ok) {
      $data .= error(qq{Could not remove the requested partition "$part" } .
                     qq{because it does not exist or is not a valid } .
                     qq{partition name.});
      return data($host, $data);
   }

   my $hasSwap = partHasActiveSwapFile($host, $disk, $part);

   if ($hasSwap) {
      $data .= error(qq{Could not remove the requested partition "$part" } .
                     qq{because it contains an active swap file.});
      return data($host, $data);
   }

   my $partitionTable = $disk->getPartitionTable();

   my $type = $partitionTable->getPartType($number);
   if (VMware::HConfig::RawDisk::isPartExtended($type) &&
       $partitionTable->hasLogicalParts()) {
      $partitionTable->removeAllLogicalParts();
   }

   $data .= preDiskOperation($host, $DISK_SET_VMKDUMP);

   $partitionTable->removePart($number);

   if (VMware::HConfig::RawDisk::isLogicalPart($number)) {
      $partitionTable->reorderLogicalParts();
   }

   if (scalar($partitionTable->getPartNums()) < 1) {
      $data .= remAllPartitions($host, $disk);
   } else {
      $data .= writePartitionTable($host, $disk, $partitionTable);

      $q->delete("op");
      $q->delete("part");
   }

   $data .= postDiskOperation($host, $DISK_SET_VMKDUMP);

   return data($host, $data);
}


sub remAllPartitions {
   my $host = shift();
   my $disk = shift();

   if (! $disk->deleteAllPartitions()) {
      return error(qq{Could not remove all partitions. Reason: $!.});
   }

   return "";
}


# ---------------------------------------------------------------------------

sub partHasActiveSwapFile {
   my $host = shift(); # SSHost object instance
   my $disk = shift(); # RawDisk object instance
   my $part = shift(); # VMK part name (e.g., vmhba0:0:0:1)
   my $data = "";

   #   $name ... VMK LUN name (e.g., vmhba0:0:0)
   # $number ... Partition number (e.g., 1)
   my ($ok, $name, $number) = isExistingPartitionName($disk, $part);

   my $partitionTable = $disk->getPartitionTable();
   my $type = $partitionTable->getPartType($number);
   if (VMware::HConfig::RawDisk::isPartExtended($type)) {
      if (! $partitionTable->hasLogicalParts()) {
         return 0;
      }

      my $n = 5;

      # Loop until we find an unused number.
      while (isExistingPartitionNumber($disk, $n)) {
         if (hasActiveSwapFile($host, $name, $n)) {
            return 1;
         }

         $n++;
      }
   } else {
      if (hasActiveSwapFile($host, $name, $number)) {
         return 1;
      }
   }

   return 0;
}


# ---------------------------------------------------------------------------

sub hasActiveSwapFile {
   my $host = shift(); # SSHost object instance
   my $disk = shift(); # VMK LUN name (e.g., vmhba0:0:0)
   my $part = shift(); # Partition number (e.g., 1)
   my $data = "";

   my $vmkn = VMware::HConfig::RawDisk::makeLogVmkPartName($disk, $part);

   if ($host->getMFS($vmkn)) {
      my (undef, undef, undef, undef, $swapVol, undef) = $host->getSwapInfo();
      if (defined($swapVol) && $swapVol ne "") {
         my $swapMfs = $host->getMFSLookup($swapVol);
         if (defined($swapMfs) &&
             $disk eq $swapMfs->getDisk() &&
             $part eq $swapMfs->getPartition()) {
            return 1;
         }
      }
   }

   return 0;
}


# ---------------------------------------------------------------------------

sub placeDumpPartition {
   my $disk = shift();
   my $startCylinder = shift();
   my $endCylinder = shift();

   my $cylinders = $disk->getGeoCylinder() + 1;

   # if ($cylinders - 1 > $endCylinder ||
   if (! isFreeSpace($disk->getPartitionTable(), $startCylinder, $endCylinder))
   {

      return (undef, error(qq{The requested VMkernel core dump partition } .
                           qq{was not created because the specified space } .
                           qq{is not at the end of the disk or it is not } .
                           qq{available.}));
   }

   my $spc = $disk->getChsToLba(1, 0, 1) + 1; # Sectors per cylinder
   my $kpc = $spc / 2;                        # Kilobytes per cylinder

   # Round cylinder length up.
   my $cylinderLength = int($gVmkDumpPartSize / $kpc);
   if ($gVmkDumpPartSize % $kpc > 0) {
      $cylinderLength++;
   }

   if ($cylinderLength > ($cylinders - 1) - $startCylinder) {
      return (undef, error(qq{The requested VMkernel core dump partition } .
                           qq{was not created because there is not enough } .
                           qq{space available at the end of the disk to } .
                           qq{contain the mandatory } .
                           int($gVmkDumpPartSize / 1024) . qq{ megabyte } .
                           qq{core dump partition.}));
   }

   # Return cylinder rang at the end of the disk for the dump partition.
   return ($endCylinder - $cylinderLength + 1, $endCylinder);
}


# ---------------------------------------------------------------------------
# https://mikol-esx.vmware.com/storage?op=typical&disk=vmhba0:1:0&sc=0&ec=1114

sub createTypicalPartitionTable {
   my $host = shift();
   my $disk = shift();
   my $data = "";

   my $wantPartitions = $q->param("dump") =~ /^true$/i ? 2 : 1;
   my $startCylinder = $q->param("sc");
   my $endCylinder = $q->param("ec");
   my $volumeLabel = $q->param("vl");

   my @availablePartNumbers = maybeCreateExtendedPartition($disk,
                                                           $wantPartitions,
                                                           $startCylinder,
                                                           $endCylinder);

   if (@availablePartNumbers == 2 && ! defined($availablePartNumbers[0])) {
      return html($availablePartNumbers[1]); # Return error page.
   }

   my $cylinders = $disk->getGeoCylinder() + 1;
   my $heads = $disk->getGeoHead();
   my $sectors = $disk->getGeoSector();
   my ($partitionTable, $number, $cylinderLength, $sectorLength);

   # Reserve space for a dump partition -------------------------------------
   my $dumpCylinderLength = 0;
   my $dumpOnly = 0;
   if ($q->param("dump") =~ /^true$/i) {
      my $spc = $disk->getChsToLba(1, 0, 1) + 1; # Sectors per cylinder
      my $kpc = $spc / 2;                        # Kilobytes per cylinder

      # Round cylinder length up.
      $cylinderLength = int($gVmkDumpPartSize / $kpc);
      if ($gVmkDumpPartSize % $kpc > 0) {
         $cylinderLength++;
      }

      my @range = placeDumpPartition($disk, $startCylinder, $endCylinder);
      if (! defined($range[0])) {
         $data .= $range[1];
      } else {
         $cylinderLength = $range[1] - $range[0] + 1;

         if ($cylinderLength == ($cylinders - 1) - $startCylinder) {
            $data .= warning(qq{Of the requested $wantPartitions partitions } .
                             qq{only a dump partition was created beacuse } .
                             qq{there is not enough available space to } .
                             qq{contain both the mandatory } .
                             int($gVmkDumpPartSize / 1024) . qq{ megabyte } .
                             qq{core dump partition and a VMFS volume.});
            $dumpOnly = 1;
         }

         # Reserve space at the end for the dump partition.
         $dumpCylinderLength = $cylinderLength;
         $endCylinder -= $cylinderLength;
      }
   }

   $partitionTable = $disk->getPartitionTable();

   my $vmfsnumber;
   if (! $dumpOnly) {
      if (! isVolumeLabelUnique($host, $volumeLabel)) {
         $data .= error(qq{Create Typical Volumes: Volume label is not } .
                        qq{unique: $volumeLabel. No volumes were created.});
         return html($data);
      }


      # Create a VMFS partition ---------------------------------------------
      $number = shift(@availablePartNumbers);

      # Assume that we need a logical partition number.
      if (! defined($number)) {
         ($number) = generatePartitionNumberAndTypeCode($disk, "logical");
      }

      $cylinderLength = $endCylinder - $startCylinder + 1;
      $sectorLength = VMware::HConfig::RawDisk::fromChsToLba($cylinderLength,
                                                             0,
                                                             1,
                                                             $cylinders,
                                                             $heads,
                                                             $sectors);

      if (! $sectorLength) {
        # Translating from CHS to LBA failed.
        $data .= error("Could not write partition information to disk.");
        return html($data);
      }

      $partitionTable->addPart($number, $startCylinder, $endCylinder,
                               $gVMFSPartType, $sectorLength, 0);

      if (VMware::HConfig::RawDisk::isLogicalPart($number)) {
         if ($partitionTable->reorderLogicalParts() == 0) {
            # If partitions are reordered, we need to get the partition's new index
            my @parts = $partitionTable->lookupParts($startCylinder, $endCylinder, undef, undef);
            $number = $parts[0];
         }
      }

      $startCylinder += $cylinderLength;
      $vmfsnumber = $number;
   }

   # Create a dump partition ------------------------------------------------
   if ($q->param("dump") =~ /^true$/i) {
      # Give core dump partitions a number that is more likely to reflect their
      # position on disk.
      $number = pop(@availablePartNumbers);

      # Assume that we need a logical partition number.
      if (! defined($number)) {
         ($number) = generatePartitionNumberAndTypeCode($disk, "logical");
      }

      $cylinderLength = $dumpCylinderLength;
      $sectorLength = VMware::HConfig::RawDisk::fromChsToLba($cylinderLength,
                                                             0,
                                                             1,
                                                             $cylinders,
                                                             $heads,
                                                             $sectors);

      if (! $sectorLength) {
        # Translating from CHS to LBA failed.
        $data .= error("Could not write partition information to disk.");
        return html($data);
      }

      $partitionTable->addPart($number, $startCylinder,
                               $startCylinder + $cylinderLength - 1,
                               $gVmkDumpPartType, $sectorLength, 0);

      if (VMware::HConfig::RawDisk::isLogicalPart($number)) {
         if ($partitionTable->reorderLogicalParts() == 0) {
            # If partitions are reordered, we need to get the partition's new index
            my @parts = $partitionTable->lookupParts($startCylinder, $endCylinder, undef, undef);
            $number = $parts[0];
         }
      }

      $startCylinder += $cylinderLength;
   }

   # Write partition table.
   $data .= preDiskOperation($host, $DISK_SET_VMKDUMP);

   if (! $disk->makePartitions($partitionTable)) {
      $data .= error(qq{Could not write partition table: $!.});
      $data .= postDiskOperation($host, $DISK_SET_VMKDUMP);
      return html($data);
   }

   $data .= postDiskOperation($host, $DISK_SET_VMKDUMP);

   if (! $dumpOnly) {
      # Pick a block size that allows a single file to occupy the entire volume.
      my $fsSizeGB = $disk->getCylToKBytes($cylinderLength) / 1024 / 1024;
      my $blockSize = getVmfsBlockSize($fsSizeGB) . "m";

      # Format the VMFS volume.
      if (! $disk->createNewMFS($vmfsnumber, $blockSize, 2)) {
         $data .= error(qq{Could not format partition: $!.});
         return html($data);
      }

      # XXX: Need to invalidate cache so FS change is picked up.
      $host->{MFS} = undef;

      if (defined($volumeLabel)) {
         my $partName = VMware::HConfig::RawDisk::makeLogVmkPartName($disk->getVmkName(),
                                                                     $vmfsnumber);
         $data .= setVolumeLabel($host, $disk, $partName, $volumeLabel);
      }
   }

   return html($data);
}

# Pick a block size that allows a single file to occupy the entire partition.
sub getVmfsBlockSize($) {
   my($fsSizeMb) = @_;
   my $blockSize = 1;

   while ($blockSize * $gVMFSMaxKBlocks < $fsSizeMb) {
      $blockSize *= 2;

      if ($blockSize >= $gVMFSBlockSizeLimit) {
         $blockSize = $gVMFSBlockSizeLimit;
         last;
      }
   }
   return($blockSize);
}


# ---------------------------------------------------------------------------

sub writePartitionTable {
   my $host = shift();
   my $disk = shift();
   my $partitionTable = shift();

   if (! $disk->makePartitions($partitionTable)) {
      return error("Could not write partition information to disk.");
   }

   return "";
}


# ---------------------------------------------------------------------------

sub makeFileSystem {
   my $host = shift();
   my $disk = shift();
   my $part = shift();
   my $blockSize = shift();
   my $data = "";

   $part ||= $q->param("part");
   $blockSize ||= $q->param("bs");
   $blockSize ||= "1";
   if ($blockSize !~ /m$/i) {
      $blockSize .= "m";
   }

   my $version = $q->param("tn");
   # Assume that anything not VMFS-1 should be formatted as VMFS-2.
   if (! defined($version)) {
      $version = 2;
   } elsif ($version =~ /VMFS-1/i) {
      $version = 1;
   } else {
      $version = 2;
   }

   my ($ok, $name, $number) = isExistingPartitionName($disk, $part);

   if (0 && ! $ok) {
      $data .= error(qq{Could not format the requested partition "$part" } .
                     qq{because it does not exist or is not a valid } .
                     qq{partition name.});
      return $data;
   }

   my $hasSwap = partHasActiveSwapFile($host, $disk, $part);

   if ($hasSwap) {
      $data .= error(qq{Could not format the requested partition "$part" } .
                     qq{because it contains an active swap file.});
      return $data;
   }

   if (! $disk->createNewMFS($number, $blockSize, $version)) {
      $data .= error(qq{Could not format partition "$part": $!.});
   }

   return $data;
}


# ---------------------------------------------------------------------------

sub isVolumeLabelUnique {
   my $host = shift();
   my $volumeLabel = shift();

   if (! defined($volumeLabel) || $volumeLabel eq "") {
      return 1;
   }

   for (@{$host->getAllMFS()}) {
      my $label = $_->getLogName();
      VMware::Management::Util::trace("Volume label exists: $label.", "debug");

      if (defined($label) && $label ne "") {
         if ($label eq $volumeLabel) {
            return 0;
         }
      }
   }

   return 1;
}


# ---------------------------------------------------------------------------

sub setVolumeLabel {
   my $host = shift();
   my $disk = shift();
   my $part = shift();
   my $volumeLabel = shift();
   my $data = "";

   $part ||= $q->param("part");
   $volumeLabel ||= $q->param("vl");

   # Trim white space.
   $volumeLabel =~ s/^\s*(.*?)\s*$/$1/;

   if (! isVolumeLabelUnique($host, $volumeLabel)) {
      $data .= error(qq{Set Volume Label: Label is not unique: $volumeLabel.});
      return $data;
   }

   if ($volumeLabel =~ /(:|\/)/) {
      $data .= error(qq{Character ($1) cannot be used in a VMFS volume label.});
      return $data;
   }

   my $fs = $host->getMFS($part);
   if (! defined($fs)) {
      $data .= error(qq{Could not get a VMFS file system object for } .
                     qq{partition "$part".});
      return $data;
   }

   # Make sure that we don't create a duplicate volume label.
   if ($volumeLabel ne "") {
      my $existingVmfs = $host->getMFSByLogName($volumeLabel);
      if (defined($existingVmfs) && $existingVmfs != $fs) {
         my $existingName = $existingVmfs->getVmkName();
         $data .= error(qq{Could not label the requested partition "$part" } .
                        qq{because the requested label "$volumeLabel" is } .
                        qq{already used by partition "$existingName".});
         return $data;
      }
   }

   # Noop if the requested label is not different than the existing one.
   my $origName = $fs->getLogName();
   if ($volumeLabel ne $origName) {
      if (! $fs->setVMFSLogName($volumeLabel)) {
         $data .= error(qq{Could not label the requested partition "$part".});
      }
   }

   return $data;
}


# ---------------------------------------------------------------------------

sub setClusterAccess {
   my $host = shift();
   my $disk = shift();
   my $part = shift();
   my $clusterAccess = shift();
   my $data = "";

   $part ||= $q->param("part");
   $clusterAccess ||= $q->param("ca");

   my $fs = $host->getMFS($part);
   if (! defined($fs)) {
      $data .= error(qq{Could not get a VMFS file system object for } .
                     qq{partition "$part".});
      return $data;
   }

   # Noop if the requested access mode is not different
   # than the existing one. Fixes bug 30741
   my $origClusterAccess = $fs->getAccessibility();
   if ($origClusterAccess ne $clusterAccess &&
       ! $fs->setAccessibility($clusterAccess)) {
      $data .= error(qq{Could not change cluster access for } .
                     qq{partition "$part".});
      return $data;
   }

   return $data;
}


# ---------------------------------------------------------------------------

sub isExistingPartitionName {
   my $disk = shift();
   my $part = shift();
   my $ok = 1;

   my ($name, $number) = VMware::HConfig::RawDisk::unmakeLogVmkPartName($part);

   if (! defined($name) || ! defined($number)) {
      $ok = 0;
   } elsif (! isExistingPartitionNumber($disk, $number)) {
      $ok = 0;
   }

   return wantarray ? ($ok, $name, $number) : $ok;
}


# ---------------------------------------------------------------------------

sub isExistingPartitionNumber {
   my $disk = shift();
   my $number = shift();
   my $ok = 1;

   if (! defined($disk->getPart($number))) {
      $ok = 0;
   }

   return $ok;
}


# ---------------------------------------------------------------------------

sub isFreeSpace {
   my $partitionTable = shift();
   my $startCylinder = shift();
   my $endCylinder = shift();

   for ($partitionTable->getFreeCylinderRange()) {
      if ($startCylinder >= $_->{START_CYL} && $endCylinder <= $_->{END_CYL}) {
         return 1;
      }
   }
   return 0;
}


# ---------------------------------------------------------------------------

sub isLogicalFreeSpace {
   my $partitionTable = shift();
   my $startCylinder = shift();
   my $endCylinder = shift();

   for (grep($_->{LOGICAL} == 1, $partitionTable->getFreeCylinderRange())) {
      if ($startCylinder >= $_->{START_CYL} && $endCylinder <= $_->{END_CYL}) {
         return 1;
      }
   }
   return 0;
}


# ---------------------------------------------------------------------------

sub maybeCreateExtendedPartition {
   my $disk = shift();
   my $wantPartitions = shift();
   my $startCylinder = shift();
   my $endCylinder = shift();

   my @availablePartNumbers;
   my @existingPartNumbers = $disk->getPartNums(); # Assume the list is sorted.
   my %existingParts;
   for (@existingPartNumbers) {
      $existingParts{"$_"} = 1;
   }

   my $cylinders = $disk->getGeoCylinder() + 1;
   my $heads = $disk->getGeoHead();
   my $sectors = $disk->getGeoSector();

   my $partitionTable = $disk->getPartitionTable();

   my $isLogicalFreeSpace = isLogicalFreeSpace($partitionTable,
                                               $startCylinder,
                                               $endCylinder);

   my $extendedPartExists = 0; # Is there already an extended partition?
   my $createExtendedPart = 0; # Should we create an extended partition?

   # Can we create the necessary physical partitions? -----------------------
   if (! $isLogicalFreeSpace) {
      my @potentialPartNumbers = (1, 2, 3, 4);

      # Generate the list of available physical part numbers.
      for (@potentialPartNumbers) {
         if (! $existingParts{"$_"}) {
            push(@availablePartNumbers, $_);
         }
      }

      my $availablePartitions = scalar(@availablePartNumbers);
      if ($availablePartitions < $wantPartitions) {
         $createExtendedPart = 1;
      }
   }

   for (@existingPartNumbers) {
      if (VMware::HConfig::RawDisk::isPartExtended($disk->getPartType($_))) {
         $extendedPartExists = 1;
         last;
      }
   }

   # We need to, but can't, create an extended partition.
   if ($createExtendedPart && $extendedPartExists) {
      my $availablePartitions = scalar(@availablePartNumbers);
      my $str = $wantPartitions > 1 ? "partitions were" : "partition was";
      my $data = qq{The requested $wantPartitions $str } .
                 qq{not created because it is only possible to create } .
                 qq{$availablePartitions in the selected free space.};
      return (undef, error($data));
   }

   my ($number, $cylinderLength, $sectorLength);

   # Create an extended partition -------------------------------------------
   if ($createExtendedPart) {
      $number = shift(@availablePartNumbers);

      if (! defined($number)) {
         return (undef, error("Create Volume: Could not create extended " .
                              "partition: All four physical parts " .
                              "are occupied."));
      }

      $cylinderLength = $endCylinder - $startCylinder + 1;
      $sectorLength = VMware::HConfig::RawDisk::fromChsToLba($cylinderLength,
                                                             0,
                                                             1,
                                                             $cylinders,
                                                             $heads,
                                                             $sectors);
      if (! $sectorLength) {
        # Translating from CHS to LBA failed.
        return (undef, error("Could not write partition information to disk."));
      }

      $partitionTable->addPart($number, $startCylinder,
                               $endCylinder,
                               0x05, $sectorLength, 0);
   }

   return @availablePartNumbers;
}


# ---------------------------------------------------------------------------

sub addExtents {
   my $host = shift();
   my $data = "";

   my $partName = $q->param("part");

   for ($q->param()) {
      # Look for parameters beginning with "ex:".
      if (/^ex:(.*)$/i) {
         my $extentName = $1;

         if ($q->param($_) !~ /^1$/) {
            next;
         }

         my $fs = $host->getMFS($partName);
         if (! defined($fs)) {
            $data .= error(qq/Could not get VMFS file system object for / .
                           qq/partition name "$partName"./);
         }

         if (!$fs->attachExtent($extentName)) {
            $data .= error(qq/Could not span "$extentName" with "$partName"./);
         }
      }
   }

   return $data;
}


# ---------------------------------------------------------------------------

sub preDiskOperation($$) {
   my $host = shift();
   my $opt = shift();

   if (($opt & $DISK_SET_VMKDUMP) && !$host->unsetVmkDumpPartition()) {
      return error("Could not execute vmkdump to unset the dump partition.");
   }
   return "";
}


# ---------------------------------------------------------------------------

sub postDiskOperation($$) {
   my $host = shift();
   my $opt = shift();

   if (($opt & $DISK_SET_VMKDUMP) && !$host->setVmkDumpPartition()) {
      return error("Could not execute vmkdump to set the dump partition.");
   }
   return "";
}


# ---------------------------------------------------------------------------

sub error {
   my $msg = shift();
   my ($pkg, $fn, $ln) = caller();
   VMware::Management::Util::trace($msg, "error", undef, $pkg, $fn, $ln);
   return "err.push(" . jsStringLiteral($msg) . ");$/";
}


# ---------------------------------------------------------------------------

sub warning {
   my $msg = shift();
   my ($pkg, $fn, $ln) = caller();
   VMware::Management::Util::trace($msg, "warning", undef, $pkg, $fn, $ln);
   return "wrn.push(" . jsStringLiteral($msg) . ");$/";
}


# ---------------------------------------------------------------------------
#
# jsStringLiteral --
#
#      Make sure that $msg is safe for consumption in a JavaScript program.
#
# ---------------------------------------------------------------------------

sub jsStringLiteral {
   my $msg = shift();
   $msg =~ s/\\/\\\\/g;
   $msg =~ s/("|')/\\$1/g;
   my $lf = $/;
   $msg =~ s/$lf/\\n/g;

   return qq{"$msg"};
}


# ---------------------------------------------------------------------------

sub html {
   my $data = shift();
   my $cb = shift() || "if (parent.opCb) parent.opCb(self);";

   my @warnings = VMware::ExtHelpers::getWarnings();
   for (@warnings) {
      $data .= warning($_);
   }
   VMware::ExtHelpers::clearWarnings();

   my $html = qq[<html><head><title></title><script>$/] .
      qq[var err=[];var wrn=[];$/] .
      qq[$data$/</script></head><body onload="$cb"></body></html>];

   VMware::Management::Util::trace("$html", "debug");

   return($html,
          "Pragma"        => "no-cache",
          "Cache-Control" => "no-cache",
          "Content-Type"  => "text/html; charset=utf-8");
}

1;
