# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SS::SwapConfiguration;
# File: VMware/Management/SS/SwapConfiguration.pm

use strict;

use VMware::HConfig::SSHostRemote ();
use VMware::Management::SS::ESXConfig;
use VMware::Management::Util ();

use CGI();
my $q;


# ---------------------------------------------------------------------------

sub handler {
   $q = new CGI(shift());

   my %credentials = VMware::Management::Util::hash_credentials();
   my $user = $credentials{"l"};
   my $pass = $credentials{"m"};
   my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $pass);

   if (! defined($host)) {
      my $data = error("Could not create remote host object.");
      return html($data);
   }

   if (! $host->isVMKernel()) {
      my $data = error("Could not determine swap configuration: " .
                       "The VMkernel is not loaded.");
      return html($data);
   }

   for ($q->param()) {
      VMware::Management::Util::trace("::::: $_ = " . $q->param($_), "debug");
   }

   if ($q->param("op") =~ /^(activate)$/i) {
      return activate($host);
   }
   if ($q->param("op") =~ /^(save)$/i) {
      return save($host);
   }

   return data($host);
}


# ---------------------------------------------------------------------------

sub data {
   my $host = shift();
   my $data = shift() || "";
   my $configDB = $host->getConfigDB();

   # Retrieve active swap information, with sizes in megabytes.
   my ($actMb, $actFmb, $actUsed, $actUnrsvd, $actVl, $actFn) = $host->getSwapInfo();
   $actMb = $actMb / 1024;
   $actFmb = $actFmb / 1024;

   # Generate JavaScript code for the active swap file.
   $data .= "var actSwp=";
   if ($actMb > 0) {
      $data .= qq[{vl:"$actVl",fn:"$actFn",mb:$actMb,fmb:$actFmb}];
   } else {
      $data .= "null";
   }
   $data .= ";$/";

   # Retrieve saved swap configuration information.
   my $ap = $configDB->get("swapfile.enable");
   $ap ||= "";
   $ap = $ap =~ /^yes$/i ? "true" : "false";
   my $vl = $configDB->get("swapfile.volume");
   $vl ||= "";
   my $fn = $configDB->get("swapfile.filename");
   $fn ||= "";
   my $mb = $configDB->get("swapfile.sizeMB");
   $mb ||= "0";

   # Generate JavaScript code for the configured swap file.
   $data .= qq[var cfgSwp={vl:"$vl",fn:"$fn",mb:$mb,ap:$ap};$/];

   my $suggestedSize = int($host->getTotalMemSize() / 1024);

   $data .= qq[var sgdMb=$suggestedSize;$/];

   $data .= "function FsVol(id,vl,mb,fmb,bs,ca,fl){" .
      "this.id=id;" .       # vmhbaX:X:X:X
      "this.vl=vl;" .       # Volume label
      "this.mb=mb;" .       # Capacity in megabytes
      "this.fmb=fmb;" .     # Free space in megabytes
      "this.bs=bs;" .       # block size
      "this.ca=ca;" .       # Cluster access
      "this.fl=fl;}$/";     # List of files on disk

   $data .= "function File(fn,mb,rwx,uid,gid,att,lmod){" .
      "this.fn=fn;" .       # File name
      "this.mb=mb;" .       # File size in megabytes
      "this.rwx=rwx;" .     # Permissions as a string (e.g., /[rwx-]+/)
      "this.uid=uid;" .     # User ID
      "this.gid=gid;" .     # Group ID
      "this.att=att;" .     # Attribute (e.g., cow)
      "this.lmod=lmod;}$/"; # Last modified string (e.g., "Jan 23 14:43")

   my $fsData = "";
   for (@{$host->getAllMFS()}) {
      VMware::Management::Util::trace($_->getVmkName(), "debug");
      $fsData .= ($fsData ? ",$/" : "") . fsData($_);
   }

   $data .= "var vols=[$fsData];";

   return html($data);
}


# ---------------------------------------------------------------------------

sub fsData {
   my $fs = shift();

   my $id    = $fs->getVmkName();
   my $vl    = $fs->getLogName();
   my $mb    = $fs->getCapacity();
   my $fmb   = $fs->getFreeCapacity();
   my $bs    = $fs->getBlockSize();
   my $ca    = $fs->getAccessibility();
   my @files = grep($_->[0] =~ /^.*\.vswp$/, @{$fs->listFiles()});

   my $fl = "";
   for (@files) {
      $fl .= ($fl ? ",$/" : "") . fileData($_);
   }
   $fl = "[$fl]";

   return qq{new FsVol("$id","$vl",$mb,$fmb,$bs,"$ca",$fl)};
}


# ---------------------------------------------------------------------------

sub fileData {
   # The following items don't seem to have values in the case of a swap file:
   # permission, uid, gid, attribute and last modified string.
   my ($fn, $mb) = @{shift()};

   return qq{new File("$fn",$mb)};
}


# ---------------------------------------------------------------------------

sub activate {
   my $host = shift();
   my $data = "";
   my $cfgDb = $host->getConfigDB();

#    my $rc = 1;
#    VMware::Management::Util::trace("ACTIVATED", "debug");
   my $rc = $host->startSwapper($cfgDb->get("swapfile.volume"),
                                $cfgDb->get("swapfile.filename"),
                                $cfgDb->get("swapfile.sizeMB"));

   if ($rc == 0) {
      $data .= error("Failed to activate the configured swap file.");
   }

   return data($host, $data);
}


# ---------------------------------------------------------------------------

sub save {
   my $host = shift();
   my $data = "";
   my $cfgDb = $host->getConfigDB();

   # Assume that a blank file name indicates that the swap config should be
   # deepsixed.
   if (! $q->param("chgSwpFn")) {
      $cfgDb->set("swapfile.enable", "");
      $cfgDb->set("swapfile.volume", "");
      $cfgDb->set("swapfile.filename", "");
      $cfgDb->set("swapfile.sizeMB", "");

      if (! $host->commitConfigDB()) {
         $data .= error(qq{Swap configuration changes were not committed to } .
                        qq{the config database.});
      }

      return data($host, $data);
   }

   # Validate arguments
   if (! $q->param("chgSwpVl")) {
      $data .= error("A VMFS volume must be specified.");
      return data($host, $data);
   }

   if ($q->param("chgSwpFn") !~ /.vswp$/) {
      $data .= error("Swap file names must end in '.vswp'.");
      return data($host, $data);
   }

   if (($q->param("chgSwpMb") !~ /^\d+$/) || ($q->param("chgSwpMb") <= 0)) {
      $data .= error("The swap file capcity must be a positive integer.");
      return data($host, $data);
   }

   $cfgDb->set("swapfile.enable",
               $q->param("chgSwpAp") =~ /^true$/i ? "yes" : "no");
   $cfgDb->set("swapfile.volume", $q->param("chgSwpVl"));
   $cfgDb->set("swapfile.filename", $q->param("chgSwpFn"));
   $cfgDb->set("swapfile.sizeMB", $q->param("chgSwpMb"));

   if (! $host->commitConfigDB()) {
      $data .= error(qq{Swap configuration changes were not committed to } .
                     qq{the config database.});
   }

   return data($host, $data);
}


# ---------------------------------------------------------------------------

sub error {
   my $msg = shift();
   VMware::Management::Util::trace($msg, "error");
   return "err.push(" . jsStringLiteral($msg) . ");$/";
}


# ---------------------------------------------------------------------------

sub warning {
   my $msg = shift();
   VMware::Management::Util::trace($msg, "warning");
   return "wrn.push(" . jsStringLiteral($msg) . ");$/";
}


# ---------------------------------------------------------------------------
#
# jsStringLiteral --
#
#      Make sure that $msg is safe for consumption in a JavaScript program.
#
# ---------------------------------------------------------------------------

sub jsStringLiteral {
   my $msg = shift();
   $msg =~ s/\\/\\\\/g;
   $msg =~ s/("|')/\\$1/g;
   my $lf = $/;
   $msg =~ s/$lf/\\n/g;

   return qq{"$msg"};
}


# ---------------------------------------------------------------------------

sub html {
   my $data = shift();
   my $cb = shift() || "if (parent.opCb) parent.opCb(self);";

   my $html = qq[<html><head><title></title><script>$/] .
      qq[var err=new Array();var wrn=new Array();$/] .
      qq[$data$/</script></head><body onload="$cb"></body></html>];

   return($html,
          "Pragma"        => "no-cache",
          "Cache-Control" => "no-cache",
          "Content-Type"  => "text/html; charset=utf-8");
}

1;
