#!/usr/bin/perl
# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SS::VMUtil;
# File: VMware/Management/SS/VMUtil.pm

use strict;

use VMware::Control '1.00';
use VMware::Control::Server;
use VMware::Control::VM;
use VMware::Control::Profiler;
use VMware::Management::Util;
use VMware::DOMAccess;

my $p = VMware::Control::Profiler::getInstance("mui");


sub CheckAvailableRAM
{
    my $stub = $p->profile();
    my $memsize = shift ();

    my $total = &GetHostRAM;

    my $err = "";
    print STDERR "  CheckRAM: $total < $memsize \n";
    if($total < $memsize) {
        $err = qq (The amount of RAM you have selected ($memsize MB) is
		   greater than what is available on your system ($total MB).
		   );
    }

    return $err;
}

sub GetHostRAM
{
    my $stub = $p->profile();
    my $tmp = `free -m`;

    $tmp = (split(/\n/, $tmp))[1];
    my $total = (split(/\s+/, $tmp))[1];

    return $total;
}

sub WriteFile($$)
{
    my $stub = $p->profile();
    my ($file, $data) = @_;
    my $doc = VMware::DOMAccess->newXML("<exec><op></op><in></in></exec>");
    $doc->setValue("exec.op", "File_Create");
    $doc->setValue("exec.in.file", $file);
    $doc->setValue("exec.in.data", $data);
    $doc->setValue("exec.in.mode", "0754");
    my $ret = request($doc);
    $doc->dispose();

    $doc = VMware::DOMAccess->newXML($ret);
    my $result = $doc->getValue("exec.returncode");
    $doc->dispose();

    if($result == 1) {
        return "";
    } else {
        #TODO insert the correct error sting from XML here
        return "Error writing file $file";
    }
}

sub ReadFile($)
{
    my $stub = $p->profile();
    my ($file) = @_;
    my $doc = VMware::DOMAccess->newXML("<exec><op></op><in></in></exec>");
    $doc->setValue("exec.op", "File_Read");
    $doc->setValue("exec.in.file", $file);
    my $ret = request($doc);
    $doc->dispose();

    if (!$ret) {
       return (0, "Error retrieving file $file from server");
    }

    $doc = VMware::DOMAccess->newXML($ret);
    my $result = $doc->getValue("exec.returncode");
    my $data = $doc->getValue("exec.out.data");
    $doc->dispose();

    if ($result == 1) {
        return ($result, $data);
    } else {
        #TODO insert the correct error sting from XML here
        return ($result, "Error reading file $file");
    }
}

sub ReadLastN($$)
{
   my $stub = $p->profile();
   my ($file, $numLines) = @_;
   my $doc = VMware::DOMAccess->newXML("<exec><op></op><in></in></exec>");
   $doc->setValue("exec.op", "File_ReadLastN");
   $doc->setValue("exec.in.file", $file);
   $doc->setValue("exec.in.numLines", $numLines);
   my $ret = request($doc);
   $doc->dispose();

   if (!$ret) {
      return (0, "Error retrieving file $file from server");
   }

   $doc = VMware::DOMAccess->newXML($ret);
   my $result = $doc->getValue("exec.returncode");
   my $data = $doc->getValue("exec.out.data");
   $doc->dispose();

   if ($result == 1) {
      return ($result, $data);
   } else {
      #TODO insert the correct error sting from XML here
      return ($result, "Error reading file $file");
   }
}

#
# Utility routine for issuing ESXHostInfo queries
#    $op = Name of ESXHostInfo operation to execute
#    $tag = Root tag to be used for result XML tree
#
sub GetESXHostInfo
{
    my ($op, $tag, $argname, $argval) = @_;

    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "ESXHostInfo_$op");
    if (defined($argname) && defined($argval)) {
	$doc->setValue("exec.in." . $argname, $argval);
    }
    my $ret = request($doc);
    $doc->dispose();

    $doc = VMware::DOMAccess->newXML($ret);
    my $retDoc = VMware::DOMAccess->new($tag);

    foreach my $n ($doc->listElementNames("exec.out")) {
        my $s = $doc->getSubTree("exec.out.$n");
        my $xml = $s->getValue("");
        $s->dispose();

        $s = VMware::DOMAccess->newXML($xml);
        $retDoc->addSubTree($tag, $s);
        $s->dispose();
    }

    $doc->dispose();
    return $retDoc;
}

sub GetVmkNetAdapters
{
    my $stub = $p->profile();
    my $doc = GetESXHostInfo("GetVmkNetAdapters", "netAdapters", undef, undef);
    print STDERR $doc->prettyPrint() . "\n";
    return $doc;
}

sub GetVmkNetBondAdapters
{
    my $stub = $p->profile();
    my $doc = GetESXHostInfo("GetVmkNetBondAdapters", "netAdapters", undef, undef);
    print STDERR $doc->prettyPrint() . "\n";
    return $doc;
}

sub GetVmkScsiPassThrough
{
    my $stub = $p->profile();
    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "ESXHostInfo_GetVmkSCSIAdapters");
    my $ret = request($doc);
    $doc->dispose();

    $doc = VMware::DOMAccess->newXML($ret);
    my $retDoc = VMware::DOMAccess->new("scsiPassThrough");

    foreach my $n1 ($doc->listElementNames("exec.out")) {
        foreach my $n2 ($doc->listElementNames("exec.out.$n1.nondisks")) {
            my $s = $doc->getSubTree("exec.out.$n1.nondisks.$n2");
            my $xml = $s->getValue("");
            $s->dispose();

            $s = VMware::DOMAccess->newXML($xml);
            $retDoc->addSubTree("scsiPassThrough", $s);
            $s->dispose();
        }
    }

    print STDERR $doc->prettyPrint() . "\n";
    return $retDoc;
}

sub GetVMFS($) {
    my ($perm) = @_;

    my $stub = $p->profile();
    if (defined($perm) && ($perm ne "")) {
	return GetESXHostInfo("GetVMFS", "fs", "permission", $perm);
    }

    return GetESXHostInfo("GetVMFS", "fs", undef, undef);
}

sub GetMemory
{
    my $stub = $p->profile();
    return GetESXHostInfo("GetMemory", "memory", undef, undef);
}

sub GetLicense
{
    my $stub = $p->profile();
    return GetESXHostInfo("GetLicense", "license", undef, undef);
}

# Input format:
#   partitionname := the Vmk name of the partition on which the file resides
#   file := the name of the file on the VMFS to create
#   size := the size of the file in megatypes
#
# Output format:
#   No output args
#
sub CreateVMFSFile {
    my $stub = $p->profile();
    my $partition = shift();
    my $fileName = shift();
    my $size = shift();

    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "ESXHostEdit_CreateVMFSFile");
    $doc->setValue("exec.in.partitionname", $partition);
    $doc->setValue("exec.in.file", $fileName);
    $doc->setValue("exec.in.size", $size);
    my $ret = request($doc);
    $doc->dispose();

    $doc = VMware::DOMAccess->newXML($ret);
    my $result = $doc->getValue("exec.returncode");

    if ($result == 1) {
       $doc->dispose();
       return "";
    } else {
       my $msg = "";

       for ($doc->listElementNames("exec.messages")) {
          $msg .= $doc->getValue("exec.messages.$_");
       }

       $doc->dispose();
       return "$fileName: $msg";
    }
}

sub DiskExists
{
    my $stub = $p->profile();
    my $partition = shift();
    my $filename = shift();
    my $doc = GetVMFS(undef);

    foreach my $n ($doc->listElementNames("fs")) {
        my $p = $doc->getValue("fs.$n.partition");
        if($p eq $partition) {
            foreach my $fn ($doc->listElementNames("fs.$n.files")) {
		my $file = $doc->getValue("fs.$n.files.$fn.name");
		if($file eq $filename) {
		    $doc->dispose();
		    return 1;
		}
	    }
        }
    }

    $doc->dispose();
    return 0;
}

sub GetVolumeSize
{
    my $stub = $p->profile();
    my $partition = shift();
    my $doc = GetVMFS(undef);

    foreach my $n ($doc->listElementNames("fs")) {
	my $p = $doc->getValue("exec.out.$n.partition");
        if($p eq $partition) {
	    my $ret = $doc->getValue("fs.$n.free_size");
	    $doc->dispose();
            return $ret;
	}
    }

    $doc->dispose();
    return -1;
}

sub MakeConfigDirectory
{
    my $stub = $p->profile();
    my $dir = shift( );

    my @dirs = split(/\//, $dir);
    shift(@dirs);
    my $cdir = '';
    my $d;
    my $dirExists = 1;
    foreach $d (@dirs) {
        $cdir .= '/' . $d;
        if( -d $cdir ) {
            next;
        } else {
            $dirExists = 0;
            #exit;
        }
    }

    my $err;

    if($dirExists)
    {
	#$err = qq(The directory $dir already exists.  Specify another
	#	  directory or remove the existing directory and try again.
	#	  );
	#return $err;
	return "";
    }

    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "Directory_Make");
    $doc->setValue("exec.in.directory", $dir);
    my $ret = request($doc);
    $doc->dispose();

    $doc = VMware::DOMAccess->newXML($ret);
    my $result = $doc->getValue("exec.returncode");
    $doc->dispose();

    if($result == 1) {
	$err =  "";
    } else {
	#TODO insert the correct error sting from XML here
	$err = "Error creating config directory $dir";
    }

    $doc->dispose();
    return $err
}

sub request($) {
    my $stub = $p->profile();
    my $doc = shift;
    my $err;
    my %credentials = hash_credentials( );

    my $remote = 1;
    my $xml = $doc->getValue("");

    my $server = &VMware::Control::Server::new(undef, 0, $credentials{ "l" }, $credentials{ "m" } );

    if (!$server || !$server->connect())
    {
        LOG("Cannot connect\n");
    }

    my $ret;
    if( $remote ) {
        $ret = $server->exec($xml);
        if( !defined($ret) ) {
            my ($err, $errstr) = $server->get_last_error();
            LOG("\n ret: $ret \n VMControl error $err: $errstr\n");
        }

    } else {
        $ret = &VMware::VMServerd::handleExecRequest($xml);
    }

    #my $resp = VMware::DOMAccess->newXML($ret);

    return $ret;
}

sub Register
{
    my $stub = $p->profile();
    my $cfg = shift();

    my %credentials = hash_credentials( );
    my $vms = &VMware::Control::Server::new(
                                            undef,
                                            0,
                                            $credentials{ "l" },
                                            $credentials{ "m" } );

    if (!$vms || !$vms->connect())
    {
	my ($err, $errstr) = $vms->get_last_error();
	LOG("VMControl error $err: $errstr\n");
        return $errstr;
    }

    $vms->register($cfg);
    my ($err, $errstr) = $vms->get_last_error();
    if ($err)
    {
	LOG("\n VMControl error $err: $errstr\n");
	return $errstr;
    }

    return "";
}

sub LOG {
    my($msg) = @_;
    print STDERR "VMUtil: $msg";
}

1;

