#!/usr/bin/perl -w

#############################################################
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
#############################################################
#
# VMwareConfig.pm
# 

package VMware::Management::SS::VMwareConfig;

use strict;
use VMware::HConfig::SSHostRemote;
use VMware::HConfig::Constant qw(c_getFile
                                 $URL_PCIDIVY $URL_SECURITY 
                                 $URL_OVERVIEW $URL_LICENSE);
use VMware::ExtHelpers qw(&Warning &getWarnings &clearWarnings $gPanicCallback);
use VMware::Management::Util qw(hash_credentials);
use Carp qw(confess);
$SIG{__DIE__} = \&confess;
use CGI;

use VMware::Management::SS::ESXConfig;
my $htmlTitle = "Server Configuration";

# Global variables

my $URL_PREV = $URL_LICENSE;
my $URL_NAME_PREV = "License Information";
my $URL_NEXT = $URL_SECURITY;
my $URL_NAME_NEXT = "Security Configuration";

my $RCom = "request";
my $RCom_Config = "config";
my $RCom_ViewLog = "viewLog";

# 5 minute timeout when making the exec request to run the vmware-config.pl script
my $gTimeout = 300 * 1000;

# This script is no longer interactive -- use hard-coded values
my $gStartupServer = "yes";
my $gNetworking = "no";

my ($hostname, $port, $user, $password);

#####################################################################
# HTML producing functions                                          #
#####################################################################

sub PanicCallbackHTML(@) {
    my (@error) = @_;
    if ( $error[0] ) {
       print toHTMLPanic($htmlTitle, $error[0]);
    } else {
       print toHTMLPanic($htmlTitle, "");
    }
}

sub toErrorHTML($$) {
    my ($eSummary, $eDesc) = @_;
    return toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLError($eSummary, $eDesc);
}

sub toShowConfigHTML($$) {
    my ($cgi, $host) = @_;
    my $s = "";
    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");

    my $logfile = $cgi->param('logfile');

    if( !defined($logfile) ) {
	return(toErrorHTML("Showing Log File",
			   "Log file could not be found\n"))
    }

    my $log = read_file_remote($host, $logfile);
#    my $log = read_file($logfile);

    if( !defined($logfile) ) {
	return( toErrorHTML("Showing Log File",
			    "Could not find log file '$logfile'."));
    }

    my $title = "View Log File";
    my @path = @htmlPath;
    push @path, [ $htmlTitle, $url_self ];

    my $s = 
        toHTMLPath(\@path, $title) .
        toHTMLSystemInfo($host) .
        toHTMLWarnings();

    $s .=
       ("<h3>Review the log file output.</h3>\n" .
        "\n" .

	"<pre>$log</pre>\n"
	);

    if ( $wizard ) {
       $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                              [ $URL_NEXT, $URL_NAME_NEXT ] );
    }

    return($s);
}


sub toDoneConfigHTML($$) {
    my ($cgi, $host) = @_;
    my $s = "";
    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");

    $host->checkModules();

    my $message =
	"<p class='txt'>The configuration of VMware ESX Server\n" .
        "for this running kernel completed successfully.</p>\n\n";
    
    my $logfile = $cgi->param('logfile');
    my $logURL = "$url_self?$RCom=$RCom_ViewLog&logfile=$logfile";
    $logURL = toWizardURL($logURL, $wizard);
    if( defined($logfile) ) {
	$message .=
	    "<p class='txt'>To review the log file containing the results of the\n" . 
	    "configuration, <a href=\"$logURL\">click on this link.</a></p>\n";
    }
       
    $message .=
	"<p class='txt'>If you would like more advanced configuration options,\n" .  
        "you may run the script 'vmware-config.pl' at a Console OS command prompt</p>\n";

    my $title = "Configuration Done";
    my @path = @htmlPath;
    push @path, [ $htmlTitle, $url_self ];

    my $s = 
        toHTMLPath(\@path, $title) .
        toHTMLSystemInfo($host) .
        toHTMLWarnings() .
        $message;

    if ($wizard) {
       $s .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                              [ $URL_NEXT, $URL_NAME_NEXT ] );
    }
    return $s;
}


sub toErrorSharedHTML($$) {
    my ($cgi, $host) = @_;
    my $wizard = $cgi->param("wizard");

    my $url_pcidivy = toWizardURL($URL_PCIDIVY, $wizard);
    my $message = <<END_HTML;
<p class='txt'>
This step cannot be completed because the vmkernel must be unloaded in
order for configuration to take place.  The vmkernel is currently
running with shared SCSI adapters.  In shared SCSI mode, the vmkernel
cannot be unloaded.
</p>
<p class='txt'>
However, you have already completed this step and do not need to do it
again unless you wish to change some of the settings.  If you do wish
to reconfigure the system, you should <a href="$url_pcidivy">
reallocate your devices</a>, get rid of sharing, and reboot.
</p>
END_HTML

    if ($wizard) {
       $message .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                                    [ $URL_NEXT, $URL_NAME_NEXT ] );
    }

    return(toErrorHTML("Configuration of VMware ESX Server Already Done", $message));
}

sub toErrorNotConsoleHTML($$) {
    my ($cgi, $host) = @_;
    my $wizard = $cgi->param("wizard");

    my $url_pcidivy = toWizardURL($URL_PCIDIVY, $wizard);
    my $message = <<END_HTML;
<p class='txt'>
The Console OS must be running in order to proceed.  Please return
to the <a href="$url_pcidivy">Device Allocation page</a> and reboot
your machine into the Console OS.
</p>
END_HTML
    
    if ($wizard) {
        $message .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                                     [ $URL_NEXT, $URL_NAME_NEXT ] );
    }

    return(toErrorHTML("Configuration of VMware ESX Server Failure", $message));

}

sub toErrorRunningWorldsHTML($$$) {
    my ($cgi, $host, $running) = @_;
    my $wizard = $cgi->param("wizard");

    my $message = <<END_HTML;
<p class='txt'>
There are currently $running virtual machines running on this machine.
You must stop those virtual machines before performing this operation.
</p>
<p class='txt'>
The <a href="$URL_OVERVIEW">Overview page</a> can be used to shutdown
or suspend a virtual machine.
</p>
END_HTML

    if ($wizard) {
        $message .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                                     [ $URL_NEXT, $URL_NAME_NEXT ] );
    }
    return(toErrorHTML("Configuration of VMware ESX Server Failure", $message));
}

sub toErrorNoLicenseHTML($$) {
    my ($cgi, $host) = @_;
    my $wizard = $cgi->param("wizard");

    my $message = <<END_HTML;
<p class='txt'>
Server configuration cannot proceed because the server does not have a
valid license.  You must go to the <a href="$URL_LICENSE">License
Information page</a> and enter a valid license before proceeding with
server configuration.
</p>
END_HTML

    if ($wizard) {
        $message .= toHTMLNavWizard( [ $URL_PREV, $URL_NAME_PREV ],
                                     [ $URL_NEXT, $URL_NAME_NEXT ] );
    }
    return(toErrorHTML("Configuration of VMware ESX Server Failure", $message));
}

#####################################################################
# Function unique to this version                                   #
#####################################################################

sub RunVMwareConfig() {
    my $request = 'ESXHostEdit_RunVMwareConfig';
    my $in = VMware::HConfig::SSHostRemote::makeServerdRequest($request);
    my $rc;
    my $logfile = "";
    
    $in->setValue('.in.STARTUP_SERVER', $gStartupServer);
    $in->setValue('.in.NETWORKING', $gNetworking);

    my $out = VMware::HConfig::SSHostRemote::doServerdRequestTimeout($hostname, $port,
								     $user, $password,
								     $in, $gTimeout);
    if( !$out ) {
	Warning("Serverd request failed.  Internal error.");
	$in->dispose();
	return(0, undef, undef);
    }

    my $output = $out->getValue('.out.output');

    $rc = VMware::HConfig::SSHostRemote::getReturnCode($out);
    if( !$rc ) {
	&VMware::HConfig::SSHostRemote::ServerdWarning($out);
    } else {
	$logfile = $out->getValue('.out.logfile');
    }

    $in->dispose();
    $out->dispose();

    return($rc, $logfile, $output);
}

sub read_file_remote($) {
    my ($host, $filename) = @_;
    my $s = '';

    my $request = 'ESXHostEdit_GetTempFile';
    my $in = VMware::HConfig::SSHostRemote::makeServerdRequest($request);
    $in->setValue('.in.logfile', $filename);

    my $out = $host->sendServerdRequest($in);
    if( !$out ) {
	Warning("Serverd request failed.  Internal error.");
	$in->dispose();
	return(undef);
    }

    my $rc = VMware::HConfig::SSHostRemote::getReturnCode($out);
    if( !$rc ) {
	&VMware::HConfig::SSHostRemote::ServerdWarning($out);
    } else {
	$s = $out->getValue('.out.output');
    }

    return($s);
}

sub read_file($) {
    my ($filename) = @_;
    my $s = "";
    local *FD;

    if( !open(FD, '< ' . $filename) ) {
	# Error could not open file.
	print STDERR "Error in read_file: Could not open file $filename.  Reason: $!\n";
	return(undef);
    }

    while(<FD>) {
	$s .= $_;
    }

    close(FD);

    return($s);
}


sub saveConfig($$) {
    my ($cgi, $host) = @_;
    my $args = "";           # Arguments to the vmware-config.pl script 
    my $hiddens = "";        # Hidden input values to pass to the HTML document

    my($rc, $tmpfile, $output) = RunVMwareConfig();

    if(!$rc) {
        my $message = "The configuration script failed.\n";
	if(defined($output) && $output ne "") {
            $message .= "Please look at the output below to see where the script " .
                "failed.\n<pre>\n\n" . $output . "\n\n</pre>\n";
	} else {
            $message .= "The log of the configuration script is also unavailable.\n";
	}
        return toErrorHTML("Configuration Script Failed", $message);
    }

    $cgi->param(-name=>'logfile',-value=>"$tmpfile");
    return(toDoneConfigHTML($cgi, $host));
}


sub hasSharedSCSIAdapters($) {
    my ($host) = @_;

    my @scsi = @{ $host->getVMSCSIAdapters() };
    if( $#scsi > -1 ) {
	my $adapt;
	for $adapt (@scsi) {
	    if( $adapt->isShared() ) {
		return(1);
	    }
	}
    }

    return(0);
}
			  

sub handleRequest($$) {
    my ($cgi, $host) = @_;

    my $url_self = $cgi->url(-relative=>1);
    my $wizard = $cgi->param("wizard");
    
    my $rc = 1;

    # XXX Hack
    $gPanicCallback = \&PanicCallbackHTML;

    if (!$host->isVMnix()) {
        return toErrorNotConsoleHTML($cgi, $host);
    }

    my $numWorlds = scalar(@{ $host->getRunningVmmWorlds() });
    if ($numWorlds > 0) {
        return toErrorRunningWorldsHTML($cgi, $host, $numWorlds);
    }

    my $req = $cgi->param($RCom);

    if (defined($req) && $req eq $RCom_ViewLog) {
	return(toShowConfigHTML($cgi, $host));
    }

    if (hasSharedSCSIAdapters($host)) {
	return(toErrorSharedHTML($cgi, $host));
    }

    my ($canRun, $message) = $host->checkCanRunLicense();
    if (!$canRun) {
       return(toErrorNoLicenseHTML($cgi, $host));
    }

    return(saveConfig($cgi, $host));
}

sub handler {
    # Delete the PATH to get by the tainted security requirements of Perl
    delete($ENV{'PATH'}); 

    umask 022;   # Create files that are not writable by group
    
    &clearWarnings();

    my $cgi = new CGI( shift( ) );
    my %credentials = hash_credentials( );
    
    $hostname = undef;
    $port = 0;
    $user = $credentials{ "l" };
    $password = $credentials{ "m" };
    
    my $s = toHTMLHeader($htmlTitle, $user);

    if( $user ne 'root' ) {
	my $url_self = $cgi->url(-relative=>1);
        $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLNotRoot($url_self);
        goto footer;
    }
    
    my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $password);

    if(!defined($host)) {
        $s .= toHTMLPath(\@htmlPath, $htmlTitle) . toHTMLHostFailure($htmlTitle);
        goto footer;
    }

    $s .= handleRequest($cgi, $host);

footer:
    $s .= toHTMLFooter();
    return ($s, %MimeHeaders);
}

1;
