# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::ShowDir;
# File: VMware/Management/ShowDir.pm

use strict;

use VMware::Control '1.00';
use VMware::Control::Server ();
use VMware::Management::Util qw( trace
                                 string_to_file
                                 file_to_string
                                 connect_ctrl_obj
                                 append_error
                                 hash_credentials );

use URI::Escape;
use CGI ();
use XML::DOM ();
use VMware::DOMAccess ();
use XML::Sablotron ();
use MIME::Base64 ();

my $server;

my $doc;
my $q;

sub request($$) {
  my %credentials = hash_credentials( );

  if (! $credentials{"l"} || ! defined($credentials{"l"})) {
    trace("Username credential is not defined. Aborting.", "error");
    exit(1);
  }

  $server = &VMware::Control::Server::new(undef, 0,
                                          $credentials{ "l" },
                                          $credentials{ "m" } );

  if( ! $server || ! $server->connect() ) {
    my ($err, $errstr) = $server->get_last_error();
    print STDERR "$0: Could not connect to vmware-authd\n";
    print STDERR "  (VMControl error $err: $errstr)\n";
    exit(1);
  }

  if( !$server->is_connected() ) {
    print STDERR "Server not connected. Aborting.\n";
    exit(1);
  }

  my ($doc, $returntype) = @_;
  my $xml = $doc->getValue("");

  my $ret;
  $ret = $server->exec($xml);
  return undef unless defined( $ret );

  my $resp = VMware::DOMAccess->newXML($ret);

  $doc->dispose();
  $server->disconnect();

  if ($returntype eq "xmldoc") {
    return($resp); #caller dispose
  }

  if ($returntype eq "xmlstring") {
    $ret = $resp->getValue("");
  } elsif ($returntype eq "xmlstringpretty") {
    $ret = $resp->prettyPrint();
  } else {
    $ret = "<ERROR_NOT_REACHED/>";
  }
  $resp->dispose();
  return($ret);
}

sub processResult {
  my ($xmlstring) = @_;
  my $processor = new XML::Sablotron();
  my $xsl = $ENV{ "style_DEFAULT" };

  $processor->SetBase("file://".$ENV{"vmware_BASEHREF"});
  $processor->RunProcessor("arg:/template", "arg:/data", "arg:/result",
                           undef,
                           ["template", file_to_string($xsl), "data", $xmlstring]);
  my $returnhtml = $processor->GetResultArg("result");
  return($returnhtml,
         "Pragma"        => "no-cache",
         "Cache-Control" => "no-cache",
         "Content-Type"  => "text/html; charset=iso-8859-1");
}

# ----------------------------------------------------------------
# newdir --
#
# inputs:
# - CGI object $q with params:
#   - newdir : A full or relative path to the new directory
#   - destdir: The root directory for relative paths
# ----------------------------------------------------------------
sub newdir ($)  {
  my ($q) = @_;
  my $dir = $q->param("newdir");
  if (index($dir, "/") != 0) {
    $dir = $q->param("destdir")."/".$dir;
  }
  my $returnhtml = "";

  my $op = "FileManager_MakeDir";

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", $op);
  $doc->setValue("exec.in", "");
  $doc->setValue("exec.in.directories", "");
  $doc->setValue("exec.in.directories.directory", $dir);

  if (&xmlDebugLevel()) {
    $returnhtml = &dumpXML($doc);
    $doc->dispose( );
    return (
            $returnhtml,
            "Pragma"        => "no-cache",
            "Cache-Control" => "no-cache",
            "Content-Type"  => "text/html; charset=iso-8859-1"
            );
  } else {
    $doc = request($doc, "xmldoc");

    my $rc = defined($doc) ? $doc->getValue("exec.returncode") : 0;

    if (! $rc) {
      if (defined($doc)) {
        if ($ENV{vmware_DEBUG_LEVEL} > 2) {
          string_to_file($doc->getValue(""), &VMware::Management::Util::tmpdir() . "makedir_err.xml", 1);
        }

        my $docstring = $doc->getValue("");
        $doc->dispose();
        return processResult($docstring);
      }
    }

    $doc->dispose();
    return ("", "Location" => "/showdir?op=ls&dir=" . safeFileName($dir . "/.."));
  }
}

sub xmlDebugLevel() {
  return 2 if (-e &VMware::Management::Util::tmpdir() . "xmldbg");
  return 1 if (-e &VMware::Management::Util::tmpdir() . "xmldbg_inonly");
  return 0;
}

sub dumpXML() {
  my ($doc) = @_;
  my $returnhtml = "";

  if (xmlDebugLevel() == 1) {
    $returnhtml =  "<font color=red>input xml : temporary for debugging only</font><pre>" .
      CGI::escapeHTML($doc->prettyPrint())."</pre>";
  } elsif (xmlDebugLevel() == 2) {
    $returnhtml =  "<font color=red>input xml : temporary for debugging only</font><pre>" .
      CGI::escapeHTML($doc->prettyPrint())."</pre>" .
      "<font color=red>output xml : temporary for debugging only</font><pre>".CGI::escapeHTML(request($doc, "xmlstringpretty"))."</pre>";
  }

  return $returnhtml;
}

sub append_homedir {
  my ($thedoc, $path) = @_;

  my $op = "FileManager_GetHomeDir";
  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", $op);
  $doc->setValue("exec.in", "");

  my $retxml = request($doc, "xmlstring");
  $doc->dispose( );
  return undef unless defined($retxml);

  my $do = VMware::DOMAccess->newXML($retxml);
  return undef unless defined( $do );

  my $rc = $do->getValue( "exec.returncode" );
  if ( $rc == 1 && $do->getValue( "exec.out" ) ) {
    $thedoc->setValue($path, $do->getValue("exec.out.directory"));
  } else {
    trace( "exec.returncode is $rc", "error" );
    trace( $do->getDocument( )->toString( ), "error" );
  }
  $do->dispose( );
  return 1;
}

sub errorMsg($) {
  my ($msg) = @_;
  my $returnhtml = "<html><body><font color=red><b>" .
    CGI::escapeHTML($msg) . "</b></font></body></html>";
  return (
    $returnhtml,
    "Pragma"        => "no-cache",
    "Cache-Control" => "no-cache",
    "Content-Type"  => "text/html; charset=iso-8859-1"
  );
}

# ----------------------------------------------------------------
# showdir --
#
# inputs:
# - CGI object $q with params:
#   - dir : full path to directory to display
# ----------------------------------------------------------------
sub showdir(1$) {
  my ($q) = @_;
  my $dir = $q->param("dir");
  if ($q->param("destdir") && index($dir, "/") != 0) {
    $dir = $q->param("destdir")."/".$dir;
  }

  my $returnhtml = "";
  my $op = "FileManager_ListDir";

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", $op);
  $doc->setValue("exec.in", "");
  if ($dir ne "") {
    $doc->setValue("exec.in.directory", $dir);
  } else {
    unless (&append_homedir($doc, "exec.in.directory")) {
      return errorMsg("Cannot find user directory");
    }
  }

  $doc = request($doc, "xmldoc");

  if ($ENV{vmware_DEBUG_LEVEL} > 2) {
    string_to_file($doc->getValue(""), &VMware::Management::Util::tmpdir() . "showdir.xml", 1);
  }

  my $docstring = $doc->getValue("");
  $doc->dispose();
  return processResult($docstring);
}

sub returnFrameset() {
  my $dir = $q->param("dir") || "";
  my $hostname = `hostname`;
  my %credentials = hash_credentials();
  my $username = $credentials{l};
  my $returnhtml .= <<end;
<html>
<head>
  <title>VMware MUI File Manager for $username\@$hostname</title>
  <script src="/vmware/themes/default/file_manager.js"></script>
</head>
<frameset cols="200,*">
<frame name="tree" src="/vmware/blank.html">
<frameset name="fs" rows="84,*,42,0" frameborder="0" border="0">
<frame name="pwd" src="/vmware/pwd.html" border="0" scrolling="no" marginheight="0" marginwidth="0"/>
<frame name="showdir" src="/showdir?op=ls&dir=$dir">
<frame name="glob" src="/vmware/glob.html" border="0" scrolling="no" marginheight="0" marginwidth="0"/>
<framset cols="640,*" frameborder="0" border="0">
<frame name="hidden" src="/showdir?op=hide" border="0" scrolling="no" marginheight="0" marginwidth="0"/>
<frame name="properties" src="/showdir?op=properties_form" border="0" scrolling="no" marginheight="0" marginwidth="0"/>
</framset>
</frameset>
</frameset>

</html>
end

  return ($returnhtml,
          "Content-Type"  => "text/html; charset=iso-8859-1");
}

sub init_properties {
  my $q = shift();
  my @files = split(/\|/, $q->param("files"));
  my @dirs = split(/\|/, $q->param("dirs"));

  # Count the number of files, occurrences of each permission bit and
  # occurrences of arbitrary user and group names. If the permission bit count
  # equals the number of files, the user count is 1 or the group count is 1,
  # the form presented to the user should specify the appropriate value (since
  # it applies to all the files). Otherwise, the value should be left
  # unspecified. The user can specify a value, which will make the bit, user
  # or group homogeneous for the set of files, or not, which will preserve the
  # orginal heterogeneous values.
  my $file_count = 0;
  my %rwx_count = ( 'ur' => 0, 'uw' => 0, 'ux' => 0,
                    'gr' => 0, 'gw' => 0, 'gx' => 0,
                    'or' => 0, 'ow' => 0, 'ox' => 0 );
  my %user_count;
  my %group_count;

  # Unify the lists since both require the exact same processing.
  for my $dir (@dirs) {
    push(@files, $dir);
  }

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.out.cwd", $q->param("destdir"));
  my %credentials = hash_credentials();
  $doc->setValue("exec.out.username", $credentials{"l"});

  for my $file (@files) {
    my @data = split(';', $file);
    $file = shift(@data);

    my $d = VMware::DOMAccess->new("exec");
    $d->setValue("exec.op", "FileManager_FileExists");
    $d->setValue("exec.in.file", $file);
    $d = request($d, "xmldoc");

    my $rc = defined($d) ? $d->getValue("exec.returncode") : 0;

    if (! $rc) {
      if (defined($d)) {
        $doc->setValue("exec.out.file.msg",
                       $d->getValue("exec.out.messages.msg"));
      }
    } else {
      # Don't emit the file unless it exists and there was no error stating it.
      if ($d->getValue("exec.out.exists") == 1) {
        my $name = substr($file, rindex($file, '/') + 1);

        $doc->setValue("exec.out.file[$file_count].name", $name);

        my %rwx;

        my $user = $d->getValue("exec.out.user");
        $rwx{'ur'} = $d->getValue("exec.out.read-user");
        $rwx{'uw'} = $d->getValue("exec.out.write-user"),
        $rwx{'ux'} = $d->getValue("exec.out.execute-user");

        $doc->setValue("exec.out.file[$file_count].user", $user);
        $doc->setValue("exec.out.file[$file_count].ur", $rwx{'ur'});
        $doc->setValue("exec.out.file[$file_count].uw", $rwx{'uw'});
        $doc->setValue("exec.out.file[$file_count].ux", $rwx{'ux'});

        my $group = $d->getValue("exec.out.group");
        $rwx{'gr'} = $d->getValue("exec.out.read-group");
        $rwx{'gw'} = $d->getValue("exec.out.write-group"),
        $rwx{'gx'} = $d->getValue("exec.out.execute-group");

        $doc->setValue("exec.out.file[$file_count].group", $group);
        $doc->setValue("exec.out.file[$file_count].gr", $rwx{'gr'});
        $doc->setValue("exec.out.file[$file_count].gw", $rwx{'gw'});
        $doc->setValue("exec.out.file[$file_count].gx", $rwx{'gx'});

        $rwx{'or'} = $d->getValue("exec.out.read-other");
        $rwx{'ow'} = $d->getValue("exec.out.write-other"),
        $rwx{'ox'} = $d->getValue("exec.out.execute-other");

        $doc->setValue("exec.out.file[$file_count].or", $rwx{'or'});
        $doc->setValue("exec.out.file[$file_count].ow", $rwx{'ow'});
        $doc->setValue("exec.out.file[$file_count].ox", $rwx{'ox'});

        for my $key (keys(%rwx_count)) {
          if ($rwx{$key}) {
            $rwx_count{$key}++;
          }
        }

        $user_count{$user} = 1;
        $group_count{$group} = 1;

        $file_count++;
      }
    }

    if (defined($d)) {
      $d->dispose();
    }
  }

  for my $key (keys(%rwx_count)) {
    if ($rwx_count{$key} == $file_count) {
      $doc->setValue("exec.out.$key", 1);
    } elsif ($rwx_count{$key} > 0) {
      $doc->setValue("exec.out.$key", -1);
    } else {
      $doc->setValue("exec.out.$key", 0);
    }
  }

  if (scalar(my @keys = keys(%user_count)) == 1) {
    $doc->setValue("exec.out.user", $keys[0]);
  }

  if (scalar(my @keys = keys(%group_count)) == 1) {
    $doc->setValue("exec.out.group", $keys[0]);
  }

  if ($ENV{vmware_DEBUG_LEVEL} > 2) {
    string_to_file($doc->getValue(""), &VMware::Management::Util::tmpdir() . "init_properties.xml", 1);
  }

  my $docstring = $doc->getValue("");
  $doc->dispose();
  return processResult($docstring);
}

sub edit_properties {
  my $q = shift();
  my @files = split(/\|/, $q->param("files"));
  my $tid;
  my $cwd = $q->param("cwd");
  my $user = $q->param("user");
  my $group = $q->param("group");
  my $name = $q->param("name");

  my $ur = $q->param("ur") ? $q->param("ur") : 0;
  my $uw = $q->param("uw") ? $q->param("uw") : 0;
  my $ux = $q->param("ux") ? $q->param("ux") : 0;

  my $gr = $q->param("gr") ? $q->param("gr") : 0;
  my $gw = $q->param("gw") ? $q->param("gw") : 0;
  my $gx = $q->param("gx") ? $q->param("gx") : 0;

  my $or = $q->param("or") ? $q->param("or") : 0;
  my $ow = $q->param("ow") ? $q->param("ow") : 0;
  my $ox = $q->param("ox") ? $q->param("ox") : 0;

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", "FileManager_ChangeAttribs");

  for my $i (0 .. $#files) {
    $doc->setValue("exec.in.files.file[$i]", "$cwd/$files[$i]");
  }

  $doc->setValue("exec.in.read-user", $ur);
  $doc->setValue("exec.in.write-user", $uw);
  $doc->setValue("exec.in.execute-user", $ux);

  $doc->setValue("exec.in.read-group", $gr);
  $doc->setValue("exec.in.write-group", $gw);
  $doc->setValue("exec.in.execute-group", $gx);

  $doc->setValue("exec.in.read-other", $or);
  $doc->setValue("exec.in.write-other", $ow);
  $doc->setValue("exec.in.execute-other", $ox);

  if ($user ne '') {
    $doc->setValue("exec.in.owner", $user);
  }

  if ($group ne '') {
    $doc->setValue("exec.in.group", $group);
  }

  $doc = request($doc, "xmldoc");

  my $rc = defined($doc) ? $doc->getValue("exec.returncode") : 0;

  if (! $rc) {
    if (defined($doc)) {
      if ($ENV{vmware_DEBUG_LEVEL} > 2) {
        string_to_file($doc->getValue(""), &VMware::Management::Util::tmpdir() . "edit_properties.xml", 1);
      }

      my $docstring = $doc->getValue("");
      $doc->dispose();
      return processResult($docstring);
    }
  }

  $doc->dispose();

  my $loc = "/showdir?op=ls&dir=" . safeFileName($cwd);

  if (scalar(@files) == 1 &&
      defined($name) &&
      $name ne '' &&
      $files[0] ne $name) {
    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "FileManager_FileMove");

    my $src_file = ( ((index($files[0], "/") == 0) ||
		      (index($files[0], "~") == 0)) ? '' : "$cwd/") . $files[0];
    my $dest_file = ( ((index($name, "/") == 0) ||
		       (index($name, "~") == 0)) ? '' : "$cwd/") . $name;

    $doc->setValue("exec.in.files.file[0]", "$src_file");
    $doc->setValue("exec.in.files.file[1]", "$dest_file");
    $doc = request($doc, "xmldoc");

    my $rc = defined($doc) ? $doc->getValue("exec.returncode") : 0;

    if (! $rc) {
      if (defined($doc)) {
        if ($ENV{vmware_DEBUG_LEVEL} > 2) {
          string_to_file($doc->getValue(""), &VMware::Management::Util::tmpdir() . "edit_properties.xml", 1);
        }

        my $docstring = $doc->getValue("");
        $doc->dispose();
        return processResult($docstring);
      }
    } else {
      $tid = $doc->getValue("exec.out.taskid");
    }

    if ($tid) {
      $loc = "/lro?op=init_task&taskid=". $tid .
        "&dir=" . safeFileName($cwd) . "&lro=mv";
    }

    $doc->dispose();
  }

  return ("", "Location" => $loc);
}

sub return_properties_form {
  my $returnhtml = <<end;
<html>
<body>
<form name="form" method="POST" action="/showdir" target="showdir">
<input type="text" name="files" size="80"/> files
<input type="text" name="dirs" size="80"/> directories
<input type="text" name="op" value="init_properties" size="80"/> operation
</form>
</body>
</html>
end
}

sub returnHiddenForm() {
  my $returnhtml = <<end;
<html>
<body onLoad="if (parent.showdir.doOnLoad) parent.showdir.doOnLoad();">
<form name="hiddenform" method="POST" action="/showdir" target="showdir">
<input type="text" name="destdir" size="80"/> destination
<input type="text" name="lastdir" size="80"/> last directory
<input type="text" name="dest_fs" size="80"/> destination fs
<input type="text" name="src_fs" size="80"/> source fs
<input type="text" name="files" size="80"/> files
<input type="text" name="dirs" size="80"/> directories
<input type="text" name="op" size="80"/> operation
<input type="text" name="newdir" size="80"/> mkdir
<input type="hidden" name="fmt" value="outhtml"/>
</form>
</body>
</html>
end

  return (
    $returnhtml,
    "Pragma"        => "no-cache",
    "Cache-Control" => "no-cache",
    "Content-Type"  => "text/html; charset=iso-8859-1"
  );
}

sub deleteFiles {
  my ($files) = @_;
  my @arpaths = split(/\|/, $files);

  my $returnhtml = undef;

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", "FileManager_FileRemove");
  $doc->setValue("exec.in", "");
  $doc->setValue("exec.in.force", "");
  $doc->setValue("exec.in.recursive", "1");

  $doc->setValue("exec.in.files", "");

  foreach my $path (@arpaths) {
    my $d = VMware::DOMAccess->new("file");
    my @data = split(';', $path);
    $d->setValue(".", $data[0]);
    $doc->addSubTree("exec.in.files", $d);
    $d->dispose();
  }

  my $rc = 0;
  $doc = request($doc, "xmldoc");
  $rc = $doc->getValue("exec.returncode");
  if (! $rc) {
      trace( $doc->prettyPrint( ), "error" );

      my $processor = new XML::Sablotron();
      my $xsl = $ENV{ "style_DEFAULT" };

      $processor->SetBase("file://".$ENV{"vmware_BASEHREF"});
      $processor->RunProcessor("arg:/template", "arg:/data", "arg:/result",
                               undef,
                               ["template", file_to_string($xsl), "data", $doc->getValue("")]);
      $returnhtml = $processor->GetResultArg("result");
  }

  $doc->dispose( );
  return $returnhtml;
}

sub deleteDirs {
  my ($dirs) = @_;
  my @arpaths = split(/\|/, $dirs);

  my $returnhtml = undef;

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", "FileManager_RemoveDir");
  $doc->setValue("exec.in", "");
  $doc->setValue("exec.in.force", "");
  $doc->setValue("exec.in.recursive", "");

  $doc->setValue("exec.in.directories", "");

  foreach my $path (@arpaths) {
    my $d = VMware::DOMAccess->new("directory");
    my @data = split(';', $path);
    $d->setValue(".", $data[0]);
    $doc->addSubTree("exec.in.directories", $d);
    $d->dispose();
  }

  my $rc = 0;
  $doc = request($doc, "xmldoc");
  $rc = $doc->getValue("exec.returncode");
  if (! $rc) {
      trace( $doc->prettyPrint( ), "error" );

      my $processor = new XML::Sablotron();
      my $xsl = $ENV{ "style_DEFAULT" };

      $processor->SetBase("file://".$ENV{"vmware_BASEHREF"});
      $processor->RunProcessor("arg:/template", "arg:/data", "arg:/result",
                               undef,
                               ["template", file_to_string($xsl), "data", $doc->getValue("")]);
      $returnhtml = $processor->GetResultArg("result");
  }

  $doc->dispose( );
  return $returnhtml;
}

# ----------------------------------------------------------------
# delete --
#   deletes files and directories from the current directory
#
# inputs:
# - CGI object $q with params:
#   - files   : "|" separated list of relative file name
#   - dirs    : "|" separated list of relative directory names
# ----------------------------------------------------------------
sub delete {
  my ($q) = @_;
  my $dir = $q->param("destdir");
  my $files = $q->param("files");
  my $dirs = $q->param("dirs");
  my $returnhtml;

  if ($files) {
      $returnhtml = &deleteFiles($files);
  }

  if ($dirs && !$returnhtml) {
      $returnhtml = &deleteDirs($dirs);
  }

  if ($returnhtml) {
      # error occurred
      return($returnhtml,
             "Pragma"        => "no-cache",
             "Cache-Control" => "no-cache",
             "Content-Type"  => "text/html; charset=iso-8859-1");
  } else {
      return ("", "Location" => "/showdir?op=ls&dir=" . safeFileName($dir));
  }
}

sub parse_file_metadata {
  my %hash;
  for my $kv (@_) {
    my @split = split('=', $kv);
    $hash{$split[0]} = $split[1];
  }
  return %hash;
}

# ---------------------------------------------------------------------------
# in:
#
#      $file ... The file being considered for import
#         @_ ... A list of files approved for import
#
# out:
#
#      1 if $file is also found in @_. 0 otherwise.
# ---------------------------------------------------------------------------
sub import_confirmed {
  my $file = shift();
  for my $f (@_) {
    if ($file eq $f) {
      return 1;
    }
  }
  return 0;
}

# ----------------------------------------------------------------
# copy_or_move --
#
# inputs:
# - CGI object $q with params:
#   - destdir : full path of destination directory of operation
#   - dest_fs : type is one of vmfs or cons
#   - src_fs  : type is one of vmfs or cons
#   - files   : "|" separated list of absolute file name
#   - dirs    : "|" separated list of absolute directory names
# - op : specify if it is a move or copy
#   ("FileManager_FileCopy", "FileManager_FileMove")
# ----------------------------------------------------------------
sub copy_or_move {
  my ($q, $op) = @_;
  my $dir = $q->param("destdir");
  my $files = $q->param("files");
  my $dirs = $q->param("dirs");
  my @tasks;

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", $op);
  $doc->setValue("exec.in", "");
  $doc->setValue("exec.in.recursive", "");
  $doc->setValue("exec.in.files", "");
  my @arfiles = split(/\|/, $files);
  my @ardirs = split(/\|/, $dirs);
  VMware::Management::Util::trace($q->param("confirmed_files"), "debug");
  my @confirmed = split(/\|/, $q->param("confirmed_files"));
  my $ie_count = 0;
  my $execerror = 0;

  foreach my $file (@arfiles) {
    my @data = split(';', $file);
    $file = shift(@data);

    my %meta = parse_file_metadata(@data);

    trace("File Type: $meta{type}", "debug");
    if ($meta{type} =~ /^vmdk/) {
      my $dest_fs = $q->param("dest_fs");
      my $src_fs = $q->param("src_fs");
      if ($dest_fs ne $src_fs || $q->param("op") eq "convert") {
        my $dele_src = $op eq "FileManager_FileMove" ? 1 : 0;
        my $tid;
        my $resultstr;
        my $ie = 0;

        if ($dest_fs eq 'vmfs') {
          # import
          if (($meta{type} =~ /^vmdk$/ &&
              scalar(@confirmed) &&
              import_confirmed($file, @confirmed)) ||
              $q->param("op") eq "convert") {
            ($tid, $resultstr) = &import_or_export('ESXHostEdit_ImportFromConsoleVMFSFile',
                                                $file, $dir, $dele_src, $src_fs);

            $ie = 1;
          }
        } else {
          # export
          if ($meta{type} =~ /^vmdk-vmfs$/) {
            ($tid, $resultstr) = &import_or_export('ESXHostEdit_ExportToConsoleVMFSFile',
                                                $file, $dir, $dele_src, $src_fs);
            $ie = 1;
          }
        }

        if ($ie) {
          if ($tid) {
            push(@tasks, $tid);
          } else {
            if (scalar(@tasks)) {
              $execerror = 1;
            } else {
              return processResult($resultstr);
            }
          }
          $ie_count++;
          next;
        }
      }
    }

    my $d = VMware::DOMAccess->new("file");
    $d->setValue(".", $file);
    $doc->addSubTree("exec.in.files", $d);
    $d->dispose();
  }

  foreach my $dir (@ardirs) {
    my $d = VMware::DOMAccess->new("file");
    my @data = split(';', $dir);
    $d->setValue(".", $data[0]);
    $doc->addSubTree("exec.in.files", $d);
    $d->dispose();
  }

  # Process any cp or mv operations that are not imports or exports.
  if ($ie_count < (scalar(@arfiles) + scalar(@ardirs))) {
    my $rc = 0;
    # add destination dir as final "file" argument
    my $d = VMware::DOMAccess->new("file");
    $d->setValue(".", $dir);
    $doc->addSubTree("exec.in.files", $d);
    $d->dispose();

    $doc = request($doc, "xmldoc");
    $rc = $doc->getValue("exec.returncode");


    if (! $rc) {
      trace($doc->getValue("exec.out.messages.msg"), "error");
      my $docstring = $doc->getValue("");
      $doc->dispose();
      return processResult($docstring);
    } else {
      trace($doc->prettyPrint(), "debug");
      my $tid = $doc->getValue("exec.out.taskid");
      if ($tid) {
        push(@tasks, $tid);
      }
    }
  }

  my $loc = "/showdir?op=ls&dir=" . safeFileName($dir);

  # If any operations returned LRO Task IDs, change the $loc appropriately.
  if (scalar(@tasks)) {
    $loc = "/lro?op=init_task&taskid=". join(',', @tasks) ."&dir=" . safeFileName($dir);
    my $lro = $op eq "FileManager_FileCopy" ? "cp" : "mv";
    if ($q->param("op") eq "convert") {
      $lro = "convert";
    }
    $loc .= "&lro=$lro";
    $loc .= $execerror ? "&execerror=1" : "";
  }
  $doc->dispose();

  return ("", "Location" => $loc);
}


# ---------------------------------------------------------------------------
# import_or_export
#
# in:
#
#          $op ... One of ESXHostEdit_<ExportTo|ImportFrom>ConsoleVMFSFile
#         $src ... The source file name
#        $dest ... The destination directory
#    $dele_src ... Should the source file be deleted after import/export
#      $src_fs ... One of vmfs or cons
#
# out:
#
#    out[0] ... The operation task ID if successful, 0 otherwise
#    out[1] ... undef if successful, string of the xml result doc  otherwise
# ---------------------------------------------------------------------------
sub import_or_export {
  my ($op, $src, $dest, $dele_src, $src_fs) = @_;

  my $file = substr($src, rindex($src, '/') + 1);
  my $vmk_name;

  if ($src_fs eq "vmfs") {
    # Extract vmk_name from /path-including/vmk_name/to-file
    trace("Extract vmk_name from /path-including/vmk_name/to-file", "debug");
    # XXX Only valid when looking at the the /vmfs directory.
    $src =~ m!^/.*?/(.*?)/.*?$!;
    $vmk_name = $1;
  } else {
    # Extract vmk_name from /path-to/vmk_name
    trace("Extract vmk_name from /path-to/vmk_name", "debug");
    $vmk_name = substr($dest, rindex($dest, '/') + 1);
  }

  trace(" vmk: $vmk_name", "debug");
  trace(" src: " . ($src_fs eq "vmfs" ? "$file" : "$src"), "debug");
  trace("dest: " . ($src_fs eq "vmfs" ? "$dest/" : "") . $file, "debug");

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", $op);
  $doc->setValue("exec.in.partitionname", $vmk_name);
  # If the source is on a vmfs parition, provide only the file name.
  # Otherwise, provide the complete path.
  $doc->setValue("exec.in.src", ($src_fs eq "vmfs" &&
                                 $op ne "ESXHostEdit_ImportFromConsoleVMFSFile"
                                 ? "$file" : "$src"));
  # XXX Destination file renaming should really be handled at a lower level.
  if ($src_fs eq "vmfs" &&
      $op eq "ESXHostEdit_ImportFromConsoleVMFSFile" &&
      $src eq "$dest/$file") {
    $file = "VMFS_copy_of_$file";
  }
  # If the destination is the console, provide the complete path.
  # Otherwise, provide only the file name.
  $doc->setValue("exec.in.dest", ($src_fs eq "vmfs" &&
                                  $op ne "ESXHostEdit_ImportFromConsoleVMFSFile"
                                  ? "$dest/" : "") . $file);
  $doc->setValue("exec.in.deletesrc", $dele_src);

  $doc = request($doc, "xmldoc");

  my $resultstr = undef;

  my $rc = $doc->getValue("exec.returncode");
  if ($rc) {
    $rc = $doc->getValue("exec.out.taskid");
  } else {
    $resultstr = $doc->getValue("");
  }

  $doc->dispose();
  return ($rc, $resultstr);
}

# ---------------------------------------------------------------------------
# lro (long running operation)
#
# in:
#
#    $q ... CGI query object
#      ->op     ... One of init_task, init_task_errors or stat_task
#      ->lro    ... The long running op such cp or mv
#      ->taskid ... As returned by a file op such as FileManager_FileCopy
#      ->dir    ... The working directory
#
# out:
#
#    Task monitoring page (if op is init_task*) or form  (if op is stat_task).
# ---------------------------------------------------------------------------
sub lro {
  my $q = shift();
  my @ids = split(',', $q->param("taskid"));
  my $lro = $q->param("lro");
  my $dir = $q->param("dir");
  my @tasks;

  for my $id (@ids) {
    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "TaskManager_StatusOfTask");
    $doc->setValue("exec.in", "");
    $doc->setValue("exec.in.taskid", $id);

    $doc = request($doc, "xmldoc");
    my $rc = $doc->getValue("exec.returncode");

    if (! $rc) {
      push(@tasks, {});

      $tasks[$#tasks]->{id} = $id;
      $tasks[$#tasks]->{msg} = $doc->getValue("exec.out.messages.msg");
      chomp($tasks[$#tasks]->{msg});
      $doc->dispose();
    } else {
      push(@tasks, {});

      $tasks[$#tasks]->{id} = $id;
      $tasks[$#tasks]->{status} = $doc->getValue("exec.out.status");
      $tasks[$#tasks]->{percent} = $doc->getValue("exec.out.percent");
      $tasks[$#tasks]->{exitstatus} = $doc->getValue("exec.out.exitstatus");
      $tasks[$#tasks]->{taskStderr} = $doc->getValue("exec.out.taskStderr");

      $doc->dispose();
    }
  }

  if ($q->param("op") =~ /^init_task|init_task_errors$/) {
    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.out", "");
    $doc->setValue("exec.out.lro", $lro);
    $doc->setValue("exec.out.dir", $dir);

    if ($q->param("op") =~ /^init_task_errors$/) {
      $doc->setValue("exec.out.err", 1);
    }

    if ($q->param("execerror")) {
      if ($q->param("op") =~ /^init_task_errors$/) {
        $doc->setValue("exec.out.execerror", 1);
      } else {
        $doc->setValue("exec.out.initexecerror", 1);
      }
    }

    for my $task (@tasks) {
      my $d = VMware::DOMAccess->new("task");

      $d->setAttribute(".", "id", $task->{id});
      $d->setValue(".status", $task->{status});
      $d->setValue(".percent", $task->{percent});

      if ($task->{exitstatus}) {
        $d->setValue(".exitstatus", $task->{exitstatus});
      }

      if ($task->{msg}) {
        $d->setValue(".msg", $task->{msg});
      }

      if ($task->{taskStderr}) {
        $d->setValue(".msg", $task->{taskStderr});
      }

      $doc->addSubTree("exec.out", $d);
      $d->dispose();
    }

    my $docstring = $doc->getValue("");
    $doc->dispose();
    return processResult($docstring);

  } else {
    my $execerrorstr = $q->param("execerror") ? "true" : "false";

    my $returnhtml = <<end;
<html>
<head>
<script>
var LROs = new Array();
end
    my $i = 0;
    for my $task (@tasks) {
      $returnhtml .= "LROs[$i] = new parent.LRO(".
        "'$task->{id}', " .
        "'$task->{status}', " .
        "parseInt('$task->{percent}'), " .
        "'$task->{exitstatus}');\n";
      $i++;
    }

  $returnhtml .= <<end;
</script>
</head>
<body onLoad="if (parent.showdir.stat_cb) { parent.showdir.stat_cb(LROs, $execerrorstr); } else { window.location.replace('/showdir?op=hide'); }">
</body>
</html>
end

  trace("$returnhtml", "debug");

  return($returnhtml,
         "Pragma"        => "no-cache",
         "Cache-Control" => "no-cache",
         "Content-Type"  => "text/html; charset=iso-8859-1");
  }
}

# ---------------------------------------------------------------------------
# lro_stop (stop a long running operation)
#
# in:
#
#    $q ... CGI query object
#      ->op     ... One of init_task, init_task_errors or stat_task
#      ->taskid ... As returned by a file op such as FileManager_FileCopy
#      ->dir    ... The working directory
#
# out:
#
#    Calls &lro_delete thereby redirecting to /showdir?op=ls&dir={->dir}.
# ---------------------------------------------------------------------------
sub lro_stop {
  my $q = shift();
  my @ids = split(',', $q->param("taskid"));

  for my $id (@ids) {
    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "TaskManager_StopTask");
    $doc->setValue("exec.in", "");
    $doc->setValue("exec.in.taskid", $id);

    $doc = request($doc, "xmldoc");
    # XXX Do we really care if TaskManager_StopTask fails? Since we can't do
    # anything about it, probably not.
    $doc->dispose();

    #   my $rc = $doc->getValue("exec.returncode");

    #   if ($rc == 0) {
    #     $doc->dispose();
    #     return("Error (Return Code: $rc)",
    #            "Pragma"        => "no-cache",
    #            "Cache-Control" => "no-cache",
    #            "Content-Type"  => "text/plain");
    #   }
  }

  &lro_delete($q);
}

# ---------------------------------------------------------------------------
# lro_delete (dispose of a long running operation taskid)
#
# in:
#
#    $q ... CGI query object
#      ->op     ... One of init_task, init_task_errors or stat_task
#      ->taskid ... As returned by a file op such as FileManager_FileCopy
#      ->dir    ... The working directory
#
# out:
#
#    Redirect to /showdir?op=ls&dir={->dir}.
# ---------------------------------------------------------------------------
sub lro_delete {
  my $q = shift();

  my @ids = split(',', $q->param("taskid"));

  for my $id (@ids) {
    my $doc = VMware::DOMAccess->new("exec");
    $doc->setValue("exec.op", "TaskManager_DeleteTask");
    $doc->setValue("exec.in", "");
    $doc->setValue("exec.in.taskid", $id);

    $doc = request($doc, "xmldoc");
    # XXX Do we really care if TaskManager_DeleteTask fails? Since we can't do
    # anything about it, probably not.
    $doc->dispose();

    #   if ($doc) {
    #     my $rc = $doc->getValue("exec.returncode");

    #     if ($rc == 0) {
    #       $doc->dispose();
    #       return("Error (Return Code: $rc)",
    #              "Pragma"        => "no-cache",
    #              "Cache-Control" => "no-cache",
    #              "Content-Type"  => "text/plain");
    #     }

    #     $doc->dispose();
    #   }
  }

  return("", "Location" => "/showdir?op=ls&dir=". safeFileName($q->param("dir")));
}

# ---------------------------------------------------------------------------
# Ensure that users who are copying or moving virtual disks from the console
# file system to a VMFS partition make the decision as to whether or not the
# disks should be converted to the VMFS format.
#
# Note: We assume that exports (from a VMFS to the console fs) should happen
# automatically since most VMFS files will be larger than the console maximum.
#
# in:
#
#      $q ... CGI query object
#      ->destdir ... The current working directory
#      ->files   ... Any files being moved or copied
#      ->dirs    ... Any directories being moved or copied
#      ->dest_fs ... The current working file system
#      ->src_fs  ... The source file system
#
# out:
#
#      An empty string (indicating that confirmation is not necessary) or an
#      html string for the confirmation form.
# ---------------------------------------------------------------------------
sub import_gate() {
  my ($q) = @_;

  # We have already determined which disks should be converted.
  if ($q->param("import_confirmed")) {
    return '';
  }

  my $cwd = $q->param("destdir");
  my $files = $q->param("files");
  my $dirs = $q->param("dirs");
  my @files = split(/\|/, $q->param("files"));
  my @dirs = split(/\|/, $q->param("dirs"));

  my $confirm_import = 0;

  my $dest_fs = $q->param("dest_fs");
  my $src_fs = $q->param("src_fs");

  # If the destination fs is VMFS and we are moving vmdk files, make sure we
  # ask whether or not the disk(s) should be converted during the copy/move.
  if ($dest_fs ne $src_fs && $dest_fs eq 'vmfs') {
    for my $file (@files) {
      my $f = $file;
      my @data = split(';', $f);
      $f = shift(@data);

      my %meta = parse_file_metadata(@data);

      if ($meta{type} =~ /^vmdk$/) {
        $confirm_import = 1;
        last;
      }
    }
  }

  if (! $confirm_import) {
    return '';
  }

  #   <out>
  #      <op> cut | copy </op>
  #      <relativeRootDir> absolute path of the current working directory </relativeRootDir>
  #      <sourceFileSystem> vmfs | cons </sourceFileSystem>
  #      <destinationFileSystem> vmfs | cons </sourceFileSystem>
  #      <confirmImport> 1 </confirmImport>
  #      <files>
  #         <file>
  #            <name> file name </name>
  #            <path> absolute path to file name </path>
  #            <type> vmdk | vmdk-vmfs | ... </type>
  #         </file>
  #         ...
  #      </files>
  #      <directories>
  #         <directory>
  #            <name> directory name </name>
  #            <path> absolute path to directory name </path>
  #            <type> ... </type>
  #         </directory>
  #         ...
  #      </directories>

  my $doc = VMware::DOMAccess->new("exec");
  $doc->setValue("exec.op", $q->param("op"));
  $doc->setValue("exec.out.relativeRootDir", $cwd);
  $doc->setValue("exec.out.sourceFileSystem", $src_fs);
  $doc->setValue("exec.out.destinationFileSystem", $dest_fs);
  $doc->setValue("exec.out.confirmImport", 1);

  my $i = 0;
  for my $file (@files) {
    trace("FILE: $file", "debug");
    my @data = split(';', $file);
    $file = shift(@data);

    $doc->setValue("exec.out.files.file[$i].path", "$file");

    $file = substr($file, rindex($file, '/') + 1);
    $doc->setValue("exec.out.files.file[$i].name", "$file");

    my %meta = parse_file_metadata(@data);
    trace("TYPE: $meta{type}", "debug");
    $doc->setValue("exec.out.files.file[$i].type", $meta{type});
    $i++;
  }

  $i = 0;
  for my $dir (@dirs) {
    my @data = split(';', $dir);
    $dir = shift(@data);

    $doc->setValue("exec.out.directories.directory[$i].path", "$dir");

    $dir = substr($dir, rindex($dir, '/') + 1);
    $doc->setValue("exec.out.directories.directory[$i].name", "$dir");

    my %meta = parse_file_metadata(@data);
    $doc->setValue("exec.out.directories.directory[$i].type", $meta{type});
    $i++;
  }

  if ($ENV{vmware_DEBUG_LEVEL} > 2) {
    string_to_file($doc->getValue(""), &VMware::Management::Util::tmpdir() . "import_gate.xml", 1);
  }

  my $processor = new XML::Sablotron();
  my $xsl = $ENV{style_DEFAULT};

  $processor->SetBase("file://".$ENV{vmware_BASEHREF});
  $processor->RunProcessor("arg:/template", "arg:/data", "arg:/result",
                           undef,
                           ["template", file_to_string($xsl), "data", $doc->getValue("")]);
  $doc->dispose();

  return $processor->GetResultArg("result");
}

sub convert {
  return &copy_or_move($q, "FileManager_FileCopy");
}

sub copy {
  my ($q) = @_;
  my $ret = &import_gate($q);
  if ($ret ne '') {
    return ($ret,
            "Pragma"        => "no-cache",
            "Cache-Control" => "no-cache",
            "Content-Type"  => "text/html; charset=iso-8859-1");
  }
  return &copy_or_move($q, "FileManager_FileCopy");
}

sub move {
  my ($q) = @_;
  my $ret = &import_gate($q);
  if ($ret ne '') {
    return ($ret,
            "Pragma"        => "no-cache",
            "Cache-Control" => "no-cache",
            "Content-Type"  => "text/html; charset=iso-8859-1");
  }
  return &copy_or_move($q, "FileManager_FileMove");
}

sub handler {
  $q = new CGI(shift());

  if ($q->param("op") =~ /^init_task|init_task_errors$/) {
    return &lro($q);
  } elsif ($q->param("op") eq "stat_task") {
    return &lro($q);
  } elsif ($q->param("op") eq "stop_task") {
    return &lro_stop($q);
  } elsif ($q->param("op") eq "dele_task") {
    return &lro_delete($q);
  } elsif ($q->param("op") eq "convert") {
    return &convert($q);
  } elsif ($q->param("op") eq "copy") {
    return &copy($q);
  } elsif ($q->param("op") eq "cut") {
    return &move($q);
  } elsif ($q->param("op") eq "newdir") {
    return &newdir($q);
  } elsif ($q->param("op") eq "ls") {
    return &showdir($q);
  } elsif ($q->param("op") eq "delete") {
    return &delete($q);
  } elsif ($q->param("op") eq "hide") {
    return &returnHiddenForm();
  } elsif ($q->param("op") eq "properties_form") {
    return &return_properties_form();
  } elsif ($q->param("op") eq "init_properties") {
    return &init_properties($q);
  } elsif ($q->param("op") eq "edit_properties") {
    return &edit_properties($q);
  } else {
    return &returnFrameset(0);
  }
}

sub safeFileName($) {
  my ($dir) = @_;

  my $safedir = URI::Escape::uri_escape($dir);
  $safedir = URI::Escape::uri_escape($safedir, "\+&,:;=?@");

  return $safedir;
}

1;
