# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SxLegacyOperation;

# ---------------------------------------------------------------------------
# Handle control operations such as power-on, suspend, etc.
# ---------------------------------------------------------------------------

use strict;

use VMware::Control '1.00';
use VMware::Control::Server ();
use VMware::Control::VM ();
use VMware::Control::Profiler;
use VMware::Management::Util qw( trace
                                 file_to_string
                                 string_to_file
                                 hash_credentials
                                 connect_ctrl_obj
                                 use_vmdb );

use VMware::VmdbPerl;
use VMware::Management::Session;
use CGI ();

my %operation = ( "poweroff" => "stop",
                  "suspend"  => "suspend_to_disk",
                  "resume"   => "start",
                  "poweron"  => "start",
                  "reset"    => "reset",
                  "answer"   => "answer",
                  "set"      => "set" );

my ($q, %credentials, $hash, $cfg);
my $p = VMware::Control::Profiler::getInstance("mui");

my $head = qq{<html><head><title>vmlist</title><script>$/};
my $tail = qq{</script></head><body bgcolor="#336699" onload="if (parent.opcb) parent.opcb(self);"></body></html>$/};


sub handler {
   my ($stub, $server, $server_ok, $vm, $vm_ok, $err, $errstr, $ctx, $opType);

   $stub = $p->profile();
   $q = new CGI( shift() );
   %credentials = hash_credentials( );

   $server = VMware::Control::Server::new(undef, 0, $credentials{"l"},
                                          $credentials{"m"});

   ($server_ok, $err, $errstr) = connect_ctrl_obj($server);
   unless( $server_ok ) {
      return html( undef, undef, $err, $errstr );
   }

   my $op    = $operation{$ENV{"OPERATION"}};
   my $vmId  = $q->param("vmid");
   my $state = $q->param("state");
   $opType   = $q->param("optype");
   $hash     = $vmId;

   if($state eq "poweredOff") { $op = "stop"; }
   if($state eq "poweredOn")  { $op = "start"; }
   if($state eq "suspend")    { $op = "suspend_to_disk"; }
   if($state eq "reset")      { $op = "reset"; }

   if (!use_vmdb()) {
       return (&VMware::Management::Util::error_page("VMDB is not being used.",
               __PACKAGE__, __LINE__));
   }

   # set vmdb specific variables
   my $vmCfg = "/vm/#$vmId/cfgPath/";
   $ctx =  &VMware::Management::Session::get_temp_context();

   if (!defined($ctx)) {
      return (&VMware::Management::Util::error_page("Could not get temporary context.",
             __PACKAGE__, __LINE__));
   }

   $cfg    = $ctx->GetString($vmCfg);
   # trace("\n  DEBUG: Config file - $cfg", "debug");

   # set vmdb specific variables
   my $vmPath = "/vm/#$vmId";
   $ctx =  &VMware::Management::Session::get_temp_context();

   if (!defined($ctx)) {
      return (&VMware::Management::Util::error_page("Could not get temporary context.",
             __PACKAGE__, __LINE__));
   }

  $vm = VMware::Control::VM::new($server, $cfg);

   ($vm_ok, $err, $errstr) = connect_ctrl_obj($vm);
   unless( $vm_ok ) {
      return html(undef, undef, $err, $errstr);
   }

   my $action = $op;
   trace("Calling \$vm->$action.", "debug");

   my $op_ok;

   if ($op eq "set") {
      my @names = $q->url_param();
      for my $name (@names) {
         next if ($name eq "c" || $name eq "f" || $name eq "vm");

         $op_ok = $vm->set("$name", $q->url_param($name));
         if (! $op_ok) {
            last;
         }

         # Save everytime to ensure that the config jives with what has been set
         # thus far.
         $op_ok = $vm->save_config();
         if (! $op_ok) {
            last;
         }
      }
   } elsif ($op ne "answer") {
      $op_ok = $vm->$action($opType or undef);
      # XXX Power on, resume and suspend should probably work regardless of the
      # availaibility of the appropriate tools version (especially while the
      # guest is booting). This, however, masks the fact that scripts failed to
      # run in those cases where users might expect them to (e.g., because they
      # believe that the tools are installed and capable of running scripts).
      if (! $op_ok && ! $opType &&
          ($op eq 'start' || $op eq 'suspend_to_disk')) {
         my ($err, $errstr) = $vm->get_last_error();

         if ($err == VMware::Control::VM_E_BADSTATE) {
            # trace("Caught error: $err $errstr. Retrying $$op{action}.", "debug");
            $op_ok = $vm->$action(1);
         }
      }
   } else {
      $op_ok = $vm->answer_question($q->param("question-id"),
                                    $q->param("answer-id"));
   }

   unless ($op_ok) {
      ($err, $errstr) = $vm->get_last_error();
      trace("$err: $errstr", "error");
      return html($server, $vm, $err, $errstr);
   }

   return html($server, $vm);
}


sub cleanup {
   my $stub = $p->profile();
   my ($vm, $server) = @_;
   if ($vm) {
      $vm->disconnect();
   }

   if ($server) {
      $server->disconnect();
   }
}


sub html {
   my ($server, $vm, $err, $errstr) = @_;

   if (! defined($err)) {
      $err = 0;
   }

   if (! defined($errstr)) {
      $errstr = "";
   } else {
      $errstr =~ s/\\/\\\\/g;
      $errstr =~ s/("|')/\\$1/g;
      $errstr =~ s/(\n)/<br \/>/g;
   }

   my $body = qq{var hash="$hash";var err="$err";var errstr="$errstr";$/};

   if (defined($server) && defined($vm) && $err &&
       $err == VMware::Control::VM_E_NEEDINPUT) {
      $body .= question($vm);
   }

   cleanup($server, $vm);

   return ( "$head$body$tail",
            "Pragma"        => "no-cache",
            "Cache-Control" => "no-cache, must-revalidate",
            "Content-Type"  => "text/html; charset=utf-8" );
}


sub question {
   my $vm = shift();

   my $id = $vm->get("Status.question.current");
   return "" unless (defined($id));

   my $limit = $vm->get("Status.question.$id.choices.number");
   return "" unless (defined( $limit ));

   my $st = $vm->get("Status.question.$id.text");

   # Make $st more amenable for HTML and JavaScript.
   $st =~ s/(\n\n)/<\/p><p>/g;
   $st =~ s/(\n)/ /g;
   $st =~ s/("|')/\\$1/g;
   $st = "<p>$st</p>";

   my $h = qq'var qstn = {id:"$id",$/st:"$st",$/ch:[';

   my $b = "";
   for (0..$limit - 1) {
      $b .=  ($b ne "" ? "," : "") .
         '"' . $vm->get("Status.question.$id.choices.$_") . '"';
   }

   my $t = "]};$/";

   # trace("$h$b$t", "debug");

   return "$h$b$t";
}


1;
