# ---------------------------------------------------------------------------
# Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
# ---------------------------------------------------------------------------
package VMware::Management::SxMonitor;

use strict;

use VMware::HConfig::SSHostRemote ();
use VMware::Control '1.01';
use VMware::Control::Server;
use VMware::Control::VM;
use VMware::Control::Profiler;
use VMware::Management::Util qw( trace
                                 file_to_string
                                 string_to_file
                                 hash_credentials
                                 connect_ctrl_obj
                                 error_page
                                 use_vmdb
                               );
use VMware::VmdbPerl;
use VMware::Management::Session;
use VMware::Management::VmdbUtil ();
use VMware::Config();
use VMware::Management::VmConfiguration ();
use VMware::Management::VmOptions ();

use CGI ();

# XXX FIX: Remove when values from xml aren't required for the vm configs
use VMware::Management::SS::VMUtil;

my $q;
my $doc;
my %credentials = ();
my $xml;
my $product;
my $MMDATA_ROOT = "/";   #Establish VMDB root tree context

# ---------------------------------------------------------------------------
#
# genHead --
#
#      Define JS functions, arrays & objects to be used in creating the
#      host and VM information.
#
# Prototype:
#
#      string genHead()
#
# Arguments:
#
#      none
#
# Results:
#
#      Returns an HTML string with embedded JavaScript for use by the client.
#
# ---------------------------------------------------------------------------

sub genHead {
   # Generate/define all JS function to be used for a VM and
   # its sub pages (overview, hardware, events)
   return <<HEAD;
<html><title>SxMonitor</title><head><script>
function I0(a,b,c,d,v,e,f,g,h,i,j,k,w,l,m,x,y,z,r,hwv,usr) {
   this.a=a;
   this.b=b, this.c=c; this.d=d; this.v=v; this.e=e; this.f=f; this.g=g;
   this.h=h; this.i=i; this.j=j; this.k=k; this.w=w; this.l=l; this.m=m;
   this.x=x;this.y=y;this.z=z; this.r=r; this.hwv=hwv; this.usr=usr;
};

function I1(a,b,c,d,e,f) {
   this.a=a; this.b=b, this.c=c; this.d=d; this.e=e; this.f=f;
};

function I2() {
   this.dev=new Array();
   this.opt=new Array();
   for(var i=0;i<arguments.length;i++) {
      if (arguments[i].id != null) {
         this.dev[arguments[i].id]=arguments[i];
      }
      if (arguments[i].k != null) {
         this.opt[arguments[i].k]=arguments[i].v;
      }
   }
};

function Evt(a,b,c) {
   this.a=a;this.b=b;this.c=c;
};

function Usr(a,b,c) {
   this.a=a;this.b=b;this.c=c;
};

function I4(a,b) {
   this.a=a;this.b=b;
};

function I8(a,b,c,d) {
   this.a=a;this.b=b;this.c=c;this.d=d;
};

function sxVmAutoStartup(a,b,c,d) {
  this.autostart = a;
  this.autostartDelay = b;
  this.autostartContOnTools = c;
  this.autostopDelay = d;
}

var vms=new Array();
HEAD

} ## close genHead() sub


# ---------------------------------------------------------------------------
#
# genTail --
#
#      Collect stats for the host and for each VM on the host.
#      Stats for each VM are collected by calling genVm().
#
# Prototype:
#
#      string genTail(string $ctx)
#
# Arguments:
#
#      $ctx ... vmdb context path
#
# Results:
#
#      Returns an HTML string with embedded JavaScript for use by the client.
#
# ---------------------------------------------------------------------------

sub genTail($$) {
   my ($ctx, $server) = @_;

   # check user rights
   my $admin = &VMware::Management::Session::user_is_admin() ? "true" : "false";
   my $canAddVm = &VMware::Management::Session::user_can_create_vms() ? "true" : "false";

   # retrieve host values from vmdb

   my $hstName   = $ctx->GetString("/host2/#_mui/info/hostname");
   # XXX Hard-code product ID if it's not available in VMDB (PR8862);
   my $prdId     = $ctx->GetString("/host2/#_mui/productinfo/id") or "gsx";
   my $agtId     = $ctx->GetString("/host2/#_mui/productinfo/serverAgent") or "vmserverd";
   my $prdName   = $ctx->GetString("/host2/#_mui/productinfo/name");
   $prdName     .= " " . $ctx->GetString("/host2/#_mui/productinfo/version/versionstring");
   $prdName     .= " " . $ctx->GetString("/host2/#_mui/productinfo/version/buildnum");
   my $hstOs     = $ctx->GetString("/host2/#_mui/info/os/type");
   my $cpuInfo   = $ctx->GetInt("/host2/#_mui/info/cpuNum");
   my $memInfo   = $ctx->GetInt("/host2/#_mui/info/memory/sizeMb");
   my $hstUpTime = $ctx->GetInt("/stats/host/secUptime");
   my $cosRamUsg = int($ctx->GetInt("/stats/host/memory/kbActiveCos/avg/") / 1024);
   if ($cpuInfo < 1) {
      trace("CPU information incorrect: $cpuInfo", "debug");
      $cpuInfo = 1;
   }

   # cpu hyperthreading is currently implemented in vmdb. Get through SSHostRemote
   my %credentials = VMware::Management::Util::hash_credentials();
   my $user = $credentials{"l"};
   my $pass = $credentials{"m"};

   my $maxVcpuNum = 1;
   my $maxPcpuNum = 1;

   my $cosInfo   = 0;
   my $numaInfo  = 1;

   my $cpuHtSprt = 0;
   my $cpuHtEnbl = "false";

   if ($prdId eq "ESX") {
      my $host = VMware::HConfig::SSHostRemote->new(undef, 0, $user, $pass, $ctx);

      $cpuHtSprt = $host->isCPUHyperthreadingEnabled();
      # my $cpuHtSprt = 0; # uncomment line to test cpus with no support for hprThrdng

      # XXX Fix, this should be retrieved from vmdb once implemented, instead of
      #     being read from the hwconfig file directly and defaulting to false if
      #     the hyperthreading line is not there.
      my $cfgDb = $host->getConfigDB();
      my $cpuHtEnbl = $cfgDb->get("hyperthreading");
      if ($cpuHtEnbl eq "") {
         $cpuHtEnbl = "false";
      }

      $cosInfo   = int($ctx->GetInt("/vmkernel/host/stats/memory/COS/") / 1024);
      $numaInfo  = $ctx->GetInt("/host2/#_mui/info/numa/numNodes") || 1;

      # retrieve license info
      # Only on ESX for now since the port from the server02 branch isn't done yet.
      $maxVcpuNum = $ctx->GetInt("/host2/#_mui/setup/license/limits/maxVcpusPerVm") || 0;
      $maxPcpuNum = $ctx->GetInt("/host2/#_mui/setup/license/limits/maxHostCpus") || 0;
   }

   my $license = "{maxVcpuNum:$maxVcpuNum,maxPcpuNum:$maxPcpuNum}";

   # Retrieve Sx defaults for VMs autostart and autostop

   my $autostart =
     $ctx->GetBool("/host2/#_mui/settingsState/val/autoStartEnabled/");
   my $autostartDelay =
     $ctx->GetInt("/host2/#_mui/settingsState/val/autoStartDefaultStartDelay/");
   my $autostartContinueWhenToolsRun  =
     $ctx->GetBool("/host2/#_mui/settingsState/val/autoStartDefaultContinueWhenToolsRun/");
   my $autostopPowerOff = "null"; # XXX FIX once implemented by VMDB
   my $autostopDelay =
     $ctx->GetInt("/host2/#_mui/settingsState/val/autoStartDefaultStopDelay/");

   my $sxVmAutoStartup = qq{new sxVmAutoStartup($autostart,$autostartDelay,} .
                         qq{$autostartContinueWhenToolsRun,$autostopDelay)};

   # Generate guest OS list
   my @guestList = ();
   my $path = "/host2/#_mui/cfgLimits/guest/#/";
   while (defined($path = $ctx->GetNextSibling($path))) {
      # Only show supported OSes in list
      if ($ctx->GetBool("${path}supported")) {
         my %params = ();
         $params{key} = jsStringLiteral($ctx->GetString("${path}key"));
         $params{name} = jsStringLiteral($ctx->GetString("${path}name"));
         $params{family} = jsStringLiteral($ctx->GetString("${path}family"));
         $params{defDiskMb} = int($ctx->GetInt("${path}defaultDiskSize") / 1024 / 1024);
         $params{defSizeMb} = $ctx->GetInt("${path}memory/defaultMb");
         $params{minSizeMb} = $ctx->GetInt("${path}memory/minMb");
         $params{maxSizeMb} = $ctx->GetInt("${path}memory/maxMb");
         $params{maxVcpuNum} = $ctx->GetInt("${path}maxVcpuNum");
         $params{depth} = $ctx->GetInt("${path}defaultColorDepth");
         $params{lsi} = $ctx->GetBool("${path}defaultLsiAdapter");
         $params{scsi} = $ctx->GetBool("${path}defaultScsiDisk");
         my @paramList = ();
         foreach (keys %params) {
            push @paramList, "$_:$params{$_}";
         }
         push @guestList, "{" . join(",", @paramList) . "}";
      }
   }
   my $guestInfo = "[" . join(",", @guestList) . "]";

   # Values are aggregated at bottom of sub
   my $vmCpuUsg  = 0;
   my $vmRamUsg  = 0;

   # XXX FIX: Waiting to be implemented into VMDB
   # $interval ... The time between client refreshes in milliseconds
   my $interval = 90 * 1000;  #XXX: s/b value from vmdb
   if (defined($ENV{"vmware_REFRESH_INTERVAL"}) &&
       $ENV{"vmware_REFRESH_INTERVAL"} > 0) {
      $interval = $ENV{"vmware_REFRESH_INTERVAL"} * 1000;
   }

   # Get the serverd stats collection rate.
   my $samplePeriod = $ctx->GetInt("/stats/samplePeriod/");
   my $sampleHistory = $ctx->GetInt("/stats/sampleHistory/");
   my $period = int($samplePeriod * $sampleHistory / 60);

   # User info
   my $username = &VMware::Management::Session::get_username();

   # Retrieve all VMIDs & define node
   $ctx->SetCurrentPath($MMDATA_ROOT);
   my $regVmPath = "host2/#_mui/vms/registeredvms/#";

   # will contain JS code with all VM info as arrays/objects
   my $jsVmObjs  = "";

   # Generate JS VM objects with all relevant info for a given vmId
   while (defined($regVmPath = $ctx->GetNextSibling($regVmPath))) {
      $regVmPath  = $ctx->GetAbsPath($regVmPath);

      $ctx->SetCurrentPath($regVmPath);
      my $vmCfg   = $ctx->GetString("cfgPath");

      my $vmPath  = $ctx->GetString("vm");
      my $vmId;  # hash value

      if ($vmPath !~ m{/vm/\#(.*)/$}) {
         next;  # skip to next sibling, no valid vm number was returned
      } else {
         $vmId = $1;
      }
      trace("VMID Assigned: $vmId", "debug");

      # Generate HTML/JS code for each vmId supplied
      # (A JS object will be created for each VM)

      my $vmCtxPath = $MMDATA_ROOT . "vm/#$vmId";
      $ctx->SetCurrentPath($vmCtxPath);

      # keep running totals to display on overview page
      my $vmStats = "/stats/vm/#$vmId";
      if ($ctx->IsSet($vmStats)) {
         $vmRamUsg += int($ctx->GetInt("$vmStats/memory/kbGranted/avg/") / 1024);
         # Deduct shared memory to prevent double-counting, ONLY ESX
         if ($prdId eq "ESX") {
            $vmRamUsg -= int($ctx->GetInt("$vmStats/memory/kbShared/avg/") / 1024);
         }
         my $cpuStats = "$vmStats/cpu/all";
         if ($ctx->IsSet($cpuStats)) {
            $vmCpuUsg += int(0.1 * $ctx->GetInt("$cpuStats/milUsage/avg/") / $cpuInfo);
         }
      }

      # generate JS code per VM

      $jsVmObjs .= genVm($vmCfg, $vmId, $ctx, $server, $prdId);

   } ## close while() for getting vmIds

   # Generate data for Service Console
   if ($prdId eq "ESX") {
      $jsVmObjs .= genConsole($ctx, $server);
   }

   my $ramUsgTtl = $cosRamUsg;

   if ($hstOs ne "win32") {
      # For ESX and Linux, the console OS ram usage does not include guest
      # RAM memory. In Linux case, the cached memory which including
      # the VM RAM is counted as free memory. Add it back to total usage.
      $ramUsgTtl = $vmRamUsg + $cosRamUsg;
   }

   # XXX It would be nice to fudge this number in an attempt to report
   # the total host memory including all hidden overheads (see PR26888),
   # but for now just report the total of vmkernel and console memory.
   $memInfo = $memInfo + $cosInfo;

   my $cpuUsgTtl = int(0.1 * $ctx->GetInt("/stats/host/cpu/all/milUsage/avg/") / $cpuInfo);

   # Because some of the values are one-minute averaged, it is possible
   # to get into transient cases where the total resource usage appears
   # to be greater than the capacity (which manifests itself as either
   # greater than 100% usage overall or as negative usage after doing
   # the math).  Clamp the values here to prevent that from occurring.
   $cpuUsgTtl = ($cpuUsgTtl > 100) ? 100 : $cpuUsgTtl;
   $ramUsgTtl = ($ramUsgTtl > $memInfo) ?  $memInfo : $ramUsgTtl;
   $vmCpuUsg = ($vmCpuUsg > $cpuUsgTtl) ? $cpuUsgTtl : $vmCpuUsg;
   $vmRamUsg = ($vmRamUsg > $ramUsgTtl) ? $ramUsgTtl : $vmRamUsg;


   return <<TAIL;
$jsVmObjs

var sx={a:"$hstName",b:"$prdId",c:"$prdName",d:"$cpuInfo",e:"$memInfo",f:$cpuUsgTtl,g:"$vmCpuUsg",h:"$ramUsgTtl",i:"$vmRamUsg",j:"$hstOs",k:"$hstUpTime",l:"$numaInfo",m:$license,n:$cpuHtSprt,o:$cpuHtEnbl,p:$sxVmAutoStartup,q:$guestInfo,r:"$agtId"};

var user       =  "$username";
var admin      =  $admin;
var canAddVm   =  $canAddVm;
var interval   =  "$interval";
var period     =  "$period";
</script></head><body bgcolor="#336699" onload="try { if (parent.cb) parent.cb(self); } catch (e) {;}"></body></html>
TAIL

} ## close genTail() sub


# ---------------------------------------------------------------------------
#
# genVm --
#
#      Collect stats for each VM on the host.
#      Additional stats are collected for VMs that are in 'poweredOn' state
#
# Prototype:
#
#      string genVm(string $cfg, string $vmId, string $ctx)
#
# Arguments:
#
#      $cfg  ... configuration file_path/file_name for VM
#      $vmId ... hash value of configuration file_path/file_name for VM
#      $ctx  ... vmdb context path
#
# Results:
#
#      Stats collected are then compiled into JavaScript arrays and returned
#      as a text string.
#      Returns an HTML string with embedded JavaScript for use by the client.
#
# ---------------------------------------------------------------------------

sub genVm($$$$) {
   my ($cfg, $vmId, $ctx, $server, $prdId) = @_;
   trace("VM ID: $vmId", "debug");

   # Host information required for computing VM stats
   my $cpuInfo   = $ctx->GetInt("/host2/#_mui/info/cpuNum");

   if ($cpuInfo < 1) {
      trace("CPU information incorrect: $cpuInfo", "debug");
      $cpuInfo = 1;
   }

   # Stats Section: Stat totals for installed VMs (any power state)
   my $upTime      = 0;
   my $vmCpuUsgAvg = 0;
   my $vmRamUsgAvg = 0;

   # Other VM Information
   my $hb = 0;  # value is retireved if 'poweredOn' and have tools 'running'
   my $dn = $ctx->GetString("vmx/cfgState/val/name");
   my $vcpu = 1; # XXX: Not in VMDB yet.
   my $es = $ctx->GetString("vmx/execState/val");
   my $guestOs = jsStringLiteral($ctx->GetString("vmx/cfgState/val/guestOS"));
   my $usr = jsStringLiteral($ctx->GetString("user/username"));
   my $hwv;

   # Determine hardware version by examining the virtual machine's boot device.
   my @devices = VMware::Management::VmConfiguration::devices($ctx);
   for (@devices) {
      # Devices are in sort order. The first device on SCSI controller 0 is the
      # boot device. XXX: This will pick up 0:10 before 0:2 in the unlikely
      # event that A) someone configures the boot device on something other
      # than 0 or 1 -and- B) that same someone configures more than 9 devices
      # on a single SCSI controller (which isn't even possible via the MUI).
      if (/^scsiDev0:\d+$/i) {
         my $devPath = "vmx/cfgState/val/dev/#_$_/class/scsiDev/subclass/";
         my $subclass = $ctx->GetString($devPath);
         $devPath = "${devPath}$subclass/";

         if ($subclass eq "disk") {
            $hwv = $ctx->GetInt("${devPath}hwVersion");

            if ($hwv eq "-1") {
               my $file = $ctx->GetString("${devPath}file");

               $file = &VMware::Management::Util::getDiskPath($file,
                                                              $ctx->GetCurrentPath());
               if (!defined($file)) {
                 trace("Failed to get diskpath for disk", "error");
                 $hwv = 0;
               } else {
                 my $trnCtx = &VMware::Management::Session::get_temp_context();
                 $trnCtx->SetCurrentPath($ctx->GetCurrentPath());
                 $trnCtx->BeginTransaction();

                 # Initialize VMDB request.
                 my $reqPath = $trnCtx->NewArrayIndex("/host2/#_mui/util/disk/cmd/##/");
                 $reqPath = $trnCtx->GetAbsPath($reqPath);
                 $trnCtx->SetString("${reqPath}op", "getDiskInfo");
                 $trnCtx->SetString("${reqPath}user",
                                    &VMware::Management::Session::get_usertoken());
                 $reqPath = "${reqPath}op/getDiskInfo/";
                 $trnCtx->SetString("${reqPath}in/file", $file);

                 my ($ok, $msg) = VMware::Management::VmdbUtil::vmdbEndCmd($trnCtx,
                                                                           $reqPath);
                 if (! $ok) {
                   # Unable to determine hardware version.
                   $hwv = 0;
                 } else {
                   $hwv = $trnCtx->GetInt("${reqPath}out/hwVersion/");
                   if ($hwv == -1) {
                     $hwv = 0;
                   }
                 }

                 $reqPath =~ s{^(.*)(cmd/\#\#.*?/).*}{$1$2};
                 $trnCtx->LocalArrayUnset($reqPath, 1);

                 # Cache the hwVersion so we do not perform expensive lookups in
                 # the future.
                 $trnCtx->SetInt("${devPath}hwVersion", $hwv);
               }
             }
         }

         # Either the boot device is not a virtual disk, or we should have
         # grabbbed the hardware version from the disk.
         last;
      }
   }
   # Grab the hardware version from the config file if all else fails.
   $hwv ||= $ctx->GetString("vmx/hwVersion");
   $hwv ||= "null";

   if ($dn eq "") {
      $dn = $cfg;
   }

   # Var Clean: Make sure backslashes are usable in JavaScript.
   $cfg = jsStringLiteral($cfg);
   $dn  = jsStringLiteral($dn);

   # Stable virtual machine info
   my $pid  = $ctx->GetInt("vmx/pid");
   my $wid  = ($prdId =~ /esx/i) ? $ctx->GetInt("vmx/id") : 0;
   my $r = 4; # We're looking at the VM in question, aren't we?
   my $w = $ctx->CheckAccess("vmx/cfgState/val/name", 1) == 0 ? 2 : 0;
   my $x = $ctx->CheckAccess("vmx/execState/val", 1) == 0 ? 1 : 0;
   my $mode = $r + $w + $x;

   # extra info bit: 1 -> VM Overview, 2 -> Hardware, 4 -> Users & Events
   my $extra = $q->param($vmId);
   trace("EXTRA BIT: $extra", "debug");

   my $allVmsExtra = $q->param("allVms");
   trace("EXTRA BIT ACCUMULATIVE ALL VMS : $allVmsExtra", "debug");


   my $wfi;

   my $path = "msg/event/#";
   while (defined($path = $ctx->GetNextSibling($path))) {
      if ($ctx->GetBool("${path}prompt") &&
	  ! $ctx->GetBool("${path}revoked") &&
	  $ctx->GetString("${path}choiceType") ne "ok") {
	 if ($ctx->GetString("${path}input/choice") ne "") {
	    # Ignore messages that have already been answered.
	    next;
	 }

	 $wfi = getVmdbMsgEvent($ctx->GetAbsPath($path));
	 last;
      }
   }

   # Ensure that each variable is suitable for use in a JavaScript program.
   $pid   ||= 0;
   $wid   ||= 0;
   $extra ||= 0;
   $wfi     = "{id:null}" unless defined($wfi);

   # Virtual machine power state and statistics
   my $guestState  = $ctx->GetString("vmx/guestTools/runningStatus");
   my $toolsRdy    = "false";

   # XXX FIX: Need to impelment power status into vmdb to derive tools status
   #    It was derived by the softPowerCapable & softPowerPending values
   if ($guestState =~ /^running$/i) {
      $toolsRdy   = "true";     # required to see hb
   }

   my $I1      = "null";
   my $I2      = "null";
   my $I4      = "null";
   my $I8      = "null";

   # Get statistics for running VMs
   my $vmStats = "/stats/vm/#$vmId";
   my $cpuStats = "$vmStats/cpu/all";
   my ($vmIp, $vmMaxCpu, $vmMinCpu, $vmMaxRam, $vmMinRam) = ("", 0, 0, 0, 0);
   if ($es eq "poweredOn") {
      # Required values for the overview page
      $vmRamUsgAvg = int($ctx->GetInt("$vmStats/memory/kbGranted/avg/") / 1024);
      if ($ctx->IsSet($cpuStats)) {
         $vmCpuUsgAvg = int(0.1 * $ctx->GetInt("$cpuStats/milUsage/avg/") / $cpuInfo);
      }
      $upTime      = $ctx->GetInt("$vmStats/guest/secUptime/val/");

      # Retrieve only if tools are running
      if ($guestState =~ /^running$/i) {
         $hb = int(0.1 * $ctx->GetInt("$vmStats/guest/milHeartbeat/avg/"));
      }

      # Clean up values for initial vm startup
      if ($hb > 100) { $hb=100; }; if ($hb < 0) { $hb=0; };
   }

   # generate Overview information per VM if extra bit '1' is set
   if (($extra & 1) == 1) {
      trace("INFO: Get Overview For VM ID: $vmId", "debug");
      if ($es eq "poweredOn") {
         # Extra virtual machine info (vmSummary)
         $vmIp     = $ctx->GetString("guest/guestInfo/ip/");
         if ($ctx->IsSet($cpuStats)) {
            $vmMaxCpu = int(0.1 * $ctx->GetInt("$cpuStats/milUsage/max/") / $cpuInfo);
            $vmMinCpu = int(0.1 * $ctx->GetInt("$cpuStats/milUsage/min/") / $cpuInfo);
         }
         $vmMaxRam = int($ctx->GetInt("$vmStats/memory/kbGranted/max/") / 1024);
         $vmMinRam = int($ctx->GetInt("$vmStats/memory/kbGranted/min/") / 1024);
      }

      my $vmRamUsg = $ctx->GetInt("vmx/cfgState/val/dev/#_memory/class/memory/sizeMb/");
      $I1 = qq{new I1($vmMaxCpu,$vmMinCpu,$vmMaxRam,$vmMinRam,$vmRamUsg,"$vmIp")};
      trace("INFO: Get JS I1 Object: $I1", "debug");
   }

   my %ret = VMware::Management::VmConfiguration::memory($ctx, "memory");
   $vcpu = $ret{num};
   # XXX: Sometimes (as after answering a waiting for input question) this is
   # empty. Why? --mikol June 12, 2003 9:41 PM
   $vcpu ||= 1;

   # generate Hardware information per VM if extra bit '2' is set,
   # or was set to 2 for any other virtual machine.
   # XXX FIX: once vcpu is in VMDB, it shouldn't be passed into
   #     getDevices.
   if (($extra & 2) == 2 || ($allVmsExtra & 2) == 2) {
      $I2 = getDevices($ctx);
      $I2 .= ($I2 ? ",$/" : "") . getOptions($ctx);
      $I2 = qq{new I2($I2)};

      trace("INFO: Get Devices For VM ID: $vmId", "debug");
      trace("INFO: Get JS I2 Object: $I2", "debug");
   }

   # generate Users & Events information per VM if extra bit '4' is set
   if (($extra & 4) == 4) {
      $I4 = getUsersEvents($ctx, $vmId, $server);
   }

   # generate Resource information per VM if extra bit '8' is set
   # XXX FIX: once vcpu is in VMDB, it shouldn't be passed into
   #     getResources.
   if (($extra & 8) == 8) {
      $I8 = getResources($ctx, $vmId, $server, $vcpu);
   }


   # The following mapping outlines how the JS function s/b populated
   # Since this is a mapping of old code to new, I have retained the initial
   # single character
   # mapping for reference only (i.e. new_var  old_var ... parameter
   # description)
   # $cfg         a ... Config path
   # $dn          b ... Display name
   # $es          c ... Execution state (poweredOn, poweredOff or suspended)
   # $hb          d ... Heartbeat
   # $vcpu        v ... # of CPUs allocated to VM
   # $vmCpuUsgAvg e ... CPU usage
   # $vmRamUsgAvg f ... RAM usage
   # $guestOs     g ... Guest operating system
   # $upTime      h ... Up time
   # $guestState  i ... Guest state (unknown, running, halting, rebooting or
   #                    halting/rebooting)
   # $toolsRdy    j ... Tools ready (i.e. available, soft power capable and
   #                    not currently executing a soft power op)
   # $pid         k ... VMX process ID
   # $wid         w ... VMX world ID
   # $mode        l ... Effective permissions (mode)
   # $wfi         m ... Waiting for input
   # $I1          x ... I1: Extra info for the virtual machine details
   #                    overview tab
   # $I2          y ... I2: Extra info for the virtual machine details
   #                    hardware tab
   # $I4          z ... I4: Extra info for the virtual machine details
   #                    users and events tab
   # $I8          r ... I8: Extra info for the virtual machine details
   #                    resource management tabs
   # $hwv       hwv ... Hardware version
   # $usr       usr ... User to run VM as (empty string for interactive user)
   return <<HTML;
vms['$vmId'] = new I0($cfg,$dn,'$es',$hb,$vcpu,$vmCpuUsgAvg,$vmRamUsgAvg,$guestOs,$upTime,'$guestState',$toolsRdy,$pid,$wid,$mode,$wfi,$I1,$I2,$I4,$I8,$hwv,$usr);
HTML

} ## close genVm() sub


# ---------------------------------------------------------------------------
#
# genConsole --
#
#      Collect stats for the Service Console.
#
# Prototype:
#
#      string genConsole(string $ctx)
#
# Arguments:
#
#      $ctx  ... vmdb context path
#
# Results:
#
#      Stats collected are then compiled into JavaScript arrays and returned
#      as a text string.
#      Returns an HTML string with embedded JavaScript for use by the client.
#
# ---------------------------------------------------------------------------

sub genConsole($$$$) {
   my ($ctx, $server) = @_;
   my $cfg = qq{''};
   my $vmId = "console";
   trace("VM ID: $vmId", "debug");

   # Stats Section: Stat totals for installed VMs (any power state)
   my $upTime      = 0;
   my $vmCpuUsgAvg = 0;
   my $vmRamUsgAvg = 0;

   # VM information applicable to the console
   my $dn = qq{'Service Console'};
   my $cosPath = $ctx->GetString("/vmkernel/console");
   $cosPath =~ m{/vmkernel/world/\#(\d+)};
   my $wid  = $1;
   $wid ||= -1; # e.g., when the vmkernel is not running

   # VM information not applicable to the console
   my $hb = 0;
   my $vcpu = 1;
   my $es = "";
   my $guestOs = qq{''};
   my $wfi = "{id:null}";
   my $guestState  = "";
   my $toolsRdy    = "false";
   my $pid  = 0;

   # Permissions: read/write for admin, read for other
   my $username = &VMware::Management::Session::get_username();
   my $mode = &VMware::Management::Session::user_is_admin() ? 6 : 4;

   # extra info bit: 1 -> VM Overview, 2 -> Hardware, 4 -> Users & Events
   my $extra = $q->param($vmId) || 0;
   trace("EXTRA BIT: $extra", "debug");

   # Extra information for specific tabs
   my $I1      = "null"; # VM Status Monitor tab
   my $I2      = "null"; # Hardware tab -- not used by console
   my $I4      = "null"; # Users and events -- not used by console
   my $I8      = "null"; # Resource tabs

   # Overview information for console if extra bit '1' is set
   if (($extra & 1) == 1) {
      my ($vmIp, $vmMaxCpu, $vmMinCpu, $vmMaxRam, $vmMinRam, $vmRamUsg) = ("", 0, 0, 0, 0, 0);
      $I1 = qq{new I1($vmMaxCpu,$vmMinCpu,$vmMaxRam,$vmMinRam,$vmRamUsg,"$vmIp")};
      trace("INFO: Get JS I1 Object: $I1", "debug");
   }

   # Resource information for console if extra bit '8' is set
   if (($extra & 8) == 8) {
      $I8 = getResources($ctx, $vmId, $server, $vcpu);
   }


   # See genVm for details on the argument mapping
   return <<HTML;
vms['$vmId'] = new I0($cfg,$dn,'$es',$hb,$vcpu,$vmCpuUsgAvg,$vmRamUsgAvg,$guestOs,$upTime,'$guestState',$toolsRdy,$pid,$wid,$mode,$wfi,$I1,$I2,$I4,$I8);
HTML

} ## close genConsole() sub


# XXX Fix:
# Remove the $vcpu requirement from this sub once its available in VMDB

# ---------------------------------------------------------------------------
#
# getDevices --
#
#      Creates JS arrays for the hardware tab on the overview page for each
#      VM id supplied
#
# Prototype:
#
#      string getDevices($ctx, $vcpu)
#
# Arguments:
#
#      $ctx  ... pre-set control source to a vm
#      $vcpu ... # cpus assigned to a VM
#
# Results:
#
#      Returns JS arrays to the client.
#
# ---------------------------------------------------------------------------

sub getDevices {
   my $ctx = shift();
   my $vcpu = shift();
   my $data = "";

   my @devices = VMware::Management::VmConfiguration::devices($ctx);

   for (@devices) {
      $data .= ($data ? ",$/" : "");

      if (/^serial.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::serial($ctx, $_));
         next;
      }

      if (/^parallel.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::parallel($ctx, $_));
         next;
      }

      if (/^usbCtlr.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::usbCtlr($ctx, $_));
         next;
      }

      if (/^floppy.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::floppy($ctx, $_));
         next;
      }

      if (/^memory.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::memory($ctx, $_));
         next;
      }

      if (/^scsiCtlr[0-9]+$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::scsiCtlr($ctx, $_));
         next;
      }

      if (/^(ide|scsi)Dev[0-9]+:[0-9:]+$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::storageDev($ctx, $_));
         next;
      }

      if (/^nic.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::nic($ctx, $_));
         next;
      }

      if (/^video.*$/i) {
         $data .= VMware::Management::VmConfiguration::devData(VMware::Management::VmConfiguration::video($ctx, $_));
         next;
      }

      # Remove ",$/".
      chomp($data);
      chop($data);
   }

   return "$data";
}


# ---------------------------------------------------------------------------

sub getOptions {
   my $ctx = shift();
   my $optData = "";

   for (VMware::Management::VmOptions::options($ctx)) {
      $optData .= ($optData ? ",$/" : "") .
         VMware::Management::VmOptions::optData($ctx, $_);
   }

   return $optData;
}


# XXX FIX:
# This sub is only needed for generating users and events. When U&E is
# being retrieved from VMDB, this sub should be removed or replaced.

# ---------------------------------------------------------------------------
#
# getUsersEvents --
#
#      Creates JS arrays for the users/events tab on the overview page for
#      each supplied VM id
#
# Prototype:
#
#      string getUsersEvents($ctx, $vmId)
#
# Arguments:
#
#      $ctx  ... control source
#      $vmId ... hash value of vm
#
# Results:
#
#      Returns JS arrays to the client.
#
# ---------------------------------------------------------------------------

sub getUsersEvents($$$) {
   my ($ctx, $vmId, $server) = @_;

   my $vmCfg = getCfgPath($ctx, $vmId);
   my $log = getEventLog($server, $vmCfg);
   my $usr = getConnectedUsers($ctx);
   # trace("VM U/E Log: $log", "debug");
   # trace("VM U/E Users: $usr", "debug");

   return "new I4($log, $usr)";

} ## close sub getUsersEvents()


# XXX FIX:
# This sub is only needed for generating users and events. When U&E is
# being retrieved from VMDB, this sub should be removed or replaced.

sub getEventLog {
   my ($server, $cfg) = @_;
   my ($r, $n, $p, $di, $do, $argstr);
   my $lastn = $ENV{"vmware_EVENTLOG_LASTN"};
   $lastn ||= 15;

   my $ret = VMware::DOMAccess->newXML("<exec></exec>");
   $di = $ret->getDocument();
   $r = $di->getDocumentElement();
   $n = $r->appendChild($di->createElement("op"));
   $n->appendChild($di->createTextNode("VM_GetLastNEvents"));

   $r = $r->appendChild($di->createElement("in"));
   $n = $r->appendChild($di->createElement("cfg"));
   $n->appendChild($di->createTextNode($cfg));

   $n = $r->appendChild($di->createElement("lastn"));
   $n->appendChild($di->createTextNode($lastn));

   $do = $server->exec($di->toString());
   $di->dispose();

   return undef unless defined($do);
   $do = VMware::DOMAccess->newXML($do);

   return undef unless defined($do);

   if ($ENV{ "vmware_DEBUG_LEVEL" } > 2) {
      my $xmlfile = &VMware::Management::Util::tmpdir() .
      "SxMonitor-EventLog.xml";
      string_to_file($do->prettyPrint(), $xmlfile, 1);
   }

   my $rc = $do->getValue("exec.returncode");
   if ($rc == 1 && $do->getValue("exec.out")) {
      my @events = $do->get("exec.out.vm")->getElementsByTagName("event");
      @events = reverse(@events);

      foreach my $event (@events) {
         if (! defined($event)) {
            next;
         }

         my ($type, $time, $msg);

         $type = $event->getAttribute("type");
         $time = getVal($event, "time");

         if ($type eq "answer") {
            $msg = getVal($event, "subject");
            $msg =~ s/^(\w)(.*?)(OK)(.*?)$/$1\L$2\E$3\L$4\E/g;
            $msg = qq{<div class="lead">$msg.</div>};
         } else {
            my $subj = getVal($event, "subject");

            if ($subj =~ /^power-off$/i) {
               $msg = '<div class="lead">Power off</div>';
            } elsif ($subj =~ /^power-on$/i) {
               $msg = '<div class="lead">Power on</div>';
            } elsif ($subj =~ /^suspend$/i) {
               $msg = '<div class="lead">Suspend</div>';
            } else {
               my @p = $event->getElementsByTagName("p");

               for (@p) {
                  if (! defined($_) || ! defined($_->getFirstChild())) {
                     next;
                  }

                  $msg .= ($msg ? '<div>' : '<div class="lead">') .
                  $_->getFirstChild()->getNodeValue() . '</div>';
               }
            }
         }

         if ($type =~ /^info$/i &&
         (getVal($event, "subject") =~ /^vmware warning$/i ||
         getVal($event, "subject") =~ /^vmware informational message$/i)) {
            $type = 2; # Warning
         } elsif ($type =~ /^(warning|question)$/i) {
            $type = 2; # Warning
         } elsif ($type =~ /^error$/i) {
            $type = 4; # Error
         } else {
            $type = 1; # Notice
         }

         $msg =~ s/\\/\\\\/g;
         $msg =~ s/("|')/\\$1/g;
         my $lf = $/;
         $msg =~ s/$lf$lf/<br \/><br \/>/g;
         $msg =~ s/$lf/ /g;

         $argstr .= ($argstr ? ",$/" : "") . qq{new Evt($type,$time,"$msg")};
      }
   } else {
      trace("exec.returncode is $rc", "error");
      trace($do->getDocument()->toString(), "error");
   }
   $do->dispose();

   $argstr = "[$argstr]";
   return $argstr;
}


sub getConnectedUsers {
   my $ctx = shift();
   my $argstr = "";

   my $path = "vmx/connections/#";
   while (defined($path = $ctx->GetNextSibling($path))) {
      my $user = $ctx->GetString("${path}username");
      my $from = $ctx->GetString("${path}ipAddress");
      my $time = $ctx->GetString("${path}connectedTime");

      $argstr .= ($argstr ? "," : "") .
         qq{new Usr("$user",$time,"$from")};
   }

   $argstr = "[$argstr]";
   return $argstr;
}


# XXX FIX:"
# This sub is only needed for generating users and events. When U&E is
# being retrieved from VMDB, this sub should be removed or replaced.

sub getQuestion {
   my $top = shift();
   trace("Get Question Attribute: $top", "debug");

   my $q = $top->getElementsByTagName("question");
   if (! $q->getLength() > 0) {
      #trace("Died #1", "debug");
      return undef;
   }

   $q = $q->item(0);

   my $a = $top->getElementsByTagName("answer");
   if (! $a->getLength() > 0) {
      #trace("Died #2", "debug");
      return undef;
   }

   my $id = $q->getAttribute("id");
   my $qstr = $q->getFirstChild()->getNodeValue();
   $qstr =~ s/\\/\\\\/g;
   $qstr =~ s/("|')/\\$1/g;
   my $lf = $/;
   $qstr =~ s/$lf$lf/<br \/><br \/>/g;
   $qstr =~ s/$lf/ /g;

   my $astr = "";
   for (0 .. $a->getLength() - 1) {
      my $n = getElementByAtt($a, "id", $_);
      if (! defined($n) || $n->getAttribute("ref") ne $id) {
         trace("$n is not a node or $_ is not an answer id.", "error");
         next;
      }
      my $str = $n->getFirstChild()->getNodeValue();
      $str =~ s/\\/\\\\/g;
      $str =~ s/("|')/\\$1/g;
      $str =~ s/(\n)/<br \/>/g;

      $astr .= ($astr ? "," : "") . qq{"$str"};
   }
   $astr = "[$astr]";

   my $muiQst = qq[{id:$id,q:"$qstr",a:$astr}];

   trace("MUI Question: $muiQst", "debug");
   return $muiQst;

}


sub getVmdbMsgEvent {
   VMware::Management::Util::trace("IN getVmdbMsgEvent [", "debug");

   my $ctx = &VMware::Management::Session::get_temp_context();
   my $msgPath = shift();

   my $eventType = $ctx->GetString("${msgPath}type");
   my $choiceType = $ctx->GetString("${msgPath}choiceType");
   VMware::Management::Util::trace("CHOICE TYPE: $choiceType", "debug");
   my $cancelChoiceId = $ctx->GetInt("${msgPath}cancelChoice");

   # Use the event array index as the unique message instance identifier.
   # ${msgPath}id identifies the message class.
   my $jsQuestionId = $msgPath;
   $jsQuestionId =~ s/^.*\/#(_.*)\/$/$1/;
   $jsQuestionId = jsStringLiteral($jsQuestionId);
   
   my $jsQuestionStr = jsStringLiteral($ctx->GetString("${msgPath}text"));

   VMware::Management::Util::trace("QUESTION $jsQuestionId: $jsQuestionStr", "debug");
   
   my @jsAnswerArr = ();
   # custom, ok, okCancel, retryAbort, retryCancel, yesNo
   if ($choiceType =~ /^custom$/i) {
      VMware::Management::Util::trace("IN CHOICE TYPE: $choiceType", "debug");
      # There are one or more non-standard buttons.
      my $cancelChoiceExists = 0;

      # Built up an array of button labels.
      my $path = "${msgPath}choiceType/custom/choice/#";
      while (defined($path = $ctx->GetNextSibling($path))) {
	 my $choiceId = $ctx->GetString("${path}id");
	 my $choiceStr = $ctx->GetString("${path}text");
	 VMware::Management::Util::trace("CHOICE STR: $choiceStr", "debug");

	 if ($choiceId == $cancelChoiceId) {
	    $cancelChoiceExists = 1;
	 }

	 $jsAnswerArr[$choiceId] = qq{"$choiceStr"};
      }

      # If the cancel button does not yet have a label, give it one.
      if (! $cancelChoiceExists && $cancelChoiceId >= 0) {
	 $jsAnswerArr[$cancelChoiceId] = qq{"Cancel"};
      }
   } elsif ($choiceType =~ /^(ok|retry|yes)(|abort|cancel|no)$/i) {
      # There are one or two standard buttons.
      my $OK = $1;
      if ($OK eq "ok") {
	 $OK = "OK";
      }

      # XXX: I don't think that it's possible, but the cancel button could end
      # up on the left side, which is not the standard position for
      # Windows/Linux UIs.
      if ($cancelChoiceId > -1) {
	 if ($cancelChoiceId == 1) {
	    $jsAnswerArr[0] = qq{"\u$1"} || qq{"OK"};
	    $jsAnswerArr[1] = qq{"\u$2"} || qq{"Cancel"};
	 } else {
	    $jsAnswerArr[0] = qq{"\u$2"} || qq{"Cancel"};
	    $jsAnswerArr[1] = qq{"\u$1"} || qq{"OK"};
	 }
      } else {
	 $jsAnswerArr[0] = qq{"\u$1"} || qq{"OK"};
      }
   }

   my $jsAnswerStr = join(",", @jsAnswerArr);
   $jsAnswerStr = qq{[$jsAnswerStr]};

   VMware::Management::Util::trace("] OUT getVmdbMsgEvent", "debug");
   return qq{{id:$jsQuestionId,q:$jsQuestionStr,a:$jsAnswerStr}};
}


# XXX FIX:'
# This sub is only needed for generating users and events. When U&E is
# being retrieved from VMDB, this sub should be removed or replaced.

# ---------------------------------------------------------------------------
#
# getVal --
#
#      Return the node value of the first tag named by $key below the node
#      named by $top.
#
# Prototype:
#
#      string getVal(XML::DOM::Node $top, string $key)
#
# Arguments:
#
#      $top ... The node from which the search should begin
#      $key ... The tag name of the target node
#
# Results:
#
#      Returns the string value of the node specified by $key if found, or an
#      empty string.
#
# ---------------------------------------------------------------------------

sub getVal {
   my $top = shift();
   my $key = shift();
   my $val = $top->getElementsByTagName($key, 0);

   if ($val->getLength() > 0) {
      $val = $val->item(0)->getFirstChild()->getNodeValue();
   } else {
      $val = "";
   }

   return $val;

}


# XXX FIX:'
# This sub is only needed for generating users and events. When U&E is
# being retrieved from VMDB, this sub should be removed or replaced.

# ---------------------------------------------------------------------------
#
# getElementByAtt --
#
#    Find the first node in NodeList $list with attribute $att of value $val.
#
# Prototype:
#
#    XML::DOM::Node getElementByAtt(XML::DOM::NodeList $list,
#                                     string $att,
#                                     string $val)
#
# Arguments:
#
#    $list ... A list of nodes to search
#    $att  ... The name of the target attribute
#    $val  ... The desired attribute value
#
# Results:
#
#    Returns the first matching node found in $list, or undef.
#
# ---------------------------------------------------------------------------

sub getElementByAtt {
   my $list = shift();
   my $att = shift();
   my $val = shift();

   for (my $i = 0; $i < $list->getLength(); $i++) {
      if ($list->item($i)->getAttribute($att) eq $val) {
         return $list->item($i);
      }
   }

   return undef;

}


# ---------------------------------------------------------------------------
#
# getResources --
#
#      Creates JS arrays for the resources tabs on the VM details page for
#      each supplied VM id
#
# Prototype:
#
#      string getResources($ctx, $vmId, $server)
#
# Arguments:
#
#      $ctx    ... control source
#      $vmId   ... hash value of VM
#      $server ... Control::Server object
#      $vcpu   ... # cpus assigned to VM
#
# Results:
#
#      Returns JS arrays to the client.
#
# ---------------------------------------------------------------------------

sub getResources($$$) {
   my ($ctx, $vmId, $server, $vcpu) = @_;
   my ($r, $n, $di, $do);
   my ($cpuOpts, $memOpts, $dskOpts, $netOpts);

   my $vmCfg;
   if ($vmId ne "console") {
      $vmCfg = getCfgPath($ctx, $vmId);
   } else {
      # For the service console, use the world ID as the config path
      my $cosPath = $ctx->GetString("/vmkernel/console");
      $cosPath =~ m{/vmkernel/world/\#(\d+)};
      $vmCfg = $1;
   }
   trace("getResources: vmCfg = $vmCfg", "debug");
   my $vmStats = "/stats/vm/#$vmId";
   if (!$ctx->IsSet($vmStats)) {
      $vmStats = undef;
   }
   my $pcpu = $ctx->GetInt("/host2/#_mui/info/cpuNum");

   my $ret = VMware::DOMAccess->newXML("<exec></exec>");
   $di = $ret->getDocument();
   $r = $di->getDocumentElement();
   $n = $r->appendChild($di->createElement("op"));
   $n->appendChild($di->createTextNode("ESXResourceInfo_GetWorld"));

   $r = $r->appendChild($di->createElement("in"));
   $n = $r->appendChild($di->createElement("world"));
   $n->appendChild($di->createTextNode($vmCfg));

   $do = $server->exec($di->toString());
   $di->dispose();

   return undef unless defined($do);
   $do = VMware::DOMAccess->newXML($do);
   return undef unless defined($do);

   if ($ENV{ "vmware_DEBUG_LEVEL" } > 2) {
      my $xmlfile = &VMware::Management::Util::tmpdir() .
      "SxMonitor-Resources.xml";
      string_to_file($do->prettyPrint(), $xmlfile, 1);
   }

   my $rc = $do->getValue("exec.returncode");
   if ($rc == 1 && $do->getValue("exec.out")) {
      my $vm = $do->getSubTree("exec.out.server.vm");

      # Note about resources: all VMs include information about the configured
      # resource settings; running VMs also include the active settings.  So we
      # check for active settings first, falling back to configured settings if
      # they are undefined, and falling back to default values if both are unset.

      #
      # CPU
      #

      # Resources
      my @cpuOptions = ();
      push @cpuOptions, "min:" . int($vm->getValue("vm.cpu.min") ||
                                     $vm->getValue("vm.cpu.config_min") || 0);
      push @cpuOptions, "max:" . int($vm->getValue("vm.cpu.max") ||
                                     $vm->getValue("vm.cpu.config_max") ||
                                     (100 * $vcpu));

      my $cpuShares = $vm->getValue("vm.cpu.shares") ||
         $vm->getValue("vm.cpu.config_shares") ||
         "normal";

      if ($cpuShares !~ /^\d+$/) {
         $cpuShares = qq{"$cpuShares"};
      } else {
         int($cpuShares);
      }

      push @cpuOptions, "shares:$cpuShares";

      my $cpuHtStatus = $vm->getValue("vm.cpu.htsharing") ||
	                $vm->getValue("vm.cpu.config_htsharing") ||
	                "any";
      $cpuHtStatus = qq{"$cpuHtStatus"};
      push @cpuOptions, "ht:$cpuHtStatus";

      # Examples of valid affinity formats:
      # -- "all", "default", "all:all;": all VCPUs on all PCPUs
      # -- "all:0,1;": all VCPUs on PCPUs 0 and 1
      # -- "0:0,1;1:2,3;": VCPU 0 on PCPUs 0 and 1, VCPU 1 on PCPUs 2 and 3
      # Convert these affinity specifiers into a list of VM-wide (if 'all')
      # or per-VCPU (if individually specified) affinity masks (in which
      # the nth bit is set iff the VM/VCPU can run on CPU n).  For example,
      # given a 4-CPU server and the above affinity specifiers, we get:
      # -- "all" -> [0xf]
      # -- "all:0,1;" -> [0x3]
      # -- "0:0,1;1:2,3;" -> [0x3, 0xc]
      my $pcpu = $ctx->GetInt("/host2/#_mui/info/cpuNum");
      my @affinityList = ();

      # XXX Serverd currently doesn't do the above for running VMs;
      # it reports affinity for the group leader world only, as a
      # comma-separated list of valid CPUs.  Need to extend it to
      # handle per-VCPU affinity, but in the meantime assume that
      # all VCPUs are scheduled on the same set of CPUs.
      # Note: Can't use "world || config || default" idiom here since "0"
      # is a valid affinity value, which would be interpreted as false and
      # cause us to pick up the default value of "all" (PR29885).
      my $affinity = $vm->getValue("vm.cpu.affinity.string");
      if (!defined($affinity) || ($affinity eq "")) {
         $affinity = $vm->getValue("vm.cpu.config_affinity.string");
      }
      if (!defined($affinity) || ($affinity eq "")) {
         $affinity = "all";
      }

      # Backward-compatibility with old-style affinity
      # (comma-separated set of PCPUs applied to all VCPUs)
      if ($affinity =~ /^[0-9,]+$/) {
         $affinity = "all:$affinity";
      }
      # The VMkernel treats 'default' and 'all' as synonyms
      $affinity =~ s/default/all/g;

      if ($affinity eq "all") {
         push @affinityList, (1 << $pcpu) - 1;
      } else {
         foreach my $vaffin (split(/;/, $affinity)) {
            if (!$vaffin) {
               next;
            } elsif ($vaffin =~ /^(all|\d):(all|[\d,]+)$/) {
               my $vcpuid = ($1 eq "all") ? 0 : $1;
               my $mask = 0;
               if ($2 eq "all") {
                  $mask = (1 << $pcpu) - 1;
               } else {
                  foreach (split(/,/, $2)) {
                     $mask |= (1 << $_);
                  }
               }
               $affinityList[$vcpuid] = $mask;
            } else {
               trace("getResources: can't parse CPU affinity '$affinity'", "error");
            }
         }
      }
      if ((scalar(@affinityList) != 1) && (scalar(@affinityList) != $vcpu)) {
         trace("getResources: only found CPU affinity for " .
               scalar(@affinityList) . " of " . $vcpu . " VCPUs", "error");
      }
      push @cpuOptions, "affinity:[" . join(",", @affinityList) . "]";

      # Statistics
      my @cpuUsage = ();
      foreach my $p (0 .. $vcpu - 1) {
         if ($vmStats) {
            push @cpuUsage, $ctx->GetInt("$vmStats/cpu/instance/#$p/milUsage/avg/") / 10;
         } else {
            push @cpuUsage, 0;
         }
      }
      push @cpuOptions, "usage:[" . join(',', @cpuUsage) . "]";

      $cpuOpts = join(",", @cpuOptions);
      trace("getResources: cpuOpts = $cpuOpts", "debug");

      #
      # Memory
      #

      # Resources
      my @memOptions = ();
      my $memMin = int($vm->getValue("vm.memory.min") ||
                       $vm->getValue("vm.memory.config_min") || 0);
      my $memMax = int($vm->getValue("vm.memory.max") ||
                       $vm->getValue("vm.memory.config_max") || 0);

      my $memShares = $vm->getValue("vm.memory.shares") ||
         $vm->getValue("vm.memory.config_shares") ||
         "normal";

      if ($memShares !~ /^\d+$/) {
         $memShares = qq{"$memShares"};
      } else {
         int($memShares);
      }

      push @memOptions, "min:$memMin";
      push @memOptions, "max:$memMax";
      push @memOptions, "shares:$memShares";
      # NUMA affinity is just a comma-separated node list, or all/default
      # Note: Can't use "world || config || default" idiom here since "0"
      # is a valid affinity value, which would be interpreted as false and
      # cause us to pick up the default value of "all" (PR29885).
      $affinity = $vm->getValue("vm.memory.affinity.string");
      if (!defined($affinity) || ($affinity eq "")) {
         $affinity = $vm->getValue("vm.memory.config_affinity.string");
      }
      if (!defined($affinity) || ($affinity eq "")) {
         $affinity = "all";
      }
      $affinity =~ s/default/all/g;
      push @memOptions, "affinity:'$affinity'";

      # Statistics
      my %memFields = ( "kbGranted" => "granted",
                        "kbActive" => "active",
                        "kbShared" => "shared",
                        "kbSwapped" => "swapped",
                        "kbVmmemctl" => "vmmemctl",
                        "kbOverhead" => "overhead",
                        );
      foreach (keys %memFields) {
         my $val = $vmStats ? $ctx->GetInt("$vmStats/memory/$_/avg/") : 0;
         push @memOptions, $memFields{$_} . ":" . $val;
      }
      $memOpts = join(",", @memOptions);
      trace("getResources: memOpts = $memOpts", "debug");

      #
      # Disk
      #

      my @dskOptions = ();
      foreach my $elt ($vm->listElementNames("vm", "disk")) {
         my @tgtOptions = ();
         my $tgt = $vm->getAttribute("vm.$elt", "name");

         # Resources
         my $tgtShares = $vm->getValue("vm.$elt.shares") ||
            $vm->getValue("vm.$elt.config_shares") ||
            "normal";

         if ($tgtShares !~ /^\d+$/) {
            $tgtShares = qq{"$tgtShares"};
         } else {
            int($tgtShares);
         }

         push @tgtOptions, "shares:$tgtShares";

         # Statistics
         my %dskFields = ( "kbpsRead" => "read", "kbpsWritten" => "write" );
         foreach (keys %dskFields) {
            my $val = $vmStats ?
                $ctx->GetInt("$vmStats/disk/instance/#$tgt/$_/avg/") : 0;
            push @tgtOptions, $dskFields{$_} . ":" . $val;
         }
         push @dskOptions, "'$tgt':{" . join(",", @tgtOptions) . "}";
      }

      $dskOpts = join(",", @dskOptions);
      trace("getResources: dskOpts = $dskOpts", "debug");

      #
      # Network
      #

      my @netOptions = ();

      # Resources (global across all NICs)
      my @fltOptions = ();
      foreach my $elt ($vm->listElementNames("vm.network", "filter")) {
         my $filter = "vm.network.filter";
         if ($vm->getValue("$filter.class") eq "nfshaper") {
            @fltOptions = ("enable:true",
                           "avg:" . $vm->getValue("$filter.avg_bandwidth"),
                           "peak:" . $vm->getValue("$filter.peak_bandwidth"),
                           "burst:" . $vm->getValue("$filter.burst_size"));
            last;
         }
      }
      if (!@fltOptions) {
         my $filter = "vm.network.config_filter";
         if ($vm->getValue("$filter.class") eq "nfshaper") {
            @fltOptions = ("enable:true",
                           "avg:" . $vm->getValue("$filter.avg_bandwidth"),
                           "peak:" . $vm->getValue("$filter.peak_bandwidth"),
                           "burst:" . $vm->getValue("$filter.burst_size"));
         } else {
            @fltOptions = ( "enable:false" );
         }
      }
      push @netOptions, "nfshaper:{" . join(",", @fltOptions) . "}";

      # Statistics (per NIC)
      # XXX Need to get nic info elsewhere since stats are only available if VM is running
      if ($vmStats) {
         my $nicStats = "$vmStats/network/instance/#/";
         while (defined($nicStats = $ctx->GetNextSibling($nicStats))) {
            next if ($nicStats =~ m{_rollup_});
            my @nicOptions = ();
            # VMDB reports bytes per second; convert to bits per second
            push @nicOptions, "rx:" . (8 * $ctx->GetInt("${nicStats}kbpsRx/avg/"));
            push @nicOptions, "tx:" . (8 * $ctx->GetInt("${nicStats}kbpsTx/avg/"));

            $nicStats =~ m|network/instance/\#([^/]+)/|;
            my $nic = $1;
            push @netOptions, "'$nic':{" . join(",", @nicOptions) . "}";
         }
      }

      $netOpts = join(",", @netOptions);
      trace("getResources: netOpts = $netOpts", "debug");

      $vm->dispose();
   } else {
      trace("exec.returncode is $rc", "error");
      trace($do->getDocument()->toString(), "error");
   }
   $do->dispose();

   # Build and return the constructor for the I8 object
   return "new I8({$cpuOpts},{$memOpts},{$dskOpts},{$netOpts})";
} ## close sub getResources()


# ---------------------------------------------------------------------------
#
# getCfgPath --
#
#
#      Returns the config file path for the VM with the specified id.
#
# Prototype:
#
#      string getCfgPath($ctx, $vmId)
#
# Arguments:
#
#      $ctx    ... control source
#      $vmId   ... hash value of VM
#
# Results:
#
#      Returns the config file path, or undef if no such registered VM exists
#
# ---------------------------------------------------------------------------

sub getCfgPath {
   my ($ctx, $vmId) = @_;
   my $regVmPath = $MMDATA_ROOT . "host2/#_mui/vms/registeredvms/#/";
   while (defined($regVmPath = $ctx->GetNextSibling($regVmPath))) {
      if ($ctx->GetString("${regVmPath}vm") =~ /\#\Q$vmId\E/) {
         return $ctx->GetString("${regVmPath}cfgPath");
      }
   }
   return undef;
}


# ---------------------------------------------------------------------------
#
# handler --
#
#      Main sub which calls the genHead() and genTail subs to generate
#      javascript for embedding into the client's browser.
#      The JS code will contain all necessary information for the
#      host and it's VMs.
#
# Prototype:
#
#      string handler()
#
# Arguments:
#
#      none
#
# Results:
#
#      Returns an HTML string with embedded JavaScript for use by the client.
#
# ---------------------------------------------------------------------------

sub handler {
   $q = new CGI(shift());

   if (!use_vmdb()) {
      return (&VMware::Management::Util::error_page("VMDB is not being used.",
      __PACKAGE__, __LINE__));
   }

   my $ctx =  &VMware::Management::Session::get_temp_context();

   if (!defined($ctx)) {
      return (&VMware::Management::Util::error_page("Could not get temporary context.",
      __PACKAGE__, __LINE__));
   }

   my $html = genHead();
   $ctx->SetCurrentPath($MMDATA_ROOT);

   # XXX: The following code still relies on lagacy infrastructure [
   %credentials = hash_credentials();
   my $server = &VMware::Control::Server::new(undef, 0,
					      $credentials{ "l" },
					      $credentials{ "m" });
   my $version = VMware::Control::version( );
   my $want_version = "1.01";

   if (defined($server) && $server->connect()) {
      trace("Connected to server object ($server).", "notice");
      $html .= genTail($ctx, $server);
   } else {
      my ($err, $errstr) = $server->get_last_error();
      return (&VMware::Management::Util::error_page
      ("Unable to connect to server: $err - $errstr", __PACKAGE__, __LINE__));
   }

   if ($server && $server->is_connected()) {
      $server->disconnect();
      trace("server->disconnect( ) called ($server).", "notice");
   }
   # ] The preceding code still relies on lagacy infrastructure :XXX

   # DEBUG: Prints JS content to a file
   if ($ENV{"vmware_DEBUG_LEVEL"} > 2) {
      my $file = &VMware::Management::Util::tmpdir() . "monitor.html";
      string_to_file($html, $file, 1);
   }

   return ($html,
   "Pragma"        => "no-cache",
   "Cache-Control" => "no-cache",
   "Content-Type"  => "text/html; charset=iso-8859-1" );
} ## close handler() sub


# ---------------------------------------------------------------------------
#
# jsStringLiteral --
#
#      Make sure that $str is safe for consumption in a JavaScript program.
#
# ---------------------------------------------------------------------------

sub jsStringLiteral {
   my $str = shift();
   $str =~ s/\\/\\\\/g;
   $str =~ s/("|')/\\$1/g;
   $str =~ s/(\r\n|\r|\n)/\\n/g;

   return qq{"$str"};
}

1;
